Package: rtracklayer
Title: R interface to genome annotation files and the UCSC genome
        browser
Version: 1.73.0
Author: Michael Lawrence, Vince Carey, Robert Gentleman
Depends: R (>= 3.5), methods, GenomicRanges (>= 1.37.2)
Imports: XML (>= 1.98-0), BiocGenerics (>= 0.35.3), S4Vectors (>=
        0.23.18), IRanges (>= 2.13.13), XVector (>= 0.19.7), Seqinfo,
        Biostrings (>= 2.77.2), curl, httr, Rsamtools (>= 1.31.2),
        GenomicAlignments (>= 1.15.6), BiocIO, tools, restfulr (>=
        0.0.13)
Suggests: GenomeInfoDb, BSgenome (>= 1.33.4), humanStemCell, microRNA
        (>= 1.1.1), genefilter, limma, org.Hs.eg.db, hgu133plus2.db,
        GenomicFeatures, BSgenome.Hsapiens.UCSC.hg19,
        TxDb.Hsapiens.UCSC.hg19.knownGene, RUnit
LinkingTo: S4Vectors, IRanges, XVector
Description: Extensible framework for interacting with multiple genome
        browsers (currently UCSC built-in) and manipulating annotation
        tracks in various formats (currently GFF, BED, bedGraph, BED15,
        WIG, BigWig and 2bit built-in). The user may export/import
        tracks to/from the supported browsers, as well as query and
        modify the browser state, such as the current viewport.
Maintainer: Michael Lawrence <lawremi@gmail.com>
License: Artistic-2.0 + file LICENSE
Collate: io.R web.R ranges.R trackDb.R browser.R ucsc.R readGFF.R gff.R
        bed.R wig.R utils.R bigWig.R bigBed.R chain.R quickload.R
        trackhub.R twobit.R fasta.R tabix.R bam.R trackTable.R index.R
        test_rtracklayer_package.R ncbi.R igv.R zzz.R
biocViews: Annotation,Visualization,DataImport
Config/pak/sysreqs: make libbz2-dev liblzma-dev libxml2-dev libssl-dev
        xz-utils zlib1g-dev
Repository: https://bioc.r-universe.dev
Date/Publication: 2026-06-11 02:10:19 UTC
RemoteUrl: https://github.com/bioc/rtracklayer
RemoteRef: HEAD
RemoteSha: c524fdf64dfe7ecf2178a4ac85a2344ced17e158
NeedsCompilation: yes
Packaged: 2026-06-11 18:26:26 UTC; root
Built: R 4.6.0; x86_64-apple-darwin20; 2026-06-11 18:33:06 UTC; unix

useDynLib(rtracklayer, .registration = TRUE)

## ============================================================
## Import
## ============================================================

import(methods)
import(BiocGenerics)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### base packages
###

importFrom("stats", offset, setNames)
importFrom("utils", count.fields, URLencode, browseURL,
           download.file, read.table, read.csv, type.convert, write.table,
           packageVersion, strcapture)
importFrom("tools", file_path_as_absolute, file_ext, file_path_sans_ext)
importFrom("grDevices", col2rgb, rgb)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Non-bioconductor packages
###

importFrom("XML", getNodeSet, xmlValue, xmlAttrs, htmlTreeParse, xmlTreeParse,
           xmlInternalTreeParse, parseURI, newXMLNode, xmlChildren,
           addChildren, removeChildren, readHTMLTable)
importMethodsFrom("XML", saveXML)

importFrom("curl", curl_unescape)

importFrom("httr", GET, HEAD, POST, config, user_agent, set_cookies, content,
           upload_file)

import(restfulr)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Bioconductor packages
###

import(S4Vectors)
import(XVector)
import(Seqinfo)
import(IRanges)
import(GenomicRanges)
import(BiocIO)

importFrom("Biostrings", get_seqtype_conversion_lookup, writeXStringSet,
           DNAStringSet, DNA_BASES, DNA_ALPHABET, alphabetFrequency,
           reverseComplement, uniqueLetters)
importMethodsFrom("Biostrings", masks, "masks<-", getSeq)
importClassesFrom("Biostrings", DNAStringSet, XStringSet)

importFrom("Rsamtools", indexTabix, bgzip, TabixFile, index)
importMethodsFrom("Rsamtools", ScanBamParam, asBam, headerTabix, isOpen,
                  scanTabix)
importClassesFrom("Rsamtools", RsamtoolsFile, TabixFile, BamFile)

importMethodsFrom("GenomicAlignments", readGAlignments, cigar,
                  last, readGAlignmentPairs)
importClassesFrom("GenomicAlignments", GAlignments)

## ============================================================
## Export
## ============================================================

exportClasses(BrowserSession, BrowserView, BrowserViewList,
              UCSCSession, UCSCView,
              UCSCData, TrackLine, BasicTrackLine, GraphTrackLine,
              Bed15TrackLine, UCSCTrackModes, BigWigSelection,
              BigBedSelection, UCSCSchema, Quickload, QuickloadGenome,
              TrackHub, TrackHubGenome, Track, TrackContainer)

## File classes
exportClasses(GFFFile, UCSCFile, BEDFile,
              WIGFile, ChainFile, FastaFile, GFF1File, GFF2File, GFF3File,
              BEDGraphFile, BED15File, GTFFile, GVFFile, BigWigFile,
              BigWigFileList, BigBedFile, TwoBitFile, BEDPEFile)

exportMethods(activeView, "activeView<-", blocks, browseGenome,
              browserSession, "browserSession<-",
              browserView, browserViews,
              close, export, export.bed, export.bed15,
              export.bedGraph, export.gff, export.gff1, export.gff2,
              export.gff3, export.ucsc, export.wig, export.bw,
              export.bb, export.2bit,
              import, import.bed, import.bed15, import.bedGraph,
              import.gff, import.gff1, import.gff2, import.gff3,
              import.ucsc, import.wig, import.bw, import.chain,
              import.2bit, import.bb,
              exportToTabix,
              "track<-",
              track, trackNames, "trackNames<-", getTable, hubUrl, "hubUrl<-",
              tableNames,
              tableName, "tableName<-",
              ucscTrackModes, "ucscTrackModes<-",
              ucscSchema,
              coerce, initialize,
              show, summary, "[", ucscTableQuery,
              genome, "genome<-", chrom, "chrom<-", range, "range<-",
              visible, "visible<-",
              liftOver, offset, reversed, nrow,
              referenceSequence, "referenceSequence<-",
              asBED, asGFF,
              split, fileFormat,
              hub, "hub<-", shortLabel, "shortLabel<-", longLabel, "longLabel<-",
              genomesFile, "genomesFile<-", email, "email<-", descriptionUrl,
              "descriptionUrl<-", genomeField, "genomeField<-", getTracks , writeTrackHub,
              trackField, "trackField<-", genomeInfo, "genomeInfo<-",
              ## from IRanges
              start, end, "start<-", "end<-",
              score, "score<-",
              as.data.frame, space, mcols,
              ## from GenomicRanges
              strand, seqinfo, "seqinfo<-",
              ## from BSgenome
              organism, releaseDate, getSeq
              )

export(genomeBrowsers, start, end, strand, "start<-", "end<-", width, "width<-",
       ranges, values, GenomicData, GenomicSelection,
       score, "score<-", as.data.frame, space,
       ucscGenomes, ucscTables,
       readGFFPragmas, sniffGFFVersion, GFFcolnames, readGFF, readGFFAsGRanges,
       BigWigSelection, BigBedSelection,
       GRangesForUCSCGenome, GRangesForBSGenome,
       summary, seqinfo, genome, "genome<-", Genome,
       uri, Quickload, quickload, QuickloadGenome,
       organism, releaseDate, mcols, TrackHub, trackhub, TrackHubGenome,
       Track, TrackContainer, wigToBigWig,
       SeqinfoForBSGenome, SeqinfoForUCSCGenome, resource, path,
       cleanupBigWigCache, cleanupBigBedCache, viewURL)

export(GFFFile, UCSCFile, BEDFile, WIGFile,
       ChainFile, FastaFile, GFF1File, GFF2File, GFF3File, BEDGraphFile,
       BED15File, GTFFile, GVFFile, BigWigFile, BigWigFileList, TwoBitFile,
       BEDPEFile, BigBedFile)

test_wig <- function() {
  test_path <- system.file("tests", package = "rtracklayer")
  test_wig <- file.path(test_path, "step.wig")

  createCorrectGR <- function(si) {
    score <- c(seq(10, 20, by = 2.5), seq(17.5, 10, by = -2.5),
               seq(1000, 100, by = -100))
    start <- c(59104701 + cumsum(c(0, 200, 500, 200, 300, 180, 220, 390, 1180)),
               59108021 + cumsum(c(0, rep(300, 9))))
    width <- c(rep(1, 9), rep(200, 10))
    space <- factor(c(rep("chr19", 9), rep("chr18", 10)), seqlevels(si))
    correct_gr <- GRanges(space, IRanges(start, width = width), score = score)
    if (!any(is.na(genome(si))))
      genome(correct_gr) <- unname(genome(si)[1])
    seqinfo(correct_gr) <- si
    correct_gr
  }
  createCorrectUCSC <- function(gr) {  
    track_line <- new("GraphTrackLine", type = "wig", name = "test",
                      description = "test track", visibility = "full",
                      autoScale = FALSE, viewLimits = c(0, 1000),
                      color = c(0L, 200L, 100L),
                      maxHeightPixels = c(100L, 50L, 20L),
                      graphType = "points", priority = 30)
    new("UCSCData", gr, trackLine = track_line)
  }

  correct_gr <- createCorrectGR(Seqinfo(c("chr19", "chr18")))
  correct_ucsc <- createCorrectUCSC(correct_gr)

  ## TEST: basic import
  test <- import(test_wig)
  checkIdentical(correct_ucsc, test)
  test <- import.wig(test_wig)
  checkIdentical(correct_ucsc, test)
  test_wig_file <- WIGFile(test_wig)
  checkException(import(test_wig_file, format = "bed"))
  test <- import(test_wig_file, format = "wig")
  checkIdentical(correct_ucsc, test)
  test <- import(test_wig_file)
  checkIdentical(correct_ucsc, test)
  test_wig_con <- file(test_wig)
  test <- import(test_wig_con, format = "wig")
  checkIdentical(correct_ucsc, test)
  test_wig_con <- file(test_wig)
  test <- import(WIGFile(test_wig_con))
  checkIdentical(correct_ucsc, test)
  
  ## TEST: 'genome'
  hg19_seqinfo <- SeqinfoForBSGenome("hg19")
  correct_genome <- createCorrectUCSC(createCorrectGR(hg19_seqinfo))
  test <- import(test_wig, genome = "hg19")
  checkIdentical(correct_genome, test)

  ## TEST: trackLine = FALSE
  test <- import(test_wig, trackLine = FALSE)
  checkIdentical(correct_gr, test)

  ## TEST: which
  which <- correct_gr[3:4]
  correct_which <- subsetByOverlaps(correct_ucsc, which)
  test <- import(test_wig, which = which)
  checkIdentical(correct_which, test)

  ## TEST: basic export
  test_wig_out <- file.path(tempdir(), "test.wig")
  on.exit(unlink(test_wig_out))
  export(correct_ucsc, test_wig_out)
  test <- import(test_wig_out)
  checkIdentical(correct_ucsc, test)

  export.wig(correct_ucsc, test_wig_out)
  test <- import(test_wig_out)
  checkIdentical(correct_ucsc, test)
  test_foo_out <- file.path(tempdir(), "test.foo")
  export(correct_ucsc, test_foo_out, format = "wig")
  on.exit(unlink(test_foo_out))
  test <- import(test_wig_out)
  checkIdentical(correct_ucsc, test)
  test_wig_out_file <- WIGFile(test_wig_out)
  export(correct_ucsc, test_wig_out_file)
  test <- import(test_wig_out)
  checkIdentical(correct_ucsc, test)
  checkException(export(correct_ucsc, test_wig_out_file, format = "gff"))

  ## TEST: append
  correct_ucsc2 <- initialize(correct_ucsc,
                              trackLine = initialize(correct_ucsc@trackLine,
                                                     name = "test2"))
  export(correct_ucsc2, test_wig_out_file, append = TRUE)
  test <- import(test_wig_out_file)
  correct_list <- GRangesList(test = correct_ucsc,
                              test2 = correct_ucsc2,
                              compress = FALSE)
  checkIdentical(correct_list, test)

  ## TEST: track line parameters
  export(correct_ucsc, test_wig_out, name = "test2")
  test <- import(test_wig_out)
  checkIdentical(correct_ucsc2, test)

  ## TEST: export trackLine
  export(correct_ucsc, test_wig_out, trackLine = FALSE)
  test <- import(test_wig_out)
  checkIdentical(test, correct_gr)

  ## TEST: Plain GRanges / bedGraph
  export.ucsc(correct_gr, test_wig_out) # mixture of steps leads to bedGraph
  test <- import(test_wig_out)
  default_line <- new("GraphTrackLine", name = "R Track", type = "bedGraph")
  correct_default <- new("UCSCData", correct_gr, trackLine = default_line)
  checkIdentical(test, correct_default)
  correct_chr18 <- GenomeInfoDb::keepSeqlevels(correct_gr, "chr18",
                                               pruning.mode="coarse")
  export.ucsc(correct_chr18, test_wig_out)
  test <- import(test_wig_out)
  default_line <- new("GraphTrackLine", name = "R Track", type = "wig")
  correct_default <- new("UCSCData",
                         GenomeInfoDb::keepSeqlevels(correct_gr, "chr18",
                                                     pruning.mode="coarse"),
                         trackLine = default_line)
  checkIdentical(test, correct_default)
  
  ## TEST: GenomicRangesList
  export(correct_list, test_wig_out)
  test <- import(test_wig_out)
  checkIdentical(correct_list, test)
  
  ## TEST: gzip
  test_wig_gz <- paste(test_wig_out, ".gz", sep = "")
  on.exit(unlink(test_wig_gz))
  export(correct_ucsc, test_wig_gz)
  test <- import(test_wig_gz)
  checkIdentical(correct_ucsc, test)
  export(correct_ucsc2, test_wig_gz, append = TRUE)
  test <- import(test_wig_gz)
  checkIdentical(correct_list, test)
  
  ## TEST: Using connection to add comment header
  test_wig_con <- file(test_wig_out)
  open(test_wig_con, "w")
  comment <- "# test comment"
  writeLines(comment, test_wig_con)
  export(correct_ucsc, test_wig_con)
  close(test_wig_con)
  checkIdentical(comment, readLines(test_wig_out, n = 1))
  test <- import(test_wig_out)
  checkIdentical(correct_ucsc, test)
}

### =========================================================================
### BigWig support
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Classes
###

setClass("BigWigFile", contains = "BiocFile")
setClass("BWFile", contains = "BigWigFile")

BigWigFile <- function(path) {
  if (!isSingleString(path))
    stop("'filename' must be a single string, specifying a path")
  new("BigWigFile", resource = path)
}
BWFile <- BigWigFile

setMethod("seqinfo", "BigWigFile", function(x) {
  seqlengths <- .Call(BWGFile_seqlengths, expandPath(path(x)))
  Seqinfo(names(seqlengths), seqlengths) # no circularity information
})

setClass("BigWigFileList", contains = "BiocFileList",
    prototype = prototype(elementType = "BigWigFile"))

BigWigFileList <- function(path)
{
    new("BigWigFileList", listData = (lapply(path, BigWigFile)))
}

setMethod("path", "BigWigFileList",
    function(object, ...)
{
    sapply(as.list(object), path)
})

.allowedColNames <- list(bigWig = "score")

.validateColNames <- function(object, format) {
  allowedColNames <- .allowedColNames[[format]]
  invalidColNames <- setdiff(colnames(object), allowedColNames)
  if (length(invalidColNames))
    paste("Column names",
          paste("'", invalidColNames, "'", sep = "", collapse=", "),
          "are invalid for the", format, "format.")
  else NULL
}

setClass("BigWigSelection", prototype = prototype(colnames = "score"),
         contains = "RangedSelection")

setValidity("BigWigSelection",
            function(object) {
              .validateColNames(object, "bigWig")
            })

BigWigSelection <- function(ranges=IRangesList(), colnames = "score") {
  if (!is.character(colnames) ||
      (length(colnames) && !identical(colnames, "score")))
    stop("'score' is the only valid column for BigWig")
  if (is.character(ranges))
    new("BigWigSelection", GenomicSelection(ranges, colnames = colnames))
  else {
    if (is(ranges, "BigWigFile"))
      ranges <- seqinfo(ranges)
    new("BigWigSelection", ranges = as(ranges, "IntegerRangesList"),
        colnames = colnames)
  }
}

setAs("IntegerRangesList", "BigWigSelection", function(from) {
  new("BigWigSelection", as(from, "RangedSelection"), colnames = "score")
})

setAs("GenomicRanges", "BigWigSelection", function(from) {
  as(as(from, "IntegerRangesList"), "BigWigSelection")
})

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export
###

setGeneric("export.bw", function(object, con, ...) standardGeneric("export.bw"))

setMethod("export.bw", "ANY",
          function(object, con, ...)
          {
            export(object, con, "BigWig", ...)
          })

setMethod("export", c("ANY", "BigWigFile"),
          function(object, con, format, ...)
          {
            object <- as(object, "GRanges")
            callGeneric()
          })

setMethod("export", c("GenomicRanges", "BigWigFile"),
          function(object, con, format,
                   dataFormat = c("auto", "variableStep", "fixedStep",
                                  "bedGraph"),
                   compress = TRUE, fixedSummaries = FALSE)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            con <- path.expand(path(con))
            object <- sortBySeqnameAndStart(object)
            score <- score(object)
            if (isValidScore(score))
              stop("The score must be numeric, without any NA's")
            if (!isTRUEorFALSE(compress))
              stop("'compress' must be TRUE or FALSE")
            seqlengths <- seqlengths(object)
            if (any(is.na(seqlengths)))
              stop("Unable to determine seqlengths; either specify ",
                   "'seqlengths' or specify a genome on 'object' that ",
                   "is known to BSgenome or UCSC")
            sectionPtr <- NULL # keep adding to the same linked list
            .bigWigWriter <- function(chromData, con, dataFormat, append) {
              if (length(chromData) == 0L)
                return()
              if (any(tail(start(chromData), -1) <= head(end(chromData), -1)))
                stop("BigWig ranges cannot overlap")
              sectionPtr <<- .Call(BWGSectionList_add, sectionPtr,
                                   as.vector(seqnames(chromData)[1]),
                                   as(ranges(chromData), "IRanges"),
                                   as.numeric(score(chromData)), dataFormat)
            }
            dataFormat <- match.arg(dataFormat)
            if (dataFormat == "auto")
              format <- chooseGraphType(object)
            else format <- dataFormat
            on.exit(.Call(BWGSectionList_cleanup, sectionPtr))
            if (format == "bedGraph")
              lapply(split(object, seqnames(object)), .bigWigWriter, con,
                     dataFormat)
            else export.wig(object, con, dataFormat = dataFormat,
                            writer = .bigWigWriter, trackLine = FALSE)
            storage.mode(seqlengths) <- "integer"
            invisible(BigWigFile(.Call(BWGSectionList_write, sectionPtr,
                                       seqlengths, compress, fixedSummaries,
                                       con)))
          })

setMethod("export", c("List", "BigWigFile"),
          function(object, con, format, compress = TRUE, fixedSummaries = FALSE)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            con <- path.expand(path(con))
            if (!isTRUEorFALSE(compress))
              stop("'compress' must be TRUE or FALSE")
            if (is.null(names(object)))
                stop("'object' must have names")
            seqlengths <- elementNROWS(object)
            sectionPtr <- NULL # keep adding to the same linked list
            on.exit(.Call(BWGSectionList_cleanup, sectionPtr))
            writer <- BigWigWriter(object)
            for(chr in names(object)) {
              sectionPtr <- writer(chr, sectionPtr)
            }
            invisible(BigWigFile(.Call(BWGSectionList_write, sectionPtr,
                                       seqlengths, compress, fixedSummaries,
                                       con)))
          })

setGeneric("BigWigWriter", function(x) standardGeneric("BigWigWriter"))

setMethod("BigWigWriter", "RleList", function(x) {
  function(chr, sectionPtr) {
    .Call(BWGSectionList_add, sectionPtr,
          chr, ranges(x[[chr]]), as.numeric(runValue(x[[chr]])),
          "bedGraph")
  }
})

setMethods("BigWigWriter", list("IntegerList", "NumericList"), function(x) {
  function(chr, sectionPtr) {
    .Call(BWGSectionList_add, sectionPtr,
          chr, NULL, as.numeric(x[[chr]]),
          "fixedStep")
  }
})

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import
###

setGeneric("import.bw", function(con, ...) standardGeneric("import.bw"))

setMethod("import.bw", "ANY",
          function(con, ...)
          {
            import(con, "BigWig", ...)
          })

### Consider the use cases:
### - Extract whole genome or whole chromosome coverage as Rle (ChIP-seq)
###   - Use coverage() fast path
### - Extract single position coverage for millions of variants
###   - Slow to query by position, so extract whole genome/chromosome (above),
###     then use findRun trick.
### - Extract single position coverage for a hundred hot-spot variants
###   - coverage() should be fast enough, followed by findRun trick.
### - Extract coverage for one/all genes and summarize (RNA-seq?)
###   - probably want summarize,BigWigFile


setMethod("import", "BigWigFile",
          function(con, format, text, selection = BigWigSelection(which, ...),
                   which = con,
                   as = c("GRanges", "RleList", "NumericList"), ...)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            as <- match.arg(as)
            if (is(which, "GenomicRanges") && as == "NumericList") {
                orig_order <- order(seqnames(which))
            }
            selection <- as(selection, "BigWigSelection")
            validObject(selection)
            si <- seqinfo(con)
            which <- ranges(selection)
            badSpaces <- setdiff(names(which)[lengths(which) > 0L],
                                 seqlevels(si))
            if (length(badSpaces) > 0L)
              warning("'which' contains seqnames not known to BigWig file: ",
                      paste(badSpaces, collapse = ", "))
            which <- which[names(which) %in% seqlevels(si)]
            flatWhich <- unlist(which, use.names = FALSE)
            if (is.null(flatWhich))
              flatWhich <- IRanges()
            which_rl <- split(flatWhich, factor(space(which), seqlevels(si)))
            if (as != "NumericList") {
                which_rl <- as(which, "NormalIRangesList")
            }
            which <- GRanges(which_rl)
            names(which) <- names(unlist(which_rl, use.names=FALSE))
            C_ans <- .Call(BWGFile_query, expandPath(path(con)),
                           as.character(seqnames(which)), ranges(which),
                           identical(colnames(selection), "score"), 
                           as == "NumericList")
            if (as == "NumericList") {
              ans <- as(C_ans, "NumericList")
              names(ans) <- names(which)
              metadata(ans) <- list(ranges = as(which, "GRanges"))
              if (exists("orig_order")) {
                  ans[orig_order] <- ans
              }
              ans
            } else {
              nhits <- C_ans[[3L]]
              gr <- GRanges(rep(seqnames(which), nhits), C_ans[[1L]],
                            seqinfo=si)
              gr$score <- C_ans[[2L]]
              if (as == "RleList") {
                coverage(gr, weight="score")
              } else {
                strand(gr) <- "*"
                gr
              }
            }
           })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Summary
###

setMethod("summary", "BigWigFile",
          function(object, which = as(seqinfo(object), "GenomicRanges"),
                   size = 1L, type = c("mean", "min", "max", "coverage", "sd"),
                   defaultValue = NA_real_, asRle = FALSE,
                   as = c("GRangesList", "RleList", "matrix"), ...)
          {
            ### FIXME: could do with "GenomicRanges" here, but
            ### coercions generally only exist for GRanges specifically
            which <- as(which, "GRanges")
            if (!is.numeric(size))
              stop("'size' must be numeric")
            size <- recycleIntegerArg(size, "size", length(which))
            as <- match.arg(as)
            if (any(size > width(which)) && as != "matrix")
              stop("some 'which' are smaller than 'size'; ",
                   "consider passing as='matrix'")
            if (as == "matrix" && (length(size) == 0L || any(size != size[1L])))
                stop("for as='matrix', there must be one unique 'size'")
            type <- match.arg(type)
            if (!missing(asRle))
              warning("argument asRle is deprecated; use as='RleList'")
            if (asRle)
              as <- "RleList"
            if (type == "sd") type <- "std"
            if (!isSingleNumberOrNA(defaultValue))
              stop("'defaultValue' must be a single number or NA")
            summaryList <- .Call(BWGFile_summary, path.expand(path(object)),
                                 as.character(seqnames(which)),
                                 ranges(which), size, type,
                                 as.numeric(defaultValue))
            if (as == "matrix") {
                return(do.call(rbind, summaryList))
            }
            tiles <- tile(which, n = size)
            if (asRle) {
              setNames(RleList(mapply(Rle, summaryList, as.list(width(tiles))),
                               compress=FALSE),
                       names(which))
            } else {
              tiles <- unlist(tiles)
              tiles$score <- unlist(summaryList)
              relist(tiles, summaryList)
            }
          })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Conversion
###

wigToBigWig <-
  function(x, seqinfo,
           dest = paste(file_path_sans_ext(x, TRUE), "bw", sep = "."),
           clip = FALSE)
  {
    if (!isSingleString(x))
      stop("'x' must be a single string, the path to a WIG file")
    if (!isSingleString(dest))
      stop("'dest' must be a single string, the path to the BigWig output")
    if (!is(seqinfo, "Seqinfo"))
      stop("'seqinfo' must be NULL or a Seqinfo object")
    if (!isTRUEorFALSE(clip))
      stop("'clip' must be TRUE or FALSE")
    seqlengths <- seqlengths(seqinfo)
    if (any(is.na(seqlengths)))
      stop("'seqlengths(seqinfo)' must not contain any 'NA' values")
    x <- path.expand(x)
    dest <- path.expand(dest)
    ans <- .Call(BWGFile_fromWIG, x, clip, seqlengths, dest)
    invisible(BigWigFile(ans))
  }

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Utilities
###

## Remote data are cached locally. Need a way to cleanup.

cleanupBigWigCache <- function(maxDays = 0) {
  stopifnot(isSingleNumber(maxDays))
  dir <- "/tmp/udcCache"
  if (file.exists(dir)) {
      invisible(.Call(R_udcCleanup, maxDays))
  }
}

### =========================================================================
### BigBed support
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Classes
###

setClass("BigBedFile", contains = "BiocFile")
setClass("BBFile", contains = "BigBedFile")

BigBedFile <- function(path) {
  if (!isSingleString(path))
    stop("'filename' must be a single string, specifying a path")
  new("BigBedFile", resource = path)
}
BBFile <- BigBedFile

setMethod("seqinfo", "BigBedFile", function(x) {
  seqlengths <- .Call(BBDFile_seqlengths, expandPath(path(x)))
  Seqinfo(names(seqlengths), seqlengths)
})

.defaultColNames <- c("chrom", "chromStart", "chromEnd", "name", "score",
                      "strand", "thickStart", "thickEnd", "itemRgb",
                      "blockCount", "blockSizes", "blockStarts")

setClass("BigBedSelection", prototype = prototype(colnames = .defaultColNames),
         contains = "RangedSelection")

BigBedSelection <- function(ranges=IRangesList(), colnames = .defaultColNames) {
  if (is.character(ranges))
    new("BigBedSelection", GenomicSelection(ranges, colnames = colnames))
  else {
    if (is(ranges, "BigBedFile"))
      ranges <- seqinfo(ranges)
    new("BigBedSelection", ranges = as(ranges, "IntegerRangesList"),
     colnames = colnames)
  }
}

setAs("IntegerRangesList", "BigBedSelection", function(from) {
  new("BigBedSelection", as(from, "RangedSelection"), colnames = .defaultColNames)
})

setAs("GenomicRanges", "BigBedSelection", function(from) {
  as(as(from, "IntegerRangesList"), "BigBedSelection")
})

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import
###

setGeneric("import.bb", function(con, ...) standardGeneric("import.bb"))

setMethod("import.bb", "ANY", function(con, ...) {
  import(con, "BigBed", ...)
})

setMethod("import", "BigBedFile",
          function(con, format, text, selection = BigBedSelection(which, ...),
                   which = con, ...) {

            if (!missing(format))
              checkArgFormat(con, format)
            si <- seqinfo(con)

            # If which is NULL, create a GRanges spanning all sequences full length
            if (is.null(which)) {
              full_ranges <- IRanges(start = rep(1, length(seqlengths(si))),
                                    end = seqlengths(si))
              which <- GRanges(seqnames = seqlevels(si), ranges = full_ranges)
              selection <- BigBedSelection(which)
            } else {
              selection <- as(selection, "BigBedSelection")
            }

            col_names <- colnames(selection)
            # Get all available field names from bigBed file
            fields_name <- .Call(BBDFile_fieldnames, expandPath(path(con)))
            fields_name <- c(fields_name[[1L]], fields_name[[2L]])
            # always prefer file field names if the user did not provide any
            if (identical(col_names, .defaultColNames))
                col_names <- fields_name
            required_fields <- c("chrom", "chromStart", "chromEnd")
            missing_fields <- setdiff(required_fields, col_names)
            col_names <- c(col_names, missing_fields)
            # set the selected colnames back
            colnames(selection) <- col_names
            unavailable_fields <- setdiff(col_names, fields_name)
            unavailable_fields <- setdiff(unavailable_fields, .defaultColNames)
            if (length(unavailable_fields) > 0L)
                warning(paste("Unavailable field(s):",
                        paste(unavailable_fields, collapse = ", ")))

            # Extract and sanitize ranges against seqlevels in file
            ranges <- ranges(selection)
            badSpaces <- setdiff(names(ranges)[lengths(ranges) > 0L], seqlevels(si))
            if (length(badSpaces) > 0L)
              warning("'which' contains seqnames not known to BigBed file: ",
                      paste(badSpaces, collapse = ", "))
            ranges <- ranges[names(ranges) %in% seqlevels(si)]

            # Flatten and split by seqlevels for querying
            flatranges <- unlist(ranges, use.names = FALSE)
            if (is.null(flatranges))
              flatranges <- IRanges()
            which_rl <- split(flatranges, factor(space(ranges), seqlevels(si)))
            which <- GRanges(which_rl)

            C_ans <- .Call(BBDFile_query, expandPath(path(con)),
                           as.character(seqnames(which)), ranges(which),
                           colnames(selection))


            nhits <- C_ans[["n_qhits"]]
            # Reconstruct GRanges with genomic coordinates and seqinfo
            result_seqnames <- C_ans[["chrom"]]
            chromStart <- C_ans[["chromStart"]]
            chromEnd <- C_ans[["chromEnd"]]
            # to 1 based
            chromStart <- chromStart + 1L
            result_ranges <- IRanges(start = chromStart, chromEnd)
            gr <- GRanges(result_seqnames, result_ranges, seqinfo = si)

            if ("strand" %in% names(C_ans) && !is.null(C_ans[["strand"]])) {
                cleaned_strand <- gsub("\\.", "*", C_ans[["strand"]])
                strand(gr) <- factor(cleaned_strand, levels = c("+", "-", "*"))
            }

            if ("thickStart" %in% names(C_ans) && !is.null(C_ans[["thickStart"]])) {
                # convert 0-based to 1-based start
                thickStart <- C_ans[["thickStart"]] + 1L
                thickEnd <- C_ans[["thickEnd"]]

                if (!is.null(thickEnd) && length(thickStart) == length(thickEnd)) {
                    thick <- IRanges(start = thickStart, end = thickEnd)
                    mcols(gr)$thick <- thick
                } else {
                    warning("thickEnd not found or length mismatch with thickStart")
                }
            }

            if ("blockCount" %in% names(C_ans) && !is.null(C_ans[["blockCount"]])) {
                blockCount <- C_ans[["blockCount"]]
                blockSizes <- C_ans[["blockSizes"]]
                blockStarts <- C_ans[["blockStarts"]]
                # Convert blockStarts from 0-based to 1-based coordinates
                ir_list <- IRanges(start = blockStarts + 1L, width = blockSizes)
                blocks <- relist(ir_list, PartitioningByWidth(blockCount))
                names(blocks) <- NULL
                mcols(gr)$blocks <- blocks
            }

            if ("itemRgb" %in% names(C_ans) && !is.null(C_ans[["itemRgb"]])) {
                color <- C_ans[["itemRgb"]]
                spec <- color != 0L
                ans <- rep(NA_character_, length(color))
                ans[spec] <- sprintf("#%06X", bitwAnd(color[spec], 0xFFFFFF))
                mcols(gr)$itemRgb <- ans
            }

            processed_fields <- c("n_qhits", "chrom", "chromStart", "chromEnd",
                                  "strand", "thickStart", "thickEnd",
                                  "blockCount", "blockSizes", "blockStarts",
                                  "itemRgb")
            remaining_fields <- setdiff(names(C_ans), processed_fields)
            for (field in remaining_fields)
                mcols(gr)[[field]] <- C_ans[[field]]

           gr
          })


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export
###

setGeneric("export.bb", function(object, con, ...) standardGeneric("export.bb"))

setMethod("export.bb", "ANY",
          function(object, con, ...)
          {
            export(object, con, "BigBed", ...)
          })

setMethod("export", c("ANY", "BigBedFile"),
          function(object, con, format, ...)
          {
            object <- as(object, "GRanges")
            callGeneric()
          })

setMethod("export", c("GenomicRanges", "BigBedFile"),
          function(object, con, format, compress = TRUE, extraIndexes = "")
          {
            if (!missing(format))
              checkArgFormat(con, format)
            con <- path.expand(path(con))
            object <- sortBySeqnameAndStart(object)
            seqlengths <- seqlengths(object)
            stopIfNotValidForExport(object)
            if (!is.character(extraIndexes))
              stop("The extraIndexes must be character")
            if (any(is.na(seqlengths)))
              stop("Unable to determine seqlengths; either specify ",
                   "'seqlengths' or specify a genome on 'object' that ",
                   "is known to BSgenome or UCSC")
            if (!isTRUEorFALSE(compress))
              stop("'compress' must be TRUE or FALSE")
            seqlengths <- seqlengths(object)
            bedString <- bedString(object)
            autoSqlString <- autoSqlString(object)
            extraIndexes <- gsub("[\n\t ]", "", extraIndexes, perl = TRUE)
            invisible(BigBedFile(.Call(BBDFile_write, seqlengths, bedString, autoSqlString,
                                       extraIndexes, compress, con)))
          })

stopIfNotValidForExport <- function(x) {
  elementMetadata <- elementMetadata(x)
  name <- elementMetadata$name
  score <- elementMetadata$score
  itemRgb <- elementMetadata$itemRgb
  thick <- elementMetadata$thick
  blocks <- elementMetadata$blocks
  if (!is.null(name) && (!is.character(name) || any(is.na(name))))
    stop("The name must be character, without any NA's")
  if (isValidScore(score))
    stop("The score must be numeric, without any NA's")
  if (!is.null(itemRgb) && (!is.character(itemRgb) || any(is.na(itemRgb))))
    stop("The itemRgb must be character, without any NA's")
  if (!is.null(thick) && !is(thick, "IRanges"))
    stop("The thick must be IRanges")
  if (!is.null(blocks) && !is(blocks, "IRangesList"))
    stop("The blocks must be IRangesList")
}

bedString <- function(x) {
  elementMetadata <- elementMetadata(x)
  name <- elementMetadata$name
  elementMetadata$name <- NULL
  score <- elementMetadata$score
  elementMetadata$score <- NULL
  strand <- as.character(strand(x))
  strand <- gsub("*", ".", strand, fixed = TRUE)
  thick <- elementMetadata$thick
  thickStart <- NULL
  thickEnd <- NULL
  if (!is.null(thick)) {
    thickStart <- start(ranges(thick)) - 1L
    thickEnd <- end(ranges(thick))
    elementMetadata$thick <- NULL
  }
  itemRgb <- as.data.frame(t(col2rgb(elementMetadata$itemRgb)))
  itemRgb <- do.call(paste, c(itemRgb, sep=","))
  elementMetadata$itemRgb <- NULL
  blocks <- elementMetadata$blocks
  blockCount  <- NULL
  blockSizes  <- NULL
  blockStarts <- NULL
  if (!is.null(blocks)) {
    length <- length(blocks)
    blockCount <- lengths(blocks)
    blockSizes <- lapply(width(blocks), function(x) paste(x, collapse=","))
    blockStarts <- lapply(start(blocks) - 1L, function(x) paste(x, collapse=","))
    elementMetadata$blocks <- NULL
  }
  extraColumnsString <- do.call(paste, as.list(elementMetadata))
  paste(as.character(seqnames(x)), start(ranges(x)) - 1L, end(ranges(x)), name, score,
                     strand, thickStart, thickEnd, itemRgb, blockCount, blockSizes,
                     blockStarts, extraColumnsString, collapse = "\n")
}

autoSqlString <- function(x) {
  asString <- c('table bed "Browser Extensible Data" (\n',
                'string chrom; "Reference sequence chromosome or scaffold"\n',
                'uint chromStart; "Start position in chromosome"\n',
                'uint chromEnd; "End position in chromosome"\n')

  names <- c("name", "itemRgb", "score", "thick", "blocks", "double", "integer", "character", "raw")
  values <- c('string name; "Name of item."\n',
              'uint reserved; "Used as itemRgb as of 2004-11-22"\n',
              'uint score; "Score (0-1000)"\nchar[1] strand; "+ or - for strand"\n',
              paste0('uint thickStart; "Start of where display should be thick (start codon)"\n',
                     'uint thickEnd; "End of where display should be thick (stop codon)"\n'),
              paste0('int blockCount; "Number of blocks"\n',
                     'int[blockCount] blockSizes; "Comma separated list of block sizes"\n',
                     'int[blockCount] chromStarts; "Start positions relative to chromStart"\n'),
              "double ", "int ", "string ", "uint ")
  mapping <- setNames(values, names)
  metadata <- elementMetadata(x)
  names <- names(metadata)
  defaultFields <- colnames(BigBedSelection())
  fieldsString <- lapply(names, function(y) {
    if (y %in% defaultFields)
      mapping[y]
    else {
      typeString <- mapping[storage.mode(metadata[[y]])]
      paste(typeString, y, '; ""\n')
    }
  })
  asString <- c(asString, fieldsString, ')')
  paste(asString, collapse = "")
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Utilities
###

cleanupBigBedCache <- cleanupBigWigCache

### =========================================================================
### FASTA support
### -------------------------------------------------------------------------
###
### Since we have 2bit, we might as well have FASTA
###

setClass("FastaFile", contains = "BiocFile")

FastaFile <- function(resource) {
  new("FastaFile", resource = resource)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export
###

setMethod("export", c("ANY", "FastaFile"), function(object, con, ...) {
  export(as(object, "DNAStringSet"), con, ...)
})

setMethod("export", c("XStringSet", "FastaFile"),
          function(object, con, format, ...)
          {
            writeXStringSet(object, path(con), format = "fasta", ...)
          })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import
###

setMethod("import", "FastaFile",
          function(con, format, text, type = c("DNA", "RNA", "AA", "B"), ...)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            readFun <- get(paste0("read", match.arg(type), "StringSet"),
                           getNamespace("Biostrings"))
            readFun(path(con), format = "fasta", ...)
          })

### =========================================================================
### readGFF()
### -------------------------------------------------------------------------


.make_filexp_from_filepath <- function(filepath)
{
    if (isSingleString(filepath))
        return(open_input_files(filepath)[[1L]])
    if (!inherits(filepath, "connection"))
        stop(wmsg("'filepath' must be a single string or a connection"))
    if (!base::isSeekable(filepath))
        stop(wmsg("connection is not seekable"))
    filepath
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### readGFFPragmas()
###

readGFFPragmas <- function(filepath)
{
    filexp <- .make_filexp_from_filepath(filepath)
    if (inherits(filexp, "connection")) {
        if (!base::isOpen(filexp)) {
            base::open(filexp)
            on.exit(base::close(filexp))
        }
        if (base::seek(filexp) != 0) {
            warning(wmsg("connection is not positioned at the start ",
                         "of the file, rewinding it"), immediate.=TRUE)
            base::seek(filexp, where=0)
        }
    }
    .Call(read_gff_pragmas, filexp)
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### sniffGFFVersion()
###

.get_version_from_pragmas <- function(pragmas)
{
    attrcol_fmt <- attr(pragmas, "attrcol_fmt")
    idx <- grep("^##gff-version", pragmas)
    if (length(idx) == 0L) {
        if (is.null(attrcol_fmt))
            stop(wmsg("'attr(pragmas, \"attrcol_fmt\")' is NULL"))
        return(attrcol_fmt)
    }
    version <- sub("^##gff-version", "", pragmas[idx])
    version <- unique(version)
    if (length(version) > 1L) {
        warning(wmsg("more than one GFF version specified in the file, ",
                     "returning the first one"))
        version <- version[[1L]]
    }
    version <- suppressWarnings(as.integer(version))
    if (is.na(version)) {
        warning(wmsg("unrecognized GFF version specified in the file"))
        if (is.null(attrcol_fmt))
            stop(wmsg("'attr(pragmas, \"attrcol_fmt\")' is NULL"))
        return(attrcol_fmt)
    }
    version
}

sniffGFFVersion <- function(filepath)
{
    pragmas <- readGFFPragmas(filepath)
    .get_version_from_pragmas(pragmas)
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### GFFcolnames()
###
### Return the 9 standard GFF columns as specified at:
###   http://www.sequenceontology.org/resources/gff3.html
###

GFFcolnames <- function(GFF1=FALSE)
{
    if (!isTRUEorFALSE(GFF1))
        stop(wmsg("'GFF1' must be TRUE or FALSE"))
    .Call(gff_colnames, GFF1)
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### readGFF()
###

### Returns 0L, 1L, 2L, or 3L.
.normarg_version <- function(version=0)
{
    if (is.character(version)) {
        ## For compatibility with "import" method for GFFFile objects.
        if (length(version) == 0L)
            return(0L)
        if (isSingleString(version)) {
            IMPORT_STYLE_VERSIONS <- c("", "1", "2", "3")
            m <- match(version, IMPORT_STYLE_VERSIONS)
            if (is.na(m))
                stop(wmsg("when a single string, 'version' must ",
                          "be \"\", \"1\", \"2\", or \"3\""))
            version <- m - 1L
            return(version)
        }
    }
    if (isSingleNumber(version)) {
        if (!is.integer(version))
            version <- as.integer(version)
        if (version < 0L || version > 3L)
            stop(wmsg("'version' must be 0, 1, 2, or 3"))
        return(version)
    }
    stop(wmsg("'version' must be a single number"))
}

.prepare_colmap_and_tags <- function(columns=NULL, tags=NULL, attrcol_fmt=0L)
{
    ## Check 'columns'.
    if (!(is.null(columns) || is.character(columns)))
        stop(wmsg("'columns' must be NULL or a character vector"))

    GFF_colnames <- GFFcolnames(attrcol_fmt == 1L)

    ## Check 'tags'.
    if (!is.null(tags)) {
        if (!is.character(tags))
            stop(wmsg("'tags' must be NULL or character vector"))
        if (any(is.na(tags)) || anyDuplicated(tags))
            stop(wmsg("'tags' cannot contain NAs or duplicates"))
        if (attrcol_fmt == 1L) {
            ## Move any GFF colname found in 'tags' to 'columns'.
            columns <- union(columns, intersect(tags, GFF_colnames))
            tags <- setdiff(tags, GFF_colnames)
            if (length(tags) != 0L)
                warning(wmsg("trying to extract tags from a GFF1 file"))
        }
    }

    ## Prepare 'colmap'.
    if (is.null(columns)) {
        colmap <- seq_along(GFF_colnames)
        if (attrcol_fmt != 1L) {
            stopifnot(GFF_colnames[[length(GFF_colnames)]] == "attributes")
            ## We don't load the "attributes" column unless the user
            ## requested no tags (i.e. by setting 'tags' to character(0)).
            if (!(is.character(tags) && length(tags) == 0L))
                colmap[[length(GFF_colnames)]] <- NA_integer_
        }
    } else if (is.character(columns)) {
        if (!all(columns %in% GFF_colnames)) {
            in1string <- paste0(GFF_colnames, collapse=", ")
            stop(wmsg("'columns' must contain valid GFF columns. ",
                  "Valid GFF columns are: ", in1string))
        }
        if (anyDuplicated(columns))
            stop(wmsg("'columns' cannot contain duplicates"))
        colmap <- match(GFF_colnames, columns)
    } else {
        stop(wmsg("'columns' must be NULL or a character vector"))
    }

    list(colmap=colmap, tags=tags)
}

.normarg_filter <- function(filter, attrcol_fmt=0L)
{
    GFF_colnames <- GFFcolnames(attrcol_fmt == 1L)
    if (is.null(filter))
        return(NULL)
    if (!is.list(filter))
        stop(wmsg("'filter' must be NULL or a named list"))
    filter_names <- names(filter)
    if (is.null(filter_names))
        stop(wmsg("'filter' must have names"))
    if (attrcol_fmt == 1L) {
        valid_filter_names <- GFF_colnames
    } else {
        valid_filter_names <- head(GFF_colnames, n=-1L)
    }
    if (!all(filter_names %in% valid_filter_names)) {
        in1string <- paste0(valid_filter_names, collapse=", ")
        if (attrcol_fmt == 1L) {
            excluding_note <- ""
        } else {
            excluding_note <- "(excluding \"attributes\")"
        }
        stop(wmsg("The names on 'filter' must be valid GFF columns ",
                  excluding_note, ". ",
                  "Valid 'filter' names: ", in1string))
    }
    if (anyDuplicated(filter_names))
        stop(wmsg("the names on 'filter' must be unique"))
    unname(filter[valid_filter_names])
}

### 'df' must be a data-frame-like object (typically an ordinary data frame or
### a DataFrame object).
.is_multi_tag <- function(df, ntag, attrcol_fmt=0L)
{
    if (ntag == 0L || attrcol_fmt != 3L)
        return(logical(ntag))
    multi_tags <- c("Parent", "Alias", "Note",
                    "Dbxref", "Ontology_term")
    sapply(seq_len(ntag) + ncol(df) - ntag,
           function(j)
               colnames(df)[[j]] %in% multi_tags ||
                   any(grepl(",", df[[j]], fixed=TRUE)))
}

### 'df' must be a data-frame-like object (typically an ordinary data frame or
### a DataFrame object). 'decode_idx' must be a non-empty integer vector
### indicating which columns to decode. The columns to decode must be character
### vectors.
.url_decode_cols <- function(df, decode_idx)
{
    decoded_cols <- lapply(setNames(decode_idx, colnames(df)[decode_idx]),
                           function(j)
                             urlDecode(df[[j]], na.strings=NA_character_))
    df[decode_idx] <- decoded_cols
    df
}

### 'df' must be a data-frame-like object (typically an ordinary data frame or
### a DataFrame object). 'split_idx' must be a non-empty integer vector
### indicating which columns to split. The columns to split must be character
### vectors. Split values are passed thru urlDecode() unless 'raw_data' is
### TRUE. Always returns a DataFrame.
.strsplit_cols <- function(df, split_idx, raw_data)
{
    split_cols <- lapply(setNames(split_idx, colnames(df)[split_idx]),
        function(j) {
            col <- df[[j]]
            ## Probably the most efficient way to create an empty CharacterList
            ## of arbitrary length.
            split_col <- relist(character(0),
                                PartitioningByEnd(rep.int(0L, length(col))))
            not_na <- !is.na(col)
            tmp <- strsplit(col[not_na], ",", fixed=TRUE)
            split_col[not_na] <- CharacterList(tmp)
            if (raw_data)
                return(split_col)
            relist(urlDecode(unlist(split_col)), split_col)
        })
    ## Surprisingly sticking the CharacterList cols back into 'df' works
    ## even if 'df' is an ordinary data frame!
    df[split_idx] <- split_cols
    ans <- DataFrame(df, check.names=FALSE)
    ## "show" method for DataFrame is broken if some colnames are the empty
    ## string so we rename this column (in our case, we know there can only
    ## be one).
    m <- match("", colnames(ans))
    if (!is.na(m))
        colnames(ans)[m] <- "__empty_tag__"
    ans
}

readGFF <- function(filepath, version=0, columns=NULL, tags=NULL,
                    filter=NULL, nrows=-1, raw_data=FALSE)
{
    ## Check 'filepath'.
    filexp <- .make_filexp_from_filepath(filepath)
    if (inherits(filexp, "connection")) {
        if (!base::isOpen(filexp)) {
            base::open(filexp)
            on.exit(base::close(filexp))
        }
        if (base::seek(filexp) != 0) {
            warning(wmsg("connection is not positioned at the start ",
                         "of the file, rewinding it"), immediate.=TRUE)
            base::seek(filexp, where=0)
        }
    }

    ## Check 'version'.
    version <- .normarg_version(version)

    ## Get pragmas lines.
    pragmas <- .Call(read_gff_pragmas, filexp)

    ## Rewind file.
    if (inherits(filexp, "connection")) {
        base::seek(filexp, where=0)
    } else {
        XVector:::rewind_filexp(filexp)
    }

    if (version == 0L) {
        attrcol_fmt <- .get_version_from_pragmas(pragmas)
    } else {
        attrcol_fmt <- version
    }

    ## Prepare 'colmap' and normalize 'tags'.
    colmap_and_tags <- .prepare_colmap_and_tags(columns, tags, attrcol_fmt)
    colmap <- colmap_and_tags$colmap
    tags <- colmap_and_tags$tags

    ## Normalize 'filter'.
    filter <- .normarg_filter(filter, attrcol_fmt)

    ## Normalize 'nrows'.
    if (!isSingleNumber(nrows))
        stop(wmsg("'nrows' must be a single number"))
    if (!is.integer(nrows))
        nrows <- as.integer(nrows)

    ## Check 'raw_data'.
    if (!isTRUEorFALSE(raw_data))
        stop(wmsg("'raw_data' must be TRUE or FALSE"))

    ## 1st pass.
    scan_ans <- .Call(scan_gff, filexp, attrcol_fmt, tags, filter, nrows)
    if (is.null(tags))
        tags <- scan_ans[[1L]]
    nrows <- scan_ans[[2L]]

    ## Rewind file.
    if (inherits(filexp, "connection")) {
        base::seek(filexp, where=0)
    } else {
        XVector:::rewind_filexp(filexp)
    }

    ## 2nd pass: return 'ans' as an ordinary data frame.
    ans <- .Call(load_gff, filexp, attrcol_fmt, tags, filter,
                           nrows, pragmas,
                           colmap, raw_data)
    ncol0 <- attr(ans, "ncol0")
    ntag <- attr(ans, "ntag")          # should be the same as 'length(tags)'

    ## Post-process standard GFF cols.
    if (!raw_data) {
        if (attrcol_fmt == 1L) {
            #factor_colnames <- c("seqid", "source", "type", "strand", "group")
            factor_colnames <- c("seqid", "source", "type", "group")
        } else {
            #factor_colnames <- c("seqid", "source", "type", "strand")
            factor_colnames <- c("seqid", "source", "type")
        }
        m <- match(factor_colnames, head(colnames(ans), n=ncol0))
        m <- m[!is.na(m)]
        factor_cols <- lapply(setNames(m, colnames(ans)[m]),
                              function(j)
                                factor(ans[[j]], levels=unique(ans[[j]])))
        ans[m] <- factor_cols
    }

    ## Post-process tags.
    if (ntag != 0L) {
        is_multi_tag <- .is_multi_tag(ans, ntag, attrcol_fmt)
        if (!raw_data) {
            decode_idx <- which(!is_multi_tag) + ncol0
            if (length(decode_idx) != 0L)
                ans <- .url_decode_cols(ans, decode_idx)
        }
        split_idx <- which(is_multi_tag) + ncol0
        if (length(split_idx) != 0L) {
            ## Returns 'ans' as a DataFrame.
            ans <- .strsplit_cols(ans, split_idx, raw_data)
        }
    }

    ## 'ans' could have lost its readGFF-specific attributes (e.g. if it was
    ## turned into a DataFrame), so we restore them and cross our fingers that
    ## they won't clash with the DataFrame slots the day the internals of
    ## DataFrame objects happen to change (very unlikely though).
    if (is.null(attr(ans, "pragmas")))
        attr(ans, "pragmas") <- pragmas
    if (is.null(attr(ans, "attrcol_fmt")))
        attr(ans, "attrcol_fmt") <- attrcol_fmt
    if (is.null(attr(ans, "ncol0")))
        attr(ans, "ncol0") <- ncol0
    if (is.null(attr(ans, "ntag")))
        attr(ans, "ntag") <- ntag
    if (is.null(attr(ans, "raw_data")))
        attr(ans, "raw_data") <- raw_data
    ans
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### readGFFAsGRanges()
###
### Used in "import" method for GFFFile objects.
### Not exported (user should use import()).
###

### sequence-region => Seqinfo -- by Michael
.parseSequenceRegionsAsSeqinfo <- function(lines) {
    sr <- grep("##sequence-region", lines, value=TRUE)
    srcon <- file()
    on.exit(base::close(srcon))
    writeLines(sr, srcon)
    srt <- read.table(srcon, comment.char="",
                      colClasses=list(NULL, "character", "integer",
                          "integer"))
    if (any(srt[[2L]] != 1L)) {
        warning("One or more ##sequence-region directives do not start at 1. ",
                "The assumptions made by 'sequenceRegionsAsSeqinfo=TRUE' ",
                "have been violated.")
    }
    Seqinfo(srt[[1L]], srt[[3L]])
}

### -- by Michael
.parseSpeciesAsMetadata <- function(lines) {
    species <- unique(grep("##species", lines, fixed=TRUE, value=TRUE))
    if (length(species) > 1L) {
        stop("multiple species definitions found")
    }
    metadata <- list()
    if (length(species) == 1L) {
        species <- sub("##species ", "", species, fixed=TRUE)
        if (isNCBISpeciesURL(species)) {
            ncbiError <- function(e) {
                warning("failed to retrieve organism information from NCBI")
            }
            metadata <- tryCatch(metadataFromNCBI(species), error = ncbiError)
        }
    }
    metadata
}

readGFFAsGRanges <- function(filepath, version=0, colnames=NULL, filter=NULL,
                             genome=NA,
                             sequenceRegionsAsSeqinfo=FALSE,
                             speciesAsMetadata=FALSE)
{
    if (!(isSingleStringOrNA(genome) || is(genome, "Seqinfo")))
        stop(wmsg("'genome' must be a single string or NA, ",
                  "or a Seqinfo object"))
    if (!isTRUEorFALSE(sequenceRegionsAsSeqinfo))
        stop(wmsg("'sequenceRegionsAsSeqinfo' must be TRUE or FALSE"))
    if (!isTRUEorFALSE(speciesAsMetadata))
        stop(wmsg("'speciesAsMetadata' must be TRUE or FALSE"))

    ## Read as data frame.
    if (is.null(colnames)) {
        df <- readGFF(filepath, version=version, filter=filter)
    } else {
        if (!is.character(colnames))
            stop(wmsg("'colnames' must be a character vector"))
        ## Split 'colnames' between 'columns' and 'tags'.
        GFF_colnames <- GFFcolnames()
        columns <- intersect(colnames, GFF_colnames)
        tags <- setdiff(colnames, GFF_colnames)
        core_columns <- c("seqid", "start", "end", "strand")
        columns <- union(columns, core_columns)
        df <- readGFF(filepath, version=version,
                      columns=columns, tags=tags, filter=filter)
    }

    ## Turn data frame into GRanges.
    ## TODO: Maybe we should be able to pass the metadata to
    ## makeGRangesFromDataFrame()?
    if (is.null(colnames)) {
        ans <- makeGRangesFromDataFrame(df, keep.extra.columns=TRUE,
                                            seqnames.field="seqid")
    } else {
        ans <- makeGRangesFromDataFrame(df, seqnames.field="seqid")
        mcols(ans) <- df[ , colnames, drop=FALSE]
    }

    ## Set seqinfo.
    ans_seqinfo <- NULL
    pragmas <- attr(df, "pragmas")
    attrcol_fmt <- attr(df, "attrcol_fmt")
    if (sequenceRegionsAsSeqinfo && attrcol_fmt == 3L) {
        ## Get 'ans_seqinfo' from pragmas.
        ans_seqinfo <- .parseSequenceRegionsAsSeqinfo(pragmas)
    } else if (is(genome, "Seqinfo")) {
        ans_seqinfo <- genome
        if (!all(seqlevels(ans) %in% seqlevels(ans_seqinfo)))
            stop(wmsg("the sequence names in the GTF or GFF file are in ",
                      "disagreement with the Seqinfo object specified via ",
                      "the 'genome' argument"))
    } else if (isSingleString(genome)) {
        ans_seqinfo <- seqinfoForGenome(genome)  # can return NULL
        if (!is.null(ans_seqinfo) &&
            !all(seqlevels(ans) %in% seqlevels(ans_seqinfo)))
        {
            warning(wmsg("cannot set the seqlengths or circularity flags on ",
                         "the GRanges object to return because the sequence ",
                         "names in the GTF or GFF file are in disagreement ",
                         "with the sequence names implied by the genome ",
                         "assembly (", genome, ") specified via the 'genome' ",
                         "argument"))
            ans_seqinfo <- NULL
        }
    }
    if (!is.null(ans_seqinfo)) {
        seqlevels(ans) <- seqlevels(ans_seqinfo)
        seqinfo(ans) <- ans_seqinfo
    }
    if (isSingleString(genome))
        genome(ans) <- genome

    ## Get 'ans_metadata' from pragmas.
    if (speciesAsMetadata) {
        ans_metadata <- .parseSpeciesAsMetadata(pragmas)
    } else {
        ans_metadata <- list()
    }

    metadata(ans) <- ans_metadata
    ans
}

.onUnload <- function(libpath)
{
    library.dynam.unload("rtracklayer", libpath)
}

setUserUdcDir <- function() {
    dir <- paste0("/tmp/udcCache_", Sys.info()[["user"]])
    .Call(R_setUserUdcDir, dir)
}

.onLoad <- function(libname, pkgname) 
{
    setUserUdcDir()
}

### =========================================================================
### Separate index support
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### queryForResource
###

## There is some question as to whether the import routines should
## recurse through import.tabix, or perform the query
## internally. Currently, we are taking the latter route, because it
## is consistent with (gzip) decoding and it allows the parsers more
## flexibility.

setGeneric("queryForResource",
           function(manager, x, which = NULL, ...)
               standardGeneric("queryForResource"),
           signature="x")

## Attaches 'usedWhich' attribute, an optimization hint indicating
## that subsetting by 'which' has been performed and is no longer
## necessary. Probably premature.

setMethod("queryForResource", "BiocFile", function(manager, x, which = NULL, ...)
{
  r <- resource(x)
  ans <- structure(r, usedWhich = FALSE)
  if (!is.null(which) && is.character(r)) {
    x_tbi <- paste(r, "tbi", sep = ".")
    if (file.exists(x_tbi))
      ans <- queryForResource(manager, TabixFile(r), which = which, ...)
  }
  ans
})

setMethod("queryForResource", "TabixFile",
          function(manager, x, which, header = TRUE, ...) {
            tabixHeader <- headerTabix(x)
            si <- Seqinfo(tabixHeader$seqnames)
            if (is.null(which)) {
              buffer <- connectionForResource(manager, path(x), "r")
              if (!header)
                readLines(buffer, tabixHeader$skip)
            } else {
              buffer <- manage(manager, file())
              if (header) {
                skippedLines <- readLines(path(x), tabixHeader$skip)
                writeLines(skippedLines, buffer)
              }
              lines <- unlist(scanTabix(x, param = which), use.names = FALSE)
              writeLines(lines, buffer)
              si <- merge(si, seqinfo(which))
            }
            structure(buffer, usedWhich = TRUE, seqinfo = si)
          })

queryForConnection <- function(manager, x, which = NULL, ...) {
  resource <- queryForResource(manager, x, which = which, ...)
  con <- connectionForResource(manager, resource, open = "r")
  structure(con, usedWhich = attr(resource, "usedWhich"))
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### indexTrack: attempt to build an index for a track
###
### This is an obscene hack
###

indexTrack <- function(con, ...) {
  indexed <- NULL
  formats <- eval(formals(indexTabix)$format)
  uri <- path(con)
  parsed_uri <- .parseURI(uri)
  if (!uriIsLocal(parsed_uri))
    stop("'con' must be a path to a local file")
  original_path <- parsed_uri$path
  path <- bgzip(original_path, overwrite = TRUE)
  format <- Find(function(f) {
    is(con, paste(toupper(f), "File", sep = ""))
  }, formats)
  if (!is.null(format)) {
    indexTabix(path, format, ...)
  } else {
    indexTabix(path, ...)
  }
  indexed <- TabixFile(path)
  unlink(original_path)
  invisible(indexed)
}

### =========================================================================
### BAM support (wrappers around Rsamtools)
### -------------------------------------------------------------------------

setMethod("import", "BamFile",
          function(con, format, text, paired = FALSE, use.names = FALSE,
                   param = ScanBamParam(...), genome = NA_character_, ...)
          {
            if (!missing(format))
                checkArgFormat(con, format)
            stopifnot(isSingleStringOrNA(genome) || is(genome, "Seqinfo"))
            stopifnot(isTRUEorFALSE(paired))
            if (paired) {
                ans <- readGAlignmentPairs(con, use.names = use.names,
                                           param = param)
            } else {
                ans <- readGAlignments(con, use.names = use.names,
                                       param = param)
            }
            if (isSingleStringOrNA(genome)) {
                genome <- Seqinfo(genome=genome)
            }
            seqinfo(ans) <- merge(seqinfo(ans), genome)
            ans
          })

fillColumn <- function(x, filler) {
    if (is.null(x))
        filler
    else if (anyNA(x))
        ifelse(is.na(x), filler, x)
    else x
}

setMethod("export", c("GAlignments", "BamFile"),
          function(object, con, format, index = TRUE) {
            sam_path <- paste(file_path_sans_ext(path(con)), ".sam", sep = "")
            sam_con <- file(sam_path, "w")
            on.exit(close(sam_con))
            si <- seqinfo(object)
            has_info <-
              seqlevels(si)[!is.na(seqlevels(si)) & !is.na(seqlengths(si))]
            si <- si[has_info]
            if (length(si)) {
              header <- paste0("@SQ",
                               "\tSN:", seqlevels(si),
                               "\tLN:", seqlengths(si))
              has_genome <- !is.na(genome(si))
              header[has_genome] <-  paste0(header[has_genome], "\tAS:",
                                            genome(si)[has_genome])
              writeLines(header, sam_con)
            }
            emd <- mcols(object)
            aln <- paste(fillColumn(names(object), "*"),
                         fillColumn(emd[["flag"]],
                                    ifelse(strand(object) == "-", "16", "0")),
                         seqnames(object), start(object),
                         fillColumn(emd[["mapq"]], "255"),
                         cigar(object),
                         fillColumn(emd[["mrnm"]], "*"),
                         fillColumn(emd[["mpos"]], "0"),
                         fillColumn(emd[["isize"]], "0"),
                         if (is(object, "GappedReads")) object@qseq
                         else fillColumn(emd[["seq"]], "*"),
                         fillColumn(emd[["qual"]], "*"),
                         sep = "\t")
            custom <- emd[nchar(names(emd)) == 2L]
            if (length(custom) > 0L) {
              type.map <- c(integer = "i", numeric = "f", character = "Z",
                            factor = "Z")
              custom.class <- sapply(custom, class)
              custom.type <- type.map[custom.class]
              unknown.class <- custom.class[is.na(custom.type)]
              if (length(unknown.class) > 0L) {
                warning("these classes are not yet valid for BAM tag export: ",
                        paste(unknown.class, collapse=", "))
                custom <- custom[!is.na(custom.type)]
              }
              tags <- mapply(paste0, names(custom), ":", custom.type, ":",
                             as.list(custom), SIMPLIFY=FALSE)
              aln <- do.call(paste, c(list(aln), tags, sep = "\t"))
            }
            writeLines(aln, sam_con)
            close(sam_con)
            on.exit()
            bam <- asBam(sam_path, file_path_sans_ext(sam_path),
                         overwrite = TRUE, indexDestination = index)
            unlink(sam_path)
            invisible(bam)
          })

setMethod("export", c("GAlignmentPairs", "BamFile"),
          function(object, con, format, ...) {
            ga <- as(object, "GAlignments")
            collate_subscript <-
              S4Vectors:::make_XYZxyz_to_XxYyZz_subscript(length(object))
            getMateAttribute <- function(FUN) {
              c(FUN(last(object)), FUN(first(object)))[collate_subscript]
            }
            if (is.null(mcols(ga)$mrnm)) {
              mcols(ga)$mrnm <- getMateAttribute(seqnames)
            }
            if (is.null(mcols(ga)$mpos)) {
              mcols(ga)$mpos <- getMateAttribute(start)
            }
### FIXME: we cannot infer whether the pair is 'proper' (0x2)
### nor whether the alignment is 'primary' (0x100)
            if (is.null(mcols(ga)$flag)) {
              mcols(ga)$flag <- 0x1 + # all paired
                ifelse(strand(ga) == "-", 0x10, 0) +
                  ifelse(getMateAttribute(strand) == "-", 0x20, 0) +
                    c(0x40, 0x80) # left vs. right
            }
            if (is.null(names(ga))) {
              names(ga) <- as.character(rep(seq_len(length(object)), each=2L))
            }
            export(ga, con, ...)
          })

setMethod("export", c("ANY", "BamFile"),
          function(object, con, format, ...) {
            export(as(object, "GAlignments"), con, ...)
          })

test_bb <- function() {
  if (.Platform$OS.type == "windows")
    return()

  test_path <- system.file("tests", package = "rtracklayer")
  test_bb <- file.path(test_path, "test.bb")
  start <- c(237640, 521500 ,565725, 565900, 566760,
             119905, 122525, 173925, 179865, 180185)
  start <- start + 1L
  ir <- IRanges(start, width = 150)
  space <- factor(c(rep("chr1", 5), rep("chr10", 5)))
  name <- rep(".", 10)
  score <- seq.int(70L, 700L, length = 10)
  signalValue <- seq(10, 100, length = 10)
  peak <- rep(-1L, 10)
  correct_fixed <- GRanges(space, ir, name = name, score = score,
                           signalValue = signalValue , peak = peak)
  si <- SeqinfoForBSGenome("hg19")
  seqlengths(correct_fixed) <- seqlengths(si)[levels(space)]

  ## TEST: import whole file
  test <- import(test_bb)
  checkIdentical(test, correct_fixed)

  ## TEST: 'which'
  which <- GRanges(c("chr10"), IRanges(c(180185, 180335)))
  correct_which <- subsetByOverlaps(correct_fixed, which)
  test <- import(test_bb, which = which)
  checkIdentical(test, correct_which)

  ## TEST: empty which
  which <- GRanges()
  correct_which <- subsetByOverlaps(correct_fixed, which)
  test <- import(test_bb, which = which)
  checkIdentical(test, correct_which)

  ## TEST: BigBedSelection (GRanges, no field)
  which <- GRanges(c("chr10"), IRanges(173925, width = 151))
  test <- import(test_bb,
                 selection = BigBedSelection(which, colnames = character()))
  correct_subset <- subsetByOverlaps(correct_fixed, which)
  correct_which <- correct_subset[, character()]
  checkIdentical(test, correct_which)

  ## TEST: BigBedSelection (GRanges, 1 default field)
  test <- import(test_bb,
                 selection = BigBedSelection(which, colnames = c("name")))
  correct_which <- correct_subset[, c("name")]
  checkIdentical(test, correct_which)

  ## TEST: BigBedSelection (GRanges, 1 extra field)
  test <- import(test_bb,
                 selection = BigBedSelection(which, colnames = c("peak")))
  correct_which <- correct_subset[, c("peak")]
  checkIdentical(test, correct_which)

  ## TEST: BigBedSelection (GRanges, 1 default field and 1 extra field)
  colnames <- c("name", "peak")
  test <- import(test_bb,
                 selection = BigBedSelection(which, colnames =colnames))
  correct_which <- correct_subset[, colnames]
  checkIdentical(test, correct_which)

  # TEST: export
  test_bb_out <- file.path(tempdir(), "test_out.bb")
  export(correct_fixed, test_bb_out)
  on.exit(unlink(test_bb_out))
  test <- import(test_bb_out)
  checkIdentical(test, correct_fixed)
}

### =========================================================================
### Utilities
### -------------------------------------------------------------------------

pasteCollapse <- function(x, collapse = ",") {
  if (!is(x, "CharacterList"))
    stop("'x' must be a CharacterList")
  if (!isSingleString(collapse) || nchar(collapse) != 1L)
    stop("'collapse' must be a single string, with a single character")
  x <- as.list(x)
  .Call(CharacterList_pasteCollapse, x, collapse)
}

test_ucsc <- function(x) {

    ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ### TEST ucscTableQuery Class
    ###

    genome <- "hg38"
    table_name <- "gold"
    full_range <- as(Seqinfo(genome = genome), "GRanges")
    custom_range <- GRangesForUCSCGenome(genome, "chr1", IRanges(67003232, 67132477))
    selected_table <- data.frame(bin = 963, chrom = "chr5", chromStart = 49656261,
                                 chromEnd = 49661871, ix = 804, type ="W", frag="ABBA01004242.1",
                                 fragStart = 0, fragEnd = 5610, strand = "+")

    # creating a track and a table for UCSCSession and genome identifier
    elementMetadata  <- list(bin = 0, ix = 1060, type = "F", frag = "AL133320.8",
                             fragStart = 2000, fragEnd = 131245)
    track <- GRanges("chr1", IRanges(67003232, 67132477), "+", elementMetadata)
    genome(track) <- genome
    table <- data.frame(bin = 0, chrom = "chr1", chromStart = 67003232, chromEnd = 67132477,
                        ix = 1060, type = "F", frag = "AL133320.8", fragStart = 2000, 
                        fragEnd = 131245,strand = "+")


    test_trackhub_path <- system.file("tests", "trackhub", package = "rtracklayer")
    trackhub_genome <- "hg19"
    trackhub_custom_range <- GRangesForUCSCGenome("hg19", "chr1", IRanges(237640, 237791))
    trackhub_full_range <- as(Seqinfo(genome = trackhub_genome), "GRanges")
    trackhub_table_name <- "wgEncodeUWDukeDnaseGM12878FdrPeaks"

    # creating a track for Track Hub
    start <- c(237640, 521500 ,565725, 565900, 566760,
               119905, 122525, 173925, 179865, 180185)
    start <- start + 1L
    ir <- IRanges(start, width = 150)
    space <- factor(c(rep("chr1", 5), rep("chr10", 5)))
    name <- rep(".", 10)
    score <- seq.int(70L, 700L, length = 10)
    signalValue <- seq(10, 100, length = 10)
    peak <- rep(-1L, 10)
    trackhub_track <- GRanges(space, ir, name = name, score = score,
                              signalValue = signalValue , peak = peak)
    si <- Seqinfo(genome = "hg19")
    seqlengths(trackhub_track) <- seqlengths(si)[levels(space)]

    ## TEST: ucscTableQuery with UCSCSession with NAMES selection
    session <- browserSession()
    genome(session) <- "hg38"
    query <- ucscTableQuery(session, table = table_name, names = "ABBA01004242.1")
    checkIdentical(range(query), full_range)
    checkIdentical(getTable(query), selected_table)

    ## TEST: ucscTableQuery with UCSCSession without any custom range selection
    query <- ucscTableQuery(session, table = table_name)
    checkIdentical(range(query), full_range)

    ## TEST: ucscTableQuery with UCSCSession with custom range selection
    query <- ucscTableQuery(session, table = table_name, range = custom_range)
    checkIdentical(range(query), custom_range)

    ## TEST: ucscTableQuery with genome idenetifer(character) without any custom range selection
    query <- ucscTableQuery(genome, table = table_name)
    checkIdentical(genome(query), genome)
    checkIdentical(tableName(query), table_name)
    checkIdentical(range(query), full_range)

    ## TEST: ucscTableQuery with genome idenetifer(character) with custom range selection
    query <- ucscTableQuery(genome, table = table_name, range = custom_range)
    checkIdentical(genome(query), genome)
    checkIdentical(tableName(query), table_name)
    checkIdentical(range(query), custom_range)
    checkIdentical(track(query), track)
    checkIdentical(getTable(query), table)

    if (.Platform$OS.type == "windows")
        return()

    ## TEST: ucscTableQuery with Track Hub(character) without any custom range selection
    query <- ucscTableQuery(test_trackhub_path, table = trackhub_table_name,
                            genome = trackhub_genome)
    checkIdentical(genome(query), trackhub_genome)
    checkIdentical(tableName(query), trackhub_table_name)
    checkIdentical(range(query), trackhub_full_range)
    checkIdentical(track(query), trackhub_track)
    checkIdentical(getTable(query), as.data.frame(trackhub_track))

    ## TEST: ucscTableQuery with Track Hub(character) with custom range selection
    query <- ucscTableQuery(test_trackhub_path, table = trackhub_table_name,
                            genome = trackhub_genome, range = trackhub_custom_range)
    checkIdentical(genome(query), trackhub_genome)
    checkIdentical(tableName(query), trackhub_table_name)
    checkIdentical(range(query), trackhub_custom_range)
    checkIdentical(track(query), trackhub_track[1])
    checkIdentical(getTable(query), as.data.frame(trackhub_track)[1,])




    # TEST: UCSCSession gets initialize
    ucscsession <- new("UCSCSession")
    checkTrue(is.character(ucscsession@hguid) && length(ucscsession@hguid) != 0L)

    # TEST ucscCart returns details of the activeView
    ucscCart <- rtracklayer:::ucscCart(new("UCSCSession"))
    checkTrue(length(ucscCart) != 0L)

    # TEST ucscTracks retrieves the modes and ids
    tracks <- rtracklayer:::ucscTracks(new("UCSCSession"))
    checkIdentical(length(tracks@ids), length(tracks@modes))
    checkTrue(length(tracks@ids) != 0L)

    # TEST browserView with browse = F
    # caveat:we have to test manually whether with `browse = T`
    # UCSC sets the correct view onto the browser.
    bView <- browserView(new("UCSCSession"), browse = FALSE)
    checkTrue(!is.null(bView@session@views$instances))
    checkTrue(is.character(bView@hgsid) && length(bView@hgsid) != 0L)

    # TEST : range(UCSCSession)<-
    session <- new("UCSCSession")
    range <- GRanges("chr7:200000-250000")
    range(session) <- range
    returned <- range(session)
    checkIdentical(as.character(seqnames(range)), as.character(seqnames(returned)))
    checkIdentical(start(range), start(returned))
    checkIdentical(end(range), end(returned))
    checkIdentical(strand(range), strand(returned))
    checkIdentical(elementMetadata(range), elementMetadata(returned))

    # TEST : genome(UCSCSession)<-
    session <- new("UCSCSession")
    genome(session) <- "hg18"
    checkIdentical(genome(session), "hg18")

    # TEST : track(GRanges)<-
    session <- new("UCSCSession")
    track(session, "test_track") <- GRanges("chr7:1-50")
    checkTrue("test_track" %in% names(trackNames(session)))
}

### =========================================================================
### WIG (Wiggle) support (fixedStep and variableStep, legacy bedGraph)
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Classes
###

setClass("WIGFile", contains = "BiocFile")

WIGFile <- function(resource) {
  new("WIGFile", resource = resource)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export
###

setGeneric("export.wig",
           function(object, con, ...) standardGeneric("export.wig"))

setMethod("export.wig", "ANY",
          function(object, con, ...)
          {
            export(object, con, "wig", ...)
          })

setMethod("export", c("ANY", "WIGFile"),
          function(object, con, ...)
          {
            cl <- class(object)
            track <- try(as(object, "GRanges"), silent = TRUE)
            if (class(track) == "try-error") {
              track <- try(as(object, "SimpleGRangesList"), silent = TRUE)
              if (is(track, "try-error"))
                stop("cannot export object of class '", cl, "': ", track)
            }
            export(track, con, ...)
          })

.wigWriter <- function(chromData, con, dataFormat, append) {
  m <- manager()
  con <- connection(m, con, if (append) "a" else "w")
  on.exit(release(m, con))
  cat(dataFormat, file = con)
  cat(" chrom=", as.character(seqnames(chromData)[1]), file = con, sep = "")
  data <- score(chromData)
  starts <- start(chromData)
  if (dataFormat == "variableStep")
    data <- cbind(starts, data)
  else {
    cat(" start=", starts[1], file = con, sep = "")
    step <- if (length(starts) == 1) 0 else starts[2] - starts[1]
    cat(" step=", step, file = con, sep = "")
  }  
  cat(" span=", width(chromData)[1], file = con, sep = "")
  cat("\n", file = con, sep = "")
  write.table(data, con, sep = "\t", col.names = FALSE,
              row.names = FALSE, quote = FALSE)
}

isValidScore <- function(score) {
    !(is.numeric(score) || is(score, "Rle") && is.numeric(runValue(score))) ||
        any(is.na(score))
}

setMethod("export", c("GenomicRanges", "WIGFile"),
          function(object, con, format,
                   dataFormat = c("auto", "variableStep", "fixedStep"),
                   writer = .wigWriter, append = FALSE, ...)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            if (isValidScore(score(object)))
              stop("The score must be numeric, without any NA's")
            scipen <- getOption("scipen")
            options(scipen = 100) # prevent use of scientific notation
            on.exit(options(scipen = scipen))
            dataFormat <- match.arg(dataFormat)            
            doBlock <- function(chromData) {
              if (length(chromData) == 0L)
                  return()
              if (is.unsorted(start(chromData)))
                chromData <- chromData[order(start(chromData)),]
              starts <- start(chromData)
              ends <- end(chromData)
              if (!all(tail(starts, -1) - head(ends, -1) > 0))
                stop("Features cannot overlap. ",
                     "Note that WIG does not distinguish between strands - ",
                     "try exporting two tracks, one for each strand.")
              if (length(starts) == 1)
                steps <- 0
              else steps <- diff(starts)
              fixedSpan <- is(object, "GPos") || {
                  spans <- ends - starts + 1L
                  all(spans[1L] == spans)
              }
              if (!fixedSpan)
                stop("The span must be uniform for Wiggle export. ",
                     "Consider exporting to bedGraph/bigWig, ",
                     "or coerce data to a GPos object first.")
              fixedStep <- all(steps[1] == steps)
              if (dataFormat == "auto") {
                dataFormat <- "variableStep"
                if (fixedStep)
                  dataFormat <- "fixedStep"
              }
              if (dataFormat != "variableStep" && !fixedStep)
                stop("Step not uniform: consider variableStep")
              ans <- writer(chromData, con, dataFormat, append)
              append <<- TRUE
              ans
            }
            dataFormat <- match.arg(dataFormat)
            lapply(split(object, seqnames(object)), doBlock)
            invisible(con)
          })

setMethod("export", c("UCSCData", "WIGFile"),
          function(object, con, format, trackLine = TRUE, ...)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            if (trackLine) {
              export.ucsc(object, con, ...)
            } else {
              callNextMethod()
            }
            invisible(con)
          })

setMethod("export", c("SimpleGRangesList", "WIGFile"),
          .export_SimpleGRangesList_BiocFile)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import
###

setGeneric("import.wig", function(con, ...) standardGeneric("import.wig"))

setMethod("import.wig", "ANY",
          function(con, ...)
          {
            import(con, "wig", ...)
          })

setMethod("import", "WIGFile",
          function(con, format, text, genome = NA,
                   trackLine = TRUE, which = NULL, seqinfo = NULL, ...)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            file <- con
            m <- manager()
            con <- connection(m, con, "r")
            on.exit(release(m, con))
            ## check for a track line
            line <- scanTrackLine(con)
            if (!is.null(line) && trackLine) {
              pushBack(line, con)
              return(import.ucsc(initialize(file, resource = con), drop = TRUE,
                                 trackLine = FALSE, genome = genome,
                                 which = which, seqinfo = seqinfo,
                                 ...))
            }
            lines <- readLines(con, warn = FALSE)
            formatInds <- grep("chrom=", lines, fixed=TRUE)
            formatLines <- lines[formatInds]
            starts <- formatInds + 1L
            ends <- c(tail(formatInds, -1) - 1L, length(lines))
            format <- gsub("^([^ ]*) .*", "\\1", formatLines)
            parsedFormat <- lapply(formatLines, ucscParsePairs)
            chrom <- sapply(parsedFormat, `[[`, "chrom")
            if (is.null(seqinfo) && !is.null(genome) && !is.na(genome))
              seqinfo <- seqinfoForGenome(genome)
            if (!is.null(seqinfo))
              seqlevels <- seqlevels(seqinfo)
            else seqlevels <- unique(chrom)
            if (length(formatLines)) {
              parseData <- function(i) {
                ## parse format line
                formatVals <- parsedFormat[[i]]
                # parse the data values
                block_lines <- window(lines, starts[i], ends[i])
                if (!length(block_lines)) {
                  gr <- GRanges(score = numeric())
                  seqlevels(gr) <- seqlevels
                  return(gr)
                }
                block_lines <- grep("^#", block_lines, invert = TRUE,
                                    value = TRUE)
                if (format[i] == "variableStep") {
                  ## assume the same white space for every row
                  sep <- sub(".*?([[:space:]]+).*", "\\1", block_lines[1])
                  split_lines <- strsplit(block_lines, sep, fixed=TRUE)
                  mat <- matrix(as.numeric(unlist(split_lines)), nrow = 2)
                  start <- mat[1,]
                  score <- mat[2,]
                } else {
                  start <- seq(as.integer(formatVals["start"]),
                               by = as.integer(formatVals["step"]),
                               length.out = length(block_lines))
                  score <- as.numeric(block_lines)
                }
                span <- formatVals["span"]
                if (is.na(span))
                  span <- 1
                end <- start + as.integer(span) - 1
                gd <- GenomicData(IRanges(start, end), score = score,
                                  chrom = formatVals[["chrom"]],
                                  which = which)
                seqlevels(gd) <- seqlevels
                gd
              }
              parseInds <- seq_along(formatInds)
              if (!is.null(which)) {
                message("For efficiency, consider converting this WIG file",
                        " to a BigWig file;\nsee ?wigToBigWig")
                parseInds <- parseInds[chrom %in% seqlevels(which)]
              }
              resultList <- lapply(parseInds, parseData)
              gd <- do.call(c, resultList)
              if (!is.null(seqinfo))
                seqinfo(gd) <- seqinfo
              else if (!is.null(genome))
                genome(gd) <- genome
              gd
            } else {
              import(text = lines, format = "bedGraph",
                     genome = genome, which = which, seqinfo = seqinfo)
            }
        })

### =========================================================================
### Genome-oriented methods for GRanges and IntegerRangesList objects
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### GRanges convenience constructor
###

GenomicData <- function(ranges, ..., strand = NULL, chrom = NULL, genome = NA,
                        seqinfo = NULL, which = NULL, metadata = list())
{
  if (is.null(genome))
    genome <- NA
  if (!isSingleStringOrNA(genome))
    stop("'genome' must be a single string, or NULL, or NA")
  if (!is.null(seqinfo) && !all(genome == genome(seqinfo), na.rm=TRUE)) {
      stop("'genome' ", genome, "' does not match that in 'seqinfo'")
  }
  if (length(seqinfo) == 0L)
    seqinfo <- seqinfoForGenome(genome)
  if (length(chrom) > length(ranges))
    stop("length of 'chrom' greater than length of 'ranges'")
  if (length(chrom) > 0 && (length(ranges) %% length(chrom) != 0))
    stop("length of 'ranges' not a multiple of 'chrom' length")
  normStrand <- function(strand) {
    strand <- as.character(strand)
    strand[is.na(strand)] <- "*"
    strand(strand)
  }
  if (!(is.null(strand) || is(strand, "Rle")))
    strand <- normStrand(strand)
  if (is.null(chrom))
    chrom <- droplevels(Rle(factor("1"), length(ranges)))
  dots <- list(...)
  if (length(dots) == 1) {
    dots <- dots[[1L]]
    if ((is(dots, "data.frame") || is(dots, "DataFrame")) &&
        !is.null(dots[["strand"]])) {
      strand <- dots[["strand"]]
      dots[["strand"]] <- NULL
      return(GenomicData(ranges = ranges, dots, strand = strand,
                         chrom = chrom, genome = genome,
                         seqinfo = seqinfo, which = which,
                         metadata = metadata))
    }
  }
  if (is.null(strand))
    strand <- Rle("*", length(ranges))
  df <- DataFrame(...)
  invalidNames <- names(df) %in% GenomicRanges:::INVALID.GR.COLNAMES
  names(df)[invalidNames] <- paste0(".", names(df)[invalidNames])
  gd <- GRanges(seqnames = chrom, ranges = ranges, strand = strand, df,
                seqinfo = seqinfo)
  if (!is.na(genome))
    genome(gd) <- genome
  if (!is.null(which)) {
    if (is(which, "IntegerRangesList"))
      which <- as(which, "GRanges")
    gd <- subsetByOverlaps(gd, which)
  }
  metadata(gd) <- metadata
  gd
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Automatic seqinfo lookup
###

seqinfoForGenome <- function(genome, method = c("auto", "BSgenome", "UCSC")) {
  method <- match.arg(method)
  if (is.na(genome))
    return(NULL)
  if (method == "auto" || method == "BSgenome")
    sl <- SeqinfoForBSGenome(genome)
  if (method == "UCSC" || (method == "auto" && is.null(sl)))
    sl <- SeqinfoForUCSCGenome(genome)
  sl
}

BSGenomeForID <- function(genome) {
  if (!suppressWarnings(requireNamespace("BSgenome", quietly=TRUE)))
    return(NULL)
  bsgenome <- try(BSgenome::getBSgenome(genome), silent=TRUE)
  if (inherits(bsgenome, "try-error"))
    return(NULL)
  bsgenome
}

SeqinfoForBSGenome <- function(genome) {
  bsgenome <- BSGenomeForID(genome)
  if (!is.null(bsgenome))
    seqinfo(bsgenome)
  else NULL
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Convience constructor for GRanges over an entire genome
###

GRangesForGenome <- function(genome, chrom = NULL, ranges = NULL, ...,
                             method = c("auto", "BSgenome", "UCSC"),
                             seqinfo = NULL)
{
  if (missing(genome) || !isSingleString(genome))
    stop("'genome' must be a single string identifying a genome")
  if (is.null(seqinfo))
    seqinfo <- seqinfoForGenome(genome, match.arg(method))
  if (is.null(seqinfo))
    stop("Failed to obtain information for genome '", genome, "'")
  if (!is.null(ranges) && !is(ranges, "IntegerRanges"))
    stop("'ranges' must be NULL or an IntegerRanges object")
  if (is.null(chrom))
    chrom <- seqnames(seqinfo)
  else {
    badChrom <- setdiff(chrom, seqnames(seqinfo))
    if (length(badChrom))
      stop("Chromosome(s) ", paste(badChrom, collapse = ", "),
           "are invalid for: ", genome)
  }
  if (is.null(ranges))
    ranges <- IRanges(1L, seqlengths(seqinfo)[chrom])
  gr <- GRanges(chrom, ranges, seqlengths = seqlengths(seqinfo), ...)
  seqinfo(gr) <- seqinfo
  gr
}

GRangesForBSGenome <- function(genome, chrom = NULL, ranges = NULL, ...)
{
  GRangesForGenome(genome, chrom = chrom, ranges = ranges, method = "BSgenome",
                   seqinfo = NULL, ...)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### score: for internal convenience
###

setMethod("score", "ANY", function(x) NULL)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### chrom(): Returns chromosome name vector of length 'length(x)'.
###          Not to be confused with 'seqnames', which returns a List for
###          IntegerRangesList.
###          More or less a pre-GenomicRanges relic.
###

setGeneric("chrom", function(x, ...) standardGeneric("chrom"))
setMethod("chrom", "GRanges", function(x) seqnames(x))
setMethod("chrom", "IntegerRangesList", function(x) {
  names(x)
})

setGeneric("chrom<-", function(x, ..., value) standardGeneric("chrom<-"))
setReplaceMethod("chrom", "GRanges", function(x, value) {
  seqnames(x) <- value
  x
})

### =========================================================================
### Genome-oriented conveniences for RangedSelection classes
### -------------------------------------------------------------------------

GenomicSelection <- function(genome, chrom = NULL, colnames = character(0))
{
  if (missing(genome) || !isSingleString(genome))
    stop("'genome' must be a single string identifying a genome")
  si <- seqinfoForGenome(genome)
  if (is.null(si))
    stop("Failed to obtain information for genome '", genome, "'")
  lens <- seqlengths(si)
  if (is.null(chrom))
    chrom <- names(lens)
  else {
    if (!is.character(chrom))
      stop("'chrom' must be NULL or a character vector")
    invalidChroms <- setdiff(chrom, names(lens))
    if (length(invalidChroms))
      stop("'chrom' contains invalid chromosomes: ",
           paste(invalidChroms, collapse = ", "))
    lens <- lens[chrom]
  }
  RangedSelection(split(IRanges(1, lens), factor(chrom, chrom)), colnames)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Utilities
###

## Individual sessions/views only support a single genome at a time,
## whereas range data structures can have multiple genomes. We try to
## rectify this here.
singleGenome <- function(x) {
  x1 <- unname(x[1])
  if (!all(is.na(x)) && (any(is.na(x)) || any(x != x1)))
      stop("Multiple genomes encountered; only one supported")
  x1
}

## normalize 'range', using 'seqinfo' for default genome
normGenomeRange <- function(range, seqinfo, max.length = 1L) {
  ## the user can specify a portion of the genome in several ways:
  ## - String identifying a genome
  ## - IntegerRangesList
  ## - GRanges, the preferred way, possibly from GRangesForUCSCGenome()
  ## - We do not allow IntegerRanges, since it does not make sense to have
  ##   one range over many chromosomes
  if (is.character(range)) {
    range <- singleGenome(range)
    seqinfo <- Seqinfo(genome = range)
    return(GRangesForGenome(range, seqinfo = seqinfo))
  }
  genome <- genome(seqinfo)[1]
  if (is(range, "Seqinfo"))
    range <- as(range, "GRanges")
  if (!is(range, "IntegerRangesList") && !is(range, "GenomicRanges"))
    stop("'range' should be a genome string, IntegerRangesList, GRanges or Seqinfo")

  if (length(seqinfo(range)) == 0L) {
    ## hack: need to avoid calling seqlengths(session) here, so use 'foo'
    seqinfo(range) <- Seqinfo("foo", genome = genome)
  } else {
    rangeGenome <- singleGenome(genome(range))
    if (!is.na(rangeGenome) && rangeGenome != genome) {
      genome <- rangeGenome
    }
    seqinfo <- si <- Seqinfo(genome = genome)
    seqinfo(range, new2old = match(seqlevels(si), seqlevels(range))) <-
      merge(si, seqinfo(range))
  }
  if (is(range, "IntegerRangesList")) {
    range <- GRangesForGenome(singleGenome(genome(range)), names(range),
                              unlist(range), seqinfo = seqinfo)
  } else if (is(range, "GenomicRanges")) {
    strand(range) <- "*"
    mcols(range) <- NULL
  }
  if (length(range) > max.length) {
    warning("number of ranges (", length(range),
            ") exceeds limit of ", max.length)
  }
  if (max.length == 1L) {
    if (length(unique(seqnames(range))) > max.length)
      stop("'range' must contain ranges on a single chromosome")
    range(range)
  } else {
    head(range, max.length)
  }
}

spansGenome <- function(x) {
  strand(x) <- "*"
  x <- reduce(x)
  w <- structure(width(x), names = as.character(seqnames(x)))
  identical(w[names(seqlengths(x))], seqlengths(x))
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Questionable utilities
###

# replace fields in 'x' with those in 'y', if given
## mergeRange <- function(x, y)
## {
##   if (length(genome(y)))
##     genome(x) <- genome(y)
##   if (length(y))
##     x[[1]] <- y[[1]]
##   if (length(names(y)))
##     names(x)[1] <- names(y)[1]
##   x
## }

## take 'y', filling in info from 'x' if necessary
mergeRange <- function(x, y)
{
  if (!length(genome(y))) {
    genome(y) <- genome(x)
    if (!length(names(y)) && length(y) == length(x)) {
      names(y) <- names(x)
      if (!length(unlist(y)) && length(x))
        y[[1]] <- x[[1]]
    }
  }
  if (!length(genome(y)))
    stop("Genome must be specified")
  if (length(unlist(y)) && !length(names(y)))
    stop("Chromosome must be specified if an interval is specified")
  y
}

### =========================================================================
### TrackHub support
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### TrackContainer class
###

setClass("TrackContainer",
         representation("SimpleList"),
         prototype(elementType = "Track")
)

setMethod("names", "TrackContainer", function(x) {
    vapply(x, function(y) y@track, character(1L) ,USE.NAMES = FALSE)
})

TrackContainer <- function(...) {
    args <- list(...)
    if (length(args) == 1 && is.list(args[[1L]]))
        args <- args[[1L]]
    if (!all(vapply(args, is, logical(1L), "Track")))
        stop("all elements in '...' must be Track objects")
    S4Vectors:::new_SimpleList_from_list("TrackContainer", args)
}

setClass("Track",
         representation(
                        # common trackDb settings
                        track = "character",
                        type = "character",
                        shortLabel = "character",
                        longLabel = "character",
                        bigDataUrl = "character",
                        html = "character",
                        visibility = "character",
                        meta = "character",


                        # common optional settings
                        color = "character",
                        priority = "numeric",
                        altColor = "character",
                        boxedCfg = "logical",
                        chromosomes = "character",
                        darkerLabels = "logical",
                        dataVersion = "character",
                        directUrl = "character",
                        iframeUrl = "character",
                        iframeOptions = "character",
                        mouseOverField = "character",
                        otherDb = "character",
                        pennantIcon = "character",
                        tableBrowser = "character",
                        url = "character",
                        urlLabel = "character",
                        urls = "character",
                        skipEmptyFields = "logical",
                        skipFields = "character",
                        sepFields = "character",


                        ##settings by track type

                        # bam settings
                        refUrl = "character",
                        bigDataIndex = "character",
                        bamColorMode = "character",
                        bamGrayMode = "character",
                        aliQualRange = "character",
                        baseQualRange = "character",
                        bamColorTag = "character",
                        noColorTag = "character",
                        bamSkipPrintQualScore = "character",
                        indelDoubleInsert = "logical",
                        indelQueryInsert = "logical",
                        indelPolyA = "logical",
                        minAliQual = "character",
                        pairEndsByName = "character",
                        pairSearchRange = "character",
                        showNames = "logical",
                        doWiggle = "logical",
                        maxWindowToDraw = "integer",

                        # bigBarChart settings
                        barChartBars = "character",
                        barChartColor = "character",
                        barChartLabel = "character",
                        barChartMaxSize = "character",
                        barChartSizeWindows = "character",
                        barChartMetric = "character",
                        barChartUnit = "character",
                        barChartMatrixUrl = "character",
                        barChartSampleUrl = "character",
                        maxLimit = "character",
                        labelFields = "character",
                        defaultLabelFields = "character",
                        itemRgb = "logical",
                        colorByStrand = "character",
                        denseCoverage = "integer",
                        labelOnFeature = "logical",
                        exonArrows = "logical",
                        exonNumbers = "logical",
                        scoreFilter = "character",
                        scoreFilterLimits = "character",
                        maxItems = "integer",
                        minGrayLevel = "character",
                        noScoreFilter = "logical",
                        spectrum = "logical",
                        scoreMax = "integer",
                        scoreMin = "integer",
                        thickDrawItem = "logical",
                        searchIndex = "character",
                        searchTrix = "character",
                        labelSeparator = "character",
                        # UNSUPPORTED fields
                        # filter.<fieldName>
                        # filterByRange.<fieldName>
                        # filterLimits.<fieldName>
                        # filterText.<fieldName>
                        # filterType.<fieldName>
                        # filterValues.<fieldName>
                        # filterValuesDefault.<fieldName>
                        # filterType.<fieldName>
                        # filterLabel.<fieldName>
                        bedNameLabel = "character",
                        exonArrowsDense = "logical",
                        itemImagePath = "character",
                        itemBigImagePath = "character",
                        mergeSpannedItems = "logical",
                        linkIdInName = "logical",
                        nextExonText = "character",
                        prevExonText = "character",
                        scoreLabel = "character",
                        showTopScorers = "character",

                        # bigChain settings
                        linkDataUrl = "character",

                        # bigInteract settings
                        interactDirectional = "character",
                        interactUp = "character",
                        interactMultiRegion = "character",
                        maxHeightPixels = "character",
                        speciesOrder = "character",
                        frames = "character",
                        summary = "character",

                        # bigNarrowPeak settings
                        # UNSUPPORTED fields
                        #scoreFilter
                        #pValueFilter
                        #qValueFilter
                        #signalFilter
                        #<column>FilterLimits
                        #<column>FilterByRange

                        # bigPsl settings
                        baseColorUseCds = "character",
                        baseColorUseSequence = "character",
                        baseColorDefault = "character",
                        showDiffBasesAllScales = "logical",

                        # bigWig settings
                        autoscale = "character",
                        autoScale = "character",
                        viewLimits = "character",
                        viewLimitsMax = "character",
                        alwaysZero = "logical",
                        graphTypeDefault = "character",
                        maxWindowToQuery = "integer",
                        negateValues = "logical",
                        smoothingWindow = "character",
                        transformFunc = "character",
                        windowingFunction = "character",
                        yLineMark = "character",
                        yLineOnOff = "logical",
                        gridDefault = "logical",


                        # hic settings
                        showSnpWidth = "integer",
                        otherSpecies = "character",


                        # vcfTabix settings
                        minQual = "character",
                        minFreq = "character",
                        hapClusterEnabled = "character",
                        hapClusterColorBy = "character",
                        hapClusterTreeAngle = "character",
                        hapClusterHeight = "character",
                        applyMinQual = "character",


                        ##Grouping tracks into sets and hierarchies

                        # Supertrack
                        superTrack = "character",
                        parent = "character",

                        # Composite Tracks
                        compositeTrack = "logical",
                        allButtonPair = "logical",
                        centerLabelsDense = "logical",
                        dragAndDrop = "character",
                        hideEmptySubtracks = "logical",
                        hideEmptySubtracksMultiBedUrl = "character",
                        hideEmptySubtracksSourcesUrl = "character",
                        hideEmptySubtracksLabel = "character",


                        # Subgroups
                        subGroup1 = "character",
                        subGroup2 = "character",
                        subGroup3 = "character",
                        subGroup4 = "character",
                        subGroup5 = "character",
                        subGroup6 = "character",
                        subGroup7 = "character",
                        subGroup8 = "character",
                        subGroup9 = "character",
                        subGroups = "character",
                        dimensions = "character",
                        filterComposite = "character",
                        dimensionAchecked = "character",
                        dimensionBchecked = "character",
                        sortOrder = "character",

                        # Subgroups Views
                        view = "character",
                        viewUi = "logical",
                        configurable = "logical",

                        # multiWig
                        container = "character",
                        aggregate = "character",
                        showSubtrackColorOnUi = "logical",

                        # Miscellaneous Deprecated Settings
                        metadata = "character",
                        noInherit = "logical",
                        useScore = "integer"
                        )
)

Track <- function(...) {
    new("Track", ...)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Genome class
###

setClass("Genome",
         representation(
                        genome = "character",
                        trackDb = "character",
                        metaDb = "character",
                        metaTab = "character",
                        twoBitPath = "character",
                        groups = "character",
                        description = "character",
                        organism = "character",
                        defaultPos = "character",
                        orderKey = "character",
                        htmlPath = "character"
         ),
         prototype(
                   genome = NA_character_,
                   trackDb = NA_character_,
                   metaDb = NA_character_,
                   metaTab = NA_character_,
                   twoBitPath = NA_character_,
                   groups = NA_character_,
                   description = NA_character_,
                   organism = NA_character_,
                   defaultPos = NA_character_,
                   orderKey = NA_character_,
                   htmlPath = NA_character_
         )
)

Genome <- function(...) {
        new("Genome", ...)
}

stopIfNotGenome <- function(x) {
    if (!is(x, "Genome"))
        stop("value must be Genome object")
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### GenomeContainer class
###

setClass("GenomeContainer",
         representation("SimpleList"),
         prototype(elementType = "Genome")
)

setMethod("names", "GenomeContainer", function(x) {
    vapply(x, function(y) y@genome, character(1L), USE.NAMES = FALSE)
})

setMethod("getListElement", "GenomeContainer", function(x, i, exact = TRUE) {
    genome <- x[names(x) == i]
    if (length(genome) == 1L) unlist(genome)[[1L]]
})

GenomeContainer <- function(...) {
    args <- list(...)
    if (length(args) == 1 && is.list(args[[1L]]))
        args <- args[[1L]]
    if (!all(vapply(args, is, logical(1L), "Genome")))
        stop("all elements in '...' must be Genome objects")
    S4Vectors:::new_SimpleList_from_list("GenomeContainer", args)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### TrackHub class
###

setGeneric("hub", function(x) standardGeneric("hub"))
setGeneric("hub<-", function(x, value) standardGeneric("hub<-"))
setGeneric("shortLabel", function(x) standardGeneric("shortLabel"))
setGeneric("shortLabel<-", function(x, value) standardGeneric("shortLabel<-"))
setGeneric("longLabel", function(x) standardGeneric("longLabel"))
setGeneric("longLabel<-", function(x, value) standardGeneric("longLabel<-"))
setGeneric("genomesFile", function(x) standardGeneric("genomesFile"))
setGeneric("genomesFile<-", function(x, value) standardGeneric("genomesFile<-"))
setGeneric("email", function(x) standardGeneric("email"))
setGeneric("email<-", function(x, value) standardGeneric("email<-"))
setGeneric("descriptionUrl", function(x) standardGeneric("descriptionUrl"))
setGeneric("descriptionUrl<-", function(x, value) standardGeneric("descriptionUrl<-"))
setGeneric("writeTrackHub", function(x) standardGeneric("writeTrackHub"))
setGeneric("genomeField", function(x, name, key) standardGeneric("genomeField"))
setGeneric("genomeField<-", function(x, name, key, value) standardGeneric("genomeField<-"))
setGeneric("genomeInfo", function(x, name) standardGeneric("genomeInfo"))
setGeneric("genomeInfo<-", function(x, value) standardGeneric("genomeInfo<-"))

setClass("TrackHub",
         representation(
                        uri = "character",
                        hub = "character",
                        shortLabel = "character",
                        longLabel = "character",
                        genomesFile = "character",
                        email = "character",
                        descriptionUrl = "character",
                        genomeContainer = "GenomeContainer"),
         prototype(
                   hub = NA_character_,
                   shortLabel = NA_character_,
                   longLabel = NA_character_,
                   genomesFile = NA_character_,
                   email = NA_character_,
                   descriptionUrl = NA_character_),
         contains = "List")

hubFile <- function(x) paste(trimSlash(uri(x)), "hub.txt", sep = "/")

stopIfNotLocal <- function(x) {
    if (!uriIsWritable(x)) {
        stop("Repository is read only; cannot write on remote repository")
    }
}

getHubContent <- function(x) {
    content <- readLines(hubFile(x), warn = FALSE)
    rexp <- "^(\\w+)\\s?(.*)$"
    contentVec <- c(sub(rexp, "\\2", content))
    names(contentVec) <- sub(rexp, "\\1", content)
    x@hub <- contentVec["hub"]
    x@shortLabel <- contentVec["shortLabel"]
    x@longLabel <- contentVec["longLabel"]
    x@genomesFile <- contentVec["genomesFile"]
    x@email <- contentVec["email"]
    x@descriptionUrl <- contentVec["descriptionUrl"]
    x
}

setHubContent <- function(x) {
    file = hubFile(x)
    cat("", file = file)
    if (!is.na(x@hub))
        cat("hub ", x@hub, "\n", append = TRUE, sep = "", file = file)
    if (!is.na(x@shortLabel))
        cat("shortLabel ", x@shortLabel, "\n", append = TRUE, sep = "", file = file)
    if (!is.na(x@longLabel))
        cat("longLabel ", x@longLabel, "\n", append = TRUE, sep = "", file = file)
    if (!isFieldEmpty(x@genomesFile))
        cat("genomesFile ", x@genomesFile, "\n", append = TRUE, sep = "", file = file)
    if (!is.na(x@email))
        cat("email ", x@email, "\n", append = TRUE, sep = "", file = file)
    if (!is.na(x@descriptionUrl))
        cat("descriptionUrl ", x@descriptionUrl, "\n", append = TRUE, sep = "", file = file)
}

getGenomesContent <- function(x) {
    if (uriExists(hubFile(x))) {
        genomesFileValue <- x@genomesFile
        if (!isFieldEmpty(genomesFileValue)) {
            genomesFilePath <- combineURI(uri(x), unname(genomesFileValue))
            if (isFileEmpty(genomesFilePath)) {
                return(list())
            }
            content <- readLines(genomesFilePath, warn = FALSE)
            content_df <- read.csv(text = sub(" ", ",", content), header = FALSE)
            genomesIndex <- grep("\\bgenome\\b", content_df$V1)
            totalGenomes <- length(genomesIndex)
            genomesIndex[length(genomesIndex) + 1] <- length(content_df$V1) + 1
            genomes <- lapply(1:totalGenomes, function(x) {
                startPosition <- genomesIndex[x]
                endPosition <- genomesIndex[x + 1] - 1
                genome <- setNames(data.frame(content_df$V1[startPosition:endPosition],
                                    content_df$V2[startPosition:endPosition]),
                         c("field", "value"))
                genome <- setNames(as.list(genome$value), genome$field)
                genome <- do.call(Genome, genome)
            })
            genomes
        }
        else message("hub.txt: 'genomesFile' does not contain valid reference to genomes file")
    }
}

setGenomesContent <- function(x, genomeContainer) {
    genomesFilePath <- combineURI(uri(x), x@genomesFile)
    slots <- slotNames(Genome())
    genomesFields <- c("twoBitPath", "groups", "htmlPath", "metaDb", "trackDb" , "metaTab")
    genomes <- vapply(genomeContainer, function(y) {
        uri <- combineURI(uri(x), y@genome)
        if (!uriExists(uri))
            createResource(uri, dir = TRUE)
        genome <- vapply(slots, function(slotName) {
            slotValue <- slot(y, slotName)
            if (!isEmpty(slotValue) && !is.na(slotValue)) {
                if (slotName %in% genomesFields) {
                    filePath <- combineURI(uri(x), slotValue)
                    if (!uriExists(filePath)) {
                        createResource(filePath)
                    }
                }
                paste0(slotName, " ", slotValue)
            }
            else ""
        }, character(1L))
    }, character(11L))
    genomes <- genomes[genomes != ""]
    genomes <- gsub("\\bgenome\\b", "\ngenome", genomes)
    writeLines(genomes, genomesFilePath)
}

getGenome <- function(x, name) {
    genome <- x@genomeContainer[[name]]
    if (length(genome) == 1L) genome
    else if (length(genome) == 0L) stop("Genome '", name, "' does not exist")
    else if (length(genome) > 1L) stop("Multiple genomes match ", name)
}

setGenome <- function(x, name, value) {
    stopIfNotGenome(value)
    genome <- x@genomeContainer[[name]]
    if (length(genome) == 1L) genome <- value
    else if (length(genome) == 0L) stop("Genome '", name, "' does not exist")
    else if (length(genome) > 1L) stop("Multiple genomes match ", name)
    x@genomeContainer[[name]] <- genome
    x
}

setMethod("uri", "TrackHub", function(x) {
    x@uri
})

setMethod("hub", "TrackHub", function(x) {
    unname(x@hub)
})

setReplaceMethod("hub", "TrackHub", function(x, value) {
    x@hub <- value
    x
})

setMethod("shortLabel", "TrackHub", function(x) {
    unname(x@shortLabel)
})

setReplaceMethod("shortLabel", "TrackHub", function(x, value) {
    x@shortLabel <- value
    x
})

setMethod("longLabel", "TrackHub", function(x) {
    unname(x@longLabel)
})

setReplaceMethod("longLabel", "TrackHub", function(x, value) {
    x@longLabel <- value
    x
})

setMethod("genomesFile", "TrackHub", function(x) {
    unname(x@genomesFile)
})

setReplaceMethod("genomesFile", "TrackHub", function(x, value) {
    x@genomesFile <- value
    x
})

setMethod("email", "TrackHub", function(x) {
    unname(x@email)
})

setReplaceMethod("email", "TrackHub", function(x, value) {
    x@email <- value
    x
})

setMethod("descriptionUrl", "TrackHub", function(x) {
    unname(x@descriptionUrl)
})

setReplaceMethod("descriptionUrl", "TrackHub", function(x, value) {
    x@descriptionUrl <- value
    x
})

setMethod("genome", "TrackHub", function(x) {
    genomes <- x@genomeContainer
    names(genomes)
})

setMethod("getListElement", "TrackHub", function(x, i, exact = TRUE) {
    TrackHubGenome(x, i)
})

setMethod("names", "TrackHub", function(x) genome(x))

setMethod("length", "TrackHub", function(x) length(names(x)))

setMethod("genomeInfo", "TrackHub", function(x, name) {
    names <- names(x@genomeContainer)
    genome <- x@genomeContainer[[name]]
    if (length(genome) == 0L) stop("Genome '", name, "' does not exist")
    else genome
})

setReplaceMethod("genomeInfo", "TrackHub", function(x, value) {
    stopIfNotGenome(value)
    name <- value@genome
    names <- names(x@genomeContainer)
    genome <- x@genomeContainer[[name]]
    if (length(genome) == 1L) stop("NOTE: Genome '", name, "' already exists")
    else if (length(genome) > 1L) stop("Multiple genomes match ", name)
    else {
        if (!identical(value, Genome())) {
            if (length(x@genomeContainer) == 0L) genomes <- value
            else genomes <- c(unlist(x@genomeContainer), value)
            x@genomeContainer <- GenomeContainer(genomes)
        }
    }
    x
})

setMethod("genomeField", "TrackHub", function(x, name, key) {
    genome <- getGenome(x, name)
    slot(genome, key)
})

setReplaceMethod("genomeField", "TrackHub", function(x, name, key, value) {
    genome <- getGenome(x, name)
    slot(genome, key) <- value
    setGenome(x, name, genome)
})

setMethod("writeTrackHub", "TrackHub", function(x) {
    stopIfNotLocal(hubFile(x))
    setHubContent(x)
    genomesFilePath <- combineURI(uri(x), genomesFile(x))
    if (!uriExists(genomesFilePath) && !is.na(genomesFile(x)))
        createResource(genomesFilePath)
    if (uriExists(genomesFilePath))
        setGenomesContent(x, x@genomeContainer)
})

setMethod("show", "TrackHub", function(object) {
    cat(class(object), "repository\nuri:", uri(object), "\n")
    cat(S4Vectors:::labeledLine("genomes", genome(object)))
    cat("hub:", hub(object), "\n")
    cat("shortLabel:", shortLabel(object), "\n")
    cat("longLabel:", longLabel(object), "\n")
    cat("genomesFile:", genomesFile(object), "\n")
    cat("email:", email(object), "\n")
    cat("descriptionUrl:", descriptionUrl(object), "\n")
})

TrackHub <- function(uri, create = FALSE) {
    if (!isTRUEorFALSE(create))
        stop("'create' must be TRUE or FALSE")
    if (create) {
        if (uriExists(uri)) {
            message("NOTE: '", uri, "' already exists")
            create <- FALSE
        }
        else createResource(uri, dir = TRUE)
    }
    th <- new("TrackHub")
    th@uri <- checkURI(uri)
    if (create && !uriExists(hubFile(th))) {
        createResource(hubFile(th))
    } else {
        th <- getHubContent(th)
        genomes <- getGenomesContent(th)
        if (is.list(genomes) && length(genomes) >= 1L)
            th@genomeContainer <- GenomeContainer(unlist(genomes))
    }
    th
}

setAs("character", "TrackHub", function(from) TrackHub(from))

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### TrackHubGenome class
###

setGeneric("getTracks", function(x) standardGeneric("getTracks"))
setGeneric("trackField", function(x, name, key) standardGeneric("trackField"))
setGeneric("trackField<-", function(x, name, key, value) standardGeneric("trackField<-"))

setClass("TrackHubGenome",
         representation(trackhub = "TrackHub",
                        genome = "character",
                        tracks = "TrackContainer",
                        levels = "integer"),
         contains = "TrackDb")

trackhub <- function(x) x@trackhub

createTrack <- function(trackDf) {
    fieldToType <- list(
        track = "character", type = "character", shortLabel = "character", longLabel = "character",
        bigDataUrl = "character", html = "character", visibility = "character", meta = "character",
        color = "character", priority = "numeric", altColor = "character", boxedCfg = "logical",
        chromosomes = "character", darkerLabels = "logical", dataVersion = "character",
        directUrl = "character", iframeUrl = "character", iframeOptions = "character",
        mouseOverField = "character", otherDb = "character", pennantIcon = "character",
        tableBrowser = "character", url = "character", urlLabel = "character", urls = "character",
        skipEmptyFields = "logical", skipFields = "character", sepFields = "character",
        refUrl = "character", bigDataIndex = "character", bamColorMode = "character",
        bamGrayMode = "character", aliQualRange = "character", baseQualRange = "character",
        bamColorTag = "character", noColorTag = "character", bamSkipPrintQualScore = "character",
        indelDoubleInsert = "logical", indelQueryInsert = "logical", indelPolyA = "logical",
        minAliQual = "character", pairEndsByName = "character", pairSearchRange = "character",
        showNames = "logical", doWiggle = "logical", maxWindowToDraw = "integer",
        barChartBars = "character", barChartColor = "character", barChartLabel = "character",
        barChartMaxSize = "character", barChartSizeWindows = "character", barChartMetric = "character",
        barChartUnit = "character", barChartMatrixUrl = "character", barChartSampleUrl = "character",
        maxLimit = "character", labelFields = "character", defaultLabelFields = "character",
        itemRgb = "logical", colorByStrand = "character", denseCoverage = "integer",
        labelOnFeature = "logical", exonArrows = "logical", exonNumbers = "logical",
        scoreFilter = "character", scoreFilterLimits = "character", maxItems = "integer",
        minGrayLevel = "character", noScoreFilter = "logical", spectrum = "logical",
        scoreMax = "integer", scoreMin = "integer", thickDrawItem = "logical", searchIndex = "character",
        searchTrix = "character", labelSeparator = "character", bedNameLabel = "character",
        exonArrowsDense = "logical", itemImagePath = "character", itemBigImagePath = "character",
        mergeSpannedItems = "logical", linkIdInName = "logical", nextExonText = "character",
        prevExonText = "character", scoreLabel = "character", showTopScorers = "character",
        linkDataUrl = "character", interactDirectional = "character", interactUp = "character",
        interactMultiRegion = "character", maxHeightPixels = "character", speciesOrder = "character",
        frames = "character", summary = "character", baseColorUseCds = "character",
        baseColorUseSequence = "character", baseColorDefault = "character",
        showDiffBasesAllScales = "logical", autoscale = "character", autoScale = "character",
        viewLimits = "character", viewLimitsMax = "character", alwaysZero = "logical",
        graphTypeDefault = "character", maxWindowToQuery = "integer", negateValues = "logical",
        smoothingWindow = "character", transformFunc = "character", windowingFunction = "character",
        yLineMark = "character", yLineOnOff = "logical", gridDefault = "logical",
        showSnpWidth = "integer", otherSpecies = "character", minQual = "character", minFreq = "character",
        hapClusterEnabled = "character", hapClusterColorBy = "character", hapClusterTreeAngle = "character",
        hapClusterHeight = "character", applyMinQual = "character", superTrack = "character",
        parent = "character", compositeTrack = "logical", allButtonPair = "logical",
        centerLabelsDense = "logical", dragAndDrop = "character",
        hideEmptySubtracks = "logical", hideEmptySubtracksMultiBedUrl = "character",
        hideEmptySubtracksSourcesUrl = "character", hideEmptySubtracksLabel = "character",
        subGroup1 = "character", subGroup2 = "character", subGroup3 = "character", subGroup4 = "character",
        subGroup5 = "character", subGroup6 = "character", subGroup7 = "character", subGroup8 = "character",
        subGroup9 = "character", subGroups = "character", dimensions = "character",
        filterComposite = "character", dimensionAchecked = "character", dimensionBchecked = "character",
        sortOrder = "character", view = "character", viewUi = "logical", configurable = "logical",
        container = "character", aggregate = "character", showSubtrackColorOnUi = "logical",
        metadata = "character", noInherit = "logical", useScore = "integer")
    trackDf$value <- gsub("\\b[Oo]n\\b", "TRUE", trackDf$value)
    trackDf$value <- gsub("\\b[Oo]ff\\b", "FALSE", trackDf$value)
    extrafields <- setdiff(trackDf$field, names(fieldToType))
    selectedRows <- !(trackDf$field == extrafields)
    if (length(selectedRows))
        trackDf <- trackDf[selectedRows,]
    args <- Map(as, trackDf$value, fieldToType[trackDf$field])
    names(args) <- trackDf$field
    track <- do.call(Track, args)
    track
}

getTabCountList <- function(filepath) {
    fileContent <- readLines(filepath, warn = FALSE)
    fileContent <- gsub("^\\t*\\s*#(.)*", "", fileContent)
    fileContent <- fileContent[fileContent != ""]
    matches <- gregexpr("^\\s+", fileContent)
    tabCountList <- vapply(matches, attr, integer(1L), "match.length")
    indexes <- grep(-1, tabCountList)
    tabCountList <- replace(tabCountList, indexes, 0)
    as.integer(tabCountList)
}

readFileAsDf <- function(filepath) {
    fileContent <- readLines(filepath, warn = FALSE)
    fileContent <- gsub("^(\\t)*#(.)*", "", fileContent) # to avoid reading commented tracks
    regex <- "^\\s*\\t*([a-zA-Z.]+)\\s?(.*)$"
    field <- sub(regex, "\\1", fileContent)
    value <- sub(regex, "\\2", fileContent)
    contentDf <- data.frame(field, value)
    contentDf <- contentDf[contentDf$field != "",]
}

getTrackDbContent <- function(x, trackDbFilePath) {
    if (isFileEmpty(trackDbFilePath)) {
        x@tracks <- TrackContainer()
        return(x)
    }
    contentDf <- readFileAsDf(trackDbFilePath)
    levels <- getTabCountList(trackDbFilePath)
    tracksIndex <- grep("\\btrack\\b", contentDf$field)
    levels <- levels[tracksIndex]
    totalTracks <- length(tracksIndex)
    tracksIndex[length(tracksIndex) + 1] <- length(contentDf$field) + 1 # to read last track from file
    contentDf$field <- gsub("^(\\t)+", "", contentDf$field)
    # to speed up, reading track by track
    tracks <- lapply(seq_len(totalTracks), function(x) {
        startPosition <- tracksIndex[x]
        endPosition <- tracksIndex[x + 1] - 1
        trackDf <- setNames(data.frame(contentDf$field[startPosition:endPosition],
                                       contentDf$value[startPosition:endPosition]),
                          c("field", "value"))
        track <- createTrack(trackDf)
    })
    x@tracks <- TrackContainer(tracks)
    x@levels <- levels
    x
}

setMethod("genome", "TrackHubGenome", function(x) x@genome)

setMethod("uri", "TrackHubGenome", function(x)
          paste(trimSlash(uri(trackhub(x))), genome(x), sep = "/"))

setMethod("getTracks", "TrackHubGenome", function(x) {
    x@tracks
})

setMethod("names", "TrackHubGenome", function(x) {
    as.character(names(getTracks(x)))
})

setMethod("trackNames", "TrackHubGenome", function(object) {
    names(object)
})

setMethod("trackField", "TrackHubGenome", function(x, name, key) {
    names <- names(x@tracks)
    track <- x@tracks[names == name]
    if (length(track) == 0L) stop("Track '", name, "' does not exist")
    else if (length(track) > 1L) stop("Multiple tracks match ", name)
    slot(track[[1L]], key)
})

setReplaceMethod("trackField", "TrackHubGenome", function(x, name, key, value) {
    names <- names(x@tracks)
    track <- x@tracks[names == name]
    slot(track[[1L]], key) <- value
    x@tracks[names == name] <- track
    x
})

setMethod("organism", "TrackHubGenome", function(object) {
    genome <- getGenome(trackhub(object), genome(object))
    genome@organism
})

setMethod("referenceSequence", "TrackHubGenome", function(x) {
    trackhub <- trackhub(x)
    genome <- getGenome(trackhub, genome(x))
    twoBitPathValue <- genome@twoBitPath
    if (!isFieldEmpty(twoBitPathValue)) {
        twoBitFilePath <- combineURI(uri(trackhub), twoBitPathValue)
        import(twoBitFilePath)
    }
    else stop("genome.txt: 'twoBitPath' does not contain a reference to a file")
})

setReplaceMethod("referenceSequence", "TrackHubGenome", function(x, value) {
    trackhub <- trackhub(x)
    genomesFilePath <- combineURI(uri(trackhub), trackhub@genomesFile)
    stopIfNotLocal(genomesFilePath)
    genome <- getGenome(trackhub, genome(x))
    twoBitPathValue <- genome@twoBitPath
    if (!isFieldEmpty(twoBitPathValue)) {
        twoBitFilePath <- combineURI(uri(trackhub(x)), twoBitPathValue)
        export.2bit(value, twoBitFilePath)
        x
    }
    else stop("genome.txt: 'twoBitPath' does not contain a reference to a file")
})

setMethod("length", "TrackHubGenome", function(x) {
    length(names(x))
})

setMethod("writeTrackHub", "TrackHubGenome", function(x) {
    trackhub <- trackhub(x)
    stopIfNotLocal(hubFile(trackhub))
    genome <- getGenome(trackhub, genome(x))
    trackDbValue <- genome@trackDb
    trackDbFilePath <- combineURI(uri(trackhub), trackDbValue)
    if (!isFileEmpty(trackDbFilePath) || length(x@tracks)) {
        tabStrings <- vapply(x@levels, function(y) {
            paste(rep("\t", y), collapse = "")
        },character(1L))
        if (length(tabStrings) == 0)
            tabStrings <- rep("", length(x@tracks))
        slots <- slotNames(Track())
        tracks <- vapply(seq_len(length(x@tracks)), function(i) {
            track <- vapply(slots, function(slotName) {
                slotValue <- slot(x@tracks[[i]], slotName)
                if (!isEmpty(slotValue)) {
                    if (is.na(tabStrings[i])) tabStrings[i] <- ""
                    trackline <- paste0(tabStrings[i], slotName, " ", slotValue)
                    if (slotName == "track") trackline <- paste0("\n", trackline)
                    trackline
                }
                else ""
            }, character(1L))
        }, character(155L))
        tracks <- tracks[tracks != ""]
        tracks <- gsub("\\bTRUE\\b", "on", tracks)
        tracks <- gsub("\\bFALSE\\b", "off", tracks)
        writeLines(tracks, trackDbFilePath)
    }
})

setMethod("show", "TrackHubGenome", function(object) {
    cat(class(object), "track database\ngenome:", genome(object), "\ntrackhub:",
        uri(trackhub(object)), "\n")
    cat(S4Vectors:::labeledLine("names", names(object)))
})

TrackHubGenome <- function(trackhub, genome, create = FALSE) {
    trackhub <- as(trackhub, "TrackHub")
    thg <- new("TrackHubGenome")
    thg@trackhub <- trackhub
    thg@genome <- genome
    genome <- getGenome(trackhub(thg), genome(thg))
    trackDbValue <- genome@trackDb
    if (!isFieldEmpty(trackDbValue)) {
        trackDbFilePath <- combineURI(uri(trackhub(thg)), trackDbValue)
        if (!uriExists(trackDbFilePath) && create) {
            createResource(trackDbFilePath)
        }else if (!isFileEmpty(trackDbFilePath) && uriExists(trackDbFilePath)) {
            thg <- getTrackDbContent(thg, trackDbFilePath)
        }
    }
    thg
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import of tracks from Track Hub
###

setMethod("track", "TrackHubGenome", function(object, name, ...) {
    names <- names(object@tracks)
    track <- object@tracks[names == name]
    if (length(track) == 0L) stop("Track '", name, "' does not exist")
    else if (length(track) > 1L) stop("Multiple tracks match ", name)
    bigDataUrl <- track[[1L]]@bigDataUrl
    parsed <- .parseURI(bigDataUrl)
    if (isEmpty(bigDataUrl)) {
        stop("Track '", name, "' does not contain any data file")
    } else if (uriIsLocal(parsed)) {
        import(paste0(uri(object), "/", bigDataUrl), ...)
    } else {
        import(bigDataUrl, ...)
    }
})

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export of tracks to Track Hub
###

copyResourceToTrackHub <- function(object, uri) {
    parsed_uri <- .parseURI(uri)
    if (parsed_uri$scheme == "")
        uri <- paste0("file://", uri)
    filename <- basename(uri)
    trackhub <- trackhub(object)
    object_uri <- .parseURI(uri(trackhub))
    if (uriIsLocal(object_uri)) {
        genome <- getGenome(trackhub, genome(object))
        trackDbValue <- genome@trackDb
        trackDbValue <- sub(basename(trackDbValue), "", trackDbValue)
        trackDbValue <- sub("/$", "", trackDbValue)
        dest_file <- paste(object_uri$path, trackDbValue, filename, sep = "/")
        dest_file <- sub("^/", "", dest_file)
        if (paste(uri(object), filename, sep = "/") != uri)
            ### FIXME: curl_unescape() here because of R bug
            download.file(curl_unescape(uri), dest_file)
    }
    else stop("TrackHub is not local; cannot copy track")
    filename
}

.exportToTrackHub <- function(object, name,
                              format = bestFileFormat(value, object),
                              index = TRUE, ..., value)
{
    filename <- paste(name, format, sep = ".")
    path <- paste(uri(object), filename, sep = "/")
    file <- export(value, path, format = format, index = index, ...)
    track(object, name, index = FALSE) <- file
    object
}

setReplaceMethod("track",signature(object = "TrackHubGenome", value = "ANY"),
                 .exportToTrackHub)

setReplaceMethod("track",
                 signature(object = "TrackHubGenome", value = "RsamtoolsFile"),
                 function(object, name, ..., value)
                 {
                     if (missing(name))
                         name <- basename(path(value))
                     track(object, name) <- URLencode(path(value))
                     copyResourceToTrackHub(object, URLencode(index(value)))
                     object
                 })

setReplaceMethod("track",
                 signature(object = "TrackHubGenome", value = "BiocFile"),
                 function(object, name, ..., value)
                 {
                     if (missing(name))
                         name <- basename(path(value))
                     track(object, name) <- URLencode(path(value))
                     object
                 })

setReplaceMethod("track",
                 signature(object = "TrackHubGenome", value = "character"),
                 function(object, name = basename(object), ..., value)
                 {
                     filename <- copyResourceToTrackHub(object, value)
                     genome <- getGenome(trackhub(object), genome(object))
                     trackDbValue <- genome@trackDb
                     trackDbValue <- sub(basename(trackDbValue), "", trackDbValue)
                     trackDbValue <- sub("/$", "", trackDbValue)
                     bigDataUrlValue <- paste(trackDbValue, filename, sep = "/")
                     bigDataUrlValue <- sub("^/", "", bigDataUrlValue)
                     names <- names(object@tracks)
                     trackPosition <- which(names == name)
                     trackDf <- setNames(data.frame(c("track", "bigDataUrl"),
                                                    c(name, bigDataUrlValue)),
                                         c("field", "value"))
                     track <- createTrack(trackDf)
                     if (isEmpty(trackPosition)) {
                         trackPosition <- length(object@tracks) + 1
                         object@tracks[[trackPosition]] <- track
                     }else {
                         object@tracks[[trackPosition]] <- track
                     }
                     object
                 })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Utilities
###

combineURI <- function(x,y) paste(trimSlash(x), y, sep = "/")

isFileReference <- function(x) {
    formats <- c("txt", "2bit", "html")
    tools::file_ext(x) %in% formats
}

isFieldEmpty <- function(x) {
    if ((isFileReference(x) && !is.na(x)) && !is.null(x)) {
        return(FALSE)
    }
    return(TRUE)
}

trimSlash <- function(x) {
    sub("/$", "", x)
}

### Note that prior to rtracklayer 1.63.1, isFileEmpty() was returning an
### error on an empty remote file. Also it was always returning FALSE on a
### local file, even on an empty one!
isFileEmpty <- function(path) {
    url <- .parseURI(path)
    if (uriIsLocal(url)) {
        size <- file.size(url$path)
        size == 1L || size == 0L
    } else {
        contentLength <- HEAD(path)$headers[["content-length"]]
        is.null(contentLength) || contentLength == "0"
    }
}

### =========================================================================
### Chain file parsing and lift over
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Classes
###

setClass("ChainBlock",
         representation(ranges = "IRanges", # start in A, width
                        offset = "integer", # offset to start in B
                        score = "integer", # rle scores
                        space = "character", # rle spaces
                        reversed = "logical", # rle reversal
                        length = "integer")) # lengths for rle slots

##setMethod("length", "ChainBlock", function(x) length(x@offset))

setClass("Chain",
         prototype = prototype(elementType = "ChainBlock"),
         contains = "SimpleList")

setClass("ChainFile", contains = "BiocFile")

ChainFile <- function(path) {
  if (!isSingleString(path))
    stop("'filename' must be a single string, specifying a path")
  new("ChainFile", resource = path)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import
###

setGeneric("import.chain",
           function(con, ...) standardGeneric("import.chain"))

setMethod("import.chain", "ANY", function(con, ...) {
  import(con, "chain", ...)
})

setMethod("import", "ChainFile", function(con, format, text, exclude = "_") {
  if (!missing(format))
    checkArgFormat(con, format)
  ### FIXME: use readLines() to read the file, parse lines in C
  if (!isSingleString(resource(con)) || isURL(resource(con))) {
    stop("chain import currently only handles local, uncompressed file paths")
  }
  .Call("readChain", path.expand(path(con)), as.character(exclude),
        PACKAGE="rtracklayer")
})

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Accessors
###

setMethod("ranges", "ChainBlock",
    function(x, use.names=TRUE, use.mcols=FALSE) x@ranges
)
setMethod("offset", "ChainBlock", function(object) object@offset)
setMethod("score", "ChainBlock", function(x) Rle(x@score, x@length))
setMethod("space", "ChainBlock", function(x) Rle(x@space, x@length))

setGeneric("reversed", function(x, ...) standardGeneric("reversed"))
setMethod("reversed", "ChainBlock", function(x) Rle(x@reversed, x@length))

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Liftover
###

flipStrandSimple <- function(strand, flip) {
  strand <- as.vector(strand)
  flipped <- ifelse(flip, ifelse(strand == "+", "-",
                                 ifelse(strand == "-", "+", strand)),
                    strand)
  strand(flipped)
}

flipStrandTricky <- function(strand, flip) {
  strandCodes <- c("+" = 1L, "-" = -1L, "*" = 0L)
  strandInt <- strandCodes[as.vector(strand)]
  flipped <- ifelse(flip, strandInt * -1L, strandInt) + 2L
  strandRevCodes <- factor(c("-", "*", "+"), levels(strand()))
  strandRevCodes[as.vector(flipped)]
}

smoothGaps <- function(qhits, ranges, offsets) {
    congruent_gaps <- width(gaps(ranges)) == abs(offsets)
    congruent_gaps_rle <- Rle(congruent_gaps)
    congruent_rle <- c(Rle(FALSE), congruent_gaps_rle)
    group_rle <- Rle(cumsum(!congruent_rle))
    group_ranges <- disjoin(ranges(Rle(qhits)), ranges(group_rle))
    ans <- range(relist(ranges, group_ranges))
    mcols(ans)$qhits <- qhits[start(group_ranges)]
    mcols(ans)$offsets <- offsets[start(group_ranges)]
    ans
}


setGeneric("liftOver", function(x, chain, ...) standardGeneric("liftOver"))
setMethod("liftOver", c("GenomicRanges", "Chain"),
          function(x, chain)
          {
            liftOverSpace <- function(gr, chain, subind) {
              r <- ranges(gr)
              ol <- findOverlaps(r, ranges(chain))
              shits <- subjectHits(ol)
              r <- overlapsRanges(r, ranges(chain), ol)
              rev <- as.vector(reversed(chain)[shits])
              starts <- ifelse(rev,
                               start(reflect(r, ranges(chain)[shits])),
                               start(r))
              strand <- flipStrandTricky(strand(gr)[queryHits(ol)], rev)
              r <- IRanges(starts, width=width(r))
              offsets <- offset(chain)[shits]
              spaces <- space(chain)[shits]
              ind[[as.character(seqnames(gr)[1])]] <<- subind[queryHits(ol)]
              GRanges(spaces,
                      IRanges(start(r) - offsets, end(r) - offsets),
                      strand = strand,
                      values(gr)[queryHits(ol),,drop=FALSE])
            }
            rl <- split(x, seqnames(x), drop = TRUE)
            unchainedNames <- setdiff(names(rl), names(chain))
            if (length(unchainedNames))
              message("Discarding unchained sequences: ",
                      paste(unchainedNames, collapse = ", "))
            sharedNames <- intersect(names(rl), names(chain))
            ind <- split(seq_len(length(x)),
                         as.vector(seqnames(x)))[sharedNames]
            liftedList <- mapply(liftOverSpace, rl[sharedNames],
                                 chain[sharedNames], ind, SIMPLIFY=FALSE)
            lifted <- unlist(GRangesList(liftedList), use.names=FALSE)
            f <- structure(as.integer(unlist(ind, use.names=FALSE)),
                           levels=seq_len(length(x)), class="factor")
            setNames(split(lifted, f), names(x))
          })

setMethod("liftOver", c("GRangesList", "Chain"),
          function(x, chain)
          {
              lifted <- liftOver(unlist(x), chain)
              IRanges:::regroupBySupergroup(lifted, PartitioningByEnd(x))
          })

setMethod("liftOver", c("Pairs", "Chain"),
          function(x, chain) {
              Pairs(liftOver(first(x), chain),
                    liftOver(second(x), chain))
          })

setMethod("liftOver", c("ANY", "ANY"),
          function(x, chain) {
    chain <- as(chain, "Chain")
    x <- granges(x)
    callGeneric()
})

# UCSC genome browser interface

# every UCSC session is identified by a 'hguid'
setClass("UCSCSession",
         representation(url = "character", hguid = "character",
                        views = "environment"),
         contains = "BrowserSession")

# gets an 'hgsid' to initialize the session
setMethod("initialize", "UCSCSession",
          function(.Object, url = "https://genome.ucsc.edu/cgi-bin/",
                   user = NULL, session = NULL, force = FALSE, ...)
          {
            .Object@url <- url
            .Object@views <- new.env()
            gwURL <- ucscURL(.Object, "gateway", force=force)
            config <- config(followlocation = 1, nobody = 1,
                             cookiefile = tempfile(),
                             useragent = "rtracklayer", ...)
            ## nobody = 1 seems to be ignored and HEAD() doesn't work here
            response <- GET(gwURL, config)
            cookie <- response$headers[["set-cookie"]]
            if (is.null(cookie))
              stop("Failed to obtain 'hguid' cookie")
            hguid <- strsplit(cookie, split = ";", fixed = TRUE)[[1L]][[1L]]
            .Object@hguid <- hguid
            if (!is.null(user) && !is.null(session)) { ## bring in other session
              ucscGet(.Object, "tracks",
                      list(hgS_doOtherUser = "submit", hgS_otherUserName = user,
                           hgS_otherUserSessionName = session))
            }
            .Object
          })

setMethod("seqlengths", "UCSCSession", function(x) {
  seqlengths(Seqinfo(genome = genome(x)))
})

setMethod("seqnames", "UCSCSession", function(x) names(seqlengths(x)))

setMethod("seqinfo", "UCSCSession", function(x) {
  Seqinfo(genome = genome(x)) # no circularity information
})

normArgTrackData <- function(value, session) {
  genomes <- vapply(value, function(x) singleGenome(genome(x)), character(1L))
  genomes[is.na(genomes)] <- ""
  tapply(value, unlist(genomes),
         function(tracks)
         {
           genome <- singleGenome(genome(tracks[[1]]))
           if (!is.na(genome))
             genome(session) <- genome
           spaces <- do.call(c, unname(lapply(tracks, seqnames)))
           badSpaces <- setdiff(spaces, seqnames(session))
           if (length(badSpaces) > 0L)
             stop("Invalid chromosomes for ", genome(session), ": ",
                  paste(badSpaces, collapse = ", "))
         })
  value
}

handleError <- function(response) {
    msg <- getNodeSet(response, "//span[text()='Error']/../text()")
    if (length(msg) == 2L)
        stop(sub(".*? - ", "", xmlValue(msg[[2L]])))
}

setReplaceMethod("track", c("UCSCSession", "SimpleGRangesList"),
          function(object, name = names(value),
                   format = c("auto", "bed", "wig", "gff1", "bed15",
                     "bedGraph"), ..., value)
          {
            format <- match.arg(format)
            if (length(value)) {
              ## upload values in blocks, one for each genome
              value <- normArgTrackData(value, object)
              names(value) <- name
              genomes <- sapply(value, function(x) singleGenome(genome(x)))
              genomes[is.na(genomes)] <- ""
              tapply(value, unlist(genomes),
                     function(tracks)
                     {
                       form <- ucscForm(tracks, format, ...)
                       response <- ucscPost(object, "custom", form)
                       handleError(response)
                     })
            }
            object
          })

setReplaceMethod("track", c("UCSCSession", "BiocFile"),
                 function(object, name = names(value), ..., value)
                 {
                   form <- ucscForm(value, genome(object), ...)
                   response <- ucscPost(object, "custom", form)
                   object
                 })

setMethod("browserViews", "UCSCSession",
          function(object) object@views$instances)

## get the list of track names
setMethod("trackNames", "UCSCSession",
          function(object) ucscTracks(object)@ids)

## get the current range
setMethod("range", "UCSCSession",
          function(x, ..., na.rm) range(ucscCart(x)))

setReplaceMethod("range", "UCSCSession",
                 function(x, value) {
                   ucscGet(x, "cart", ucscForm(normGenomeRange(value, seqinfo(x))))
                   x
                 })

setMethod("genome", "UCSCSession", function(x) {
  genome(ucscCart(x))
})

setReplaceMethod("genome", "UCSCSession",
                 function(x, value) {
                   if (!isSingleString(value))
                     stop("'genome' must be a single non-NA string")
                   ucscGet(x, "gateway", list(db = value))
                   if (genome(x) != value)
                     stop("Failed to set session genome to '", value, "'")
                   x
                 })

SeqinfoForUCSCGenome <- function(genome) {
  tryCatch({
    session <- browserSession("UCSC")
    genome(session) <- genome
    seqinfo(session)
  }, error = function(cond) NULL)
}

GRangesForUCSCGenome <- function(genome, chrom = NULL, ranges = NULL, ...)
{
  GRangesForGenome(genome, chrom = chrom, ranges = ranges, method = "UCSC",
                   seqinfo = NULL, ...)
}


## context for querying UCSC tables
setClass("UCSCTableQuery",
         representation(genome = "character",
                        table = "character_OR_NULL",
                        range = "GRanges",
                        NAMES = "character_OR_NULL",
                        url = "character",
                        hubUrl = "character_OR_NULL",
                        track = "character_OR_NULL"))

setMethod("show", "UCSCTableQuery",
          function(object) {
            cat("Get ")
            if (!is.null(tableName(object)))
              cat("table '", tableName(object), "' from ", sep = "")
            range <- range(object)
            if (length(range) > 1)
              start <- end <- chrom <- "*"
            else {
              chrom <- as.character(seqnames(range))
              start <- start(range)
              end <- end(range)
            }
            cat(genome(range)[1], ":", chrom, ":", start, "-", end, sep="")
            cat("\n")
          })

setMethod("genome", "UCSCTableQuery", function(x) {
  x@genome
})

setReplaceMethod("genome", "UCSCTableQuery", function(x, value) {
  x@genome <- value
  x
})

setMethod("browserSession", "UCSCTableQuery", function(object) {
    .Defunct(msg = "UCSCTableQuery is no longer associated with a UCSCSession")
})

setGeneric("browserSession<-",
           function(object, ..., value) standardGeneric("browserSession<-"))
setReplaceMethod("browserSession", "UCSCTableQuery",
                 function(object, value) {
                   .Defunct(msg = "UCSCTableQuery is no longer associated with a UCSCSession")
                 })

setMethod("range", "UCSCTableQuery", function(x, ..., na.rm) x@range)
setReplaceMethod("range", "UCSCTableQuery",
                 function(x, value) {
                   x@range <- normTableQueryRange(value, x@genome, 1L)
                   x
                 })

setGeneric("tableName", function(x, ...) standardGeneric("tableName"))
setMethod("tableName", "UCSCTableQuery", function(x) x@table)

normArgTable <- function(name, query) {
  if (!is.null(name)) {
    if (!isSingleString(name))
      stop("table name must be a single string or NULL")
    if (!tableExists(query, name))
      stop("Table '", name, "' is unavailable")
  }
  name
}

setGeneric("tableName<-", function(x, ..., value)
           standardGeneric("tableName<-"))
### FIXME: we need '...' in the formals due to a bug in R 3.5, remove for 3.6
setReplaceMethod("tableName", "UCSCTableQuery",
                 function(x, check=TRUE, ..., value)
                 {
                   if (!missing(...))
                     warning("arguments in '...' ignored")
                   if (check)
                       value <- normArgTable(value, x)
                   x@table <- value
                   x
                 })

setMethod("names", "UCSCTableQuery", function(x) x@NAMES)
setReplaceMethod("names", "UCSCTableQuery", function(x, value) {
  x@NAMES <- value
  x
})

setGeneric("intersectTrack", function(x, ...)
           standardGeneric("intersectTrack"))
setMethod("intersectTrack", "UCSCTableQuery", function(x) {
  .Defunct(msg = "intersectTrack is no longer supported")
})
setGeneric("intersectTrack<-", function(x, ..., value)
           standardGeneric("intersectTrack<-"))
setReplaceMethod("intersectTrack", "UCSCTableQuery", function(x, value) {
  .Defunct(msg = "intersectTrack is no longer supported")
})

normTableQueryRange <- function(range, genome, max.length = 1000L) {
  seqinfo <- Seqinfo(genome = genome)
  normGenomeRange(range, seqinfo, max.length)
}

ucscTables <- function(genome, track) {
    .Defunct("tableNames",
             msg = paste("Query UCSC data using the table identifier,",
                         "found by using the UCSC table browser interactively."))
}

setGeneric("ucscTableQuery", function(x, ...) standardGeneric("ucscTableQuery"))
setMethod("ucscTableQuery", "UCSCSession",
            function(x, ...) {
              ucscTableQuery(genome(x), ...)
          })

normArgTrack <- function(name, trackids) {
  if (!isSingleString(name))
    stop("'track' must be a single string")
  if (name %in% names(trackids))
      unname(trackids[name])
  else
      stop("Unknown track: ", name)
}

setMethod("ucscTableQuery", "character",
          function(x, track = NULL, range =  NULL, table = NULL,
                   names = NULL, intersectTrack = NULL, check = TRUE, hubUrl = NULL,
                   genome = NULL, url = "https://genome.ucsc.edu/cgi-bin/") {
              stopifnot(isSingleString(x))
              if (!is.null(intersectTrack))
                stop("intersectTrack is no longer supported")
              if (!is(names, "character_OR_NULL"))
                stop("'names' must be 'NULL' or a character vector")
              if (uriExists(x)) { # if x is URI and it exits that means it's a trackHub
                if (is.null(genome))
                  stop("'genome' is a mandatory parameter and must be a single string")
                hubUrl <- x
              } else genome <- x
              # if the table is provied then it will not try to identify the table from the track
              if (!is.null(track) && is.null(table)) {
                .Defunct(msg = "'track' argument is defunct; use 'table' instead")
                tables <- ucscTables(genome, track)
                table <- tables[1]
              }
              if (is.null(range)) {
                range <- Seqinfo(genome = genome)
                range <- as(range, "GRanges")
              } else range <- normTableQueryRange(range, genome)
              query <- new("UCSCTableQuery", genome = genome, range = range,
                          NAMES = names, url = url, hubUrl = hubUrl, track = track)
              if (is.null(table))
                check <- FALSE
              tableName(query, check=check) <- table
              query
          })

isTrackHub <- function(x) {
  status <- FALSE
  if (!is.null(x@hubUrl)) {
    if (uriExists(x@hubUrl))
      status <- TRUE
    else stop(paste("TrackHub", x@hubUrl, "does not exists"))
  }
  status
}

tableExists <- function(query, name) {
  if (isTrackHub(query))
    return(name %in% tableNames(query))

  url <- RestUri(paste0(query@url, "hubApi"))
  response <- tryCatch(
    read(url$list$schema, genome = query@genome, track = name),
    error = function(...) NULL
  )
  !is.null(response$columnTypes)
}

dropCookie <- function(object) {
    object@hguid <- character()
    object
}

stopIfTableEmpty <- function(object) {
  if (is.null(object@table))
    stop("'table' is a mandatory value and must be a single string, Use tableName()<- to set it")
}

setGeneric("hubUrl", function(x) standardGeneric("hubUrl"))
setGeneric("hubUrl<-", function(x,value) standardGeneric("hubUrl<-"))
setMethod("hubUrl", "UCSCTableQuery", function(x) x@hubUrl)
setReplaceMethod("hubUrl", "UCSCTableQuery", function(x, value) {
  x@hubUrl <- value
  x
})

setMethod("trackNames", "UCSCTableQuery",
          function(object) {
              .Defunct("tableNames",
                       msg = "querying UCSC is now based on the table identifier")
          })

## returns a character vector of table names for a given track name + range
setGeneric("tableNames", function(object, ...)
           standardGeneric("tableNames"))

setMethod("tableNames", "UCSCTableQuery",
          function(object, trackOnly = FALSE)
          {
            if (trackOnly)
              warning("track is meaningless now you only go by the table")
            genome <- object@genome
            if (!is.null(object@track)) {
              ucscTables(object@genome, object@track)
            } else if (isTrackHub(object)) {
              th <- TrackHub(object@hubUrl)
              names(th[[genome]])
            } else {
              url <- RestUri(paste0(object@url, "hubApi"))
              response <- read(url$list$tracks, genome = genome, trackLeavesOnly = 1)
              names <- names(response[[genome]])
              tables <- mapply(function(name, response) {
                                   if (!is.null(response$protectedData)) NULL
                                   else if (!is.null(response$table)) response$table
                                   else name
                          }, names, response[[genome]])
              Filter(Negate(is.null), tables)
            }
          })

setClass("UCSCSchema",
         representation(genome = "character",
                        tableName = "character",
                        rowCount = "integer"),
         contains = "DFrame")

setMethod("genome", "UCSCSchema", function(x) {
  x@genome
})

setMethod("tableName", "UCSCSchema", function(x) {
  x@tableName
})

setMethod("nrow", "UCSCSchema", function(x) {
  x@rowCount
})

setGeneric("ucscSchema",
           function(object, ...) standardGeneric("ucscSchema"))

setMethod("ucscSchema", "UCSCTableQuery", function(object) {
  genome <- object@genome
  tableName <- tableName(object)
  stopifnot(isSingleString(tableName))
  url <- RestUri(paste0(object@url, "hubApi"))
  response <- read(url$list$schema, genome = genome, track = tableName)
  rowCount <- as.integer(response[["itemCount"]])
  if (length(rowCount) == 0L)
    rowCount <- NA_integer_
  listOfDf <- lapply(response[["columnTypes"]], function(x) {
    DataFrame(x$name, x$sqlType, x$jsonType, x$description)
  })
  schemaDf <- do.call(rbind, listOfDf)
  names(schemaDf) <- c("field", "SQL.type", "JSON.type", "description")
  new("UCSCSchema", schemaDf, genome = genome, tableName = tableName,
      rowCount = rowCount)
})

setMethod("show", "UCSCSchema", function(object) {
  if (!is.null(tableName(object)))
    cat("UCSCSchema table '", tableName(object), "' with ", object@rowCount,
        " rows and ", length(object@listData), " columns\n", sep = "")
  show(DataFrame(object))
})

setMethod("track", "UCSCSession",
          function(object, name, ...)
          {
              track(ucscTableQuery(object, table=name, ...))
          })

## download a trackSet by name
setMethod("track", "UCSCTableQuery",
          function(object)
          {
            stopIfTableEmpty(object)
            tables <- tableNames(object)
            table <- tableName(object)
            if (!is.null(table) && !(table %in% tables))
              stop("Unknown table: '", table, "'. Valid table names: ", tables)
            if (isTrackHub(object)) {
              th <- TrackHub(object@hubUrl)
              thg <- TrackHubGenome(th, object@genome)
               if (length(object@range) == 1L) {
                start <- start(object@range)
                end <- end(object@range)
                seqname <- as.vector(seqnames(object@range))
                which <- GRanges(seqname, IRanges(start, end))
                track <- track(thg, table, which = which)
               } else track <- track(thg, table)
               track
            } else {
              table <- getTable(object)
              if (nrow(table) == 1000000)
                stop("Output is incomplete: ",
                    "track may have more than 100,000 elements. ",
                    "Try downloading the data via the UCSC FTP site.")
              # normalize column names
              names(table)[names(table) == "chromStart"] <- "start"
              names(table)[names(table) == "chromEnd"] <- "end"
              if(!"strand" %in% names(table)) {
                strand <- rep("*", dim(table)[1])
                cbind(table,strand)
              }

              if (is.null(table[["chrom"]]))
                  table[["chrom"]] <- as.character(seqnames(object@range))

              # create GRange object
              output <- GRanges(seqnames = table[["chrom"]],
                                ranges = IRanges(start = as.numeric(table[["start"]]),
                                                 end = as.numeric(table[["end"]])),
                                                 strand = table[["strand"]])

              # remove used columns
              table[["chrom"]] <- table[["strand"]] <- table[["start"]] <- table[["end"]] <- NULL
              # add left columns to the GRange object
              elementMetadata(output) <- table
              genome(output) <- object@genome
              output
            }
          })

## grab sequences for features in 'track' at 'range'
## setMethod("getSeq", "UCSCSession",
##           function(object, range, table = "gold")
##           {
##             followup <- list(hgta_doGenomicDna = "get sequence",
##                              hgSeq.casing = "upper",
##                              hgSeq.repMasking = "lower")
##             output <- ucscExport(object, range, "gold", table, "sequence",
##                                  followup)
##             con <- file()
##             writeLines(output, con)
##             set <- read.DNAStringSet(con, "fasta")
##             close(con)
##             set
##           })

parseResponse <- function(response, tableName, chrom) {
  results <- response[[tableName]]
  if (is.null(results)) {
    return(as.data.frame(do.call(rbind, response[[chrom]])))
  }
  if (is.null(names(results))) {
    df <- do.call(rbind.data.frame, results)
  } else {
    chromosomes <- names(results)
    listOfDf <- lapply(chromosomes, function(x) {
      do.call(rbind.data.frame, results[[x]])
    })
    df <- do.call(rbind, listOfDf)
  }
  rownames(df) <- NULL
  df
}

## get a data.frame from a UCSC table
## think about taking specific columns
setGeneric("getTable",
           function(object, ...) standardGeneric("getTable"))
setMethod("getTable", "UCSCTableQuery",
          function(object)
          {
            stopIfTableEmpty(object)
            tableName <- tableName(object)
            genome <- object@genome
            query <- list(genome = genome, track = tableName)
            if (length(object@range) == 1L) {
              start <- start(object@range)
              end <- end(object@range)
              seqname <- as.vector(seqnames(object@range))
              query <- c(query, chrom = seqname, start = start, end = end)
            }
            if (isTrackHub(object)) {
              th <- TrackHub(object@hubUrl)
              thg <- TrackHubGenome(th, genome)
               if (length(object@range) == 1L) {
                which <- GRanges(seqname, IRanges(start, end))
                track <- track(thg, tableName, which = which)
               } else track <- track(thg, tableName)
               as.data.frame(track)
            } else {
              url <- RestUri(paste0(object@url, "hubApi"))
              response <- read(url$getData$track, query)
              seqnames <- as.character(seqnames(object@range))
              output <- parseResponse(response, tableName, seqnames)
              NAMES <- names(object)
              if (!is.null(NAMES)) { # filter by NAMES
                if (is.null(output$name))
                  output <- output[output$frag %in% NAMES,] # as in some tables `frag` field is used instead of `name`
                else output <- output[output$name %in% NAMES,]
                rownames(output) <- seq(length=nrow(output))
              }
              output
            }
          })
setMethod("getTable", "UCSCSession",
          function(object, name, range = base::range(object), table = NULL) {
            getTable(ucscTableQuery(object, range = range, table = table))
          })

## UCSC genome view
setClass("UCSCView", representation(hgsid = "character", form = "list"),
         contains = "BrowserView")

## create a view for the given session, position and track visibility settings
## if 'tracks' is a character vector (but not a UCSCTrackModes instance) it is
## assumed to name the tracks that should be in the view. otherwise, an
## attempt is made to coerce it to a UCSCTrackModes instance.
setMethod("browserView", "UCSCSession",
          function(object, range, track, imagewidth = 800, browse = TRUE, ...)
          {
            stopifnot(isTRUEorFALSE(browse))
            form <- list()
            if (!missing(range)) {
              if (is(range, "IntegerRangesList"))
                range <- range[elementNROWS(range) > 0L]
              if (length(range) > 1) {
                ranges <- range
                views <- vector("list", length(ranges))
                for (i in seq(length(ranges))) {
                  range <- ranges[i]
                  views[[i]] <- callGeneric()
                }
                return(BrowserViewList(views))
              }
              range <- normGenomeRange(range, seqinfo(object))
              form <- c(form, ucscForm(range))
            }
            view <- new("UCSCView", session = object)
            ## new hgsid for each browser launch
            doc <- ucscGet(object, "gateway")
            hgsid <- sub(".*=", "",
                         grep("hgsid=", getNodeSet(doc, "//a/@href"),
                              value=TRUE)[1L])
            view@hgsid <- as.character(hgsid)
            ## figure out track modes
            origModes <- modes <- ucscTrackModes(view)
            if (!missing(track)) {
              if (class(track) == "character")
                trackNames(modes) <- track
              else {
                userModes <- as(track, "UCSCTrackModes")
                modes[names(userModes)] <- userModes
              }
            }
            argModes <- ucscTrackModes(...)
            modes[names(argModes)] <- argModes
            modes <- modes[modes != origModes]
            form <- c(form, ucscForm(modes), ucscForm(view))
            if (!missing(imagewidth))
                form <- c(form, pix = imagewidth)
            if (browse) {
                ## launch a web browser
                ucscShow(object, "tracks", form)
            }
            view@form <- form
            ## save this view
            object@views$instances <- c(object@views$instances, view)
            view
          })

viewURL <- function(x) {
    urlForm(ucscURL(browserSession(x), "tracks"), x@form)
}

# every view has a "mode" (hide, dense, pack, squish, full) for each track
### FIXME: probably should be merged with ucscTracks
### Or just leave it; ucscTracks might become more complex, while we
### need a simple way to manipulate track modes.
setClass("UCSCTrackModes", representation(labels = "character"),
         contains = "character")

# get/set track modes to/from e.g. a view
setGeneric("ucscTrackModes",
           function(object, ...) standardGeneric("ucscTrackModes"))

# convenience constructor for track mode object
setMethod("ucscTrackModes", "character",
          function(object, labels, hide = character(),
                   dense = character(), pack = character(),
                   squish = character(), full = character())
          {
            object[hide] <- "hide"
            object[dense] <- "dense"
            object[pack] <- "pack"
            object[squish] <- "squish"
            object[full] <- "full"
            if (missing(labels))
              labels <- names(object)
            new("UCSCTrackModes", object, labels = as.character(labels))
          })
setMethod("ucscTrackModes", "missing",
          function(object, ...) ucscTrackModes(character(), ...))

setMethod("ucscTrackModes", "UCSCView",
          function(object)
          {
            ucscTrackModes(ucscTracks(object))
          })

setMethod("ucscTrackModes", "UCSCSession",
          function(object)
          {
            ucscTrackModes(ucscTracks(object))
          })

setGeneric("ucscTrackModes<-",
           function(object, value) standardGeneric("ucscTrackModes<-"))
setReplaceMethod("ucscTrackModes", c("UCSCView", "UCSCTrackModes"),
                 function(object, value)
                 { # FIXME: needs to be more extensible
                   browserView(object@session, range(object), value)
                 })
setReplaceMethod("ucscTrackModes", c("UCSCView", "character"),
                 function(object, value)
                 {
                   ucscTrackModes(object) <- ucscTrackModes(value)
                   object
                 })

## subsetting UCSCTrackModes

## if not in ids, try labels
resolveTrackIndex <- function(object, i) {
  if (is.character(i)) {
    missing <- !(i %in% names(object))
    matching <- match(i[missing], object@labels)
    if (any(is.na(matching))) {
      unmatched <- i[missing][is.na(matching)]
      stop("Unknown track(s): ", paste(unmatched, collapse = ", "))
    }
    i[missing] <- names(object)[matching]
  }
  i
}

setMethod("[", "UCSCTrackModes", function(x, i, j, ..., drop=FALSE) {
  vec <- x@.Data
  names(vec) <- names(x)
  names(x@labels) <- names(x)
  ind <- resolveTrackIndex(x, i)
  initialize(x, vec[ind], labels = x@labels[ind])
})

setReplaceMethod("[", "UCSCTrackModes", function(x, i, j, ..., value) {
  vec <- x@.Data
  names(vec) <- names(x)
  vec[resolveTrackIndex(x, i)] <- value
  x@.Data <- as.vector(vec)
  x
})

# handle simple track show/hide

setMethod("trackNames", "UCSCTrackModes",
          function(object)
          {
            visible <- object != "hide"
            tracks <- names(object)[visible]
            names(tracks) <- object@labels[visible]
            tracks
          })
setReplaceMethod("trackNames", "UCSCTrackModes",
                 function(object, value)
                 {
                   value <- resolveTrackIndex(object, value)
                   spec <- names(object) %in% value
                   object[!spec] <- "hide"
                   object[spec & object == "hide"] <- "full"
                   object
                 })

setMethod("trackNames", "UCSCView",
          function(object)
          {
            tracks <- ucscTracks(object)
            modes <- ucscTrackModes(tracks)
            tracks@ids[tracks@ids %in% trackNames(modes)]
          })
setReplaceMethod("trackNames", "UCSCView",
                 function(object, value)
                 {
                   trackNames(ucscTrackModes(object)) <- value
                   object
                 })


setMethod("visible", "UCSCView", function(object) {
  modes <- ucscTrackModes(object)
  vis <- modes != "hide"
  names(vis) <- modes@labels
  vis
})
setReplaceMethod("visible", "UCSCView", function(object, value) {
  modes <- ucscTrackModes(object)
  modes[value & modes == "hide"] <- "full"
  modes[!value] <- "hide"
  ucscTrackModes(object) <- modes
  object
})


setMethod("range", "UCSCView",
          function(x, ..., na.rm) range(ucscCart(x)))
setReplaceMethod("range", "UCSCView",
                 function(x, value)
                 {
                   browserView(x@session, value, ucscTrackModes(x))
                 })

# only one view per session; a view is always active
setMethod("activeView", "UCSCView", function(object) TRUE)

# ucscTrackSet

# visual properties are specified by a "track line" for UCSC
setClass("TrackLine",
         representation(name = "character", description = "character",
                        visibility = "character", color = "integer",
                        priority = "numeric"),
         prototype(name = "R Track"))

setMethod("show", "TrackLine",
          function(object)
          {
            cat(as(object, "character"), "\n")
          })

setClass("BasicTrackLine",
         representation(itemRgb = "logical", useScore = "logical",
                        group = "character", db = "character",
                        offset = "numeric", url = "character",
                        htmlUrl = "character", colorByStrand = "matrix"),
         contains = "TrackLine")

ucscPair <- function(key, value) paste(key, value, sep = "=")

# to a text line
setAs("TrackLine", "character",
      function(from)
      {
        checkString <- function(str, len) {
          ## These are more annoying than useful
          ## if (nchar(gsub("[a-zA-Z0-9_ ]", "", str)))
          ##   warning("The string '", str,
          ##           "' contains non-standard characters.")
          ## if (nchar(str) > len) {
          ##   str <- substring(str, 1, len)
          ##   warning("The string '", str, "' must be less than ", len,
          ##           " characters; it has been truncated.")
          ## }
          if (regexpr(" ", str)[1] != -1)
            str <- paste("\"", str, "\"", sep="")
          str
        }
        str <- "track"
        name <- from@name
        if (length(name))
          str <- paste(str, " name=", checkString(name, 15), sep="")
        desc <- from@description
        if (length(desc))
          str <- paste(str, " description=", checkString(desc, 60), sep="")
        vis <- from@visibility
        if (length(vis))
          str <- paste(str, " visibility=", vis, sep="")
        color <- from@color
        if (length(color))
            str <- paste0(str, " color=\"", paste0(color, collapse=","), "\"")
        priority <- from@priority
        if (length(priority))
          str <- paste(str, " priority=", priority, sep="")
        str
      })

setAs("BasicTrackLine", "character",
      function(from)
      {
        str <- as(as(from, "TrackLine"), "character")
        itemRgb <- from@itemRgb
        if (length(itemRgb))
          str <- paste(str, " itemRgb=", if (itemRgb) "on" else "off", sep = "")
        useScore <- from@useScore
        if (length(useScore))
          str <- paste(str, " useScore=", if (useScore) "1" else "0", sep = "")
        group <- from@group
        if (length(group))
          str <- paste(str, " group=", group, sep="")
        db <- from@db
        if (length(db))
          str <- paste(str, " db=", db, sep="")
        offset <- from@offset
        if (length(offset))
          str <- paste(str, " offset=", offset, sep="")
        url <- from@url
        if (length(url))
          str <- paste(str, " url=", "\"", url, "\"", sep="")
        htmlUrl <- from@htmlUrl
        if (length(htmlUrl))
          str <- paste(str, " htmlUrl=", "\"", htmlUrl, "\"", sep="")
        colorByStrand <- from@colorByStrand
        if (length(colorByStrand)) {
          colors <- paste(colorByStrand[1,], colorByStrand[2,],
                          colorByStrand[3,], sep = ",", collapse = " ")
          str <- paste(str, " colorByStrand=\"", colors, "\"", sep = "")
        }
        str
      })

ucscParsePairs <- function(str)
{  
  str <- sub("^[[:alpha:]]*[[:blank:]]", "", str)
  split <- as.character(read.table(sep = "=", comment.char = "", as.is = TRUE,
                                   strip.white = TRUE, text = str))
  vals <- character(0)
  if (length(split)) {
    mixed <- tail(head(split, -1), -1)
    tags <- head(split, 1)
    vals <- tail(split, 1)
    if (length(mixed)) {
      tags <- c(tags, sub(".*[[:space:]]([[:alnum:]]*)$", "\\1", mixed))
      vals <- c(sub("(.*)[[:space:]][[:alnum:]]*$", "\\1", mixed), vals)
    }
    names(vals) <- tags
  }
  vals
}

# from a text line
setAs("character", "TrackLine",
      function(from)
      {
        line <- new("TrackLine")
        vals <- ucscParsePairs(from)
        if (!is.na(vals["name"]))
          line@name <- vals[["name"]]
        if (!is.na(vals["description"]))
          line@description <- vals[["description"]]
        if (!is.na(vals["visibility"]))
          line@visibility <- vals[["visibility"]]
        if (!is.na(vals["color"]))
          line@color <- as.integer(strsplit(vals[["color"]], ",")[[1]])
        if (!is.na(vals["priority"]))
          line@priority <- as.numeric(vals[["priority"]])
        line
      })

setAs("character", "BasicTrackLine",
      function(from)
      {
        line <- new("BasicTrackLine", as(from, "TrackLine"))
        vals <- ucscParsePairs(from)
        if (!is.na(vals["itemRgb"]))
          line@itemRgb <- tolower(vals[["itemRgb"]]) == "on"
        if (!is.na(vals["useScore"]))
          line@useScore <- vals[["useScore"]] == "1"
        if (!is.na(vals["group"]))
          line@group <- vals[["group"]]
        if (!is.na(vals["db"]))
          line@db <- vals[["db"]]
        if (!is.na(vals["offset"]))
          line@offset <- as.integer(vals[["offset"]])
        if (!is.na(vals["url"]))
          line@url <- vals[["url"]]
        if (!is.na(vals["htmlUrl"]))
          line@htmlUrl <- vals[["htmlUrl"]]
        if (!is.na(vals["colorByStrand"])) {
          colorToken <- strsplit(strsplit(vals[["colorByStrand"]], " ")[[1]],
                                 ",")
          line@colorByStrand <- matrix(as.integer(unlist(colorToken)), nrow = 3)
        }
        line
      })


setClass("GraphTrackLine",
         representation(altColor = "integer", autoScale = "logical",
                        alwaysZero = "logical",
                        gridDefault = "logical", maxHeightPixels = "integer",
                        graphType = "character", viewLimits = "numeric",
                        yLineMark = "numeric", yLineOnOff = "logical",
                        windowingFunction = "character",
                        smoothingWindow = "numeric", type = "character"),
         contains = "TrackLine")

setAs("GraphTrackLine", "character",
      function(from)
      {
        str <- as(as(from, "TrackLine"), "character")
        type <- if (from@type == "wig") "wiggle_0" else "bedGraph"
        str <- paste(str, " type=", type, sep = "")
        color <- from@altColor
        if (length(color))
          str <- paste(str, " altColor=", paste(color, collapse=","), sep="")
        autoScale <- from@autoScale
        onoff <- function(x) if (x) "on" else "off"
        if (length(autoScale))
          str <- paste(str, " autoScale=", onoff(autoScale), sep = "")
        alwaysZero <- from@alwaysZero
        if (length(alwaysZero))
          str <- paste(str, " alwaysZero=", onoff(alwaysZero), sep = "")
        gridDefault <- from@gridDefault
        if (length(gridDefault))
          str <- paste(str, " gridDefault=", onoff(gridDefault), sep = "")
        maxHeightPixels <- from@maxHeightPixels
        if (length(maxHeightPixels))
          str <- paste(str, " maxHeightPixels=",
                       paste(maxHeightPixels, collapse=":"), sep = "")
        graphType <- from@graphType
        if (length(graphType))
          str <- paste(str, " graphType=", graphType, sep = "")
        viewLimits <- from@viewLimits
        if (length(viewLimits))
          str <- paste(str, " viewLimits=", paste(viewLimits, collapse = ":"),
                       sep = "")
        yLineMark <- from@yLineMark
        if (length(yLineMark))
          str <- paste(str, " yLineMark=", yLineMark, sep = "")
        yLineOnOff <- from@yLineOnOff
        if (length(yLineOnOff))
          str <- paste(str, " yLineOnOff=", onoff(yLineOnOff), sep = "")
        windowingFunction <- from@windowingFunction
        if (length(windowingFunction))
          str <- paste(str, " windowingFunction=", windowingFunction, sep = "")
        smoothingWindow <- from@smoothingWindow
        if (length(smoothingWindow))
          str <- paste(str, " smoothingWindow=", smoothingWindow, sep = "")
        str
      })

setAs("character", "GraphTrackLine",
      function(from)
      {
        line <- new("GraphTrackLine", as(from, "TrackLine"))
        vals <- ucscParsePairs(from)
        type <- vals[["type"]]
        if (!(type %in% c("wiggle_0", "bedGraph")))
          stop("Unknown graph track type: ", type)
        line@type <- if (type == "wiggle_0") "wig" else "bedGraph"
        if (!is.na(vals["altColor"]))
          line@altColor <- as.integer(strsplit(vals[["altColor"]], ",")[[1]])
        if (!is.na(vals["autoScale"]))
          line@autoScale <- tolower(vals[["autoScale"]]) == "on"
        if (!is.na(vals["alwaysZero"]))
          line@alwaysZero <- tolower(vals[["alwaysZero"]]) == "on"
        if (!is.na(vals["gridDefault"]))
          line@gridDefault <- tolower(vals[["gridDefault"]]) == "on"
        if (!is.na(vals["maxHeightPixels"]))
          line@maxHeightPixels <-
            as.integer(strsplit(vals[["maxHeightPixels"]], ":")[[1]])
        if (!is.na(vals["graphType"]))
          line@graphType <- vals[["graphType"]]
        if (!is.na(vals["viewLimits"]))
          line@viewLimits <-
            as.numeric(strsplit(vals[["viewLimits"]], ":")[[1]])
        if (!is.na(vals["yLineMark"]))
          line@yLineMark <- as.numeric(vals[["yLineMark"]])
        if (!is.na(vals["yLineOnOff"]))
          line@yLineOnOff <- tolower(vals[["yLineOnOff"]]) == "on"
        if (!is.na(vals["windowingFunction"]))
          line@windowingFunction <- vals[["windowingFunction"]]
        if (!is.na(vals["smoothingWindow"]))
          line@smoothingWindow <- as.numeric(vals[["smoothingWindow"]])
        line
      })

setAs("BasicTrackLine", "GraphTrackLine",
      function(from) new("GraphTrackLine", from))

setAs("GraphTrackLine", "BasicTrackLine",
      function(from) new("BasicTrackLine", from))

setClass("UCSCData",
         representation(trackLine = "TrackLine"),
         prototype(trackLine = new("BasicTrackLine")),
         "GRanges")

UCSCData <- function(ranges, trackLine = NULL) {
  ucsc <- as(ranges, "UCSCData")
  ucsc@trackLine <- trackLine
  ucsc
}

setMethod("show", "UCSCData",
          function(object)
          {
            if (!is.null(object@trackLine@name))
              cat("UCSC track '", object@trackLine@name, "'\n", sep = "")
            callNextMethod()
          })

chooseGraphType <- function(from) {
    if (is(from, "GPos")) {
        return(if (is(from, "StitchedGPos")) "bedGraph" else "wig")
    }
  r <- ranges(from)
  type <- "bedGraph"
  ## decide whether compression is a good idea
  steps <- diff(sort(start(r)))
  if (length(unique(width(r))) == 1L && # all spans must be the same for WIG
      (length(unique(steps)) == 1L || # fixed-step makes sense
       ((3L * length(unique(width(r)))) < length(r) && # makes sense wrt size
        mean(steps) < 100))) # dense enough for UCSC efficiency
    type <- "wig"
  type
}

setAs("GRanges", "UCSCData", function(from) {
  line <- metadata(from)$trackLine
  if (is.null(line)) {
    if (is.numeric(score(from))) { # have numbers, let's plot them
      type <- chooseGraphType(from)
      line <- new("GraphTrackLine", type = type)
    } else {
      line <- new("BasicTrackLine")
      db <- unique(genome(from))
      if (length(db) == 1 && !is.na(db))
        line@db <- db
    }
  } else {
    metadata(from)$trackLine <- NULL
  }
  new("UCSCData", as(from, "GRanges"), trackLine = line)
})

## We want 'as(UCSCData, "GRanges", strict=FALSE)' to do the right thing (i.e.
## be a no-op) but as() won't do that if a coerce,UCSCData,GRanges method
## exists (this is a serious flaw in as() current design/implementation).
## The workaround is to support the 'strict=FALSE' case at the level of
## the coerce() method but setAs() doesn't let us do that so we use
## setMethod("coerce", ...) to define the method.
setMethod("coerce", c("UCSCData", "GRanges"),
  function(from, to="GRanges", strict=TRUE) {
  if (strict) {
    gr <- new("GRanges")
    for (what in slotNames(gr))
      slot(gr, what) <- slot(from, what)
    metadata(gr)$trackLine <- from@trackLine
    gr
  } else from
})

splitUCSCData <- function(x, f, drop=FALSE, ...) {
  GRangesList(
    lapply(split(seq_along(x), f, drop=drop, ...),
           function(i) x[i]),
    compress=FALSE
  )
}

setMethod("split", "UCSCData", splitUCSCData)
setMethod("split", c("UCSCData", "Vector"), splitUCSCData)

setClass("UCSCFile", contains = "BiocFile")

UCSCFile <- function(resource) {
  new("UCSCFile", resource = resource)
}

## the 'ucsc' format is a meta format with a track line followed by
## features formatted as 'wig', 'bed', 'bed15', 'bedGraph', 'gff', or
## really any text track format.

setGeneric("export.ucsc",
           function(object, con, ...) standardGeneric("export.ucsc"))

setMethod("export.ucsc", c("ANY", "BiocFile"),
          function(object, con, subformat = "auto", ...)
          {
            if (subformat == "auto" && !is(con, "UCSCFile"))
              subformat <- fileFormat(con)
            export(object, UCSCFile(resource(con)), subformat=subformat, ...)
          })

setMethod("export.ucsc", c("ANY", "ANY"),
          function(object, con, ...)
          {
            export(object, con, "ucsc", ...)
          })

.export_SimpleGRangesList_BiocFile <- function(object, con, format, ...) {
  export(object, UCSCFile(resource(con)), subformat = fileFormat(con), ...)
}

setMethod("export", c("GRangesList", "UCSCFile"),
          function(object, con, format, append = FALSE, index = FALSE, ...)
          {
            if (isTRUE(index) && length(object) > 1)
              stop("Cannot index multiple tracks in a single file")
            trackNames <- names(object)
            if (is.null(trackNames))
              trackNames <- paste("R Track", seq_len(length(object)))
            ucsc <- unlist(lapply(object, is, "UCSCData"))
            lines <- unlist(lapply(object[ucsc], slot, "trackLine"))
            trackNames[ucsc] <- as.character(sapply(lines, slot, "name"))
            tracks <- vector("list", length(object))
            for (i in seq_len(length(object))) {
              tracks[[i]] <- export(object[[i]], con, name = trackNames[i],
                                    append = append, index = index, ...)
              append <- TRUE
            }
            BiocFileList(tracks)
          })

trackLineClass <- function(subformat)
{
  subformat <- tolower(subformat)
  if (subformat == "wig" || subformat == "bedgraph")
    "GraphTrackLine"
  else if (subformat == "bed15")
    "Bed15TrackLine"
  else "BasicTrackLine"
}

setMethod("fileFormat", "TrackLine", function(x) "bed")
setMethod("fileFormat", "GraphTrackLine", function(x) x@type)

setMethod("bestFileFormat", c("UCSCData", "ANY"), function(x, dest) {
  fileFormat(x@trackLine)
})

setMethod("export", c("ANY", "UCSCFile"),
          function(object, con, format, ...)
          {
            cl <- class(object)
            track <- try(as(object, "GRanges"), silent = TRUE)
            if (class(track) == "try-error") {
              track <- try(as(object, "SimpleGRangesList"), silent = TRUE)
              if (is(track, "try-error"))
                stop("cannot export object of class '", cl, "': ", track)
            }
            object <- track
            callGeneric()
          })

setMethod("export", c("GenomicRanges", "UCSCFile"),
          function(object, con, format, ...)
          {
            object <- as(object, "UCSCData")
            callGeneric()
           })

setMethod("export", c("UCSCData", "UCSCFile"),
          function(object, con, format, subformat = "auto", append = FALSE,
                   index = FALSE, ...)
          {
            auto <- FALSE
            if (subformat == "auto") {
              auto <- TRUE
              subformat <- bestFileFormat(object, con)
            }
            graphFormat <- tolower(subformat) %in% c("wig", "bedgraph")
            if (graphFormat) {
              strand <- as.character(strand(object))
              strand[is.na(strand)] <- "NA"
              isStrandDisjoint <- function(track) {
                all(tapply(ranges(track), seqnames(track), function(r) {
                  isDisjoint(r) && all(width(r) > 0)
                }), na.rm=TRUE)
              }
              if (!all(unlist(lapply(split(object, strand), isStrandDisjoint))))
              {
                if (auto) {
                  subformat <- "bed"
                  graphFormat <- FALSE
                }
                else stop("Track not compatible with WIG/bedGraph: ",
                          "Overlapping features must be on separate strands",
                          " and every feature width must be positive")
              }
            }
            lineClass <- trackLineClass(subformat)
            if (!is(object@trackLine, lineClass))
              object@trackLine <- as(object@trackLine, lineClass)
            if (is(object@trackLine, "GraphTrackLine"))
              object@trackLine@type <- subformat
            args <- list(...)
            lineArgs <- names(args) %in% slotNames(lineClass)
            for (argName in names(args)[lineArgs])
              slot(object@trackLine, argName) <- args[[argName]]
            if (is(object@trackLine, "BasicTrackLine") &&
                length(object@trackLine@offset))
              ranges(object) <- shift(ranges(object), -object@trackLine@offset)
            trackLine <- NULL
            if (graphFormat) {
              strand <- as.character(strand(object))
              strand[is.na(strand)] <- "NA"
              if (!all(strand[1] == strand)) {
                nameMap <- c("+" = "p", "-" = "m", "NA" = "NA")
                strand <- factor(strand)
                name <- paste(object@trackLine@name, nameMap[levels(strand)])
                tracks <- split(object, strand)
                export(tracks, con, subformat, append,
                       trackNames = name, ...)
                return()
              }
            } else if (subformat == "bed15") {
              if (is.null(object@trackLine@expNames))
                object@trackLine@expNames <- colnames(object)
              trackLine <- object@trackLine
            }
            file <- con
            m <- manager()
            con <- connection(m, con, if (append) "a" else "w")
            cat(as(object@trackLine, "character"), "\n", file=con, sep = "")
            do.call(export, c(list(as(object, "GRanges"), con, subformat),
                              args[!lineArgs], trackLine = trackLine))
            release(m, con)
            if (index)
              indexTrack(BiocIO::FileForFormat(resource(file), subformat),
                         skip = 1L)
            else invisible(file)
          })

setGeneric("import.ucsc", function(con, ...) standardGeneric("import.ucsc"))

setMethod("import.ucsc", "ANY",
          function(con, ...)
          {
            import(con, "ucsc", ...)
          })

setMethod("import.ucsc", "BiocFile",
          function(con, subformat = "auto", ...)
          {
            if (!is(con, "UCSCFile")) {
              format <- fileFormat(con)
              if (subformat != "auto" && format != subformat)
                stop("Attempt to import '", class(con), "' as ", subformat)
              subformat <- format
            }
            import.ucsc(resource(con), subformat = subformat, ...)
          })

parseFormatFromTrackLine <- function(x) {
  if (!grepl("type=", x))
    NULL
  else {
    type <- sub(".*type=\"(.*?)\".*", "\\1", x)
    if (type == "array")
      "bed15"
    else if (type == "wiggle_0")
      "wig"
    else type
  }
}

setMethod("import", "UCSCFile",
          function(con, format, text, subformat = "auto", drop = FALSE,
                   genome = NA, ...)
          {
            lines <- readLines(resource(con), warn = FALSE)
            tracks <- grep("^track", lines)
            trackLines <- lines[tracks]
            starts <- tracks + 1L
            ends <- c(tail(tracks, -1) - 1L, length(lines))
            makeTrackSet <- function(i)
            {
              if (subformat == "auto") {
                subformat <- parseFormatFromTrackLine(trackLines[i])
                if (is.null(subformat)) {
                  p <- resourceDescription(con)
                  subformat <- file_ext(p)
                }
              }
              line <- as(trackLines[i], trackLineClass(subformat))
              if (starts[i] <= ends[i]) {
                text <- window(lines, starts[i], ends[i])
              } else {
                text <- character()
              }
              if (is.na(genome) && is(line, "BasicTrackLine") &&
                  length(line@db))
                genome <- line@db
              if (subformat == "bed15") { # need to pass track line
                ucsc <- import(format = "bed15", text = text,
                               trackLine = line,
                               genome = genome, ...)
              } else {
                ucsc <- import(format = subformat, text = text,
                               genome = genome, ...)
              }
              if (is(line, "BasicTrackLine") && length(line@offset))
                ranges(ucsc) <- shift(ranges(ucsc), line@offset)
              ucsc <- as(ucsc, "UCSCData", FALSE)
              ucsc@trackLine <- line
              ucsc
            }
            tsets <- lapply(seq_along(trackLines), makeTrackSet)
            trackNames <- sapply(tsets, function(x) x@trackLine@name)
            if (!any(is.na(trackNames)))
              names(tsets) <- trackNames
            if (drop && length(tsets) == 1)
              return(tsets[[1]])
            GRangesList(tsets, compress=FALSE)
          })

############ INTERNAL API ############

## every cgi variable is stored in the 'cart'
setClass("ucscCart", contains = "character")

setGeneric("ucscCart", function(object, ...) standardGeneric("ucscCart"))

setMethod("ucscCart", "UCSCSession",
          function(object, form = ucscForm(activeView(object)))
          {
            node <- ucscGet(object, "cart", form)
            raw <- xmlValue(getNodeSet(node, "//pre/text()")[[1]])
            lines <- strsplit(raw, "\n")[[1]]
            fields <- strsplit(lines, " ")
            pairs <- fields[sapply(fields, length) == 2]
            mat <- matrix(unlist(pairs), nrow = 2)
            vals <- mat[2,]
            names(vals) <- mat[1,]
            vals[["db"]] <- sub(".*_", "", vals[["db"]])
            new("ucscCart", vals)
          })
setMethod("ucscCart", "UCSCView",
          function(object)
          {
            ucscCart(object@session, ucscForm(object))
          })

setMethod("genome", "ucscCart", function(x) x[["db"]])

setMethod("range", "ucscCart",
          function(x, ..., na.rm)
          {
            pos <- x["position"]
            posSplit <- strsplit(pos, ":")[[1]]
            range <- as.numeric(gsub(",", "", strsplit(posSplit[2], "-")[[1]]))
            GRangesForUCSCGenome(x[["db"]], posSplit[1],
                                 IRanges(range[1], range[2]))
          })

### track information

setClass("ucscTracks",
         representation(ids = "character", modes = "character"))

setGeneric("ucscTracks", function(object, ...) standardGeneric("ucscTracks"))

setMethod("ucscTracks", "UCSCSession",
          function(object, form = list(pix=800))
          {
            tracks <- ucscGet(object, "tracks", form)
            nodes <- getNodeSet(tracks, "//select/option[@selected]/text()")
            trackModes <- sapply(nodes, xmlValue)
            nodes <- getNodeSet(tracks, "//select/@name")
            trackIds <- unlist(nodes)
            ##trackIds <- sapply(nodes, xmlValue)
            nodes <- getNodeSet(tracks, "//select/../a[not(@class='pennantIconText')]/text()")
            nms <- sapply(nodes, xmlValue)
            names(trackIds) <- sub("^ ", "", nms)
            new("ucscTracks", ids = trackIds, modes = trackModes)
          })

setMethod("ucscTracks", "UCSCView",
          function(object)
          {
            ucscTracks(object@session, ucscForm(object))
          })

setMethod("ucscTrackModes", "ucscTracks",
          function(object)
          {
            modes <- object@modes
            labels <- names(object@ids)
            names(modes) <- object@ids
            ucscTrackModes(modes, labels)
          })

ucscGenomes <- function(organism=FALSE) {
  stopifnot(isTRUEorFALSE(organism))
  names <- c("db", "species", "date", "name", "organism")
  url <- RestUri("https://api.genome.ucsc.edu/")
  response <- read(url$list$ucscGenomes)
  genomes <- response[[5]]
  genomeNames <- names(genomes)
  listOfDf <- Map(function(name, x) {
    date <- sub("\\s*\\([^\\)]+\\)", "", x$description)
    data.frame(name, x$genome, date, x$sourceName, x$scientificName)
  }, genomeNames, genomes)
  df <- do.call(rbind, listOfDf)
  names(df) <- names
  rownames(df) <- NULL
  if (!organism) df$organism <- NULL
  df
}

# form creation

setGeneric("ucscForm", function(object, ...) standardGeneric("ucscForm"))

setMethod("ucscForm", "IntegerRangesList",
          function(object)
          {
            form <- list()
            genome <- singleGenome(genome(object))
            if (!is.na(genome))
              form <- c(form, db = genome)
            chrom <- space(object)
            if (!is.null(chrom)) {
              if (!length(chrom))
                chrom <- levels(chrom)[1]
              scipen <- getOption("scipen")
              options(scipen = 100) # prevent use of scientific notation
              on.exit(options(scipen = scipen))
              position <- chrom
              if (length(unlist(start(object))))
                position <- paste(chrom, ":",
                                  unlist(start(object)), "-",
                                  unlist(end(object)), sep = "")
              form <- c(form, position = position)
            }
            form
          })
setMethod("ucscForm", "GRanges",
          function(object)
          {
            scipen <- getOption("scipen")
            options(scipen = 100) # prevent use of scientific notation
            on.exit(options(scipen = scipen))
            form <- list()
            genome <- singleGenome(genome(object))
            if (!is.na(genome))
              form <- c(form, db = genome)
            if (length(object) > 0L) {
              object <- object[1]
              c(form, position = paste(seqnames(object), ":",
                        unlist(start(object)), "-",
                        unlist(end(object)), sep = ""))
            } else form
          })

setMethod("ucscForm", "UCSCTrackModes",
          function(object)
          {
            as.list(object)
          })
setMethod("ucscForm", "UCSCView",
          function(object)
          {
            if (length(object@hgsid))
              list(hgsid = as.character(object@hgsid))
            else list()
          })
setOldClass("form_file")
setMethod("ucscForm", "form_file",
          function(object, genome = NA_character_, ...)
          {
            form <- list(Submit = "Submit", hgt.customFile = object)
            if (!is.na(genome))
              form <- c(form, db = genome)
            form
          })
setMethod("ucscForm", "SimpleGRangesList",
          function(object, format, ...)
          {
            path <- tempfile()
            export(object, path, format = "ucsc", subformat = format, ...)
            upload <- upload_file(path, type = "text/plain")
            genome <- singleGenome(genome(object))
            ucscForm(upload, genome)
          })
setMethod("ucscForm", "BiocFile",
          function(object, genome, ...)
          {
            upload <- upload_file(path(object), type = "text/plain")
            ucscForm(upload, genome)
          })

setMethod("ucscForm", "NULL", function(object) list())

# Transforming to a cookie string

setGeneric("ucscCookie", function(object, ...) standardGeneric("ucscCookie"))
setMethod("ucscCookie", "UCSCSession",
          function(object)
          {
            object@hguid
          })

# HTTP wrappers

# URL constants for UCSC
ucscURLTable <- c(gateway = "hgGateway", tracks = "hgTracks",
                  custom = "hgCustom", tables = "hgTables",
                  cart = "cartDump")

ucscURL <-
  function(object, key, force=TRUE)
  {
    path <- ucscURLTable[key]
    if (is.na(path))
        stop("Key '", key, "' does not match a known URL")
    if (force && key == "gateway") {
        path <- paste0(path, '?redirect="manual"')
    }
    paste(object@url, path, sep="")
  }

# convenience wrappers for _initialized_ sessions
ucscShow <- function(object, key, .form = list(), ...)
  httpShow(ucscURL(object, key), .form, ...)
ucscPost <- function(object, key, .form = list()) {
  cookies <- set_cookies(hguid=gsub("^hguid=", "", ucscCookie(object)))
  rtracklayerPOST(ucscURL(object, key), cookies, body = .form)
}
ucscGet <- function(object, key, .form = list()) {
  cookies <- set_cookies(hguid=gsub("^hguid=", "", ucscCookie(object)))
  rtracklayerGET(ucscURL(object, key), cookies, query = .form)
}

### =========================================================================
### Tabix support
### -------------------------------------------------------------------------
###

setMethod("import", "TabixFile",
          function(con, format, text,
                   which = if (is.na(genome)) NULL
                           else as(seqinfoForGenome(genome), "GenomicRanges"),
                   genome = NA, header = TRUE, ...)
          {
            if (missing(format)) {
                  format <- file_ext(file_path_sans_ext(path(con)))
            }
            m <- manager()
            buffer <- queryForResource(m, con, which, header = header)
            on.exit(release(m, buffer))
            file <- try(BiocIO::FileForFormat(buffer, format), silent = TRUE)
            if (is(file, "try-error")) {
              tabixHeader <- headerTabix(con)
              args <- list(...)
              if (is.null(args$comment.char))
                args$comment.char <- tabixHeader$comment
              skip <- tabixHeader$skip
              if (header) {
                skip <- skip - 1L
              }
              do.call(import.tabSeparated,
                      c(list(buffer), genome = genome, tabixHeader$indexColumns,
                        skip = skip, header = header, args))
            } else import(file, genome = genome, ...)
          })

setGeneric("exportToTabix",
           function(object, con, ...) standardGeneric("exportToTabix"))

setMethod("exportToTabix", c("ANY", "character"),
          function(object, con, ...) {
            con <- TabSeparatedFile(con)
            export(sort(object, ignore.strand=TRUE), con,
                   row.names=FALSE, col.names=TRUE, ...)
            indexTrack(con, seq=1L, start=2L, end=3L, skip=1L)
          })

### =========================================================================
### GFF (General Feature Format) support (all three versions, plus GTF)
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Classes
###

setClass("GFFFile", contains = "BiocFile")

## private
GFFFile <- function(resource, version = c("", "1", "2", "3")) {
  version <- match.arg(version)
  new(gffFileClass(version), resource = resource)
}

setClass("GFF1File", contains = "GFFFile")
GFF1File <- function(resource) {
  GFFFile(resource, "1")
}

setClass("GFF2File", contains = "GFFFile")
GFF2File <- function(resource) {
  GFFFile(resource, "2")
}

setClass("GFF3File", contains = "GFFFile")
GFF3File <- function(resource) {
  GFFFile(resource, "3")
}

setClass("GTFFile", contains = "GFF2File")
GTFFile <- function(resource) {
  new("GTFFile", GFF2File(resource))
}

setClass("GVFFile", contains = "GFF3File")
GVFFile <- function(resource) {
  new("GVFFile", GFF3File(resource))
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export
###

setGeneric("export.gff",
           function(object, con, ...) standardGeneric("export.gff"))

setMethod("export.gff", "ANY",
          function(object, con, ...)
          {
            export(object, con, ...)
          })

setMethod("export", c("ANY", "GFFFile"),
          function(object, con, format, ...)
          {
            if (hasMethod("asGFF", class(object)))
              object <- asGFF(object)
            res <- try(as(object, "GRanges"), silent = TRUE)
            if (is(res, "try-error")) {
              res <- try(as(object, "GenomicRangedDataList"), silent = TRUE)
              if (is(res, "try-error"))
                stop("cannot export object of class '", class(object), "'")
            }
            object <- res
            if (!missing(format))
              checkArgFormat(con, format)
            export(object, con, ...)
          })

setMethod("export", c("CompressedGRangesList", "GFFFile"),
          function(object, con, format, ...)
          {
            object <- asGFF(object)
            callGeneric()
          }
          )

setMethod("export", c("GRangesList", "GTFFile"),
          function(object, con, format, ...) {
              stop("export of GRangesList to GTF is not yet supported")
              ## there is a start on asGTF() later in this file
              ## object <- asGTF(object)
              ## callGeneric()
          }
          )

setMethod("export", c("GenomicRanges", "GFFFile"),
          function(object, con, format, version = c("1", "2", "3"),
                   source = "rtracklayer", append = FALSE, index = FALSE)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            if (!missing(version) || !length(gffFileVersion(con)))
              con <- asGFFVersion(con, match.arg(version))
            version <- gffFileVersion(con)
            
            file <- con
            con <- resource(con)
            
            if (!append) {
              cat("", file = con) # clear any existing file
              gffComment(con, "gff-version", version)
              sourceVersion <- try(packageVersion(source), TRUE)
              if (!inherits(sourceVersion, "try-error"))
                gffComment(con, "source-version", source, sourceVersion)
              gffComment(con, "date", base::format(Sys.time(), "%Y-%m-%d"))
              genome <- singleGenome(genome(object))
              if (!is.na(genome))
                gffComment(con, "genome-build", paste(".", genome, sep = "\t"))
            }
            
            if (index)
              object <- sortBySeqnameAndStart(object)

            seqname <- seqnames(object)
            if (is.null(mcols(object)$ID))
              mcols(object)$ID <- names(object)
            if (version == "3")
              seqname <- urlEncode(seqname, "a-zA-Z0-9.:^*$@!+_?|-")
            if (!is.null(mcols(object)$source) && missing(source))
              source <- mcols(object)$source
            else source <- rep(source, length(object))
            if (version == "3")
              source <- urlEncode(source, "\t\n\r;=%&,", FALSE)
            feature <- mcols(object)$type
            if (is.null(feature))
              feature <- rep("sequence_feature", length(object))
            score <- score(object)
            if (is.null(score)) {
              score <- rep(NA_real_, length(object))
            } else {
              if (!("score" %in% colnames(mcols(object))))
                ## avoid outputting as attribute
                colnames(mcols(object))[1] <- "score" 
            }
            strand <- strand(object)
            if (is.null(strand))
                strand <- rep(strand(NA_character_), length(object))
            strand[strand == "*"] <- NA_integer_
            frame <- mcols(object)$phase
            if (is.null(frame)) {
              frame <- rep(NA_integer_, length(object))
              if ("CDS" %in% feature)
                warning(wmsg("The phase information is missing. ",
                             "The written file will contain CDS with ",
                             "no phase information."))
            } else {
              if (anyNA(frame[feature %in% "CDS"]))
                warning(wmsg("The phase information is missing for some CDS. ",
                             "The written file will contain some CDS with ",
                             "no phase information."))
            }
            
            table <- data.frame(seqname, source, feature, start(object),
                                end(object), score, strand, frame)

            attrs <- NULL
            if (version == "1") {
              attrs <- mcols(object)$group
              if (is.null(attrs))
                attrs <- as.vector(seqname)
            } else {
              builtin <- c("type", "score", "phase", "source")
              custom <- setdiff(colnames(mcols(object)), builtin)
              if (length(custom)) {
                if (version == "3") tvsep <- "=" else tvsep <- " "
                attrs <- mcols(object)
                attrs <- as.data.frame(sapply(custom, function(name) {
                  x <- attrs[[name]]
                  x_flat <- if (is(x, "List")) unlist(x, use.names=FALSE) else x
                  x_char <- as.character(x_flat)
                  x_char <- sub(" *$", "", sub("^ *", "", as.character(x_char)))
                  if (version == "3")
                    x_char <- urlEncode(x_char, "%\t\n\r;=&,", FALSE)
                  if (is(x, "List")) {
                    x_char[is.na(x_char)] <- "."
                    x_char <- pasteCollapse(relist(x_char, x))
                    x_char[elementNROWS(x) == 0] <- NA
                  }
                  ## FIXME: add option so these become "." instead of removing
                  x_char[is.na(x_char)] <- "\r"
                  if (!is.numeric(x_flat) && version != "3")
                    x_char <- paste0("\"", x_char, "\"")
                  paste(name, x_char, sep = tvsep)
                }, simplify = FALSE))
                if (version == "3") sep <- ";" else sep <- "; "
                attrs <- do.call(paste, c(attrs, sep = sep))
                attrs <- gsub("[^;]*?\r\"?(;|$)", "", attrs)
                attrs[nchar(attrs) == 0] <- NA
              }
            }
            
            scipen <- getOption("scipen")
            options(scipen = 100) # prevent use of scientific notation
            on.exit(options(scipen = scipen))
            
            if (!is.null(attrs)) { # write out the rows with attributes first
              write.table(cbind(table, attrs)[!is.na(attrs),], con, sep = "\t",
                          na = ".", quote = FALSE, col.names = FALSE,
                          row.names = FALSE, append = TRUE)
              table <- table[is.na(attrs),]
            }
            
            write.table(table, con, sep = "\t", na = ".", quote = FALSE,
                        col.names = FALSE, row.names = FALSE, append = TRUE)
            if (index)
              indexTrack(file)
            invisible(NULL)
          })

setMethod("export", c("SimpleGRangesList", "GFFFile"),
          .export_SimpleGRangesList_BiocFile)

setGeneric("export.gff1",
           function(object, con, ...) standardGeneric("export.gff1"))
setMethod("export.gff1", "ANY",
          function(object, con, ...) export(object, con, "gff1", ...))

setGeneric("export.gff2",
           function(object, con, ...) standardGeneric("export.gff2"))
setMethod("export.gff2", "ANY",
          function(object, con, ...) export(object, con, "gff2", ...))

setGeneric("export.gff3",
           function(object, con, ...) standardGeneric("export.gff3"))
setMethod("export.gff3", "ANY",
          function(object, con, ...) export(object, con, "gff3", ...))

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import
###

setGeneric("import.gff", function(con, ...) standardGeneric("import.gff"))

setMethod("import.gff", "ANY",
          function(con, ...)
          {
            import(con, "gff", ...)
          })

setMethod("import", "GFFFile",
          function(con, format, text, version = c("", "1", "2", "3"),
                   genome = NA, colnames = NULL,
                   which = NULL, feature.type = NULL,
                   sequenceRegionsAsSeqinfo = FALSE)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            if (!missing(version))
              con <- asGFFVersion(con, match.arg(version))
            stopifnot(isTRUEorFALSE(sequenceRegionsAsSeqinfo))
            
            ## download the file first if it's remote
            if (is.character(resource(con))) {
                uri <- .parseURI(resource(con))
                if (uri$scheme %in% c("ftp", "http")) {
                    destfile <- tempfile()
                    download.file(resource(con), destfile)
                    con@resource <- destfile
                }
            }

            m <- manager()
            sniff_con <- connection(m, con, "r")
            on.exit(release(m, sniff_con))
            sniffed <- .sniffGFFVersion(sniff_con)
            version <- gffFileVersion(con)
            if (!length(version)) {
              if (is.null(sniffed))
                sniffed <- "1"
              con <- asGFFVersion(con, sniffed)
            }
            
            if (length(version) && !is.null(sniffed) &&
                !identical(sniffed, version))
              warning("gff-version directive indicates version is ", sniffed,
                      ", not ", version)

            if (is(genome, "Seqinfo") && length(genome) == 0L) {
                genome <- NA_character_
            }
               
            if (is.na(genome)) {
              genome <- genome(con)
              if (is.null(genome))
                  genome <- NA
            }
            
### FIXME: a queryForLines() function would be more efficient

            ## Temporarily disable use of Tabix Index.
            ## TODO: Restore use of Tabix Index!
            #con <- queryForResource(m, con, which)
            resource <- queryForResource(m, con)
            on.exit(release(m, resource), add=TRUE)
            ans <- readGFFAsGRanges(resource,
                                    version=version,
                                    colnames=colnames,
                                    filter=list(type=feature.type),
                                    genome=genome,
                                    sequenceRegionsAsSeqinfo=
                                        sequenceRegionsAsSeqinfo,
                                    speciesAsMetadata=TRUE)
            if (!attr(resource, "usedWhich") && !is.null(which))
                ans <- subsetByOverlaps(ans, which)
            ans
          })

setGeneric("import.gff1",
           function(con, ...) standardGeneric("import.gff1"))
setMethod("import.gff1", "ANY",
          function(con, ...) import(con, "gff1", ...))

setGeneric("import.gff2",
           function(con, ...) standardGeneric("import.gff2"))
setMethod("import.gff2", "ANY",
          function(con, ...) import(con, "gff2", ...))

setGeneric("import.gff3",
           function(con, ...) standardGeneric("import.gff3"))
setMethod("import.gff3", "ANY",
          function(con, ...) import(con, "gff3", ...))

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### DNAStringSet from fasta data
###



### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Coercion
###

setGeneric("asGFF", function(x, ...) standardGeneric("asGFF"))

setMethod("asGFF", "GRangesList",
          function(x, parentType = "mRNA", childType = "exon") {
            parent_range <- range(x)
            if (!all(elementNROWS(parent_range) == 1))
              stop("Elements in a group must be on same sequence and strand")
            parents <- unlist(parent_range, use.names = FALSE)
            children <- unlist(x, use.names = FALSE)
            makeId <- function(x, prefix) {
                paste(prefix, seq_len(length(x)), sep = "")
            }
            parentIds <- makeId(parents, parentType)
            values(parents)$type <- parentType
            values(parents)$ID <- parentIds
            values(parents)$Name <- names(x)
            values(children)$type <- childType
            values(children)$ID <- makeId(children, childType)
            values(children)$Name <- names(children)
            values(children)$Parent <- rep.int(parentIds, elementNROWS(x))
            allColumns <- union(colnames(values(parents)),
                                colnames(values(children)))
            values(children) <- rectifyDataFrame(values(children), allColumns)
            values(parents) <- rectifyDataFrame(values(parents), allColumns)
            c(parents, children)
          })

rectifyDataFrame <- function(x, allColumns) {
    x[setdiff(allColumns, colnames(x))] <- DataFrame(NA)
    x[allColumns]
}

### FIXME: We wrote this but never tested it, and it is not yet
### used. People should use GFF3 instead of this.
###
### KNOWN ISSUES:
### 1) The stop codon should not be included in the CDS (annoying)
### 2) pmapFromTranscripts() does not yet support our usage of it
### 3) Needs to move to GenomicFeatures

setGeneric("asGTF", function(x, ...) standardGeneric("asGTF"))

frame <- function(x) {
    cs <- cumsum(width(x))
    ucs <- unlist(cs, use.names=FALSE)
    ucs[end(PartitioningByEnd(x))] <- 0L
    ucs <- c(0L, head(ucs, -1L))
    ucs %% 3L
}

setMethod("asGTF", "GRangesList",
          function(x) {
              tx_ids <- names(x)
              if (is.null(tx_ids)) {
                  tx_ids <- seq_along(x)
              }
              processFeatures <- function(f) {
                  ans <- unlist(f, use.names=FALSE)
                  ans$frame <- frame(f)
                  if (is.null(ans$gene_id)) {
                      ans$gene_id <- ""
                  }
                  if (is.null(ans$transcript_id)) {
                      ans$transcript_id <- tx_ids[togroup(f)]
                  }
                  ans
              }
              start_codon_tx <-
                  GenomicFeatures::pmapFromTranscripts(IRanges(1L, 3L), x)
              start_codon <- processFeatures(start_codon_tx)
              mcols(start_codon)$type <- "start_codon"
              stop_ranges <- IRanges(end=sum(width(x)), width=3L)
              stop_codon_tx <-
                  GenomicFeatures::pmapFromTranscripts(stop_ranges, x)
              stop_codon <- processFeatures(stop_codon_tx)
              mcols(stop_codon)$type <- "stop_codon"
              codons <- c(start_codon, stop_codon)
              cds <- processFeatures(x)
              mcols(cds)$type <- "CDS"
              values(codons) <- rectifyDataFrame(values(codons), colnames(cds))
              c(codons, cds)
          })

## setMethod("asGTF", "TxDb",
##           function(x, by) {
##               cds <- cds(x, columns="tx_id")
##               cds <- cds[togroup(cds$tx_id)]
##               cds$transcript_id <- unlist(cds$tx_id)
##               cds$tx_id <- NULL
##               txGene <- transcriptsBy(x)
##               txToGene <- setNames(names(txGene)[togroup(txGene)],
##                                    unlist(txGene, use.names=FALSE)$tx_id)
##               cds$gene_id <- txToGene[cds$transcript_id]
##               processUTRs <- function(utr) {
##                   ans <- unlist(utr, use.names=FALSE)
##                   ans$transcript_id <- names(utr)[togroup(utr)]
##                   ans$gene_id <- txToGene[ans$transcript_id]
##                   ans
##               }
##               three_utr <- processUTRs(threeUTRsByTranscript(x))
##               three_utr$type <- "3UTR"
##               five_utr <- processUTRs(fiveUTRsByTranscript(x))
##               five_utr$type <- "5UTR"
##               c(asGTF(split(cds, cds$transcript_id)), five_utr, three_utr)
##           })


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Utilities
###

scanGFFDirectives <- function(con, tag = NULL) {
  m <- manager()
  con <- connection(m, con, "r")
  on.exit(release(m, con))
  directives <- character()
  lines <- line <- readLines(con, n = 1)
  while(grepl("^#", line)) {
    if (grepl("^##", line)) {
        directives <- c(directives, line)
    }
    line <- readLines(con, n = 1)
    if (length(line) == 0L)
        break
    lines <- c(lines, line)
  }
  pushBack(lines, con)
  sub("^[^[:space:]]* ", "", grep(paste0("^##", tag), directives, value = TRUE))
}

gffGenomeBuild <- function(x) {
  genome_build <- scanGFFDirectives(x, "genome-build")
  unlist(strsplit(genome_build, "\t", fixed = TRUE))
}

setMethod("provider", "GFFFile", function(x) {
  gffGenomeBuild(x)[1]
})

setMethod("providerVersion", "GFFFile", function(x) {
  gffGenomeBuild(x)[2]
})
setMethod("genome", "GFFFile", function(x) providerVersion(x))

gffComment <- function(con, ...) 
  cat("##", paste(...), "\n", sep = "", file = con, append = TRUE)

.sniffGFFVersion <- function(con) {
  version <- NULL
  lines <- line <- readLines(con, n = 1)
  while(grepl("^#", line)) {
    if (grepl("^##gff-version", line)) {
      version <- sub("^##gff-version *", "", line)
      break
    }
    line <- readLines(con, n = 1)
    lines <- c(lines, line)
  }
  pushBack(lines, con)
  version
}

gffFileClass <- function(version) {
  paste("GFF", version, "File", sep = "")
}

gffFileVersion <- function(file) {
  versions <- c("1", "2", "3")
  unlist(Filter(function(v) is(file, gffFileClass(v)), versions))
}

asGFFVersion <- function(con, version) {
  if (!is(con, gffFileClass(version))) {
    if (class(con) != "GFFFile")
      warning("Treating a '", class(con), "' as GFF version '", version, "'")
    con <- GFFFile(resource(con), version)
  }
  con
}

NCBI_TAX_URL <- "http://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi"

isNCBISpeciesURL <- function(url) {
    substring(url, 1L, nchar(NCBI_TAX_URL)) == NCBI_TAX_URL
}

metadataFromNCBI <- function(url) {
    html <- rtracklayerGET(url)
    list("Taxonomy ID"=parseTaxonomyIdFromNCBI(html),
         "Organism"=parseOrganismFromNCBI(html))
}

parseTaxonomyIdFromNCBI <- function(html) {
    unlist(getNodeSet(html, "//input[@name='old_id']/@value"), use.names=FALSE)
}

taxonomyIdFromNCBI <- function(species) {
    url <- paste0(NCBI_TAX_URL, "?name=", URLencode(species))
    parseTaxonomyIdFromNCBI(rtracklayerGET(url))
}

parseOrganismFromNCBI <- function(html) {
    title <- xmlValue(getNodeSet(html, "//title/text()")[[1L]])
    sub(".*\\((.*?)\\).*", "\\1", title)
}

speciesFromNCBI <- function(id) {
    url <- paste0(NCBI_TAX_URL, "?id=", URLencode(id))
    parseOrganismFromNCBI(rtracklayerGET(url))
}

### =========================================================================
### Arbitrary track table support
### -------------------------------------------------------------------------
###
### At this point, this is internal, in support of the tabix stuff
###

setClass("TabSeparatedFile", contains = "BiocFile")
TabSeparatedFile <- function(resource) {
  new("TabSeparatedFile", resource = resource)
}

setGeneric("import.tabSeparated",
           function(con, genome = NA, seqnames = 1L, start = 2L, end = 3L, ...)
           standardGeneric("import.tabSeparated"),
           signature = "con")

setMethod("import.tabSeparated", "character_OR_connection",
          function(con, genome = NA, seqnames = 1L, start = 2L, end = 3L, ...)
          {
            tab <- read.table(con, sep = "\t", ...)
            ans <- GRanges(tab[[seqnames]], IRanges(tab[[start]], tab[[end]]),
                           strand=Rle(strand("*"), nrow(tab)),
                           tab[-c(seqnames, start, end)])
            metadata(ans) <- list(genome = genome)
            ans
          })

setGeneric("export.tabSeparated",
           function(object, con, ...) standardGeneric("export.tabSeparated"))

setMethod("export.tabSeparated", "ANY",
          function(object, con, ...) {
            export(object, con, "tabSeparated", ...)
          })

setMethod("export", c("ANY", "TabSeparatedFile"),
          function(object, con, ...) {
            df <- as.data.frame(object)
            write.table(df, path(con), sep="\t", ...)
          })

setMethod("export", c("GenomicRanges", "TabSeparatedFile"),
          function(object, con, ...) {
            object <- as.data.frame(object, row.names=NULL)
            object$width <- NULL
            export(object, con, ...)
          })

test_trackhub <- function() {
    test_trackhub_path <- system.file("tests", "trackhub", package = "rtracklayer")

    correct_trackhub_uri <- test_trackhub_path
    correct_trackhub_genome <- "hg19"
    correct_trackhub_length <- 1L
    correct_hub <- "test_hub"
    correct_shortLabel <- "test_hub"
    correct_longLabel <- "test_hub"
    correct_genomesFile <- "genomes.txt"
    correct_email <- "user@domain.com"
    correct_descriptionUrl <- "http://www.somedomain.com/articles/h19"
    correct_trackDb <- "hg19/trackDb.txt"

    ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ### TEST TrackHub Class
    ###

    th <- TrackHub(test_trackhub_path)

    ## TEST: uri
    checkIdentical(uri(th), correct_trackhub_uri)

    ## TEST: genome
    checkIdentical(genome(th), correct_trackhub_genome)

    ## TEST: length
    checkIdentical(length(th), correct_trackhub_length)

    # TEST: hub
    checkIdentical(hub(th), correct_hub)

    # TEST: shortLabel
    checkIdentical(shortLabel(th), correct_shortLabel)

    # TEST: longLabel
    checkIdentical(longLabel(th), correct_longLabel)

    # TEST: genomesFile
    checkIdentical(genomesFile(th), correct_genomesFile)

    # TEST: email
    checkIdentical(email(th), correct_email)

    # TEST: descriptionUrl
    checkIdentical(descriptionUrl(th), correct_descriptionUrl)

    # TEST: hub<-
    new_hub <- "new_hub"
    hub(th) <- new_hub
    checkIdentical(hub(th), new_hub)
    hub(th) <- correct_hub

    # TEST: shortLabel<-
    new_shortLabel <- "new_hub"
    shortLabel(th) <- new_shortLabel
    checkIdentical(shortLabel(th), new_shortLabel)
    shortLabel(th) <- correct_shortLabel

    # TEST: longLabel<-
    new_longLabel <- "new_hub"
    longLabel(th) <- new_longLabel
    checkIdentical(longLabel(th), new_longLabel)
    longLabel(th) <- correct_longLabel

    # TEST: genomesFile<-
    new_genomesFile <- "newfile.txt"
    genomesFile(th) <- new_genomesFile
    checkIdentical(genomesFile(th), new_genomesFile)
    genomesFile(th) <- correct_genomesFile

    # TEST: email<-
    new_email <- "new@domail.com"
    email(th) <- new_email
    checkIdentical(email(th), new_email)
    email(th) <- correct_email

    # TEST: descriptionUrl<-
    new_descriptionUrl <- "http://newdomail.com/articles/hg19"
    descriptionUrl(th) <- new_descriptionUrl
    checkIdentical(descriptionUrl(th), new_descriptionUrl)
    descriptionUrl(th) <- correct_descriptionUrl

    # TEST: genomeField
    checkIdentical(genomeField(th, "hg19", "trackDb"), correct_trackDb)

    ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ### TEST TrackHubGenome Class
    ###

    correct_trackhubgenome_uri <- paste0(correct_trackhub_uri, "/hg19")
    correct_trackhubgenome_genome_name <- "hg19"
    correct_trackhubgenome_length <- 1L
    correct_trackhubgenome_organism <- "BigFoot"
    correct_trackhubgenome_names <- "wgEncodeUWDukeDnaseGM12878FdrPeaks"
    correct_bigDataUrl <- "wgEncodeUWDukeDnaseGM12878.fdr01peaks.hg19.bb"

    thg <- TrackHubGenome(th, "hg19")

    # TEST: uri
    checkIdentical(uri(thg), correct_trackhubgenome_uri)

    # TEST: genome
    checkIdentical(genome(thg), correct_trackhubgenome_genome_name)

    # TEST: length
    checkIdentical(length(thg), correct_trackhubgenome_length)

    # TEST: organism
    checkIdentical(organism(thg), correct_trackhubgenome_organism)

    # TEST: names
    checkIdentical(names(thg), correct_trackhubgenome_names)

    # TEST: trackNames
    checkIdentical(trackNames(thg), correct_trackhubgenome_names)

    # TEST: trackField
    checkIdentical(trackField(thg, "wgEncodeUWDukeDnaseGM12878FdrPeaks", "bigDataUrl"), correct_bigDataUrl)

    ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ### TEST TrackContainer Class
    ###

    correct_slot_type <- "Track"
    correct_track <- Track(track = "tcell", bigDataUrl = "tcell/data.bigWig")

    # TEST: slot type
    tc <- TrackContainer()
    slot_type <- slot(tc, "elementType")
    checkIdentical(slot_type, correct_slot_type)

    # TEST: wrong type slot error reporting
    checkException(tc[[1]] <- 1)

    # TEST: names()
    tc[[1]] <- correct_track
    checkIdentical(names(tc), correct_track@track)
}

manager <- BiocIO:::manager

connection <- BiocIO:::connection

resourceDescription <- BiocIO:::resourceDescription

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Utilities
###

## First checks for Windows drive letter.
## There are no known URI schemes that are only a single character.
isURL <- BiocIO:::isURL

.parseURI <- BiocIO:::.parseURI

checkURI <- function(x) {
  if (!isSingleString(x))
    stop("URI must be a single, non-NA string")
  x
}

createResource <- function(x, dir = FALSE, content = "") {
  uri <- .parseURI(x)
  if (uri$scheme == "file" || uri$scheme == "") {
    if (!file.exists(uri$path)) {
      if (dir)
        dir.create(uri$path, recursive = TRUE)
      else writeLines(content, uri$path)
    } else warning("Path '", uri$path, "' already exists")
  } else stop("Cannot create a resource that is not a local file")
}

uriExists <- function(x) {
  uri <- .parseURI(x)
  if (uriIsLocal(uri)) {
    exists <- file.exists(uri$path)
  } else {
    exists <- HEAD(x)$status_code == 200L
  }
  exists
}

uriIsLocal <- function(x) {
  x$scheme == "file" || x$scheme == ""
}

uriIsWritable <- function(x) {
  uri <- .parseURI(x)
  if (uriIsLocal(uri)) {
    !file.access(uri$path, 2) ||
    (!file.exists(uri$path) && uriIsWritable(dirname(uri$path)))
  } else FALSE
}

checkArgFormat <- function(con, format) {
  if (toupper(format) !=
      substring(toupper(sub("File$", "", class(con))), 1, nchar(format)))
    stop("Cannot treat a '", class(con), "' as format '", format, "'")
}

connectionForResource <- BiocIO:::connectionForResource

## BestFileFormat

setGeneric("bestFileFormat",
           function(x, dest, ...) standardGeneric("bestFileFormat"))

setMethod("bestFileFormat", c("GenomicRanges", "ANY"),
          function(x, dest) {
            ## have numbers on a single strand, use BigWig
            if (is.numeric(score(x)) && length(unique(strand(x))) == 1L)
              "bw"
            else "bed"
          })

setMethod("bestFileFormat", c("GRangesList", "ANY"), function(x, dest) {
  "bed" # need hierarchical structure
})

setMethod("bestFileFormat", c("RleList", "ANY"), function(x, dest) {
  "bw" # e.g., coverage
})

setMethod("bestFileFormat", c("IntegerRangesList", "ANY"), function(x, dest) {
  "bed" # just ranges...
})

## Connection management (similar to memory management)

manage <- BiocIO:::manage

managed <- BiocIO:::managed

unmanage <- BiocIO:::unmanage

release <- BiocIO:::release

htmlErrorHandler <- function(msg, code, domain, line, col, level, filename) {
  if (!length(level))
    stop("Unknown HTML parse error")
  if (level > 2)
    stop("Failed to Parse HTML [", line, ":", col, "]: ", msg)
}

htmlParse <- function(str)
  suppressWarnings(htmlTreeParse(str, asText = TRUE, useInternalNodes = TRUE,
                                 error = htmlErrorHandler))

rtracklayerGET <- function(url, ..., query = list()) {
  verbose <- getOption("rtracklayer.http.verbose", FALSE)
  verbose <- as.integer(isTRUE(as.logical(verbose)))
  response <- GET(url, user_agent("rtracklayer"), config(verbose = verbose),
                       ...,
                       query = as.list(query))
  htmlParse(content(response, as="text"))
}

rtracklayerPOST <- function(url, ..., body = list()) {
  verbose <- getOption("rtracklayer.http.verbose", FALSE)
  verbose <- as.integer(isTRUE(as.logical(verbose)))
  response <- POST(url, user_agent("rtracklayer"), config(verbose = verbose),
                        ...,
                        body = as.list(body))
  htmlParse(content(response, as = "text"))
}

httpShow <- function(url, .form = list(), ...)
  browseURL(urlForm(url, .form, ...))

urlForm <- function(url, .form = list(), ...) {
  values <- urlEncode(as.character(c(.form, ...)))
  query <- paste(names(.form), values, sep = "=", collapse = "&")
  paste(url, query, sep = "?")
}

# differs from URLencode (vectorized, only encodes specified chars)
urlEncode <- function(str, chars = "-a-zA-Z0-9$_.+!*'(),", keep = TRUE)
{
  str <- as.character(str)
  bad <- chars
  if (keep)
    bad <- gsub(paste("[", chars, "]", sep=""), "", str)
  bad <- unique(unlist(strsplit(bad, "")))
  code <- paste("%", charToRaw(paste(bad, collapse="")), sep="")
  for (i in seq_along(bad))
    str <- gsub(bad[i], code[i], str, fixed = TRUE)
  str
}

urlDecode <- function(str, na.strings="NA")
{
  ans <- curl_unescape(str)
  if (!identical(na.strings, "NA"))
      ans[is.na(str)] <- na.strings
  ans
}

expandPath <- function(x) {
    if (startsWith(x, "http") || startsWith(x, "ftp"))
        expandURL(x)
    else path.expand(x)
}

expandURL <- function(uri) {
    if(HEAD(uri)$status_code != 200L)
        return(uri)
    else {
        #opts <- list(
        #    followlocation = TRUE,  # resolve redirects
        #    ssl.verifyhost = FALSE, # suppress certain SSL errors
        #    ssl.verifypeer = FALSE,
        #    nobody = TRUE, # perform HEAD request
        #    verbose = FALSE
        #    )
        #curlhandle <- getCurlHandle(.opts = opts)
        #getURL(uri, curl = curlhandle)
        #info <- getCurlInfo(curlhandle)
        #info$effective.url

        config <- list(followlocation = 1,
                       ssl_verifyhost = 0,
                       ssl_verifypeer = 0,
                       nobody = 1,
                       verbose = 0)
        ## nobody = 1 seems to be ignored and HEAD() doesn't work here
        GET(uri, config)$url
    }
}

### =========================================================================
### TrackDb class
### -------------------------------------------------------------------------
###
### Represents a database or source of tracks, i.e, interval datasets
###

setClass("TrackDb")

setGeneric("track<-",
           function(object, ..., value) standardGeneric("track<-"))
## load a track into a database
setReplaceMethod("track", c("TrackDb", "ANY"),
                 function(object, name = deparse(substitute(value)), ..., value)
                 {
                   track(object, name, ...) <-
                     GRangesList(as(value, "GRanges", strict = FALSE),
                                 compress=FALSE)
                   object
                 })

setMethod("[[<-", c("TrackDb", value="ANY"),
          function(x, i, j, ..., value) {
            if (!missing(j))
              warning("argument 'j' ignored")
            track(x, i, ...) <- value
            x
          })

setMethod("$<-", c("TrackDb", value="ANY"),
          function(x, name, value) {
            x[[name]] <- value
            x
          })

setGeneric("track", function(object, ...) standardGeneric("track"))

setMethod("[[", "TrackDb", function (x, i, j, ...) {
  if (!missing(j))
    warning("argument 'j' ignored")
  track(x, i, ...)
})

setMethod("$", "TrackDb", function (x, name) {
  x[[name]]
})

test_gff <- function() {
  test_path <- system.file("tests", package = "rtracklayer")

  space <- c(rep("chr10", 15), rep("chr12", 16))
  start <- c(rep(92828, 4), 92997, rep(94555, 4), rep(94744, 3), rep(95122, 2),
             95348, rep(87984, 4),
             rep(c(88257, 88570, 88860, 89675, 90587, 90796), each = 2))
  end <- c(95504, 95178, 95504, rep(94054, 2), rep(94665, 2), 94615, 94665,
           rep(94852, 3), rep(95178, 2), 95504,
           rep(c(91263, 88017, 88392, 88771, 89018, 89827, 90655, 91263),
               each = 2))
  type <- c("gene", "mRNA", "mRNA", "exon", "CDS", "exon", "exon",
            "CDS", "CDS", "exon", "exon", "CDS", "exon", "CDS", "exon",
            "gene", "mRNA", "exon", "CDS", "exon", "CDS", "exon", "CDS",
            "exon", "CDS", "exon", "CDS", "exon", "CDS", "exon", "CDS")
  type <- factor(type, unique(type))
  source <- factor("rtracklayer")
  phase <- NA_integer_
  score <- c(5, rep(NA, length(type) - 1L))
  strand <- strand(c(rep("-", 14), "*", rep("+", 15), "*"))
  Alias <- CharacterList(c(list(c("FLJ40100", "TUBB8")),
                           rep(list(character()), 14), "LOC100288778",
                           rep(list(character()), 15)))
  ID <- c("GeneID:347688", "873", "872", rep(NA, 12), "GeneID:100288778",
          "4644", rep(NA, 14))
  Name <- c(rep("TUBB8", 3), rep(NA, 12), rep("LOC100288778", 2),
            rep(NA, 14))
  Parent <- CharacterList(c(list(character()), rep("GeneID:347688", 2),
                            rep(list(c("872", "873")), 2),
                            rep(c("872", "873"), 3), rep("873", 3), "872",
                            list(character()), "GeneID:100288778",
                            rep("4644", 14)))
  geneName <- c("tubulin, beta 8", rep(NA, 14),
                "WAS protein family homolog 1; pseudogene", rep(NA, 15))
  genome <- c("hg19", rep(NA, length(geneName) - 1))
  correct_gff3 <- GRanges(space, IRanges(start, end), strand,
                          source, type, score, phase,
                          ID, Name, geneName, Alias, genome, Parent)
  seqinfo(correct_gff3) <- Seqinfo(c("chr10", "chr12"))

  correct_gff1 <- correct_gff3[,c("source", "type", "score", "phase")]
  correct_gff1$group <- as.factor(seqnames(correct_gff3))

  correct_gff2 <- correct_gff3
  toCSV <- function(x) {
    csv <- sapply(x, paste, collapse = ",")
    csv[nchar(csv) == 0] <- NA
    csv
  }
  correct_gff2$Alias <- toCSV(correct_gff2$Alias)
  correct_gff2$Parent <- toCSV(correct_gff2$Parent)

  ## TEST: basic GFF3 import
  test_gff3 <- file.path(test_path, "genes.gff3")
  test <- import(test_gff3)
  checkIdentical(correct_gff3, test)

  ## TEST: import.gff*
  test <- import.gff(test_gff3)
  checkIdentical(correct_gff3, test)
  test <- import.gff3(test_gff3)
  checkIdentical(correct_gff3, test)
  #suppressWarnings(test <- import.gff2(test_gff3))
  #checkIdentical(correct_gff3, test)
  oldOpts <- options(warn = 2)
  checkException(import.gff2(test_gff3))
  options(oldOpts)
  
  ## TEST: GFF(3)File
  test_gff_file <- GFF3File(test_gff3)
  test <- import(test_gff_file)
  checkIdentical(correct_gff3, test)
  test_gff_file <- GFFFile(test_gff3)
  test <- import(test_gff_file)
  checkIdentical(correct_gff3, test)
  test_gff_file <- GFFFile(test_gff3, version = "3")
  test <- import(test_gff_file)
  checkIdentical(correct_gff3, test)
  test_gff_file <- GFF2File(test_gff3)
  #suppressWarnings(test <- import(test_gff_file))
  #checkIdentical(correct_gff3, test)
  oldOpts <- options(warn = 2)
  checkException(test <- import(test_gff_file))
  options(oldOpts)

  ## TEST: 'gff' extension
  test_gff_out <- file.path(tempdir(), "genes.gff")
  on.exit(unlink(test_gff_out))
  export(correct_gff3, test_gff_out)
  test <- import(test_gff_out)
  checkIdentical(test, correct_gff1)
  export(correct_gff3, test_gff_out, version = "1")
  test <- import(test_gff_out)
  checkIdentical(test, correct_gff1)
  export(correct_gff3, test_gff_out, version = "2")
  test <- import(test_gff_out)
  checkIdentical(test, correct_gff2)
  export(correct_gff3, test_gff_out, version = "3")
  test <- import(test_gff_out)
  checkIdentical(test, correct_gff3)
  test <- import(GFF3File(test_gff_out))
  checkIdentical(test, correct_gff3)
  test <- import(GFFFile(test_gff_out))
  checkIdentical(test, correct_gff3)
  test <- import(test_gff_out, version = "3")
  checkIdentical(test, correct_gff3)
  #suppressWarnings(test <- import(test_gff_out, version = "2"))
  #checkIdentical(test, correct_gff3)
  oldOpts <- options(warn = 2)
  checkException(test <- import(test_gff_out, version = "2"))
  options(oldOpts)

  ## TEST: 'gff2' extension
  test_gff2_out <- file.path(tempdir(), "genes.gff2")
  export(correct_gff3, test_gff2_out)
  test <- import(test_gff2_out)
  checkIdentical(test, correct_gff2)

  ## TEST: 'gff1' extension
  test_gff1_out <- file.path(tempdir(), "genes.gff1")
  export(correct_gff3, test_gff1_out)
  test <- import(test_gff1_out)
  checkIdentical(test, correct_gff1)

  ## TEST: 'format' argument
  test_gff_file <- GFF3File(test_gff3)
  test <- import(test_gff_file, format = "gff")
  checkIdentical(test, correct_gff3)
  test <- import(test_gff_file, format = "gff3")
  checkIdentical(test, correct_gff3)
  checkException(import(test_gff_file, format = "gff2"))
  checkException(import(test_gff_file, format = "bed"))
  
  ## TEST: 'genome'  
  si_hg19 <- SeqinfoForBSGenome("hg19")
  correct_hg19 <- correct_gff3
  seqlevels(correct_hg19) <- seqlevels(si_hg19)
  seqinfo(correct_hg19) <- si_hg19
  test <- import(test_gff3, genome = "hg19")
  checkIdentical(test, correct_hg19)
  
  test_gff3_out <- file.path(tempdir(), "genes.gff3")
  on.exit(unlink(test_gff3_out))
  correct_genome_hg19 <- correct_gff3
  genome(correct_genome_hg19) <- "hg19"
  correct_genome_hg19 <- as(correct_genome_hg19, "GRanges")
  export(correct_genome_hg19, test_gff3_out)
  test <- import(test_gff3_out)
  checkIdentical(test, correct_hg19)

  ## TEST: colnames empty, colnames := "geneName"
  test <- import(test_gff3, colnames = character())
  target <- correct_gff3[,character()]
  checkIdentical(target, test)
  test <- import(test_gff3, colnames = "geneName")
  target <- correct_gff3[,"geneName"]
  checkIdentical(target, test)

  ## TEST: import from connection
  test_gff_con <- file(test_gff_out)
  test <- import(test_gff_con, format = "gff")
  checkIdentical(correct_gff3, test)

  ## TEST: export to connection, with preceding comment
  test_gff_con <- file(test_gff_out)
  open(test_gff_con, "w")
  comment <- "# test comment"
  writeLines(comment, test_gff_con)
  export(correct_gff3, test_gff_con, version = "3")
  close(test_gff_con)
  checkIdentical(comment, readLines(test_gff_out, n = 1))
  test <- import(test_gff_out)
  checkIdentical(correct_gff3, test)
  
  ## TEST: 'append'
  export(correct_gff3[seqnames(correct_gff3) == "chr10", ], test_gff3_out)
  export(correct_gff3[seqnames(correct_gff3) == "chr12", ], test_gff3_out,
         append = TRUE)
  test <- import(test_gff3_out)
  checkIdentical(correct_gff3, test)

  ## TEST: 'source'
  target <- correct_gff3
  mcols(target)$source <- factor("test")
  export(correct_gff3, test_gff3_out, source = "test")
  test <- import(test_gff3_out)
  checkIdentical(target, test)

  ## TEST: 'which'
  which <- GRanges("chr10:90000-93000")
  which_target <- subsetByOverlaps(correct_gff3, which)
  test <- import(test_gff3, which = which)
  checkIdentical(which_target, test)
  
  ## TEST: 'index'
  export(correct_gff3, test_gff3_out, index = TRUE)
  test_gff_bgz <- paste(test_gff3_out, ".bgz", sep = "")
  on.exit(unlink(test_gff_bgz))
  on.exit(unlink(paste(test_gff_bgz, ".tbi", sep = "")))
  test <- import(test_gff_bgz, which = which)
  checkIdentical(which_target, test)

  ## TEST: SimpleGRangesList
  ucsc_data1 <- new("UCSCData",
                    GenomeInfoDb::keepSeqlevels(correct_gff3, "chr10",
                                                pruning.mode="coarse"),
                    trackLine = new("BasicTrackLine", name = "chr10"))
  ucsc_data2 <- new("UCSCData",
                    GenomeInfoDb::keepSeqlevels(correct_gff3, "chr12",
                                                pruning.mode="coarse"),
                    trackLine = new("BasicTrackLine", name = "chr12"))
  correct_grl <- GRangesList(ucsc_data1, ucsc_data2, compress=FALSE)
  mcols(correct_grl[[2]])$genome <- NULL
  names(correct_grl) <- seqlevels(correct_gff3)
  export(correct_grl, test_gff3_out)
  test <- import.ucsc(test_gff3_out)
  checkIdentical(correct_grl, test)
}

### =========================================================================
### BED (Browser Extended Display) support (including bedGraph and BED15)
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Classes
###

setClass("BEDFile", contains = "BiocFile")
BEDFile <- function(resource) {
  new("BEDFile", resource = resource)
}

setClass("BEDGraphFile", contains = "BEDFile")
BEDGraphFile <- function(resource) {
  new("BEDGraphFile", resource = resource)
}

setClass("BED15File", contains = "BEDFile")
BED15File <- function(resource) {
  new("BED15File", resource = resource)
}

setClass("BEDPEFile", contains = "BEDFile")
BEDPEFile <- function(resource) {
    new("BEDPEFile", resource = resource)
}

setClass("NarrowPeakFile", contains = "BEDFile")
NarrowPeakFile <- function(resource) {
  new("NarrowPeakFile", resource = resource)
}

setClass("BroadPeakFile", contains = "BEDFile")
BroadPeakFile <- function(resource) {
  new("BroadPeakFile", resource = resource)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export
###

setGeneric("export.bed",
           function(object, con, ...) standardGeneric("export.bed"))

setMethod("export.bed", "ANY",
          function(object, con, ...) {
            export(object, con, "bed", ...)
          })

setMethod("export", c("ANY", "BEDFile"),
          function(object, con, format, ...)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            cl <- class(object)
            if (hasMethod("asBED", class(object)))
              object <- asBED(object)
            track <- try(as(object, "GRanges"), silent = TRUE)
            if (class(track) == "try-error") {
              track <- try(as(object, "GenomicRangesList"), silent = TRUE)
              if (is(track, "try-error"))
                stop("cannot export object of class '", cl, "'")
            }
            export(track, con, ...)
          })

setMethod("export", c("GenomicRanges", "BEDFile"),
          function(object, con, format, append = FALSE, index = FALSE,
                   ignore.strand = FALSE, trackLine = NULL)
          {
            if (!is.null(trackLine)) {
              object <- UCSCData(object, trackLine)
              trackLine <- TRUE
              return(callGeneric())
            }
            if (!missing(format))
              checkArgFormat(con, format)
            name <- strand <- thickStart <- thickEnd <- color <- NULL
            blockCount <- blockSizes <- blockStarts <- NULL
            if (index)
              object <- sortBySeqnameAndStart(object)
            df <- data.frame(seqnames(object), start(object) - 1L, end(object))
            score <- score(object)
            if (!is.null(score)) {
              if (!is.numeric(score) || any(is.na(score)))
                stop("Scores must be non-NA numeric values")
            }
            if (is(con, "BEDGraphFile")) {
              if (is.null(score)) ## bedGraph requires score
                score <- 0
              df$score <- score
            } else {
              toCSV <- function(x) {
                if (is(x, "IntegerList")) {
                  x <- unstrsplit(CharacterList(x), ",")
                } else if (!is.character(x) && !is.null(x))
                  stop("Could not convert block coordinates to CSV")
                x
              }
              blockSizes <- blockStarts <- NULL
              if (!is.null(object$blocks)) {
                blockSizes <- toCSV(width(object$blocks))
                blockStarts <- toCSV(start(object$blocks) - 1L)
              }
              if (is(con, "BED15File") && is.null(blockSizes))
                blockStarts <- blockSizes <- "" # bed15 must have all 15 cols
              if (!is.null(blockSizes) || !is.null(blockStarts)) {
                if (is.null(blockSizes))
                  stop("'blockStarts' specified without 'blockSizes'")
                if (is.null(blockStarts))
                  stop("'blockSizes' specified without 'blockStarts'")
                lastBlock <- function(x)
                  as.integer(sub(".*,", "", sub(",$", "", x)))
                lastSize <- lastBlock(blockSizes)
                lastStart <- lastBlock(blockStarts)
                if (any(df[[2]] + lastSize + lastStart != df[[3]]) ||
                    any(sub(",.*", "", blockStarts) != "0"))
                    stop("blocks must span entire feature,",
                         " and be within its bounds")
                blockCount <- elementNROWS(object$blocks)
                if (!is.null(object$blockCount))
                  if (!identical(blockCount, as.integer(object$blockCount)))
                    stop("incorrect block counts given block sizes")
              }
              if (is.null(color))
                color <- object$itemRgb
              if (is.null(color) && !is.null(blockCount))
                color <- "0" ## blocks require color
              else if (!is.null(color)) {
                nacol <- is.na(color)
                if (!is.matrix(color))
                  colmat <- col2rgb(color)
                color <- paste(colmat[1,], colmat[2,], colmat[3,], sep = ",")
                color[nacol] <- "0"
              }
              thickStart <- object$thickStart
              thickEnd <- object$thickEnd
              if (!is.null(object$thick)) {
                thickStart <- start(object$thick)
                thickEnd <- end(object$thick)
              }
              ## color requires thick ranges
              if (is.null(thickStart) && !is.null(color)) {
                thickStart <- start(object)
                thickEnd <- end(object)
              }
              strand <- if (ignore.strand) NULL else strand(object)
              if (!is.null(strand)) {
                strand[strand == "*"] <- NA
              }
              if (!is.null(thickStart) && is.null(strand)) {
                strand <- rep(NA, length(object))
              }
              if (!is.null(strand) && is.null(score))
                score <- rep(0, length(object))
              name <- object$name
              if (is.null(name))
                name <- names(object)
              if (!is.null(score) && is.null(name))
                name <- rep(NA, length(object))
              df$name <- name
              df$score <- score
              df$strand <- strand
              df$thickStart <-
                if (!is.null(thickStart)) thickStart - 1L else NULL
              df$thickEnd <- thickEnd
              df$itemRgb <- color
              df$blockCount <- blockCount
              df$blockSizes <- blockSizes
              df$blockStarts <- blockStarts
              if (is(con, "BED15File")) {
                df$expCount <- object$expCount
                df$expIds <- object$expIds
                df$expScores <- object$expScores
              }
            }
            scipen <- getOption("scipen")
            options(scipen = 100) # prevent use of scientific notation
            on.exit(options(scipen = scipen))
            file <- con
            m <- manager()
            con <- connection(m, con, if (append) "a" else "w")
            write.table(df, con, sep = "\t", col.names = FALSE,
                        row.names = FALSE, quote = FALSE, na = ".")
            release(m, con)
            if (index)
              invisible(indexTrack(file))
            else invisible(file)
          })

setMethod("export", c("UCSCData", "BEDFile"),
          function(object, con, format, trackLine = TRUE, ...)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            if (trackLine) {
              export.ucsc(object, con, ...)
            } else {
              trackLine <- NULL
              callNextMethod()
            }
            invisible(con)
          })

setMethod("export", c("CompressedGRangesList", "BEDFile"),
          function(object, con, format, ...) {
            object <- asBED(object)
            callGeneric()
          })

setMethod("export", c("SimpleGRangesList", "BEDFile"),
          .export_SimpleGRangesList_BiocFile)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import
###

setGeneric("import.bed", function(con, ...) standardGeneric("import.bed"))

setMethod("import.bed", "ANY",
          function(con, ...)
          {
            import(con, format = "bed", ...)
          })

scanTrackLine <- function(con) {
  m <- manager()
  con <- connectionForResource(m, con, "r")
  on.exit(release(m, con))
  line <- "#"
  while(length(grep("^ *#", line))) # skip initial comments
    line <- readLines(con, 1, warn = FALSE)
  if (length(grep("^track", line)) > 0)
    line
  else {
    pushBack(line, con)
    NULL
  }
}

setMethod("import", "BEDFile",
          function(con, format, text, trackLine = TRUE,
                   genome = NA, colnames = NULL,
                   which = NULL, seqinfo = NULL, extraCols = character(),
                   sep = c("\t", ""), na.strings=character(0L))
          {
            if (!missing(format))
              checkArgFormat(con, format)
            sep <- match.arg(sep)
            stopifnot(is.character(na.strings), !anyNA(na.strings))
            file <- con
            m <- manager()
            con <- queryForConnection(m, con, which)
            on.exit(release(m, con))
            if (attr(con, "usedWhich"))
                which <- NULL
            if (is(genome, "Seqinfo")) {
                seqinfo <- genome
                genome <- NA_character_
            }
            if (is.null(seqinfo))
              seqinfo <- attr(con, "seqinfo")
            ## check for a track line
            line <- scanTrackLine(con)
            if (!is.null(line) && trackLine) {
              pushBack(line, con)
              ans <- import.ucsc(initialize(file, resource = con), drop = TRUE,
                                 trackLine = FALSE, genome = genome,
                                 colnames = colnames,
                                 which = which, seqinfo = seqinfo,
                                 extraCols = extraCols)
              return(ans)
            }
            bedClasses <- colClasses(file)
            bedNames <- names(bedClasses)
            normArgColnames <- function(validNames) {
              if (is.null(colnames))
                colnames <- validNames
              else {
                colnames <- unique(c(head(bedNames, 3), as.character(colnames)))
                if ("thick" %in% colnames)
                  colnames <- c(setdiff(colnames, "thick"), "thickStart",
                                "thickEnd")
                if ("blocks" %in% colnames)
                  colnames <- c(setdiff(colnames, "blocks"), "blockStarts",
                                "blockSizes", "blockCount")
                missingCols <- setdiff(colnames, validNames)
                if (length(missingCols))
                  stop("Requested column(s) ",
                       paste("'", missingCols, "'", sep = "", collapse = ", "),
                       " are not valid columns or were not found in the file")
              }
              colnames
            }
            ## read a single line to get ncols up-front,
            ## and thus specify all col classes
            ## FIXME: reading in 'as.is' to save memory,
            line <- readLines(con, 1, warn=FALSE)
            ## UCSC seems to use '#' at beginning to indicate comment.
            while(length(line) &&
                  (!nzchar(line) || substring(line, 1, 1) == "#"))
            {
              line <- readLines(con, 1, warn=FALSE)
            }
            if (length(line)) {
              `tail<-` <- function(x, n, value)
                if (n != 0) c(head(x, -n), value) else x
              pushBack(line, con)
              pattern <- if (sep == "") "\\s+" else "\\t"
              colsInFile <- seq_len(length(strsplit(line, pattern)[[1L]]))
              presentNames <- bedNames[colsInFile]
              lacksNames <- is.null(names(extraCols)) ||
                  any(names(extraCols) == "") ||
                  any(is.na(names(extraCols)))
              if (length(extraCols) > 0L && lacksNames)
                  stop("'extraCols' must have valid names")
              reservedNames <- c(names(bedClasses), "thick", "blocks")
              conflictingNames <- intersect(reservedNames, names(extraCols))
              if (length(conflictingNames) > 0L) {
                  stop("'names(extraCols)' contains reserved name(s): ",
                       paste0("'", conflictingNames, "'", collapse=", "))
              }
              tail(presentNames, length(extraCols)) <- names(extraCols)
              bedNames <- presentNames
              presentClasses <- bedClasses[colsInFile]
              tail(presentClasses, length(extraCols)) <- unname(extraCols)
              colnames <- normArgColnames(presentNames)
              bedClasses <- ifelse(presentNames %in% colnames,
                                   presentClasses, "NULL")
              bed <- DataFrame(read.table(con, colClasses = bedClasses,
                                          as.is = TRUE,
                                          na.strings = c(".", na.strings),
                                          comment.char = "",
                                          sep = sep,
                                          quote = ""))
            } else {
              if (is.null(colnames))
                colnames <- character()
              else colnames <- normArgColnames(bedNames)
              keepCols <- bedNames %in% colnames
              bed <- DataFrame(as.list(sapply(bedClasses[keepCols], vector)))
            }
            colnames(bed) <- bedNames[bedNames %in% colnames]
            bed <- bed[substring(bed$chrom, 1, 1) != "#",]
            if (!is.null(bed$thickStart)) {
              thickEnd <- bed$thickEnd
              if (is.null(thickEnd))
                thickEnd <- bed$end
              bed$thick <- IRanges(bed$thickStart + 1L, thickEnd)
              bed$thickStart <- bed$thickEnd <- NULL
            }
            color <- bed$itemRgb
            if (is.character(color)) { # could be NULL
              spec <- color != "0"
              cols <- unlist(strsplit(color[spec], ",", fixed=TRUE),
                             use.names=FALSE)
              cols <- matrix(as.integer(cols), 3)
              color <- rep(NA, nrow(bed))
              color[spec] <- rgb(cols[1,], cols[2,], cols[3,],
                                 maxColorValue = 255L)
              bed$itemRgb <- color
            }
            fromCSV <- function(b) {
              as.integer(unlist(strsplit(b, ",", fixed = TRUE)))
            }
            if (!is.null(bed$blockStarts)) {
              blocks <- relist(IRanges(fromCSV(bed$blockStarts) + 1L,
                                      width = fromCSV(bed$blockSizes)),
                               PartitioningByWidth(bed$blockCount))
              names(blocks) <- NULL
              bed$blockStarts <- bed$blockSizes <- bed$blockCount <- NULL
              bed$blocks <- blocks
            }
            GenomicData(IRanges(bed$start + 1L, bed$end),
                        bed[-(1:3)],
                        chrom = bed$chrom, genome = genome,
                        seqinfo = seqinfo,
                        which = which)
          })

setGeneric("colClasses", function(x) standardGeneric("colClasses"))

setMethod("colClasses", "BEDFile", function(x) {
              c(chrom="character", start="integer", end="integer",
                name="character", score="numeric", strand="character",
                thickStart="integer", thickEnd="integer",
                itemRgb="character", blockCount="integer",
                blockSizes="character", blockStarts="character")
          })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### narrowPeak support
###
setGeneric("import.narrowPeak", function(con, ...) standardGeneric("import.narrowPeak"))

setMethod("import.narrowPeak", "ANY", function(con, ...) {
    import(con, format = "narrowPeak", ...)

})

setMethod("import", "NarrowPeakFile", function(con, ...) {
    callNextMethod(con = con,
                   extraCols = c(signalValue = "numeric", pValue = "numeric", qValue = "numeric", peak = "integer"),
                   ... )
})

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### broadPeak support
###
setGeneric("import.broadPeak", function(con, ...) standardGeneric("import.broadPeak"))

setMethod("import.broadPeak", "ANY", function(con, ...) {
    import(con, format = "broadPeak", ...)

})

setMethod("import", "BroadPeakFile", function(con, ...) {
    callNextMethod(con = con,
                   extraCols = c(signalValue = "numeric", pValue = "numeric", qValue = "numeric"),
                   ... )
})

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### BED15 (Microarray) Support
###

setMethod("import", "BED15File",
          function(con, format, text, trackLine = NULL, genome = NA,
                   which = NULL)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            if (is.null(trackLine))
              return(import.ucsc(con, drop = TRUE, genome = genome,
                                 which = which))
            bed <- callNextMethod()
            if (!length(bed))
              return(bed)
            ids <- strsplit(values(bed)$expIds[1], ",", fixed=TRUE)[[1]]
            expNames <- trackLine@expNames[as.integer(ids) + 1L]
            scores <- unlist(strsplit(values(bed)$expScores, ",", fixed=TRUE),
                             use.names=FALSE)
            scores <- as.numeric(scores)
            scores[scores == -10000] <- NA # stupid UCSC convention
            scores <- split(scores, gl(length(expNames), 1, length(scores)))
            names(scores) <- expNames
            nonExpCols <- setdiff(colnames(values(bed)),
                                  c("expCount", "expScores", "expIds"))
            values(bed) <- values(bed)[,nonExpCols]
            values(bed) <- cbind(values(bed), do.call(DataFrame, scores))
            bed
          })

setGeneric("import.bed15",
           function(con, ...)
           standardGeneric("import.bed15"),
           signature = "con")

setMethod("import.bed15", "ANY",
          function(con, ...)
          {
            import(con, "bed15", ...)
          })

setGeneric("export.bed15",
           function(object, con, ...) standardGeneric("export.bed15"))

setMethod("export.bed15", "ANY",
          function(object, con, ...) {
            export(object, con, "bed15",  ...)
          })

### FIXME: dispatch will break when 'object' is a UCSCData
### Possible solution: just merge this code with the main BEDFile method?
setMethods("export",
           list(c("GenomicRanges", "BED15File"),
                c("UCSCData", "BED15File")),
          function(object, con, format, expNames = NULL, trackLine = NULL, ...)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            if (is.null(trackLine)) {
              ## ensure we do not override existing track line parameter
              if (is.null(expNames) && is(object, "UCSCData") &&
                  is(object@trackLine, "Bed15TrackLine"))
                expNames <- object@trackLine@expNames
              return(export.ucsc(object, con, expNames = expNames, ...))
            }
            expNames <- trackLine@expNames
            object$expCount <- rep(length(expNames), length(object))
            object$expIds <- rep(paste(seq_along(expNames)-1, collapse=","),
                                 length(object))
            scores <- as.list(mcols(object)[expNames])
            scores <- do.call(paste, c(scores, sep = ","))
            scores <- gsub("NA", "-10000", scores, fixed=TRUE)
            if (length(scores) == 0L) {
                scores <- rep("", length(object))
            }
            object$expScores <- scores
            callNextMethod(object, con, ...)
          })

setClass("Bed15TrackLine",
         representation(expStep = "numeric", expScale = "numeric",
                        expNames = "character_OR_NULL"),
         prototype(expStep = 0.5, expScale = 3.0),
         contains = "BasicTrackLine") # not sure which fields work

setAs("Bed15TrackLine", "character",
      function(from)
      {
        str <- as(as(from, "TrackLine"), "character")
        paste(str, " type=array ",
              " expScale=", from@expScale,
              " expStep=", from@expStep,
              " expNames=\"", paste(from@expNames, collapse=","), "\"",
              sep = "")
      })

setAs("character", "Bed15TrackLine",
      function(from)
      {
        line <- new("Bed15TrackLine", as(from, "TrackLine"))
        vals <- ucscParsePairs(from)
        line@expScale <- as.numeric(vals[["expScale"]])
        line@expStep <- as.numeric(vals[["expStep"]])
        line@expNames <- strsplit(vals[["expNames"]], ",", fixed=TRUE)[[1]]
        line
      })

setMethod("fileFormat", "Bed15TrackLine", function(x) "bed15")

setMethod("colClasses", "BED15File", function(x) {
              c(callNextMethod(), expCount="integer", expIds="character",
                expScores="character")
          })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### bedGraph (formerly subset of WIG) support
###

setGeneric("import.bedGraph",
           function(con, ...) standardGeneric("import.bedGraph"),
           signature = "con")

setMethod("import.bedGraph", "ANY",
          function(con, ...)
          {
            import(con, "bedGraph", ...)
          })

setGeneric("export.bedGraph",
           function(object, con, ...)
           standardGeneric("export.bedGraph"))

setMethod("export.bedGraph", "ANY",
          function(object, con, ...)
          {
            export(object, con, "bedGraph", ...)
          })

setMethod("colClasses", "BEDGraphFile", function(x) {
              c(chrom="character", start="integer", end="integer",
                score="numeric")
          })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### BEDPE (bedtools paired end) support
###

setMethod("import", "BEDPEFile",
          function(con, format, text, ...) {
              gr <- callNextMethod()
              df <- mcols(gr)
              mcols(gr) <- NULL
              if (is.null(df$strand2))
                  df$strand2 <- "*"
              else
                  df$strand2[is.na(df$strand2)] <- "*"
              seqlevels(gr) <- union(seqlevels(gr), unique(df$chrom2))
              gr2 <- with(df, GRanges(chrom2, IRanges(start2+1L, end2),
                                      strand2, seqinfo=seqinfo(gr)))
              df <- subset(df, select=-(chrom2:end2))
              df$strand2 <- NULL
              Pairs(gr, gr2, df)
          })

setMethod("colClasses", "BEDPEFile", function(x) {
              c(chrom="character", start="integer", end="integer",
                chrom2="character", start2="integer", end2="integer",
                name="character", score="numeric", strand="character",
                strand2="character")
          })

setMethod("export", c("ANY", "BEDPEFile"),
          function(object, con, format, ...)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            cl <- class(object)
            track <- try(as(object, "Pairs"), silent = TRUE)
            if (class(track) == "try-error") {
              stop("cannot export object of class '", cl, "': ", track)
            }
            export(track, con, ...)
          })

setMethod("export", c("Pairs", "BEDPEFile"),
          function(object, con, format, append = FALSE, ignore.strand = FALSE)
          {
            if (!missing(format))
              checkArgFormat(con, format)
            df <- data.frame(
                             seqnames(first(object)),
                             start(first(object)) - 1L,
                             end(first(object)),
                             seqnames(second(object)),
                             start(second(object)) - 1L,
                             end(second(object))
                            )
            score <- mcols(object)$score
            if (!is.null(score)) {
              if (!is.numeric(score) || any(is.na(score)))
                stop("Scores must be non-NA numeric values")
            } else {
                score <- 0
            }
            name <- mcols(object)$name
            if (is.null(name))
              name <- names(object)
            if (is.null(name))
              name <- rep(NA, length(object))
            df$name <- name
            df$score <- score
            if (ignore.strand) {
                strand1 <- NA # strand cannot be null in bedpe
                strand2 <- NA
            } else {
                strand1 <- strand(first(object))
                strand2 <- strand(second(object))
                strand1[strand1 == "*"] <- NA
                strand2[strand2 == "*"] <- NA
            }
            df$strand1 <- strand1
            df$strand2 <- strand2
            mcol_names <- setdiff(names(mcols(object)), c("name", "score"))
            df <- cbind(df, mcols(object)[,mcol_names])
            scipen <- getOption("scipen")
            options(scipen = 100) # prevent use of scientific notation
            on.exit(options(scipen = scipen))
            file <- con
            m <- manager()
            con <- connection(m, con, if (append) "a" else "w")
            on.exit(release(m, con), add=TRUE)
            write.table(df, con, sep = "\t", col.names = FALSE,
                        row.names = FALSE, quote = FALSE, na = ".")
            invisible(file)
          })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Coercion
###

setGeneric("asBED", function(x, ...) standardGeneric("asBED"))

setMethod("asBED", "GRangesList", function(x) {
  x_range <- range(x)
  if (any(elementNROWS(x_range) != 1L))
    stop("Empty or multi-strand/seqname elements not supported by BED")
  gr <- unlist(x_range, use.names=FALSE)
  mcols(gr) <- mcols(x)
  mcols(gr)$name <- names(x)
  x_ranges <- ranges(unlist(x, use.names=FALSE))
  ord_start <- order(togroup(PartitioningByEnd(x)), start(x_ranges))
  x_ranges <- shift(x_ranges, 1L - rep(start(gr), elementNROWS(x)))[ord_start]
  mcols(gr)$blocks <- relist(x_ranges, x)
  gr
})

setMethod("asBED", "GAlignments", function(x) {
    x <- grglist(x)
    callGeneric()
})

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Utilities
###

setGeneric("blocks", function(x, ...) standardGeneric("blocks"))

setMethod("blocks", "GenomicRanges",
          function(x)
          {
            if (is.null(mcols(x)$blocks)) {
                return(as(x, "GRangesList"))
            }
            block_counts <- elementNROWS(mcols(x)$blocks)
            gr <- GRanges(rep(seqnames(x), block_counts),
                          shift(unlist(mcols(x)$blocks, use.names = FALSE),
                                rep(start(x), block_counts) - 1L),
                          rep(strand(x), block_counts))
            seqinfo(gr) <- seqinfo(x)
            ans <- relist(gr, mcols(x)$blocks)
            mcols(ans) <- subset(mcols(x), select=-blocks)
            names(ans) <- if (!is.null(names(x))) names(x) else x$name
            ans
          })

### =========================================================================
### IGV Data Server Access
### -------------------------------------------------------------------------

readIGVRegistry <- function(url) {
    do.call(rbind, lapply(readLines(url), readIGVDataFile))
}

handleCategory <- function(x) {
    paths <- unlist(getNodeSet(x, "Resource/@path"))
    names(paths) <- unlist(getNodeSet(x, "Resource/@name"))
    name <- factor(unname(xmlAttrs(x)["name"]))
    categories <- lapply(getNodeSet(x, "Category"), handleCategory)
    children <- do.call(rbind, categories)
    children$parenet <- name
    resources <- BiocFileList(lapply(paths, BiocIO::FileForFormat))
    df <- DataFrame(category=name, parent=NA, name=names(paths), resources)
    rbind(df, children)
}

readIGVDataFile <- function(url) {
    lines <- readLines(url)
    doc <- xmlTreeParse(lines, asText=TRUE, useInternalNodes=TRUE)
    handleCategory(getNodeSet(doc, "/Global")[[1L]])
}

### =========================================================================
### Quickload support
### -------------------------------------------------------------------------

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Quickload class
###

setClass("Quickload", representation(uri = "character"))

Quickload_contents <- function(x) {
  read.table(contentsFile(x), sep = "\t", col.names = c("dir", "title"),
             colClasses = "character")
}

setGeneric("uri", function(x) standardGeneric("uri"))

setMethod("uri", "Quickload", function(x) {
  x@uri
})

setMethod("genome", "Quickload", function(x) {
  contents <- Quickload_contents(x)
  as.character(structure(contents$dir, names = contents$title))
})

setMethod("names", "Quickload", function(x) genome(x))

setMethod("length", "Quickload", function(x) length(names(x)))

setMethod("[[", "Quickload", function (x, i, j, ...) {
  if (!missing(j))
    warning("argument 'j' ignored")
  QuickloadGenome(x, i, ...)
})

setMethod("$", "Quickload", function (x, name) {
  QuickloadGenome(x, name)
})

Quickload <- function(uri = "quickload", create = FALSE) {
  if (!isTRUEorFALSE(create))
    stop("'create' must be TRUE or FALSE")
  if (create) {
    if (uriExists(uri)) {
      message("NOTE: '", uri, "' already exists")
      create <- FALSE
    }
    else createResource(uri, dir = TRUE)
  }
  ql <- new("Quickload", uri = checkURI(uri))
  if (create)
    createResource(contentsFile(ql))
  ql
}

setAs("character", "Quickload", function(from) Quickload(from))

setMethod("show", "Quickload", function(object) {
  cat(class(object), "repository\nuri:", uri(object), "\n")
  cat(S4Vectors:::labeledLine("genomes", genome(object)))
})

contentsFile <- function(x) file.path(uri(x), "contents.txt")

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### QuickloadGenome class
###

setClass("QuickloadGenome",
         representation(quickload = "Quickload",
                        genome = "character"),
         contains = "TrackDb")

setMethod("uri", "QuickloadGenome",
          function(x) file.path(uri(quickload(x)), genome(x)))

quickload <- function(x, ...) x@quickload

setMethod("genome", "QuickloadGenome", function(x) x@genome)

setMethod("seqinfo", "QuickloadGenome", function(x) {
  genome_info <- read.table(genomeFile(x), sep = "\t",
                            col.names = c("seqnames", "seqlengths"),
                            colClasses = c("character", "integer"))
  Seqinfo(genome_info$seqnames, genome_info$seqlengths,
          genome = rep(genome(x), nrow(genome_info)))
})

setReplaceMethod("seqinfo", "QuickloadGenome", function(x, value)
                 {
                   if (uriIsWritable(genomeFile(x))) {
                     df <- as.data.frame(value)[1]
                     write.table(df, genomeFile(x),
                                 quote = FALSE, col.names = FALSE, sep = "\t")
                   } else stop("Repository is read only; cannot write seqinfo")
                   x
                 })

setMethod("releaseDate", "QuickloadGenome", function(x) {
  sub(".*?_(.*?)_([^_]*)$", "\\1 \\2", genome(x))
})

setMethod("organism", "QuickloadGenome", function(object) {
  gsub("_", " ", sub("(.*?)_.*?_[^_]*$", "\\1", genome(object)))
})

setMethod("length", "QuickloadGenome", function(x) {
  length(names(x))
})

QuickloadGenome_annotFiles <- function(x) {
  xmlChildren(xmlInternalTreeParse(annotsFile(x)))$files
}

setMethod("names", "QuickloadGenome", function(x) {
  x_mcols <- mcols(x)
  structure(curl_unescape(as.character(x_mcols$name)),
            names = as.character(x_mcols$title))
})

setMethod("trackNames", "QuickloadGenome", function(object) {
  names(object)
})

setMethod("mcols", "QuickloadGenome", function(x) {
  files <- QuickloadGenome_annotFiles(x)
  if (!length(xmlChildren(files)))
    make_zero_col_DFrame(length(x))
  else
    Reduce(function(x, y) merge(as.data.frame(as.list(x)),
                                as.data.frame(as.list(y)), all = TRUE),
           lapply(xmlChildren(files), xmlAttrs))
})

setMethod("toString", "QuickloadGenome", function(x) {
  Quickload_contents(quickload(x))[genome(x),"title"]
})

addGenomeToContents <- function(x, title) {
  contents <- Quickload_contents(quickload(x))
  if (!genome(x) %in% contents$dir) {
    contents <- rbind(contents, data.frame(genome(x), title))
    if (uriIsWritable(contentsFile(quickload(x))))
      write.table(contents, contentsFile(quickload(x)),
                  quote = FALSE, row.names = FALSE, col.names = FALSE,
                  sep = "\t")
    else stop("Repository is read only; cannot add genome to contents")
  } else warning("Genome '", genome(x), "' already in contents; not replaced")
}

setMethod("toString", "GenomeDescription", function(x) {
  paste(organism(x), provider(x), providerVersion(x))
})

QuickloadGenome <- function(quickload, genome, create = FALSE,
                            seqinfo = GenomicRanges::seqinfo(genome),
                            title = toString(genome))
{
  if (!isTRUEorFALSE(create))
    stop("'create' must be TRUE or FALSE")
  quickload <- as(quickload, "Quickload")
  genome <- normArgGenome(genome)
  genome_id <- quickloadGenomeId(genome)
  qlg <- new("QuickloadGenome", quickload = quickload, genome = genome_id)
  if (create) {
    if (is.character(genome) && missing(seqinfo))
      stop("No seqinfo for genome '", genome, "'")
    createQuickloadGenome(qlg, seqinfo, title)
  }
  qlg
}

setMethod("show", "QuickloadGenome", function(object) {
  cat(class(object), "track database\ngenome:", genome(object), "\nquickload:",
      uri(quickload(object)), "\n")
  cat(S4Vectors:::labeledLine("names", names(object)))
})

genomeFile <- function(x) file.path(uri(x), "mod_chromInfo.txt")
annotsFile <- function(x) file.path(uri(x), "annots.xml")

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import of tracks/sequence from Quickload
###

setMethod("track", "QuickloadGenome", function(object, name, ...) {
  object_mcols <- mcols(object)
  if (!name %in% object_mcols$title)
    stop("Track '", name, "' does not exist")
  md <- as.list(object_mcols[object_mcols$title == name,])
  rd <- import(file.path(uri(object), md$name), ...)
  metadata(rd)$quickload <- md
  rd
})

## Since a QuickloadGenome can store only a
## single referenece genome, it is best to treat it as a "slot".

setGeneric("referenceSequence",
           function(x, ...) standardGeneric("referenceSequence"))

referenceSequenceFile <- function(x) {
  paste(file.path(uri(x), genome(x)), ".2bit", sep = "")
}

setMethod("referenceSequence", "QuickloadGenome",
          function(x, which = as(seqinfo(x), "GenomicRanges"), ...)
          {
            import(referenceSequenceFile(x), which = which, ...)
          })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export of tracks/sequence to Quickload
###

### FIXME: check for file URI scheme

setGeneric("sortBySeqnameAndStart",
           function(x) standardGeneric("sortBySeqnameAndStart"))

## Seems not really possible for GRangesList

setMethod("sortBySeqnameAndStart", "GenomicRanges", function(x) {
  x[order(as.factor(seqnames(x)), start(x)),]
})

setMethod("sortBySeqnameAndStart", "IntegerRangesList", function(x) {
  x.flat <- unlist(x, use.names=FALSE)
  relist(x.flat[order(space(x), start(x.flat))], x)
})

.exportToQuickload <- function(object, name,
                               format = bestFileFormat(value, object),
                               index = TRUE, metadata = character(), ..., value)
{
  if (is(object, "Annotated")) {
    value_metadata <- metadata(value)$quickload
    value_metadata[names(metadata)] <- metadata
    metadata <- value_metadata
  }
  filename <- paste(name, format, sep = ".")
  path <- paste(uri(object), filename, sep = "/")
  seqinfo(value) <- seqinfo(object)
  file <- export(value, path, format = format, index = index, ...)
  track(object, name, index = FALSE, metadata = metadata) <- file
  object
}

setReplaceMethod("track",
                 signature(object = "QuickloadGenome", value = "ANY"),
                 .exportToQuickload)

setReplaceMethod("track",
                 signature(object = "QuickloadGenome", value = "RsamtoolsFile"),
                 function(object, name, metadata = character(), ..., value)
                 {
                   if (missing(name))
                     name <- basename(path(value))
                   track(object, name, metadata = metadata) <-
                     URLencode(path(value))
                   copyResourceToQuickload(object, URLencode(index(value)))
                   object
                 })

setReplaceMethod("track",
                 signature(object = "QuickloadGenome", value = "BiocFile"),
                 function(object, name, metadata = character(), ..., value)
                 {
                   if (missing(name))
                     name <- basename(path(value))
                   track(object, name, metadata = metadata) <-
                     URLencode(path(value))
                   object
                 })

copyResourceToQuickload <- function(object, uri) {
  parsed_uri <- .parseURI(uri)
  if (parsed_uri$scheme == "")
    uri <- paste("file://", uri, sep = "")
  filename <- basename(uri)
  object_uri <- .parseURI(uri(object))
  if (uriIsLocal(object_uri)) {
    dest_file <- file.path(object_uri$path, filename)
    if (paste(uri(object), filename, sep = "/") != uri)
### FIXME: curl_unescape() here because of R bug
      download.file(curl_unescape(uri), dest_file)
  } else stop("Quickload is not local; cannot copy track")
  filename
}

setReplaceMethod("track",
                 signature(object = "QuickloadGenome",
                           value = "character"),
                 function(object, name = basename(object),
                          metadata = character(), ..., value)
                 {
                   file <- copyResourceToQuickload(object, value)
                   files <- QuickloadGenome_annotFiles(object)
                   ## make sure we have the three required attributes
                   attrs <- c(name = file, title = name)
                   attrs[names(metadata)] <- metadata
                   filenames <- getNodeSet(files, "//@name")
                   if (file %in% filenames) {
                     removeChildren(files, match(file, filenames))
                   }
                   files <- addChildren(files,
                                        newXMLNode("file", attrs = attrs))
                   saveXML(files, annotsFile(object))
                   object
                 })

setGeneric("referenceSequence<-",
           function(x, ..., value) standardGeneric("referenceSequence<-"))

setReplaceMethod("referenceSequence",
                 signature(x = "QuickloadGenome", value = "ANY"),
                 function(x, name, ..., value)
                 {
                   export.2bit(value, referenceSequenceFile(x))
                   x
                 })

### Not exported yet
setGeneric("synonyms", function(x, ...) standardGeneric("synonyms"))

synonymFile <- function(x) file.path(uri(x), "synonyms.txt")

setMethod("synonyms", "QuickloadGenome", function(x) {
  CharacterList(strsplit(readLines(synonymFile(x)), "\t"))
})

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Utilities
###

normArgGenome <- function(genome) {
  if (is.character(genome)) {
    bs_genome <- BSGenomeForID(genome)
    if (!is.null(bs_genome))
      genome <- bs_genome
  }
  genome
}

quickloadGenomeId <- function(genome) {
  if (!is.character(genome)) {
    species <- sub("^(.).*? ", "\\1_", organism(genome))
    date <- sub("\\. ", "_", releaseDate(genome))
    paste(species, date, sep = "_")
  } else genome
}

createQuickloadGenome <- function(x, seqinfo, title) {
  if (genome(x) %in% genome(quickload(x))) {
    message("NOTE: Genome '", genome(x), "' already exists")
    return()
  }
  createResource(uri(x), dir = TRUE)
  createResource(annotsFile(x), content = "<files/>")
  seqinfo(x) <- seqinfo
  addGenomeToContents(x, title)
}

### =========================================================================
### UCSC 2bit support
### -------------------------------------------------------------------------
###

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### TwoBitFile class
###

setClass("TwoBitFile", contains = "BiocFile")
setClass("2BitFile", contains = "TwoBitFile")

twoBitPath <- function(path) {
  uri <- .parseURI(path)
  if (!uriIsLocal(uri))
    stop("TwoBit driver handles only local file paths")
  path.expand(uri$path)
}

TwoBitFile <- function(path) {
  if (!isSingleString(path))
    stop("'filename' must be a single string, specifying a path")
  new("TwoBitFile", resource = twoBitPath(path))
}
`2BitFile` <- TwoBitFile

.seqlengths_TwoBitFile <- function(x) {
    .Call(TwoBitFile_seqlengths, path(x))
}

fastaSeqnames <- function(x) {
    sub(" .*", "", x)
}

setMethod("seqinfo", "TwoBitFile", function(x) {
              seqlengths <- .seqlengths_TwoBitFile(x)
              names(seqlengths) <- fastaSeqnames(names(seqlengths))
              Seqinfo(names(seqlengths), seqlengths)
          })

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export
###

setGeneric("export.2bit",
           function(object, con, ...) standardGeneric("export.2bit"))

setMethod("export.2bit", "ANY", function(object, con, ...) {
  export(object, con, "2bit", ...)
})

setMethod("export", c("ANY", "TwoBitFile"), function(object, con, format, ...) {
  object <- as(object, "DNAStringSet")
  callGeneric()
})

setMethod("export", c("DNAStringSet", "TwoBitFile"),
          function(object, con, format) {
            if (!missing(format))
              checkArgFormat(con, format)  
            seqnames <- names(object)
            if (is.null(seqnames))
              seqnames <- as.character(seq(length(object)))
            freq <- alphabetFrequency(object)
            unsupported.chars <- setdiff(DNA_ALPHABET, c(DNA_BASES, "N"))
            if (any(uniqueLetters(object) %in% unsupported.chars)) {
              stop("One or more strings contain unsupported ambiguity ",
                   "characters.\nStrings can contain only A, C, G, T or N.",
                   "\nSee Biostrings::replaceAmbiguities().")
            }
            if (any(width(object) == 0L)) {
              stop("Empty strings are not yet supported")
            }
            invisible(.TwoBits_export(mapply(.DNAString_to_twoBit, object,
                                             seqnames),
                                      twoBitPath(path(con))))
          })

## Hidden export of a list of twoBit pointers.
## NOT exported (but used in the BSgenome package).
.TwoBits_export <- function(object, con) {
  if (!isSingleString(con))
    stop("'con' must be a single, non-NA string")
  if (!all(sapply(object,
                  function(x) typeof(x) == "externalptr" && is(x, "twoBit"))))
    stop("'object' must be a list of 'twoBit' pointers")
  .Call(TwoBits_write, object, con)
}

## NOT exported (but used in the BSgenome package).
.DNAString_to_twoBit <- function(object, seqname) {
  if (!isSingleString(seqname))
    stop("'seqname' must be a single, non-NA string")
  if (!is(object, "DNAString") && !is(object, "MaskedDNAString"))
    stop("'object' must be a DNAString")
  object_masks <- masks(object)
  if (!is.null(object_masks)) {
    object_mask <- collapse(object_masks)[[1]]
    active(masks(object)) <- FALSE # so we get the real sequence
  } else object_mask <- IRanges()
  .Call(DNAString_to_twoBit, as.character(object), object_mask, seqname)
}

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Import
###

setGeneric("import.2bit", function(con, ...) standardGeneric("import.2bit"))

setMethod("import.2bit", "ANY",
          function(con, ...)
          {
            import(con, "2bit", ...)
          })

setMethod("import", "TwoBitFile",
          function(con, format, text, which = as(seqinfo(con), "GenomicRanges"),
                   ...)
          {
            lkup <- get_seqtype_conversion_lookup("B", "DNA")
            sl <- .seqlengths_TwoBitFile(con)
            sn <- extractROWS(names(sl), match(seqnames(which), seqlevels(con)))
            if (any(is.na(sn))) {
                stop("'seqnames' not in 2bit file: ",
                     paste0("'", unique(seqnames(which)[is.na(sn)]), "'",
                            collapse=", "))
            }
            ans <- .Call(TwoBitFile_read, twoBitPath(path(con)),
                         sn, as(ranges(which), "IRanges"), lkup)
            names(ans) <- names(which)
            ans
          })

setMethod("getSeq", "TwoBitFile",
          function(x, which = as(seqinfo(x), "GenomicRanges")) {
              ans <- import(x, which = which)
              rc <- strand(which) == "-"
              ans[rc] <- reverseComplement(ans[rc])
              ans
          })

## central abstraction, represents a genome browser

## Question: is a browser session a database of tracks? No question
## that a browser has a track database. The question is of
## inheritance. Historically, browser session has had all the behavior
## we expect from a track database. But it is so much more. In
## particular, it has a list of views. We want to separate the data
## from the views. It is tempting to have the model (tracks) and views
## in separate objects, composed together in a session. However, the
## reality is that browser sessions are monolithic: UCSC can view only
## data UCSC knows, IGB can view only data IGB knows. Thus, we use
## simple inheritance here, saving many refactoring headaches.

setClass("BrowserSession", contains = c("TrackDb", "VIRTUAL"))

## alias for names() for clarity in the session context
setGeneric("trackNames", function(object, ...) standardGeneric("trackNames"))
setGeneric("trackNames<-", function(object, ..., value)
           standardGeneric("trackNames<-"))

setMethod("names", "BrowserSession", function(x) trackNames(x))
setMethod("trackNames", "BrowserSession", function(object) names(object))

## a single genome view in a session
setClass("BrowserView", representation(session = "BrowserSession"),
         contains = "VIRTUAL")

# create one or more browser views
setGeneric("browserView",
           function(object, range, track, ...)
           standardGeneric("browserView"))

# get the views from the browser
setGeneric("browserViews",
           function(object, ...) standardGeneric("browserViews"))

# get/set active view
setGeneric("activeView", function(object) standardGeneric("activeView"))
setGeneric("activeView<-", function(object, value)
           standardGeneric("activeView<-"))

setMethod("activeView", "BrowserSession",
          function(object)
          {
            views <- browserViews(object)
            active <- sapply(views, activeView)
            if (any(active))
              views[[tail(which(active),1)]]
            else NULL
          })

setMethod("show", "BrowserSession",
          function(object)
          {
            cat("A genome browser session of class '", class(object), "' with ",
                length(browserViews(object)), " views and ",
                length(trackNames(object)), " tracks\n", sep ="")
          })

# close a session or view
setGeneric("close", function(con, ...) standardGeneric("close"))

# FIXME: what about isOpen?

# get/set visibility of view tracks
setGeneric("visible", function(object, ...) standardGeneric("visible"))
setMethod("visible", "BrowserView", function(object) {
  trackNames(browserSession(object)) %in% trackNames(object)
})
          
setGeneric("visible<-",
           function(object, ..., value) standardGeneric("visible<-"))
setReplaceMethod("visible", "BrowserView", function(object, value) {
  trackNames(object) <- names(value)[value]
  object
})

# get/set the selected features in e.g. a view
# this can return a list, with a logical vector for each track
setGeneric("selectedFeatures", function(object, ...)
           standardGeneric("selectedFeatures"))
setGeneric("selectedFeatures<-", function(object, value)
           standardGeneric("selectedFeatures<-"))

setMethod("show", "BrowserView", function(object)
          {
            range <- range(object)
            cat(class(object), "of",
                paste(names(range), ":", unlist(start(range)), "-",
                      unlist(end(range)), sep = ""),
                "\n")
            nms <- paste("'", names(trackNames(object)), "'", sep = "")
            cat(S4Vectors:::labeledLine("trackNames", nms))
          })

setClass("BrowserViewList", contains = "SimpleList",
         prototype = prototype(elementType = "BrowserView"))

BrowserViewList <- function(...) {
  views <- list(...)
  if (length(views) == 1 && is.list(views[[1L]]))
    views <- views[[1L]]
  if (!all(sapply(views, is, "BrowserView")))
    stop("all elements in '...' must be BrowserView objects")
  S4Vectors:::new_SimpleList_from_list("BrowserViewList", views)
}

# get genome range of active view (or default if no views)
setMethod("range", "BrowserSession",
          function(x, ..., na.rm)
          {
            if (length(...) > 0)
              stop("arguments in '...' ignored")
            view <- activeView(x)
            if (!is.null(view))
              range(view)
            else NULL
          })

setGeneric("range<-", function(x, ..., value) standardGeneric("range<-"))

## just the genome, for convenience
setMethod("genome", "BrowserSession", function(x) genome(range(x)))
setReplaceMethod("genome", "BrowserSession", function(x, value) {
  if (!isSingleString(value))
    stop("'genome' must be a single string")
  genome(range(x)) <- value
  x
})

# high-level entry point
setGeneric("browseGenome",
           function(object, ...)
           standardGeneric("browseGenome"))

setMethod("browseGenome", "missing",
          function(object, ...) browseGenome(GRanges(compress=FALSE), ...))

setMethod("browseGenome", "GenomicRanges_OR_GenomicRangesList",
          function(object, browser = "UCSC",
                   range = base::range(object), view = TRUE,
                   trackParams = list(), viewParams = list(),
                   name = "customTrack", ...)
          {
            # initialize session of type identified by 'browser'
            session <- browserSession(browser, ...)
            # load 'object'
            trackParams <- c(list(session), trackParams)
            if (is(object, "GenomicRanges"))
              trackParams <- c(trackParams, name = name)
            session <- do.call(`track<-`, c(trackParams, list(value = object)))
            # open view of 'range'
            if (view) {
              if (!missing(range) && length(range) > 0L)
                range <- normGenomeRange(range, seqinfo(session))
              viewParams <- c(list(session, range), viewParams)
              do.call(browserView, viewParams)
            }
            session
          })

# list names of available genome browsers
genomeBrowsers <- function(where = topenv(parent.frame()))
{
  cl <- getClass("BrowserSession", where = where)
  browsers <- names(cl@subclasses)
  browsers <- browsers[!sapply(browsers, isVirtualClass, where)]
  sub("Session$", "", browsers)
}

# obtain a browser session
setGeneric("browserSession",
           function(object, ...) standardGeneric("browserSession"))

# instantiate a browser session by name
setMethod("browserSession", "character",
          function(object, ...)
          {
            class <- paste(object, "Session", sep = "")
            if (!extends(class, "BrowserSession"))
              stop("Browser named '", object, "' is unsupported.")
            new(class, ...)
          })

setMethod("browserSession", "missing",
          function(object, ...) browserSession("UCSC", ...))

# get one from a view
setMethod("browserSession", "BrowserView", function(object) object@session)

\name{UCSCFile-class}
\docType{class}

%% Classes:
\alias{class:UCSCFile}
\alias{UCSCFile-class}

%% Constructor:
\alias{UCSCFile}

%% Import:
\alias{import,UCSCFile,ANY,ANY-method}
\alias{import.ucsc}
\alias{import.ucsc,ANY-method}
\alias{import.ucsc,BiocFile-method}

%% Export:
\alias{export,ANY,UCSCFile,ANY-method}
\alias{export,GenomicRanges,UCSCFile,ANY-method}
\alias{export,GenomicRangesList,UCSCFile,ANY-method}
\alias{export,GRangesList,UCSCFile,ANY-method}
\alias{export,UCSCData,UCSCFile,ANY-method}
\alias{export.ucsc}
\alias{export.ucsc,ANY,ANY-method}
\alias{export.ucsc,ANY,BiocFile-method}

\title{UCSCFile objects}

\description{
  These functions support the import and export of tracks emucscded
  within the UCSC track line metaformat, whereby multiple tracks may be
  concatenated within a single file, along with metadata mostly oriented
  towards visualization. Any \code{\linkS4class{UCSCData}} object is
  automatically exported in this format, if the targeted format is known
  to be compatible. The BED and WIG import methods check for a track
  line, and delegate to these functions if one is found. Thus, calling
  this API directly is only necessary when importing embedded GFF
  (rare), or when one wants to create the track line during the export
  process.
}

\usage{
\S4method{import}{UCSCFile,ANY,ANY}(con, format, text,
                   subformat = "auto", drop = FALSE,
                   genome = NA, ...)
import.ucsc(con, ...)

\S4method{export}{ANY,UCSCFile,ANY}(object, con, format, ...)
\S4method{export}{GenomicRanges,UCSCFile,ANY}(object, con, format, ...)
\S4method{export}{GenomicRangesList,UCSCFile,ANY}(object, con, format,
                   append = FALSE, index = FALSE, ...)
\S4method{export}{UCSCData,UCSCFile,ANY}(object, con, format,
                   subformat = "auto", append = FALSE, index = FALSE, ...)
export.ucsc(object, con, ...)
}

\arguments{
  \item{con}{A path, URL, connection or \code{UCSCFile} object. For the
    functions ending in \code{.ucsc}, the file format is indicated by
    the function name. For the base \code{export} and \code{import}
    functions, \dQuote{ucsc} must be passed as the \code{format}
    argument.
  }
  \item{object}{The object to export, should be a \code{GRanges} or
    something coercible to a \code{GRanges}. For exporting multiple
    tracks pass a \code{GenomicRangesList}, or something coercible to one.
  }
  \item{format}{If not missing, should be \dQuote{ucsc}.}
  \item{text}{If \code{con} is missing, a character vector to use as the
    input
  }
  \item{subformat}{The file format to use for the actual features,
    between the track lines. Must be a text-based format that is
    compatible with track lines (most are). If an
    \code{\link[BiocIO:BiocFile-class]{BiocFile}} subclass other than
    \code{UCSCFile} is passed as \code{con} to \code{import.ucsc} or
    \code{export.ucsc}, the subformat is assumed to be the corresponding format
    of \code{con}. Otherwise it defaults to \dQuote{auto}.  The following
    describes the logic of the \dQuote{auto} mode. For import, the subformat is
    taken as the \code{type} field in the track line. If none, the file
    extension is consulted.  For export, if \code{object} is a
    \code{UCSCData}, the subformat is taken as the \code{type}
    in its track line, if present. Otherwise, the subformat is chosen
    based on whether \code{object} contains a \dQuote{score} column. If
    there is a score, the target is either \code{BEDGraph} or
    \code{WIG}, depending on the structure of the ranges. Otherwise,
    \code{BED} is the target.
  }
  \item{genome}{The identifier of a genome, or \code{NA} if
    unknown. Typically, this is a UCSC identifier like \dQuote{hg19}. An
    attempt will be made to derive the \code{seqinfo} on the return
    value using either an installed BSgenome package or UCSC, if network
    access is available. This defaults to the \code{db} BED track line
    parameter, if any.
  }
  \item{drop}{If \code{TRUE}, and there is only one track in the file,
    return the track object directly, rather than embedding it in a list.
  }
  \item{append}{If \code{TRUE}, and \code{con} points to a file path,
    the data is appended to the file. Obviously, if \code{con} is a
    connection, the data is always appended.
  }
  \item{index}{If \code{TRUE}, automatically compress and index the
    output file with bgzf and tabix. Note that tabix indexing will
    sort the data by chromosome and start. Tabix supports a
    single track in a file.
  }
  \item{...}{Should either specify track line parameters or arguments to
    pass down to the import and export routine for the subformat. 
  }
}

\value{
  A \code{GenomicRangesList} unless \code{drop} is \code{TRUE}
  and there is only a single track in the file. In that case, the first and
  only object is extracted from the list and returned.
  The structure of that object depends on the format of the
  data. The \code{GenomicRangesList} contains \code{UCSCData} objects.
}

\details{
  The UCSC track line permits the storage of multiple tracks in a single
  file by separating them with a so-called \dQuote{track line}, a line
  belonging with the word \dQuote{track} and containing
  various \code{key=value} pairs encoding metadata, most related to
  visualization. The standard fields in a track depend on the type of
  track being annotated. See \code{\linkS4class{TrackLine}} and its
  derivatives for how these lines are represented in R. The
  class \code{\linkS4class{UCSCData}} is an extension
  of \code{GRanges} with a formal slot for a \code{TrackLine}.
  Each \code{GRanges} in the returned \code{GenomicRangesList} has the
  track line stored in its metadata, under the \code{trackLine} key. 

  For each track object to be exported, if the object is not a
  \code{UCSCData}, and there is no \code{trackLine} element in the
  metadata, then a new track line needs to be generated. This happens
  through the coercion of \code{object} to \code{UCSCData}. The track line
  is initialized to have the appropriate \code{type} parameter for the
  subformat, and the required \code{name} parameter is taken from the
  name of the track in the input list (if any). Otherwise, the default
  is simply \dQuote{R Track}. The \code{db} parameter (specific to BED
  track lines) is taken as \code{genome(object)} if not
  \code{NA}. Additional arguments passed to the export routines
  override parameters in the provided track line.

  If the subformat is either WIG or BEDGraph, and the features are
  stranded, a separate track will be output in the file for each
  strand. Neither of those formats encodes the strand and disallow
  overlapping features (which might occur upon destranding).
}

\section{UCSCFile objects}{
  The \code{UCSCFile} class extends
  \code{\link[BiocIO:BiocFile-class]{BiocFile}} and is a
  formal represention of a resource in the UCSC format.
  To cast a path, URL or connection to a \code{UCSCFile}, pass it to
  the \code{UCSCFile} constructor.
}

\author{Michael Lawrence}

\references{
  \url{https://genome.ucsc.edu/goldenPath/help/customTrack.html}
}

\keyword{methods}
\keyword{classes}

\name{TrackHubGenome-class}
\docType{class}

%% Classes:
\alias{class:TrackHubGenome}
\alias{TrackHubGenome-class}

%% Constructor:
\alias{TrackHubGenome}

%% Accessors:
\alias{uri,TrackHubGenome-method}
\alias{genome,TrackHubGenome-method}
\alias{length,TrackHubGenome-method}
\alias{names,TrackHubGenome-method}
\alias{getTracks,TrackHubGenome-method}
\alias{trackhub}
\alias{organism,TrackHubGenome-method}
\alias{trackField,TrackHubGenome-method}
\alias{trackField<-,TrackHubGenome-method}
\alias{writeTrackHub,TrackHubGenome-method}

%% Data Access:
\alias{track,TrackHubGenome-method}
\alias{track<-,TrackHubGenome,ANY-method}
\alias{track<-,TrackHubGenome,RsamtoolsFile-method}
\alias{track<-,TrackHubGenome,character-method}
\alias{referenceSequence,TrackHubGenome-method}
\alias{referenceSequence<-,TrackHubGenome-method}

%% Show
\alias{show,TrackHubGenome-method}

\title{TrackHub Genome Access}

\description{
  A TrackHub data source is a collection of tracks and sequences,
  separated by genome. This class, \code{TrackHubGenome} provides
  direct access to the data for one particular genome.
}

\section{Constructor}{
  \describe{
    \item{\code{TrackHubGenome(trackhub, genome, create = FALSE}}{
        Constructs a
        new \code{TrackHubGenome} object, representing \code{genome} in
        the repository \code{trackhub} (a URI string or a
        \code{\linkS4class{TrackHub}} object).

        The \code{genome} argument can be an ID corresponding to a
        genome (potentially) in \code{trackhub} or an installed
        \code{BSgenome} package.

        If \code{create} is \code{TRUE}, and the trackDb file does not
        already exist, it will be created.
        Creation only works if the repository is local and
        writeable.
    }
  }
}

\section{Accessor Methods}{
  In the code snippets below, \code{x} represent a \code{TrackHubGenome} object.

  \describe{
    \item{\code{uri(x)}}{
      Get the uri pointing to the genome directory in the
      TrackHub repository.
    }
    \item{\code{genome(x)}}{
      Get the name of the genome, e.g.
      \dQuote{hg19}.
    }
    \item{\code{length(x)}}{
      number of tracks
    }
    \item{\code{names(x), trackNames(x)}}{
      names of the tracks
    }
    \item{\code{getTracks(x)}}{
      Get the \code{List} of \code{Track} from the tracks
    }
    \item{\code{trackhub(x)}}{
      Get the TrackHub object that contains this
      genome.
    }
    \item{\code{organism(x)}}{
      Get the organism name for this genome,
      e.g., \dQuote{H sapiens}.
    }
    \item{\code{trackField(x, name, field)}}{
      Get the \code{value} of \code{field} for \code{name} track.
    }
    \item{\code{trackField(x, name, field) <- value}}{
      Store the \code{field} and \code{value} for \code{name} track.
    }
    \item{\code{writeTrackHub(x)}}{
      Write tracks from memory representation to the trackDb file.
    }
  }
}

\section{Data Access}{
  \describe{
    \item{\code{track(x, name), x$name}}{
      get the track called \code{name}
    }
    \item{\code{track(x, name, format = bestFileFormat(value)) <-
        value, x$name <- value}}{
        store the track \code{value} under
        \code{name}. Note that track storing is only supported
        for local repositories, i.e., those with a \code{file://} URI
        scheme.

        Currently, supported \code{value} types include a
        \code{GenomicRanges}, \code{GRangesList}, or a file resource
        (copied to the repository).  The file resource may be
        represented as a path, URL, \code{\link[BiocIO:BiocFile-class]{BiocFile}} or
        \code{\link[Rsamtools:RsamtoolsFile-class]{RsamtoolsFile}}. If
        not a file name, \code{value} is written in \code{format}. For
        generic interval data, this means a BigWig file (if there is a
        numeric \dQuote{score} column) or a BED file otherwise. An
        \code{RleList} (e.g., coverage) is output as BigWig. For
        \code{UCSCData} values, the format is chosen according to the
        type of track line. For \code{RsamtoolsFile} objects, the file
        and its index are copied.
    }
    \item{\code{referenceSequence(x)}}{
      Get the reference sequence, as a
      \code{DNAStringSet}.
    }
    \item{\code{referenceSequence(x) <- value}}{
      Set the reference sequence, as a
      \code{DNAStringSet}. It is written as a 2bit file. This only works
      on local repositories.
    }
  }
}

\author{Michael Lawrence}

\examples{
tests_dir <- system.file("tests", package = "rtracklayer")
th <- TrackHub(file.path(tests_dir, "trackhub"))
thg <- TrackHubGenome(th, "hg19")
length(thg)
organism(thg)
names(thg)

\dontrun{
th <- TrackHub(file.path(tests_dir, "trackhub"), create = TRUE)
genomesFile(th) <- "genomes.txt"
genomeInfo(th) <- Genome(genome = "hg38", trackDb = "hg38/trackDb.txt")
genomeField(th, "hg38", "twoBitPath") <- "hg38/seq.2bit"
writeTrackHub(th)
thg <- TrackHubGenome(th, "hg38", create = TRUE)
seq <- import(file.path(tests_dir, "test.2bit"))
referenceSequence(thg) <- seq
track(thg, "PeaksData") <- paste0(tests_dir, "/test.bigWig")
trackField(thg, "wgEncodeUWDukeDnaseGM12878FdrPeaks", "bigDataUrl") <- "hg38/wgEncodeCshlShortRnaSeq.bigWig"
trackField(thg, "wgEncodeUWDukeDnaseGM12878FdrPeaks", "color") <- "8,104,172"
writeTrackHub(thg)
}
}

\keyword{methods}
\keyword{classes}

\name{activeView-methods}
\docType{methods}
\alias{activeView}
\alias{activeView<-}
\alias{activeView-methods}
\alias{activeView<--methods}
\alias{activeView,BrowserSession-method}
\alias{activeView,UCSCView-method}
\title{Accessing the active view}
\description{
 Get the active view.
}
\section{Methods}{
  The following methods are defined by \pkg{rtracklayer}.
  \describe{
    \item{object = "BrowserSession"}{
      \code{activeView(object)}: Gets the active
      \code{\linkS4class{BrowserView}} from a browser session.

      \code{activeView(object) <- value}: Sets the active
      \code{\linkS4class{BrowserView}} in a browser session.
    }
}}
\keyword{methods}

\name{UCSCView-class}
\docType{class}
\alias{UCSCView-class}
\alias{range,UCSCView-method}
\alias{range<-,UCSCView-method}
\alias{visible,UCSCView-method}
\alias{visible<-,UCSCView-method}

\alias{viewURL}

\title{Class "UCSCView"}
\description{An object representing a view of a genome in the
  \acronym{UCSC} browser.}
\section{Objects from the Class}{
  Calling \code{\link{browserView}(session, range = range(object), track
    = trackNames(object), browse = TRUE, ...)}
  creates \code{\linkS4class{BrowserView}} of \code{range} with visible
  tracks specified by \code{track}. \code{track} may be an instance of
  \code{\linkS4class{UCSCTrackModes}}. Arguments in \code{...} should
  match parameters to a
  \code{\link{ucscTrackModes}} method for creating a
  \code{UCSCTrackModes} instance that will be merged with and override
    modes indicated by the \code{track} parameter. If \code{browse} is
    \code{TRUE} (default), automatically launch a browser to display the
    view. 
}
\section{Slots}{
  \describe{
    \item{\code{form}}{Form parameters representing the client state}
    \item{\code{hgsid}}{Object of class \code{"numeric"}, which
      identifies this view to \acronym{UCSC}.}
    \item{\code{session}}{Object of class \code{"BrowserSession"} to
      which this view belongs.}
  }
}
\section{Extends}{
Class \code{"\linkS4class{BrowserView}"}, directly.
}
\section{Methods}{
  \describe{
    \item{\code{\link{activeView}(object)}}{ Obtains a logical
      indicating whether this view is the active view. }
    \item{\code{\link{range}(object)}}{ Obtains the
      \code{\link[GenomicRanges:GRanges-class]{GRanges}} displayed by
      this view. } 
    \item{\code{\link{range}(object) <- value}}{ Sets the
      \code{\link[GenomicRanges]{GRanges}} or
      \code{\link[IRanges]{IntegerRangesList}} displayed by
      this view.  } 
    \item{\code{\link{trackNames}(object)}}{ Gets the names of the visible
      tracks in this view. }
    \item{\code{\link{trackNames}(object) <- value}}{ Sets the visible
      tracks by name. }
    \item{\code{visible(object)}}{Get a named logical vector indicating
      whether each track is visible.
    }
    \item{\code{visible(object) <- value}}{Set a logical vector
      indicating the visibility of each track, in the same order as
      returned by \code{visible(object)}.
    }
    \item{\code{\link{ucscTrackModes}(object)}}{ Obtains the
      \code{\linkS4class{UCSCTrackModes}} for this view. }
    \item{\code{\link{ucscTrackModes}(object) <- value}}{ Sets the
      \code{\linkS4class{UCSCTrackModes}} for this view. The
      \code{value} may be either a \code{UCSCTrackModes} instance or a
      character vector that will be coerced by a call to
      \code{\link{ucscTrackModes}}. }
    
  }
}

\section{Utilities}{
  \describe{
    \item{viewURL(x)}{Gets the URL corresponding to the view
    \code{x}. Typically used when passing \code{browse=FALSE} to the
    constructor in order to display the view in another way.}
  }
}

\author{ Michael Lawrence }
\seealso{
  \code{\link{browserView}} for creating instances of this class.
}
\keyword{classes}

\name{asBED}
\alias{asBED}
\alias{asBED,GRangesList-method}
\alias{asBED,GAlignments-method}

\title{Coerce to BED structure}
\description{
  Coerce the structure of an object to one following BED-like
  conventions, i.e., with columns for blocks and thick regions.
}
\usage{
asBED(x, ...)
\S4method{asBED}{GRangesList}(x)
\S4method{asBED}{GAlignments}(x)
}

\arguments{
  \item{x}{
    Generally, a tabular object to structure as BED
  }
  \item{\dots}{
    Arguments to pass to methods
  }
}

\details{
  The exact behavior depends on the class of \code{object}.
  \describe{
    \item{\code{GRangesList}}{This treats \code{object} as if it were a
      list of transcripts, i.e., each element contains the exons of a
      transcript. The \code{blockStarts} and
      \code{blockSizes} columns are derived from the ranges in each
      element. Also, add \code{name} column from \code{names(object)}.
    }
    \item{\code{GAlignments}}{Converts to GRangesList via \code{grglist}
      and procedes accordingly.
    }
  }
}

\value{
  A \code{GRanges}, with the columns \code{name},
  \code{blockStarts} and \code{blockSizes} added.
}

\author{
  Michael Lawrence
}

\examples{
\dontrun{
library(TxDb.Hsapiens.UCSC.hg19.knownGene)
exons <- exonsBy(TxDb_Hsapiens_UCSC_hg19_knownGene)
mcols(asBED(exons))
}
}

\name{UCSCTrackModes-class}
\docType{class}
\alias{UCSCTrackModes-class}
\alias{[,UCSCTrackModes,ANY,ANY,ANY-method}
\alias{[<-,UCSCTrackModes,ANY,ANY,ANY-method}

\title{Class "UCSCTrackModes"}
\description{A vector of view modes ("hide", "dense", "full",
  "pack", "squish") for each track in a \acronym{UCSC} view.}
\section{Objects from the Class}{
  Objects may be created by calls of the form
  \code{\link{ucscTrackModes}(object =
    character(), hide = character(), dense = character(), pack = character(),
    squish = character(), full = character())},
  where \code{object} should be a character vector of mode names (with
  its \code{names} attribute specifying the corresponding track names).
  The other parameters should contain track names that override the
  modes in \code{object}. Later parameters override earlier ones, so,
  for example, if a track is named in \code{hide} and \code{full},
  it is shown in the full view mode.
}
\section{Slots}{
  \describe{
    \item{\code{.Data}:}{Object of class \code{"character"} holding the
      modes ("hide", "dense", "full", "pack", "squish"), with its
      \code{names} attribute holding corresponding track names.}
    \item{\code{labels}:}{Object of class \code{"character"} holding
      labels (human-readable names) corresponding to each track/mode.}
  }
}
\section{Extends}{
Class \code{"\linkS4class{character}"}, from data part.
Class \code{"\linkS4class{vector}"}, by class "character", distance 2.
}

\section{Methods}{
  \describe{
    \item{\code{\link{trackNames}(object)}}{ Gets the names of the visible
      tracks (those that do not have mode "hide"). }
    \item{\code{\link{trackNames}(object) <- value}}{ Sets the names of the
      visible tracks. Any tracks named in \code{value} are set to "full"
      if the are currently set to "hide" in this object. Any tracks not
      in \code{value} are set to "hide". All other modes are preserved.}
    \item{\code{object[i]}}{Gets the track mode of the tracks indexed
      by \code{i}, which can be any type of index supported by character
      vector subsetting. If \code{i} is a character vector, it indexes
      first by the internal track IDs (the \code{names} on \code{.Data})
      and then by the user-level track names (the \code{labels} slot).}
    \item{\code{object[i] <- value}}{Sets the track modes indexed by
      \code{i} (in the same way as in \code{object[i]} above) to those
      specified in \code{value}.} 
  }
}
\author{ Michael Lawrence }
\seealso{
  \code{\linkS4class{UCSCView}} on which track view modes may be set.
}
\keyword{classes}

\name{WIGFile-class}
\docType{class}

%% Classes:
\alias{class:WIGFile}
\alias{WIGFile-class}

%% Constructor:
\alias{WIGFile}

%% Import:
\alias{import,WIGFile,ANY,ANY-method}
\alias{import.wig}
\alias{import.wig,ANY-method}

%% Export:
\alias{export,ANY,WIGFile,ANY-method}
\alias{export,GenomicRanges,WIGFile,ANY-method}
\alias{export,SimpleGRangesList,WIGFile,ANY-method}
\alias{export,UCSCData,WIGFile,ANY-method}
\alias{export.wig}
\alias{export.wig,ANY-method}

\title{WIG Import and Export}

\description{
  These functions support the import and export of the UCSC WIG (Wiggle)
  format.
}

\usage{
\S4method{import}{WIGFile,ANY,ANY}(con, format, text, genome = NA,
                    trackLine = TRUE, which = NULL, seqinfo = NULL, ...)
import.wig(con, ...)

\S4method{export}{ANY,WIGFile,ANY}(object, con, format, ...)
\S4method{export}{GenomicRanges,WIGFile,ANY}(object, con, format,
                   dataFormat = c("auto", "variableStep", "fixedStep"),
                   writer = .wigWriter, append = FALSE, ...)
\S4method{export}{SimpleGRangesList,WIGFile,ANY}(object, con, format, ...)
\S4method{export}{UCSCData,WIGFile,ANY}(object, con, format,
                   trackLine = TRUE, ...)
export.wig(object, con, ...)
}

\arguments{
  \item{con}{A path, URL, connection or \code{WIGFile} object. For the
    functions ending in \code{.wig}, the file format is indicated by the
    function name. For the base \code{export} and \code{import}
    functions, the format must be indicated another way. If \code{con}
    is a path, URL or connection, either the file extension or the
    \code{format} argument needs to be \dQuote{wig}. Compressed files
    (\dQuote{gz}, \dQuote{bz2} and \dQuote{xz}) are handled
    transparently.
  }
  \item{object}{The object to export, should be a \code{GRanges} or
    something coercible to a \code{GRanges}. For exporting multiple
    tracks, in the UCSC track line metaformat, pass a
    \code{GenomicRangesList}, or something coercible to one.
  }
  \item{format}{If not missing, should be \dQuote{wig}.
  }
  \item{text}{If \code{con} is missing, a character vector to use as the
    input
  }
  \item{trackLine}{Whether to parse/output a UCSC track line. An
    imported track line will be stored in a \code{\linkS4class{TrackLine}}
    object, as part of the returned \code{\linkS4class{UCSCData}}.
  }
  \item{genome}{The identifier of a genome, or \code{NA} if
    unknown. Typically, this is a UCSC identifier like \dQuote{hg19}. An
    attempt will be made to derive the \code{seqinfo} on the return
    value using either an installed BSgenome package or UCSC, if network
    access is available.
  }
  \item{seqinfo}{If not \code{NULL}, the \code{Seqinfo} object to set on
    the result. If the \code{genome} argument is not \code{NA}, it must
    agree with \code{genome(seqinfo)}.
  }
  \item{which}{A range data structure like \code{IntegerRangesList}
    or \code{GRanges}. Only the intervals in the file overlapping the
    given ranges are returned. This is inefficient; use BigWig for
    efficient spatial queries.
  }
  \item{append}{If \code{TRUE}, and \code{con} points to a file path,
    the data is appended to the file. Obviously, if \code{con} is a
    connection, the data is always appended.
  }
  \item{dataFormat}{Probably best left to \dQuote{auto}. Exists only
    for historical reasons.
  }
  \item{writer}{Function for writing out the blocks; for internal use
    only.
  } 
  \item{...}{Arguments to pass down to methods to other methods. For
    import, the flow eventually reaches the \code{WIGFile} method on
    \code{import}. When \code{trackLine} is
    \code{TRUE}, the arguments are passed through \code{export.ucsc}, so
    track line parameters are supported.
  }
}

\value{
  A \code{GRanges} with the score values in the \code{score} metadata column,
  which is accessible via the \code{score} function.
}

\details{
  The WIG format is a text-based format for efficiently representing a
  dense genome-scale score vector. It encodes, for each feature, a range
  and score. Features from the same sequence (chromosome) are grouped
  together into a block, with a single block header line indicating the
  chromosome. There are two block formats: fixed step and variable
  step. For fixed step, the number of positions (or step) between
  intervals is the same across an entire block. For variable step, the
  start position is specified for each feature. For both fixed and
  variable step, the span (or width) is specified in the header and thus
  must be the same across all features. This requirement of uniform
  width dramatically limits the applicability of WIG. For scored
  features of variable width, consider \link[=BEDGraphFile]{BEDGraph} or
  \link[=BigWigFile]{BigWig}, which is generally preferred over both WIG
  and BEDGraph. To efficiently convert an existing WIG or BEDGraph file
  to BigWig, call \code{\link{wigToBigWig}}. Neither WIG, BEDGraph nor
  BigWig allow overlapping features.
}

\section{WIGFile objects}{
  The \code{WIGFile} class extends \code{\link[BiocIO:BiocFile-class]{BiocFile}}
  and is a formal represention of a resource in the WIG format.
  To cast a path, URL or connection to a \code{WIGFile}, pass it to
  the \code{WIGFile} constructor.
}

\author{Michael Lawrence}

\references{
  \url{https://genome.ucsc.edu/goldenPath/help/wiggle.html}
}

\examples{
  test_path <- system.file("tests", package = "rtracklayer")
  test_wig <- file.path(test_path, "step.wig")

  ## basic import calls
  test <- import(test_wig)
  test
  import.wig(test_wig)
  test_wig_file <- WIGFile(test_wig)
  import(test_wig_file)
  test_wig_con <- file(test_wig)
  import(test_wig_con, format = "wig")
  test_wig_con <- file(test_wig)
  import(WIGFile(test_wig_con))

  ## various options
  import(test_wig, genome = "hg19")
  import(test_wig, trackLine = FALSE)
  which <- as(test[3:4,], "IntegerRangesList")
  import(test_wig, which = which)

\dontrun{
  ## basic export calls
  test_wig_out <- file.path(tempdir(), "test.wig")
  export(test, test_wig_out)
  export.wig(test, test_wig_out)
  test_foo_out <- file.path(tempdir(), "test.foo")
  export(test, test_foo_out, format = "wig")
  test_wig_out_file <- WIGFile(test_wig_out)
  export(test, test_wig_out_file)

  ## appending
  test2 <- test
  metadata(test2)$trackLine <- initialize(metadata(test)$trackLine,
                                          name = "test2")
  export(test2, test_wig_out_file, append = TRUE)

  ## passing track line parameters
  export(test, test_wig_out, name = "test2")

  ## no track line
  export(test, test_wig_out, trackLine = FALSE)
  
  ## gzip
  test_wig_gz <- paste(test_wig_out, ".gz", sep = "")
  export(test, test_wig_gz)
}
}

\keyword{methods}
\keyword{classes}

\name{GenomicData}

% constructor
\alias{GenomicData}

% accessors
\alias{chrom}
\alias{chrom,GRanges-method}
\alias{chrom<-}
\alias{chrom<-,GRanges-method}
\alias{score,ANY-method}

\title{Data on a Genome}

\description{The \code{rtracklayer} package adds
convenience methods on top of \code{GenomicRanges} and \code{IntegerRangesList}
to manipulate data on genomic ranges.}

\section{Accessors}{
  In the code snippets below,
  \code{x} is a \code{GenomicRanges} or \code{IntegerRangesList} object.

  \describe{
    \item{\code{chrom(x), chrom(x) <- value}}{
      Gets or
      sets the chromosome names for \code{x}. The length of \code{value}
      should equal the length of \code{x}.
    }
    \item{\code{score(x)}}{
      Gets the \dQuote{score} column from the element
      metadata of a \code{GenomicRanges} or \code{GRangesList}. Many
      track formats have a score column, so this is often used during
      export. The \code{ANY} fallback for this method simply returns
      \code{NULL}.
    }
  }
}

\section{Constructor}{
  \describe{
    \item{\code{GenomicData(ranges, ..., strand = NULL, chrom = NULL,
        genome = NULL)}}{
      Constructs a \code{GRanges} instance with
      the given \code{ranges} and variables in \code{...} (see the
      \code{\link[GenomicRanges]{GRanges}} constructor).

      If non-\code{NULL}, the \code{strand} argument specifies the strand
      of each range. It should be a character vector or factor of length
      equal to that of \code{ranges}. All values should be either \code{-},
      \code{+}, or \code{*}. To get the levels for \code{strand}, call
      \code{levels(strand())}.

      \code{chrom} argument is analogous to \code{seqnames} in the
      \code{GRanges} constructor.

      The \code{genome} argument should be a scalar string. See the examples.
    }
  }
}

\author{ Michael Lawrence and Patrick Aboyoun }

\examples{
  range1 <- IRanges(c(1,2,3), c(5,2,8))

  ## with some data ##
  filter <- c(1L, 0L, 1L)
  score <- c(10L, 2L, NA)
  strand <- factor(c("+", NA, "-"), levels = levels(strand()))
  ## GRanges instance
  gr <- GenomicData(range1, score, chrom = "chr1", genome = "hg18")
  mcols(gr)[["score"]]
  strand(gr) ## all '*'
  gr <- GenomicData(range1, score, filt = filter, strand = strand,
                    chrom = "chr1")
  mcols(gr)[["filt"]]
  strand(gr) ## equal to 'strand'
  ## coercion from data.frame ##
  df <- as.data.frame(gr)
}
\keyword{classes}
\keyword{methods}

\name{GRangesForUCSCGenome}
\alias{GRangesForUCSCGenome}
\alias{GRangesForBSGenome}
\alias{SeqinfoForUCSCGenome}
\alias{SeqinfoForBSGenome}

\title{
  GRanges for a Genome
}
\description{
  These functions assist in the creation of
  \code{\link[Seqinfo]{Seqinfo}} or
  \code{\link[GenomicRanges:GRanges-class]{GRanges}} for a genome.
}
\usage{
GRangesForUCSCGenome(genome, chrom = NULL, ranges = NULL, ...)
GRangesForBSGenome(genome, chrom = NULL, ranges = NULL, ...)

SeqinfoForUCSCGenome(genome)
SeqinfoForBSGenome(genome)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{genome}{
    A string identifying a genome, usually one assigned by UCSC, like "hg19".
  }
  \item{chrom}{
    A character vector of chromosome names, or \code{NULL}.
  }
  \item{ranges}{
    A \code{\link[IRanges]{IntegerRanges}} object with the intervals.
  }
  \item{\dots}{
    Additional arguments to pass to
    the \code{\link[GenomicRanges]{GRanges}} constructor.
  }
}
\details{
  The genome ID is stored in the metadata of the ranges and is
  retrievable via the \code{\link[Seqinfo]{genome}} function. The sequence
  lengths are also properly initialized for the genome. This mitigates
  the possibility of accidentally storing intervals for the wrong
  genome.

  \code{GRangesForUCSCGenome} obtains sequence information from the UCSC
  website, while \code{GRangesForBSGenome} looks for it in an
  installed \code{BSGenome} package. Using the latter is more efficient
  in the long-run, but requires downloading and installing a potentially
  large genome package, or creating one from scratch if it does not yet
  exist for the genome of interest.
}
\value{
  For the \code{GRangesFor*} functions, a \code{GRanges} object, with the
  appropriate \code{\link[Seqinfo]{seqlengths}} and
  \code{\link{genome}} ID.

  The \code{SeqinfoFor*} functions return a \code{Seqinfo} for the
  indicated genome.
}
\author{
  Michael Lawrence
}

\name{blocks-methods}
\docType{methods}
\alias{blocks}
\alias{blocks-methods}
\alias{blocks,GenomicRanges-method}
\title{Get blocks/exons}
\description{
 Obtains the block ranges (subranges, usually exons) from an object,
 such as a \code{\link[GenomicRanges]{GRanges}} imported
 from a BED file.
}
\usage{
blocks(x, ...)
}
\arguments{
  \item{x}{The instance from which to obtain the block/exon
    information. Currently must be a \code{GenomicRanges}, with a metadata
    column of name \dQuote{blocks} and of type \code{IntegerRangesList}.
    Such an object is returned by \code{\link{import.bed}} and
    \code{\link{asBED}}.
  }
  \item{...}{Additional arguments for methods}
}
\value{
  A \code{\link[GenomicRanges:GRangesList-class]{GRangesList}} with an
  element for each range in \code{x}. The original block ranges are
  relative to the start of the containing range, so the returned ranges
  are shifted to absolute coordinates. The seqname and strand are
  inherited from the containing range.
}
\author{Michael Lawrence}
\seealso{\code{\link{import.bed}} for importing a track from BED, which
  can store block information; \code{\link{asBED}} for coercing a
  \code{GenomicRanges} into a BED-like structure that can be passed to
  this function.
}
\keyword{methods}
\keyword{manip}

\name{BigBedFile-class}
\docType{class}

%% Classes:
\alias{class:BigBedFile}
\alias{BigBedFile-class}
\alias{class:BBFile}
\alias{BBFile-class}

%% Constructor:
\alias{BigBedFile}
\alias{BBFile}

%% Accessors:
\alias{seqinfo,BigBedFile-method}

%% Import:
\alias{import.bb}
\alias{import.bb,ANY-method}
\alias{import,BigBedFile,ANY,ANY-method}

%% Export:
\alias{export.bb}
\alias{export.bb,ANY-method}
\alias{export,ANY,BigBedFile,ANY-method}
\alias{export,GenomicRanges,BigBedFile,ANY-method}

\title{BigBed Import and Export}

\description{
  These functions support the import and export of the UCSC BigBed
  format, a compressed, binary form of BED with a spatial index
  and precomputed summaries. These functions do not work on Windows.
}

\usage{
\S4method{import}{BigBedFile,ANY,ANY}(con, format, text,
                   selection = BigBedSelection(which, ...),
                   which = con, ...)
import.bb(con, ...)

\S4method{export}{ANY,BigBedFile,ANY}(object, con, format, ...)
\S4method{export}{GenomicRanges,BigBedFile,ANY}(object, con, format,
                   compress = TRUE, extraIndexes = "")
export.bb(object, con, ...)
}


\arguments{
  \item{con}{A path, URL or \code{BigBedFile} object. Connections are
    not supported. For the functions ending in \code{.bb}, the file
    format is indicated by the function name. For the \code{export}
    and \code{import} methods, the format must be indicated another
    way. If \code{con} is a path, or URL, either the file
    extension or the \code{format} argument needs to be \dQuote{bigBed}
    or \dQuote{bb}.
  }
  \item{object}{The object to export, should be \code{GRanges}.
  }
  \item{format}{If not missing, should be \dQuote{bigBed} or \dQuote{bb}
    (case insensitive).
  }
  \item{text}{Not supported.
  }
  \item{selection}{A \code{\linkS4class{BigBedSelection}} object
    indicating the ranges to load.
  }
  \item{which}{A range data structure coercible to \code{IntegerRangesList},
    like a \code{GRanges}, or a \code{BigBedFile}. Only the intervals in
    the file overlapping the given ranges are returned. By default, the
    value is the \code{BigBedFile} itself. Its \code{Seqinfo} object is
    extracted and coerced to a \code{IntegerRangesList} that represents the
    entirety of the file.
  }
  \item{compress}{If \code{TRUE}, compress the data. No reason to change this.
  }
  \item{extraIndexes}{If set, make an index on each field in a comma separated list
  }
  \item{...}{Arguments to pass down to methods to other methods. For
    import, the flow eventually reaches the \code{BigBedFile} method on
    \code{import}.
  }
}

\section{\code{BigBedFile} objects}{
  A \code{BigWigFile} object, an extension of
  \code{\link[BiocIO:BiocFile-class]{BiocFile}} is a reference to a BigBed file. To cast
  a path, URL or connection to a \code{BigBedFile}, pass it to the
  \code{BigBedFile} constructor.

  BigBed files are more complex than most track files, and there are a
  number of methods on \code{BigBedFile} for accessing the additional
  information:

  \describe{
    \item{\code{seqinfo(x)}}{
      Gets the \code{\link[Seqinfo]{Seqinfo}} object
      indicating the lengths of the sequences for the intervals in the
      file. No circularity or genome information is available.
    }
  }

  When accessing remote data, the UCSC library caches data in the
  \file{/tmp/udcCache} directory. To clean the cache, call
  \code{cleanBigBedCache(maxDays)}, where any files older than
  \code{maxDays} days old will be deleted.
}

\author{Michael Lawrence}

\examples{
if (.Platform$OS.type != "windows") {
  test_path <- system.file("tests", package = "rtracklayer")
  test_bb <- file.path(test_path, "test.bb")

  ## Returns ranges with all fields
  gr <- import(test_bb)
  gr

  ## Retuns ranges only for 'chr10'
  ## between 170000-180000 with all fields
  which <- GRanges(c("chr10"), IRanges(c(170000, 180000)))
  import(test_bb, which = which)

  ## Retuns ranges only for 'chr10' 
  ## between 170000-180000 with name and peak fields
  selection <- BigBedSelection(which, colnames = c("name", "peak"))
  import(test_bb, selection = selection)

\dontrun{
  test_bb_out <- file.path(tempdir(), "test_out.bb")
  export(test, test_bb_out)

  ## make an index for 'name'
  test_bb_out <- file.path(tempdir(), "test_out.bb")
  export(test, test_bb_out, extraIndexes = "name")
}
}
}

\keyword{methods}
\keyword{classes}

\name{BrowserViewList-class}
\docType{class}
\alias{BrowserViewList-class}

% Constructor
\alias{BrowserViewList}

\title{Lists of BrowserView} 

\description{ A formal list of
  \code{\linkS4class{BrowserView}} objects. Extends and inherits all
  its methods from
  \code{\link[S4Vectors]{Vector}}. Usually generated by
  passing multiple ranges to the \code{\link{browserView}} function.
}

\section{Constructor}{
  \describe{
  \item{\code{BrowserViewList(...)}}{Concatenates the
      \code{BrowserView} objects in \code{...} into a new
      \code{BrowserViewList}.  This is rarely called by the user.
    } 
  } 
}

\author{Michael Lawrence}

\keyword{classes}

\name{browserSession-methods}
\docType{methods}
\alias{browserSession}
\alias{browserSession-methods}
\alias{browserSession,character-method}
\alias{browserSession,BrowserView-method}
\alias{browserSession,missing-method}
\title{ Get a genome browser session }
\description{
  Methods for getting browser sessions.
}
\section{Methods}{
  The following methods are defined by \pkg{rtracklayer}. 
  
  \describe{
    \item{object = "character"}{ \code{browserSession(object, ...)}: Creates a
      \code{\linkS4class{BrowserSession}} from a genome browser
      identifier. The identifier corresponds to the prefix of the
      session class name (e.g. "UCSC" in "UCSCSession"). The arguments
      in \dots are passed to the initialization function of the class.}
    
    \item{object = "browserView"}{ Gets the
      \code{\linkS4class{BrowserSession}} for the view. }

    \item{object = "missing"}{ Calls \code{browserSession("ucsc", ...)}. }
  }
}
\keyword{methods}

\name{browserView-methods}
\docType{methods}
\alias{browserView}
\alias{browserView-methods}
\alias{browserView,UCSCSession-method}
\title{Getting browser views}
\description{
  Methods for creating and getting browser views.
}
\usage{
  browserView(object, range, track, ...)
}
\arguments{
  \item{object}{The object from which to get the views.}
  \item{range}{The \code{\link[GenomicRanges]{GRanges}} or
    \code{\link[IRanges]{IntegerRangesList}} to
    display. If there are multiple elements, a view is created for each
    element and a \code{\linkS4class{BrowserViewList}} is
    returned.} 
  \item{track}{List of track names to make visible in the view.}
  \item{\dots}{Arguments to pass to methods}
}
\section{Methods}{
  The following methods are defined by \pkg{rtracklayer}. 
  \describe{
    
    \item{object = "UCSCSession"}{ \code{browserView(object, range =
        range(object), track = trackNames(object), imagewidth = 800,
        ...)}: Creates a 
      \code{\linkS4class{BrowserView}} of \code{range} with visible
      tracks specified by \code{track}. The \code{imagewidth} parameter
      specifies the width of the track image in pixels. \code{track}
      may be an instance 
      of \code{\linkS4class{UCSCTrackModes}}. Arguments in \code{...}
      are passed to \code{\link{ucscTrackModes}} to create the
      \code{UCSCTrackModes} instance that will override modes
      indicated by the \code{track} parameter.
   }
    
}}
\examples{
\dontrun{
  session <- browserSession()
  browserView(session,
              GRangesForUCSCGenome("hg19", "chr2", IRanges(20000, 50000)))
  ## only view "knownGene" track
  browserView(session, track = "knownGene")
}
}
\keyword{methods}

\name{wigToBigWig}
\alias{wigToBigWig}
\title{
  Convert WIG to BigWig
}
\description{
  This function calls the Kent C library to efficiently convert a WIG
  file to a BigWig file, without loading the entire file into
  memory. This solves the problem where simple tools write out text WIG
  files, instead of more efficiently accessed binary BigWig files.
}
\usage{
wigToBigWig(x, seqinfo,
            dest = paste(file_path_sans_ext(x, TRUE), "bw", sep = "."),
            clip = FALSE)
}
\arguments{
  \item{x}{
    Path or URL to the WIG file. Connections are not supported.
  }
  \item{seqinfo}{
    \code{\link[Seqinfo]{Seqinfo}} object, describing the genome of the
    data. All BigWig files must have this defined.
  }
  \item{dest}{
    The path to which to write the BigWig file. Defaults to \code{x}
    with the extension changed to \dQuote{bw}.
  }
  \item{clip}{
    If \code{TRUE}, regions outside of \code{seqinfo} will be clipped,
    so that no error is thrown.
  }
}

\author{
  Michael Lawrence
}

\seealso{
  \code{\link[=BigWigFile]{BigWig}} import and export support
}

\name{TabixFile-methods}

\alias{import,TabixFile,ANY,ANY-method}

\alias{exportToTabix}
\alias{exportToTabix,ANY,character-method}

% undocumented/internal for now
\alias{export,GenomicRanges,TabSeparatedFile,ANY-method}
\alias{export,ANY,TabSeparatedFile,ANY-method}

\title{TabixFile Import/Export}

\description{
  These methods support the import and export of
  \link[Rsamtools:TabixFile]{TabixFile} objects. These are generally
  useful when working with tabix-indexed files that have a non-standard
  format (i.e., not BED nor GFF), as well as exporting an object with
  arbitrary columns (like a GRanges) to an indexed, tab-separated
  file. This relies on the tabix header, which indicates the columns in
  the file that correspond to the chromosome, start and end. The BED and
  GFF parsers handle tabix transparently.
}

\usage{
\S4method{import}{TabixFile,ANY,ANY}(con, format, text,
                   which = if (is.na(genome)) NULL
                           else as(seqinfoForGenome(genome), "GenomicRanges"),
                   genome = NA, header = TRUE, ...)
exportToTabix(object, con, ...)
}
\arguments{
  \item{con}{For \code{import}, a \code{TabixFile} object;
    for \code{exportToTabix}, a string naming the destination file.
  }
  \item{object}{The object to export. It is coerced to a
    \code{data.frame}, written to a tab-separated file, and indexed with
    tabix for efficient range-based retrieval of the data using
    \code{import}.
  }
  \item{format}{If any known format, like \dQuote{bed} or \dQuote{gff}
    (or one of their variants), then the appropriate parser is
    applied. If any other value, then the tabix header is consulted for
    the format. By default, this is taken from the file extension.
  }
  \item{text}{Ignored.}
  \item{which}{A range data structure coercible to \code{IntegerRangesList},
    like a \code{GRanges}. Only the intervals in the file overlapping the
    given ranges are returned. The default is to use the range over the
    entire genome given by \code{genome}, if specified.
  }
  \item{genome}{The identifier of a genome, or \code{NA} if
    unknown. Typically, this is a UCSC identifier like \dQuote{hg19}. An
    attempt will be made to derive the \code{seqinfo} on the return
    value using either an installed BSgenome package or UCSC, if network
    access is available.
  }
  \item{header}{If \code{TRUE}, then the header in the indexed file,
    which might include a track line, is sent to the parser. Otherwise,
    the initial lines are skipped, according to the \code{skip} field in
    the tabix index header.
  }
  \item{...}{Extra arguments to pass to the underlying import routine,
    which for non-standard formats is \code{\link{read.table}} or
    \code{\link{write.table}}.
  }
}
\value{
  For \code{import}, a \code{GRanges} object.

  For \code{exportToTabix}, a \code{TabixFile} object that is directly
  passable to \code{import}.
}
\references{
  \url{http://samtools.sourceforge.net/tabix.shtml}
}
\seealso{
  \code{\link[Rsamtools]{scanTabix}} and friends
}
\author{Michael Lawrence}

\name{IntegerRangesList-methods}
\docType{methods}
\alias{IntegerRangesList-methods}

% accessors
\alias{chrom,IntegerRangesList-method}
\alias{chrom<-,IntegerRangesList-method}

\title{Ranges on a Genome}

\description{
  Genomic coordinates are often specified in terms of a genome identifier,
  chromosome name, start position and end
  position. The \code{rtracklayer} package adds convenience methods to
  \code{IntegerRangesList} for the manipulation of genomic ranges.
  The spaces (or names) of \code{IntegerRangesList} are the chromosome names.
  The \code{universe} slot indicates the genome, usually as given by UCSC
  (e.g. \dQuote{hg18}).
}

\section{Accessors}{
  In the code snippets below,
  \code{x} is a \code{IntegerRangesList} object.

  \describe{
    \item{\code{chrom(x), chrom(x) <- value}}{
      Gets or sets the chromosome
      names for \code{x}. This is an alias for
      \code{names(x)}.
    }
  }
}

\author{ Michael Lawrence }

\keyword{classes}
\keyword{methods}

\name{targets}
\Rdversion{1.1}
\alias{targets}
\docType{data}
\title{microRNA target sites}
\description{
A data frame of human microRNA target sites retrieved from MiRBase. This is a
subset of the \code{hsTargets} data frame in the \code{microRNA}
package. See the \code{rtracklayer} vignette for more details.
}
\usage{data(targets)}
\format{
  A data frame with 2981 observations on the following 6 variables.
  \describe{
    \item{\code{name}}{The miRBase ID of the microRNA.}
    \item{\code{target}}{The Ensembl ID of the targeted transcript.}
    \item{\code{chrom}}{The name of the chromosome for target site.}
    \item{\code{start}}{Target start position.}
    \item{\code{end}}{Target stop position.}
    \item{\code{strand}}{The strand of the target site, \code{"+"}, or 
         \code{"-"}.}
  }
}
\source{
The \code{microRNA} package, dataset \code{hsTargets}. Originally
MiRBase (\url{http://microrna.sanger.ac.uk/}).
}
\examples{
data(targets)
targetTrack <- with(targets, 
    GenomicData(IRanges::IRanges(start, end), 
                strand = strand, chrom = chrom))
}
\keyword{datasets}

\name{BrowserSession-class}
\docType{class}
\alias{BrowserSession-class}
\alias{show,BrowserSession-method}
\alias{range,BrowserSession-method}
\alias{genome,BrowserSession-method}
\alias{genome<-,BrowserSession-method}
\alias{names,BrowserSession-method}
\alias{trackNames,BrowserSession-method}

\title{Class "BrowserSession"}
\description{An object representing a genome browser session. As a
  derivative of \code{\linkS4class{TrackDb}}, each
  session contains a set of loaded tracks. In addition, it has
  a set of views, in the form of \code{\linkS4class{BrowserView}}
  instances, on those tracks. Note that
  this is a virtual class; a concrete implementation is provided by each
  backend driver.}
\section{Objects from the Class}{A virtual Class: No objects may be
  created from it. See \code{\link{browserSession}} for obtaining
  an instance of an implementation for a particular genome browser.}
\section{Methods}{
  This specifies the API implemented by each browser backend. Note that
  a backend is not required to support all operations, and that each
  backend often has additional parameters for each of the methods. See
  the backend-specific documentation for more details. The only built-in
  backend is \code{\linkS4class{UCSCSession}}.

  If a method is denoted as \emph{virtual}, it must be implemented by
  the backend to support the corresponding feature. Otherwise, the
  fallback behavior is described.
  
  \describe{
    \item{\emph{virtual} \code{\link{browserView}(object, range =
        range(object), track = trackNames(object), ...)}}{
      Constructs a \code{\linkS4class{BrowserView}} of \code{range}
      for this session. 
    }
    \item{\emph{virtual} \code{\link{browserViews}(object, ...)}}{
      Gets the \code{\linkS4class{BrowserView}} instances belonging to
      this session.}
    \item{\code{\link{activeView}(object, ...)}}{
      Returns the \code{\linkS4class{BrowserView}} that is currently
      active in the session. Fallback calls \code{browserViews} and
      queries each view with \code{activeView}.}
    \item{\code{\link{range}(x, ...)}}{
      Gets the \code{\link[GenomicRanges:GRanges-class]{GRanges}}
      representing the range of the genome currently displayed by the
      browser (i.e. the range shown by the active view) or a default
      value (possibly \code{NULL}) if no views exist.
    }
    \item{\emph{virtual} \code{\link[Biostrings]{getSeq}(object, range =
        range(object), ...)}}{
      gets a genomic sequence of \code{range} from this session.}
    \item{\emph{virtual} \code{\link[=sequence<-]{sequence}(object, ...) <- value}}{
      Loads a sequence into the session.}
    \item{\emph{virtual} \code{\link[=track<-]{track}(object, name =
        deparse(substitute(track)), view = TRUE, ...) <- value}}{
      Loads one or more \code{track}s into the session and optionally open a
      \code{view} of the track.
    }
    \item{\code{x[[i]] <- value}}{Loads the track \code{value} into
      session \code{x}, under the name \code{i}. Shortcut to above.}
    \item{\code{x$name <- value}}{Loads the track \code{value} into
      session \code{x}, under the name \code{name}. Shortcut to above.}
    \item{\emph{virtual} \code{\link{track}(object, ...)}}{
      Gets a track from a session.
    }
    \item{\code{x[[i]]}}{Gets the track named \code{i} from session
      \code{x}. A shortcut to \code{track}.
    }
    \item{\code{x$name}}{Gets the track named \code{name} from session
      \code{x}. A shortcut to \code{track}.}
    \item{\emph{virtual} \code{\link{trackNames}(object, ...)}}{
      Gets the names of the tracks stored in this session. }
    \item{\emph{virtual} \code{\link[Seqinfo]{genome}(x)},
      \code{genome(x) <- value}}{
      Gets or sets the genome identifier (e.g. \dQuote{hg18}) for the session.}
    \item{\emph{virtual} \code{\link{close}(con, ...)}}{ Close this session. }
    \item{\code{\link{show}(object, ...)}}{ Output a textual description
      of this session.}
  }
}
\author{ Michael Lawrence }
\seealso{
  \code{\link{browserSession}} for obtaining implementations of this
  class for a particular genome browser.
}
\keyword{classes}

\name{BigWigFile-class}
\docType{class}

%% Classes:
\alias{class:BigWigFile}
\alias{BigWigFile-class}
\alias{class:BWFile}
\alias{BWFile-class}
\alias{class:BigWigFileList}
\alias{BigWigFileList-class}

%% Constructor:
\alias{BigWigFile}
\alias{BWFile}
\alias{BigWigFileList}

%% Accessors:
\alias{seqinfo,BigWigFile-method}

%% Import:
\alias{import.bw}
\alias{import.bw,ANY-method}
\alias{import,BigWigFile,ANY,ANY-method}

%% Export:
\alias{export.bw}
\alias{export.bw,ANY-method}
\alias{export,ANY,BigWigFile,ANY-method}
\alias{export,GenomicRanges,BigWigFile,ANY-method}
\alias{export,List,BigWigFile,ANY-method}

%% Utilites:
\alias{summary,BigWigFile-method}
\alias{path,BigWigFileList-method}
\alias{cleanupBigWigCache}

\title{BigWig Import and Export}

\description{
  These functions support the import and export of the UCSC BigWig
  format, a compressed, binary form of WIG/BEDGraph with a spatial index
  and precomputed summaries. These functions do not work on Windows.
}

\usage{
\S4method{import}{BigWigFile,ANY,ANY}(con, format, text,
                   selection = BigWigSelection(which, ...),
                   which = con,
                   as = c("GRanges", "RleList", "NumericList"), ...)
import.bw(con, ...)

\S4method{export}{ANY,BigWigFile,ANY}(object, con, format, ...)
\S4method{export}{GenomicRanges,BigWigFile,ANY}(object, con, format,
                   dataFormat = c("auto", "variableStep", "fixedStep",
                     "bedGraph"), compress = TRUE, fixedSummaries = FALSE)
export.bw(object, con, ...)
}

\arguments{
  \item{con}{A path, URL or \code{BigWigFile} object. Connections are
    not supported. For the functions ending in \code{.bw}, the file
    format is indicated by the function name. For the \code{export}
    and \code{import} methods, the format must be indicated another
    way. If \code{con} is a path, or URL, either the file
    extension or the \code{format} argument needs to be \dQuote{bigWig}
    or \dQuote{bw}.
  }
  \item{object}{The object to export, should be an \code{RleList},
    \code{IntegerList}, \code{NumericList},
    \code{GRanges} or something coercible to a \code{GRanges}.
  }
  \item{format}{If not missing, should be \dQuote{bigWig} or \dQuote{bw}
    (case insensitive).
  }
  \item{text}{Not supported.
  }
  \item{as}{Specifies the class of the return object. Default is
    \code{GRanges}, which has one range per range in the file, and a
    score column holding the value for each range. For
    \code{NumericList}, one numeric vector is returned for each range in
    the \code{selection} argument. For \code{RleList}, there is one
    \code{Rle} per sequence, and that \code{Rle} spans the entire
    sequence.
  }
  \item{selection}{A \code{\linkS4class{BigWigSelection}} object
    indicating the ranges to load.
  }
  \item{which}{A range data structure coercible to \code{IntegerRangesList},
    like a \code{GRanges}, or a \code{BigWigFile}. Only the intervals in
    the file overlapping the given ranges are returned. By default, the
    value is the \code{BigWigFile} itself. Its \code{Seqinfo} object is
    extracted and coerced to a \code{IntegerRangesList} that represents the
    entirety of the file.
  }
  \item{dataFormat}{Probably best left to \dQuote{auto}. Exists only
    for historical reasons.
  }
  \item{compress}{If \code{TRUE}, compress the data. No reason to change this.
  }
  \item{fixedSummaries}{If \code{TRUE}, compute summaries at fixed
    resolutions corresponding to the default zoom levels in the Ensembl
    genome browser (with some extrapolation): 30X, 65X, 130X, 260X,
    450X, 648X, 950X, 1296X, 4800X, 19200X. Otherwise, the resolutions
    are dynamically determined by an algorithm that computes an initial
    summary size by initializing to 10X the size of the smallest feature
    and doubling the size as needed until the size of the summary is
    less than half that of the data (or there are no further gains). It
    then computes up to 10 more levels of summary, quadrupling the size
    each time, until the summaries start to exceed the sequence size.
  }
  \item{...}{Arguments to pass down to methods to other methods. For
    import, the flow eventually reaches the \code{BigWigFile} method on
    \code{import}.
  }
}

\value{
  A \code{GRanges} (default), \code{RleList} or \code{NumericList}.
  \code{GRanges} return ranges with non-zero score values in a \code{score}
  metadata column. The length of the \code{NumericList} is the same length
  as the \code{selection} argument (one list element per range).
  The return order in the  \code{NumericList} matches the order of the
  \code{BigWigSelection} object. 
}

\section{\code{BigWigFile} objects}{
  A \code{BigWigFile} object, an extension of
  \code{\link[BiocIO:BiocFile-class]{BiocFile}} is a reference to a BigWig file.
  To cast a path, URL or connection to a \code{BigWigFile}, pass it to the
  \code{BigWigFile} constructor.

  BigWig files are more complex than most track files, and there are a
  number of methods on \code{BigWigFile} for accessing the additional
  information:

  \describe{
    \item{\code{seqinfo(x)}}{
      Gets the \code{\link[Seqinfo]{Seqinfo}} object
      indicating the lengths of the sequences for the intervals in the
      file. No circularity or genome information is available.
    }
    \item{\code{summary(ranges = as(seqinfo(object), "GenomicRanges"), size
        = 1L, type = c("mean", "min", "max", "coverage", "sd"),
        defaultValue = NA_real_),
	as = c("GRangesList", "RleList", "matrix"), ...}}{
        Aggregates the intervals in the file
        that fall into \code{ranges}, which should be something
        coercible to \code{GRanges}.  The aggregation essentially
        compresses each sequence to a length of \code{size}. The
        algorithm is specified by \code{type}; available algorithms
        include the mean, min, max, coverage (percent sequence covered
        by at least one feature), and standard deviation. When a window
        contains no features, \code{defaultValue} is assumed. The result
        type depends on \code{as}, and can be a GRangesList, RleList or
        matrix, where the number elements (or rows) is equal to the
        length of \code{ranges}. For \code{as="matrix"}, there must be
        one unique value of \code{size}, which is equal to the number of
        columns in the result. The \code{as="matrix"} case is the only
        one that supports a \code{size} greater than the width of the
        corresponding element in \code{ranges}, where values are
	interpolated to yield the matrix result. The
        driving use case for this is visualization of coverage when the
        screen space is small compared to the viewed portion of the
        sequence. The operation is very fast, as it leverages cached
        multi-level summaries present in every BigWig file.

        If a summary statistic is not available / cannot be computed
        for a given range a warning is thrown and the defaultValue
        \code{NA_real_} is returned.
    }
  }

  When accessing remote data, the UCSC library caches data in the
  \file{/tmp/udcCache} directory. To clean the cache, call
  \code{cleanBigWigCache(maxDays)}, where any files older than
  \code{maxDays} days old will be deleted.
}

\section{\code{BigWigFileList} objects}{
  A \code{BigWigFileList()} provides a convenient way of managing a list 
  of \code{BigWigFile} instances.
}

\author{Michael Lawrence}

\seealso{
  \code{\link{wigToBigWig}} for converting a WIG file to BigWig.
}

\examples{
if (.Platform$OS.type != "windows") {
  test_path <- system.file("tests", package = "rtracklayer")
  test_bw <- file.path(test_path, "test.bw")

  ## GRanges
  ## Returns ranges with non-zero scores.
  gr <- import(test_bw)
  gr 

  which <- GRanges(c("chr2", "chr2"), IRanges(c(1, 300), c(400, 1000)))
  import(test_bw, which = which)

  ## RleList
  ## Scores returned as an RleList is equivalent to the coverage.
  ## Best option when 'which' or 'selection' contain many small ranges.
  mini <- narrow(unlist(tile(which, 50)), 2)
  rle <- import(test_bw, which = mini, as = "RleList")
  rle 

  ## NumericList
  ## The 'which' is stored as metadata:
  track <- import(test_bw, which = which, as = "NumericList")
  metadata(track)

\dontrun{
  test_bw_out <- file.path(tempdir(), "test_out.bw")
  export(test, test_bw_out)
}

  bwf <- BigWigFile(test_bw)
  track <- import(bwf)

  seqinfo(bwf)  

  summary(bwf) # for each sequence, average all values into one
  summary(bwf, range(head(track))) # just average the first few features
  summary(bwf, size = seqlengths(bwf) / 10) # 10X reduction
  summary(bwf, type = "min") # min instead of mean
  summary(bwf, track, size = 10, as = "matrix") # each feature 10 windows
}
}

\keyword{methods}
\keyword{classes}

\name{TrackHub-class}
\docType{class}

%% Classes:
\alias{class:TrackHub}
\alias{TrackHub-class}

%% Constructor:
\alias{TrackHub}

%% Accessors:
\alias{$,TrackHub-method}
\alias{[[,TrackHub,ANY,ANY-method}
\alias{genome,TrackHub-method}
\alias{length,TrackHub-method}
\alias{uri}
\alias{writeTrackHub}

%% Data Access:
\alias{hub,TrackHub-method}
\alias{hub<-,TrackHub-method}
\alias{shortLabel,TrackHub-method}
\alias{shortLabel<-,TrackHub-method}
\alias{longLabel,TrackHub-method}
\alias{longLabel<-,TrackHub-method}
\alias{genomeFile,TrackHub-method}
\alias{genomeFile<-,TrackHub-method}
\alias{email,TrackHub-method}
\alias{email<-,TrackHub-method}
\alias{descriptionUrl,TrackHub-method}
\alias{descriptionUrl<-,TrackHub-method}
\alias{genomeField,TrackHub-method}
\alias{genomeField<-,TrackHub-method}
\alias{genomeInfo,TrackHub-method}
\alias{genomeInfo<-,TrackHub-method}

%% Coercion:
\alias{coerce,character,TrackHub-method}

%% Show:
\alias{show,TrackHub-method}

\title{TrackHub Access}

\description{
  The \code{TrackHub} class represents a TrackHub data source,
  essentially directory layout separating tracks and sequences by
  genome, along with a few metadata files. This interface abstracts
  those details and provides access to a TrackHub at any URL supported
  by R (HTTP, FTP, and local files). This is an easy way to make data
  accessible to the UCSC Genome Browser.
}

\section{Constructor}{
  \describe{
    \item{\code{TrackHub(uri, create = FALSE)}}{
      Constructs a
      new \code{TrackHub} object, representing a repository at
      \code{uri}. If \code{create} is \code{TRUE}, and \code{uri} is
      writeable (i.e., local), the repository is created if it does not
      already exist. If it does exist, then a message is emitted to
      indicate that the repository was not recreated.
    }
  }
}

\section{Accessor Methods}{
  In the code snippets below, \code{x} represents a \code{TrackHub}
  object.

  \describe{
    \item{\code{x$genome}, \code{x[["genome"]]}}{Get
      the \code{\linkS4class{TrackHubGenome}} object for the genome named
      \code{genome}.
    }
    \item{\code{length(x)}}{number of genomes in the repository.
    }
    \item{\code{uri(x)}}{
      Get the URI pointing to the TrackHub repository.
    }
    \item{\code{genome(x)}}{
      Get the identifiers of the genomes present in the repository.
    }
    \item{\code{writeTrackHub(x)}}{
      Write hub content and genomes from memory representation to the hub file and genomes file.
      It also create resources if they are missing like genomes file and genome directory for
      newly add genome.
    }
  }
}

\section{Data Access}{
  Note that all storing methods(like \code{hub()<-}) are only supported for local repositories, i.e., those with a file:// URI scheme.
  \describe{
    \item{\code{hub(x)}}{
      get the value of hub.
    }
    \item{\code{hub(x) <- value}}{
      store the \code{value} of hub for \code{x}.
    }
    \item{\code{shortLabel(x)}}{
      get the value of hub.
    }
    \item{\code{shortLabel(x) <- value}}{
      store the \code{value} of shortLabel for \code{x}.
    }
    \item{\code{longLabel(x)}}{
      get the value of hub.
    }
    \item{\code{longLabel(x) <- value}}{
      store the \code{value} of longLabel for \code{x}.
    }
    \item{\code{genomeFile(x)}}{
      get the value of hub.
    }
    \item{\code{genomeFile(x) <- value}}{
      store the \code{value} of genomesFile for \code{x}.
    }
    \item{\code{email(x)}}{
      get the value of hub.
    }
    \item{\code{email(x) <- value}}{
      store the \code{value} of email for \code{x}.
    }
    \item{\code{descriptionUrl(x)}}{
      get the value of hub.
    }
    \item{\code{descriptionUrl(x) <- value}}{
      store the \code{value} of descriptionUrl for \code{x}.
    }
    \item{\code{genomeField(x, name, field)}}{
      Get the \code{value} of \code{field} for \code{name} genome.
    }
    \item{\code{genomeField(x, name, field) <- value}}{
      Set or Update the \code{field} and \code{value} for \code{name} genome.
    }
    \item{\code{genomeInfo(x, name)}}{
      Get the \code{Genome} object for \code{name} genome.
    }
    \item{\code{genomeInfo(x) <- value}}{
      Add \code{value} (Genome object) to existing genomes list.
      \code{Genome} takes named arguemnts of all UCSC supported fields for genome
      file(like \code{genome, trackDb, twoBitPath}, etc).
    }
  }
}

\author{Michael Lawrence}

\examples{
th <- TrackHub(system.file("tests", "trackhub", package = "rtracklayer"))
uri(th)
genome(th)
length(th)
th$hg19
th[["hg19"]]
hub(th)
email(th)

\dontrun{
hub(th) <- "new_hub"
writeTrackHub(th)
}
}

\keyword{methods}
\keyword{classes}

\name{BigBedSelection-class}
\docType{class}
\alias{BigBedSelection-class}

% constructor
\alias{BigBedSelection}

\alias{coerce,IntegerRangesList,BigBedSelection-method}
\alias{coerce,GenomicRanges,BigBedSelection-method}

\title{Selection of ranges and columns}

\description{A \code{BigBedSelection} represents a query against a
  BigBed file, see \code{\link{import.bb}}. It extends
  \link[IRanges]{RangedSelection} by including a \code{colnames} parameter.
  The \code{colnames} should be a character vector of column names
  corresponding to BigBed fields.

  By default, \code{colnames} correspond to the standard BED12 format fields
  that cover core interval data, strand, display attributes, and block (exon)
  structure information.

  If no custom \code{colnames} are provided, \pkg{rtracklayer} attempts to use
  the field names present in the BigBed file itself. Since BigBed files can
  contain non standard or additional custom fields beyond the standard ones,
  this allows flexible access to such extended fields in the query results.
}

\section{Constructor}{
  \describe{
    \item{\code{BigBedSelection(ranges = GRanges(), colnames =
        "score")}}{Constructs a \code{BigBedSelection} with the given
        \code{ranges} and \code{colnames}.
        a \code{character} identifying a genome (see
        \code{\link{GenomicSelection}}), or a
        \code{\linkS4class{BigBedFile}}, in which case the ranges are
        derived from the bounds of its sequences.
    }
  }
}

\section{Coercion}{
  \describe{
    \item{\code{as(from, "BigBedSelection")}}{Coerces \code{from} to a
      \code{BigBedSelection} object. Typically, \code{from} is a
      \code{\link[GenomicRanges]{GRanges}} or
      a \code{\link[IRanges]{IntegerRangesList}}, the ranges of
      which become the ranges in the
      new \code{BigBedSelection}.
    }
  }
}

\author{ Michael Lawrence }

\examples{
  rl <- IRangesList(chr1 = IRanges::IRanges(c(1, 5), c(3, 6)))

  BigBedSelection(rl)
  as(rl, "BigBedSelection") # same as above

  # do not select any column
  BigBedSelection(rl, character())
}

\keyword{methods}
\keyword{classes}

\name{track<--methods}
\docType{methods}
\alias{track<-}
\alias{track<--methods}
\alias{track<-,BrowserSession,ANY-method}
\title{ Laying tracks }
\description{
  Methods for loading tracks into genome browsers.
}
\usage{
track(object, ...) <- value
}
\arguments{
  \item{object}{A \code{\linkS4class{BrowserSession}} into which the
    track is loaded.}
  \item{value}{The track(s) to load.}
  \item{\dots}{Arguments to pass on to methods. Can be:
    \describe{
      \item{name}{The name(s) of the track(s) being loaded.}
      \item{view}{Whether to create a view of the track after loading it.}
    }
  }
}
\seealso{\code{\link{track}} for getting a track from a session.}
\examples{
\dontrun{
  session <- browserSession()
  track <- import(system.file("tests", "v1.gff", package = "rtracklayer"))
  track(session, "My Track") <- track
}
}
\keyword{methods}

\name{UCSCData-class}
\docType{class}
\alias{UCSCData-class}
\alias{export.bed,UCSCData,character_OR_connection-method}
\alias{export.bed15,UCSCData-method}
\alias{export.gff,UCSCData,character_OR_connection-method}
\alias{export.ucsc,UCSCData,character_OR_connection-method}
\alias{initialize,UCSCData-method}
\alias{show,UCSCData-method}
\alias{coerce,GRanges,UCSCData-method}
\alias{coerce,UCSCData,GRanges-method}
\alias{split,UCSCData,ANY-method}
\alias{split,UCSCData,Vector-method}

\title{Class "UCSCData"}
\description{Each track in \acronym{UCSC} has an associated
  \code{\linkS4class{TrackLine}} that contains metadata on the track.}
\section{Slots}{
  \describe{
    \item{\code{trackLine}:}{Object of class \code{"TrackLine"}
      holding track metadata. }
  }
}
\section{Methods}{
  \describe{
    \item{\code{\link{export.bed}(object, con, variant = c("base",
        "bedGraph", "bed15"), color, trackLine = TRUE, ...)}}{
      Exports the track and its track line (if
      \code{trackLine} is \code{TRUE}) to \code{con} in the Browser
      Extended Display (\acronym{BED}) format. The arguments
      in \code{...} are passed to \code{\link{export.ucsc}}.
    }
    \item{\code{\link{export.bed15}(object, con, expNames = NULL, ...)}}{
      Exports the track and its track line (if
      \code{trackLine} is \code{TRUE}) to \code{con} in the Bed15
      format. The data is taken from the columns named in
      \code{expNames}, which defaults to the \code{expNames} in the
      track line, if any, otherwise all column names. The arguments
      in \code{...} are passed to \code{\link{export.ucsc}}.
    }
    \item{\code{\link{export.gff}(object)}}{
      Exports the track and its track line (as a comment) to \code{con} in the
      General Feature Format (\acronym{GFF}).
    }
    \item{\code{\link{export.ucsc}(object, con, subformat, ...)}}{
      Exports the track and its track line to \code{con} in the
      \acronym{UCSC} meta-format.  }
    \item{\code{as(object, "UCSCData")}}{Constructs a \code{UCSCData}
      from a \code{GRanges} instance, by adding a default track line
      and ensuring that the sequence/chromosome names are compliant with
      UCSC conventions. If there is a numeric score, the track line type
      is either "bedGraph" or "wig", depending on the feature
      density. Otherwise, "bed" is chosen.
    }
  }
}
\author{ Michael Lawrence }
\seealso{
  \code{\link[BiocIO]{import}} and \code{\link[BiocIO]{export}} for reading and writing
  tracks to and from connections (files), respectively.
}
\keyword{classes}

\name{GenomicSelection}
\alias{GenomicSelection}
\title{
Genomic data selection
}
\description{
  Convenience constructor of a \link[IRanges]{RangedSelection} object
  for selecting a data on a per-chromosome basis for a given genome. 
}
\usage{
GenomicSelection(genome, chrom = NULL, colnames = character(0))
}
\arguments{
  \item{genome}{
    A string identifying a genome. Should match the end of a BSgenome
    package name, e.g. "hg19".
  }
  \item{chrom}{
    Character vector naming chromosomes to select.
  }
  \item{colnames}{
    The column names to select from the dataset.
  }
}
\value{
  A \link[IRanges]{RangedSelection} object, selecting entire chromosomes
}
\author{
  Michael Lawrence
}

\seealso{
  \code{\link[IRanges]{RangedSelection}}, \code{\link{BigWigSelection}}
}
\examples{
# every chromosome from hg19
GenomicSelection("hg19")
# chr1 and 2 from hg19, with a score column
GenomicSelection("hg19", c("chr1", "chr2"), "score")
}
\keyword{ manip }

\name{GFFFile-class}
\docType{class}

%% Classes:
\alias{class:GFFFile}
\alias{class:GFF1File}
\alias{class:GFF2File}
\alias{class:GFF3File}
\alias{class:GVFFile}
\alias{class:GTFFile}
\alias{GFFFile-class}
\alias{GFF1File-class}
\alias{GFF2File-class}
\alias{GFF3File-class}
\alias{GVFFile-class}
\alias{GTFFile-class}

%% Constructor:
\alias{GFFFile}
\alias{GFF1File}
\alias{GFF2File}
\alias{GFF3File}
\alias{GVFFile}
\alias{GTFFile}

%% Import:
\alias{import,GFFFile,ANY,ANY-method}
\alias{import.gff}
\alias{import.gff1}
\alias{import.gff2}
\alias{import.gff3}
\alias{import.gff,ANY-method}
\alias{import.gff1,ANY-method}
\alias{import.gff2,ANY-method}
\alias{import.gff3,ANY-method}

%% Export:
\alias{export,ANY,GFFFile,ANY-method}
\alias{export,GenomicRanges,GFFFile,ANY-method}
\alias{export,GenomicRangesList,GFFFile,ANY-method}
\alias{export,SimpleGRangesList,GFFFile,ANY-method}
\alias{export,CompressedGRangesList,GFFFile,ANY-method}
\alias{export,GRangesList,GTFFile,ANY-method}
\alias{export.gff}
\alias{export.gff,ANY-method}
\alias{export.gff1}
\alias{export.gff1,ANY-method}
\alias{export.gff2}
\alias{export.gff2,ANY-method}
\alias{export.gff3}
\alias{export.gff3,ANY-method}

% Other:
\alias{genome,GFFFile-method}

\title{GFFFile objects}

\description{
  These functions support the import and export of the GFF format, of
  which there are three versions and several flavors.
}

\usage{
\S4method{import}{GFFFile,ANY,ANY}(con, format, text,
           version = c("", "1", "2", "3"),
           genome = NA, colnames = NULL, which = NULL,
           feature.type = NULL, sequenceRegionsAsSeqinfo = FALSE)
import.gff(con, ...)
import.gff1(con, ...)
import.gff2(con, ...)
import.gff3(con, ...)

\S4method{export}{ANY,GFFFile,ANY}(object, con, format, ...)
\S4method{export}{GenomicRanges,GFFFile,ANY}(object, con, format,
                   version = c("1", "2", "3"),
                   source = "rtracklayer", append = FALSE, index = FALSE)
\S4method{export}{GenomicRangesList,GFFFile,ANY}(object, con, format, ...)
export.gff(object, con, ...)
export.gff1(object, con, ...)
export.gff2(object, con, ...)
export.gff3(object, con, ...)
}

\arguments{
  \item{con}{A path, URL, connection or \code{GFFFile} object. For the
    functions ending in \code{.gff}, \code{.gff1}, etc, the file format
    is indicated by the function name. For the base \code{export} and
    \code{import} functions, the format must be indicated another
    way. If \code{con} is a path, URL or connection, either the file
    extension or the \code{format} argument needs to be one of
    \dQuote{gff}, \dQuote{gff1} \dQuote{gff2}, \dQuote{gff3},
    \dQuote{gvf}, or \dQuote{gtf}. Compressed files (\dQuote{gz},
    \dQuote{bz2} and \dQuote{xz}) are handled transparently.
  }
  \item{object}{The object to export, should be a \code{GRanges} or
    something coercible to a \code{GRanges}. If the object has a method
    for \code{asGFF}, it is called prior to coercion. This makes it
    possible to export a \code{GRangesList} or \code{TxDb} in a
    way that preserves the hierarchical structure. For exporting
    multiple tracks, in the UCSC track line metaformat, pass a
    \code{GenomicRangesList}, or something coercible to one.
  }
  \item{format}{If not missing, should be one of \dQuote{gff}, \dQuote{gff1}
    \dQuote{gff2}, \dQuote{gff3}, \dQuote{gvf}, or \dQuote{gtf}.
  }
  \item{version}{If the format is given as \dQuote{gff}, i.e., it does
    not specify a version, then this should indicate the GFF version as
    one of \dQuote{} (for import only, from the \code{gff-version}
    directive in the file or \dQuote{1} if none), \dQuote{1}, \dQuote{2}
    or \dQuote{3}.
  }
  \item{text}{If \code{con} is missing, a character vector to use as the
    input.
  }
  \item{genome}{The identifier of a genome, or a \code{Seqinfo}, or \code{NA} if
    unknown. Typically, this is a UCSC identifier like \dQuote{hg19}. An
    attempt will be made to derive the \code{Seqinfo} on the return
    value using either an installed BSgenome package or UCSC, if network
    access is available.
  }
  \item{colnames}{A character vector naming the columns to parse. These
    should name either fixed fields, like \code{source} or
    \code{type}, or, for GFF2 and GFF3, any attribute.
  }
  \item{which}{A \code{GRanges} or other range-based object supported
    by \code{\link[IRanges]{findOverlaps}}. Only the intervals in the file
    overlapping the given ranges are returned. This is much more efficient
    when the file is indexed with the tabix utility.
  }
  \item{feature.type}{\code{NULL} (the default) or a character vector of
    valid feature types. If not \code{NULL}, then only the features of the
    specified type(s) are imported.
  }
  \item{sequenceRegionsAsSeqinfo}{If \code{TRUE}, attempt to infer the
    \code{Seqinfo} (\code{seqlevels} and \code{seqlengths}) from the
    \dQuote{##sequence-region} directives as specified by GFF3.
  }
  \item{source}{The value for the source column in GFF. This is
    typically the name of the package or algorithm that generated the
    feature.
  }
  \item{index}{If \code{TRUE}, automatically compress and index the
    output file with bgzf and tabix. Note that tabix indexing will
    sort the data by chromosome and start. Tabix supports a
    single track in a file.
  }
  \item{append}{If \code{TRUE}, and \code{con} points to a file path,
    the data is appended to the file. Obviously, if \code{con} is a
    connection, the data is always appended.
  }
  \item{...}{Arguments to pass down to methods to other methods. For
    import, the flow eventually reaches the \code{GFFFile} method on
    \code{import}. When \code{trackLine} is
    \code{TRUE} or the target format is BED15, the arguments are passed
    through \code{export.ucsc}, so track line parameters are supported.
  }
}

\value{
  A \code{GRanges} with the metadata columns described in the details.
}

\details{
  The Generic Feature Format (GFF) format is a tab-separated table of
  intervals. There are three different versions of GFF, and they all
  have the same number of columns. In GFF1, the last column is a
  grouping factor, whereas in the later versions the last column holds
  application-specific attributes, with some conventions defined for
  those commonly used. This attribute support facilitates specifying
  extensions to the format. These include GTF (Gene Transfer Format, an
  extension of GFF2) and GVF (Genome Variation Format, an extension of
  GFF3).  The rtracklayer package recognizes the \dQuote{gtf} and
  \dQuote{gvf} extensions and parses the extra attributes into columns
  of the result; however, it does not perform any extension-specific
  processing. Both GFF1 and GFF2 have been proclaimed obsolete; however,
  the UCSC Genome Browser only supports GFF1 (and GTF), and GFF2 is still
  in broad use.

  GFF is distinguished from the simpler BED format by its flexible
  attribute support and its hierarchical structure, as specified by the
  \code{group} column in GFF1 (only one level of grouping) and the
  \code{Parent} attribute in GFF3. GFF2 does not specify a convention
  for representing hierarchies, although its GTF extension provides this
  for gene structures. The combination of support for hierarchical data
  and arbitrary descriptive attributes makes GFF(3) the preferred format
  for representing gene models.

  Although GFF features a \code{score} column, large quantitative data
  belong in a format like \link[=BigWigFile]{BigWig} and alignments from
  high-throughput experiments belong in
  \link[Rsamtools:BamFile]{BAM}. For variants, the VCF format (supported
  by the VariantAnnotation package) seems to be more widely adopted than
  the GVF extension.

  A note on the UCSC track line metaformat: track lines are a means for
  passing hints to visualization tools like the UCSC Genome Browser and
  the Integrated Genome Browser (IGB), and they allow multiple tracks to
  be concatenated in the same file. Since GFF is not a UCSC format, it
  is not common to annotate GFF data with track lines, but rtracklayer
  still supports it. To export or import GFF data in the track line
  format, call \code{\link{export.ucsc}} or \code{\link{import.ucsc}}.
  
  The following is the mapping of GFF elements to a \code{GRanges} object.
  NA values are allowed only where indicated.
  These appear as a \dQuote{.} in the file. GFF requires that all columns
  are included, so \code{export} generates defaults for missing columns.

  \describe{
    \item{seqid, start, end}{the \code{ranges} component.}
    \item{source}{character vector in the \code{source}
      column; defaults to \dQuote{rtracklayer} on export.
    }
    \item{type}{character vector in the \code{type} column; defaults
      to \dQuote{sequence_feature} in the output, i.e., SO:0000110.
    }
    \item{score}{numeric vector (NA's allowed) in the \code{score}
      column, accessible via the \code{score} accessor; defaults
      to \code{NA} upon export.
    }
    \item{strand}{strand factor (NA's allowed) in the \code{strand}
      column, accessible via the \code{strand} accessor; defaults
      to \code{NA} upon export.
    }
    \item{phase}{integer vector, either 0, 1 or 2 (NA's allowed);
      defaults to \code{NA} upon export.
    }
    \item{group}{a factor (GFF1 only); defaults to the \code{seqid}
      (e.g., chromosome) on export.
    }
  }

  In GFF versions 2 and 3, attributes map to arbitrary columns in the
  result. In GFF3, some attributes (\code{Parent}, \code{Alias},
  \code{Note}, \code{DBxref} and \code{Ontology_term}) can have
  multiple, comma-separated values; these columns are thus always
  \code{CharacterList} objects.
}

\section{GFFFile objects}{
  The \code{GFFFile} class extends \code{\link[BiocIO:BiocFile-class]{BiocFile}}
  and is a formal represention of a resource in the GFF format.
  To cast a path, URL or connection to a \code{GFFFile}, pass it to
  the \code{GFFFile} constructor. The \code{GFF1File}, \code{GFF2File},
  \code{GFF3File}, \code{GVFFile} and \code{GTFFile} classes all extend
  \code{GFFFile} and indicate a particular version of the format.

  It has the following utility methods:
  \describe{
    \item{\code{genome}}{Gets the genome identifier from
      the \dQuote{genome-build} header directive.
    }
  }
}

\author{Michael Lawrence}

\references{
  \describe{
    \item{GFF1, GFF2}{
      \url{http://www.sanger.ac.uk/resources/software/gff/spec.html}
    }
    \item{GFF3}{\url{http://www.sequenceontology.org/gff3.shtml}}
    \item{GVF}{\url{http://www.sequenceontology.org/resources/gvf.html}}
    \item{GTF}{\url{http://mblab.wustl.edu/GTF22.html}}
  }
}

\examples{
  test_path <- system.file("tests", package = "rtracklayer")
  test_gff3 <- file.path(test_path, "genes.gff3")

  ## basic import
  test <- import(test_gff3)
  test

  ## import.gff functions
  import.gff(test_gff3)
  import.gff3(test_gff3)

  ## GFFFile derivatives
  test_gff_file <- GFF3File(test_gff3)
  import(test_gff_file)
  test_gff_file <- GFFFile(test_gff3)
  import(test_gff_file)
  test_gff_file <- GFFFile(test_gff3, version = "3")
  import(test_gff_file)

  ## from connection
  test_gff_con <- file(test_gff3)
  test <- import(test_gff_con, format = "gff")

  ## various arguments
  import(test_gff3, genome = "hg19")
  import(test_gff3, colnames = character())
  import(test_gff3, colnames = c("type", "geneName"))

  ## 'which'
  which <- GRanges("chr10:90000-93000")
  import(test_gff3, which = which)

\dontrun{
  ## 'append'
  test_gff3_out <- file.path(tempdir(), "genes.gff3")

  export(test[seqnames(test) == "chr10"], test_gff3_out)
  export(test[seqnames(test) == "chr12"], test_gff3_out, append = TRUE)
  import(test_gff3_out)
  
  ## 'index'
  export(test, test_gff3_out, index = TRUE)
  test_bed_gz <- paste(test_gff3_out, ".gz", sep = "")
  import(test_bed_gz, which = which)
}
}

\keyword{methods}
\keyword{classes}

\name{BasicTrackLine-class}
\docType{class}
\alias{BasicTrackLine-class}
\alias{coerce,character,BasicTrackLine-method}
\alias{coerce,BasicTrackLine,character-method}
\alias{fileFormat,TrackLine-method}

\title{Class "BasicTrackLine"}
\description{The type of \acronym{UCSC} track line used to annotate most
  types of tracks (every type except Wiggle).}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("BasicTrackLine",
  ...)} or parsed from a character vector track line with \code{as(text,
  "BasicTrackLine")} or converted from a
\code{\linkS4class{GraphTrackLine}} using \code{as(wig, "BasicTrackLine")}.
}
\section{Slots}{
  \describe{
    \item{\code{itemRgb}:}{Object of class \code{"logical"} indicating
      whether each feature in a track uploaded as \acronym{BED} should
      be drawn in its specified color. }
    \item{\code{useScore}:}{Object of class \code{"logical"} indicating
      whether the data value should be mapped to color. }
    \item{\code{group}:}{Object of class \code{"character"} naming a
      group to which this track should belong. }
    \item{\code{db}:}{Object of class \code{"character"} indicating the
      associated genome assembly. }
    \item{\code{offset}:}{Object of class \code{"numeric"}, a number
      added to all positions in the track. }
    \item{\code{url}:}{Object of class \code{"character"} referring to
      additional information about this track. }
    \item{\code{htmlUrl}:}{Object of class \code{"character"} referring
      to an \acronym{HTML} page to be displayed with this track. }
    \item{\code{name}:}{Object of class \code{"character"} specifying
      the name of the track. }
    \item{\code{description}:}{Object of class \code{"character"}
      describing the track. }
    \item{\code{visibility}:}{Object of class \code{"character"}
      indicating the default visible mode of the track, see
      \code{\linkS4class{UCSCTrackModes}}.}
    \item{\code{color}:}{Object of class \code{"integer"} representing the
      track color (as from \code{\link{col2rgb}}). }
    \item{\code{colorByStrand}:}{Object of class \code{"matrix"}
       with two columns, as from \code{col2rgb}. The two colors
       indicate the color for each strand (positive, negative). }
    \item{\code{priority}:}{Object of class \code{"numeric"} specifying
      the rank of the track. }
  }
}
\section{Extends}{
Class \code{"\linkS4class{TrackLine}"}, directly.
}
\section{Methods}{
  \describe{
    \item{as(object, "character")}{Export line to its string
      representation.}
    \item{as(object, \code{"\linkS4class{GraphTrackLine}"})}{Convert this
      line to a graph track line, using defaults for slots not held
      in common.}
    \item{\code{fileFormat(x)}}{Returns \dQuote{bed}}
  }
}
\references{
  \url{https://genome.ucsc.edu/goldenPath/help/customTrack.html#TRACK}
  for the official documentation. }
\author{ Michael Lawrence }

\seealso{
  \code{\linkS4class{GraphTrackLine}} for Wiggle/bedGraph tracks.
}
\keyword{classes}

\name{QuickloadGenome-class}
\docType{class}

%% Classes:
\alias{class:QuickloadGenome}
\alias{QuickloadGenome-class}

%% Constructor:
\alias{QuickloadGenome}

%% Accessors:
\alias{seqinfo,QuickloadGenome-method}
\alias{seqinfo,DNAStringSet-method} % sneak it in here
\alias{seqinfo<-,QuickloadGenome-method}
\alias{genome,QuickloadGenome-method}
\alias{length,QuickloadGenome-method}
\alias{names,QuickloadGenome-method}
\alias{quickload}
\alias{mcols,QuickloadGenome-method}
\alias{releaseDate,QuickloadGenome-method}
\alias{organism,QuickloadGenome-method}
\alias{uri,QuickloadGenome-method}
\alias{trackNames,QuickloadGenome-method}

%% Data Access:
\alias{track,QuickloadGenome-method}
\alias{track<-,QuickloadGenome,ANY-method}
\alias{track<-,QuickloadGenome,BiocFile-method}
\alias{track<-,QuickloadGenome,RsamtoolsFile-method}
\alias{track<-,QuickloadGenome,character-method}
\alias{referenceSequence}
\alias{referenceSequence<-}
\alias{referenceSequence,QuickloadGenome-method}
\alias{referenceSequence<-,QuickloadGenome-method}

%% Show
\alias{show,QuickloadGenome-method}

\title{Quickload Genome Access}

\description{
  A Quickload data source is a collection of tracks and sequences,
  separated by genome. This class, \code{QuickloadGenome} provides
  direct access to the data for one particular genome.
}

\section{Constructor}{
  \describe{
    \item{\code{QuickloadGenome(quickload, genome, create = FALSE,
        seqinfo = seqinfo(genome),
        title = toString(genome))}}{
        Constructs a
        new \code{QuickloadGenome} object, representing \code{genome} in
        the repository \code{quickload} (a URI string or a
        \code{\linkS4class{Quickload}} object).

        The \code{genome} argument can be an ID corresponding to a
        genome (potentially) in \code{quickload} or an installed
        \code{BSgenome} package. It can also be any instance of a class
        which has methods for \code{organism} and \code{releaseDate}. A
        good example is \code{\link[BSgenome]{BSgenome}} or any other
        derivative of \code{\link[Seqinfo]{GenomeDescription}}.
        Those items are necessary for constructing the canonical
        Quickload genome string (G_Species_Month_Year).

        If \code{create} is \code{TRUE}, and the genome does not already
        exist, the genome will be created, using \code{seqinfo} for the
        sequence lengths and \code{title} for the display name of the
        genome in a UI. Creation only works if the repository is local
        and writeable. Reasonable defaults are used for \code{seqinfo}
        and \code{title} when the necessary methods are available (and
        they are for \code{BSgenome}).
    }
  }
}

\section{Accessor Methods}{
  In the code snippets below, \code{x} and \code{object} represent a
  \code{Quickload} object.
  
  \describe{
    \item{\code{seqinfo(x)}, \code{seqinfo(x) <- value}}{
      Gets or sets the
      \code{\link[Seqinfo]{Seqinfo}} object indicating the lengths
      of the sequences in the genome. No circularity information or genome
      identifier is stored.
    }
    \item{\code{quickload(x)}}{
      Get the Quickload object that contains this
      genome.
    }
    \item{\code{uri(x)}}{
      Get the uri pointing to the genome directory in the
      Quickload repository
    }
    \item{\code{genome(x)}}{
      Get the name of the genome, e.g.
      \dQuote{H_sapiens_Feb_2009}.
    }
    \item{\code{releaseDate(x)}}{
      Get the release portion of the genome name,
      e.g., \dQuote{Feb_2009}.
    }
    \item{\code{organism(object)}}{
      Get the organism portion of the genome name,
      e.g., \dQuote{H sapiens}.
    }
    \item{\code{trackNames(object)}}{
      Get the names of the tracks present in the Quickload repository
    }
  }
}

\section{Data Access}{
  \describe{
    \item{\code{length(x)}}{
      number of datasets
    }
    \item{\code{names(x), trackNames(x)}}{
      names of the datasets
    }
    \item{\code{mcols(x)}}{
      merged metadata on the datasets
    }
    \item{\code{track(x, name), x$name}}{
      get the track called \code{name}
    }
    \item{\code{track(x, name, format = bestFileFormat(value), ...) <-
        value, x$name <- value}}{
        store the track \code{value} under
        \code{name}. Note that track storing is only supported
        for local repositories, i.e., those with a \code{file://} URI
        scheme.

        Currently, supported \code{value} types include a
        \code{GenomicRanges}, \code{GRangesList}, or a file resource
        (copied to the repository).  The file resource may be
        represented as a path, URL,
        \code{\link[BiocIO:BiocFile-class]{BiocFile}} or
        \code{\link[Rsamtools:RsamtoolsFile-class]{RsamtoolsFile}}. If
        not a file name, \code{value} is written in \code{format}. For
        generic interval data, this means a BigWig file (if there is a
        numeric \dQuote{score} column) or a BED file otherwise. An
        \code{RleList} (e.g., coverage) is output as BigWig. For
        \code{UCSCData} values, the format is chosen according to the
        type of track line. For \code{RsamtoolsFile} objects, the file
        and its index are copied.

        The arguments in \code{...} become attributes in the XML
        metadata. The \dQuote{description} attribute is standard and is
        a blurb for describing the track in a UI. For the rest, the
        interpretation is up to the client. IGB supports an ever-growing
        list; please see its documentation.
    }
    \item{\code{referenceSequence(x)}}{
      Get the reference sequence, as a
      \code{DNAStringSet}.
    }
    \item{\code{referenceSequence(x) <- value}}{
      Set the reference sequence, as a
      \code{DNAStringSet}. It is written as a 2bit file. This only works
      on local repositories.
    }
  }
}

\author{Michael Lawrence}

\examples{
tests_dir <- system.file("tests", package = "rtracklayer")
ql <- Quickload(file.path(tests_dir, "quickload"))
qlg <- QuickloadGenome(ql, "T_species_Oct_2011")
seqinfo(qlg)
organism(qlg)
releaseDate(qlg)
names(qlg)
mcols(qlg)
if (.Platform$OS.type != "windows") { # temporary
qlg$bedData
}

\dontrun{
## populating the test repository
ql <- Quickload(file.path(tests_dir, "quickload"), create = TRUE)
reference_seq <- import(file.path(tests_dir, "test.2bit"))
names(reference_seq) <- "test"
qlg <- QuickloadGenome(ql, "T_species_Oct_2011", create = TRUE,
                       seqinfo = seqinfo(reference_seq))
referenceSequence(qlg) <- reference_seq
test_bed <- import(file.path(tests_dir, "test.bed"))
names(test_bed) <- "test"
qlg$bedData <- test_bed
test_bedGraph <- import(file.path(tests_dir, "test.bedGraph"))
names(test_bedGraph) <- "test"
start(test_bedGraph) <- seq(1, 90, 10)
width(test_bedGraph) <- 10
track(qlg, "bedGraphData", format = "bw") <- test_bedGraph
}
}

\keyword{methods}
\keyword{classes}

\name{UCSCSchema-class}
\docType{class}

\alias{UCSCSchema-class}
\alias{genome,UCSCSchema-method}
\alias{tableName,UCSCSchema-method}
\alias{genome,UCSCSchema-method}
\alias{nrow,UCSCSchema-method}
\alias{ucscSchema,UCSCSchemaDescription-method}
\alias{show,UCSCSchema-method}

% Accessors:

\title{UCSC Schema}
\description{This is a preliminary class that describes a table in the
  UCSC database. The description includes the table name, corresponding
  genome, row count, and a textual description of the format. In the
  future, we could provide more table information, like the links and
  sample data frame. This is awaiting a use-case.} 

\section{Accessor methods}{
  In the code snippets below, \code{x}/\code{object} is a
  \code{UCSCSchema} object.

  \describe{
    \item{\code{genome(x)}}{Get the genome for the table. }
    \item{\code{tableName(x)}}{Get the name of the table. }
    \item{\code{nrow(x)}}{Get the number of rows in the table. }
  }
}

\author{ Michael Lawrence }

\examples{
\dontrun{
session <- browserSession()
genome(session) <- "mm9"
query <- ucscTableQuery(session, "knownGene")
schema <- ucscSchema(query)
nrow(schema)
}
}

\keyword{methods}
\keyword{classes}

\name{browseGenome}
\alias{browseGenome}
\alias{browseGenome,GenomicRanges_OR_GenomicRangesList-method}
\alias{browseGenome,missing-method}
\title{ Browse a genome }
\description{
  A generic function for launching a genome browser.
}
\usage{
browseGenome(object, ...)
\S4method{browseGenome}{GenomicRanges_OR_GenomicRangesList}(object,
  browser = "UCSC", range = base::range(object),
  view = TRUE, trackParams = list(), viewParams = list(),
  name = "customTrack", ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{object}{ A \code{\link[GenomicRanges]{GRanges}} object or a list
    of \code{\link[GenomicRanges]{GRanges}} objects (e.g. a
    \code{\link[GenomicRanges]{GenomicRangesList}} object). }
  \item{browser}{ The name of the genome browser. }
  \item{range}{ A genome identifier or a
    \code{\link[GenomicRanges]{GRanges}} or
  \code{\link[IRanges]{IntegerRangesList}} to display in the
    initial view. }
  \item{view}{ Whether to open a view. }
  \item{trackParams}{ Named list of parameters to pass to
    \code{\link{track<-}}. }
  \item{viewParams}{ Named list of parameters to pass to
    \code{\link{browserView}}. }
  \item{name}{ The name for the track. }
  \item{\dots}{ Arguments passed to \code{\link{browserSession}}. } 
}
\value{
  Returns a \code{\linkS4class{BrowserSession}}.
}
\author{ Michael Lawrence }
\seealso{ \code{\linkS4class{BrowserSession}} and
  \code{\linkS4class{BrowserView}}, the two main classes for interfacing
  with genome browsers.}
\examples{
  \dontrun{
  ## open UCSC genome browser:
  browseGenome()
  ## to view a specific range:
  range <- GRangesForUCSCGenome("hg18", "chr22", IRanges(20000, 50000))
  browseGenome(range = range)
  ## a slightly larger range:
  browseGenome(range = range, end = 75000)
  ## with a track:
  track <- import(system.file("tests", "v1.gff", package = "rtracklayer"))
  browseGenome(GRangesList(track))
  }
}
\keyword{interface}

\name{readGFF}

\alias{readGFF}
\alias{GFFcolnames}


\title{Reads a file in GFF format}

\description{
  Reads a file in GFF format and creates a data frame or
  \link[S4Vectors]{DataFrame} object from it. This is a low-level
  function that should not be called by user code.
}

\usage{
readGFF(filepath, version=0,
        columns=NULL, tags=NULL, filter=NULL, nrows=-1,
        raw_data=FALSE)

GFFcolnames(GFF1=FALSE)
}

\arguments{
  \item{filepath}{
    A single string containing the path or URL to the file to read.
    Alternatively can be a connection.
  }
  \item{version}{
    \code{readGFF} should do a pretty descent job at detecting the GFF
    version. Use this argument \emph{only} if it doesn't or if you want to
    force it to parse and import the file as if its 9-th column was in a
    different format than what it really is (e.g. specify \code{version=1}
    on a GTF or GFF3 file to interpret its 9-th column as the \code{"group"}
    column of a GFF1 file). Supported versions are 1, 2, and 3.
  }
  \item{columns}{
    The standard GFF columns to load. All of them are loaded by default.
  }
  \item{tags}{
    The tags to load. All of them are loaded by default.
  }
  \item{filter}{
  }
  \item{nrows}{
    \code{-1} or the maximum number of rows to read in (after filtering).
  }
  \item{raw_data}{
  }
  \item{GFF1}{
  }
}

\value{
  A DataFrame with columns corresponding to those in the GFF.
}

\author{
  H. Pages
}

\seealso{
  \itemize{
    \item \code{\link[BiocIO]{import}} for importing a GFF file as a
          \link[GenomicRanges]{GRanges} object.

    \item \code{\link[GenomicRanges]{makeGRangesFromDataFrame}} in the
          \pkg{GenomicRanges} package for making a
          \link[GenomicRanges]{GRanges} object from a data frame or
          \link[S4Vectors]{DataFrame} object.

    \item \code{\link[txdbmaker]{makeTxDbFromGFF}} in the
          \pkg{txdbmaker} package for importing a GFF file as a
          \link[GenomicFeatures]{TxDb} object.

    \item The \link[S4Vectors]{DataFrame} class in the \pkg{S4Vectors}
          package.
  }
}

\examples{
## Standard GFF columns.
GFFcolnames()
GFFcolnames(GFF1=TRUE)  # "group" instead of "attributes"

tests_dir <- system.file("tests", package="rtracklayer")
test_gff3 <- file.path(tests_dir, "genes.gff3")

## Load everything.
df0 <- readGFF(test_gff3)
head(df0)

## Load some tags only (in addition to the standard GFF columns).
my_tags <- c("ID", "Parent", "Name", "Dbxref", "geneID")
df1 <- readGFF(test_gff3, tags=my_tags)
head(df1)

## Load no tags (in that case, the "attributes" standard column
## is loaded).
df2 <- readGFF(test_gff3, tags=character(0))
head(df2)

## Load some standard GFF columns only (in addition to all tags).
my_columns <- c("seqid", "start", "end", "strand", "type")
df3 <- readGFF(test_gff3, columns=my_columns)
df3
table(df3$seqid, df3$type)
makeGRangesFromDataFrame(df3, keep.extra.columns=TRUE)

## Combine use of 'columns' and 'tags' arguments.
readGFF(test_gff3, columns=my_columns, tags=c("ID", "Parent", "Name"))
readGFF(test_gff3, columns=my_columns, tags=character(0))

## Use the 'filter' argument to load only features of type "gene"
## or "mRNA" located on chr10.
my_filter <- list(type=c("gene", "mRNA"), seqid="chr10")
readGFF(test_gff3, filter=my_filter)
readGFF(test_gff3, columns=my_columns, tags=character(0), filter=my_filter)
}

\keyword{manip}

\name{Bed15TrackLine-class}
\docType{class}
\alias{Bed15TrackLine-class}
\alias{coerce,character,Bed15TrackLine-method}
\alias{coerce,Bed15TrackLine,character-method}
\alias{fileFormat,Bed15TrackLine-method}

\title{Class "Bed15TrackLine"}
\description{ A \acronym{UCSC} track line for graphical tracks. }
\section{Objects from the Class}{
  Objects can be created by calls of the form \code{new("Bed15TrackLine",
    ...)} or parsed from a character vector track line with
  \code{as(text, "Bed15TrackLine")}.
}
\section{Slots}{
  \describe{
    \item{\code{expStep}:}{A \code{"numeric"} scalar indicating the step
      size for the heatmap color gradient.}
    \item{\code{expScale}:}{A positive \code{"numeric"} scalar indicating the
      range of the data to be \code{[-expScale, expScale]} for
      determining the heatmap color gradient.}
    \item{\code{expNames}:}{A \code{"character"} vector naming the
      the experimental samples.}
    \item{\code{name}:}{Object of class \code{"character"} specifying
      the name of the track. }
    \item{\code{description}:}{Object of class \code{"character"}
      describing the track. }
    \item{\code{visibility}:}{Object of class \code{"character"}
      indicating the default visible mode of the track, see
      \code{\linkS4class{UCSCTrackModes}}.}
    \item{\code{color}:}{Object of class \code{"integer"} representing the
      track color (as from \code{\link{col2rgb}}).
    }
    \item{\code{priority}:}{Object of class \code{"numeric"} specifying
      the rank of this track. }
  }
}
\section{Extends}{
  Class \code{"\linkS4class{TrackLine}"}, directly.
}
\section{Methods}{
  \describe{
    \item{as(object, "character")}{Export line to its string
      representation.}
    \item{fileFormat(x)}{Returns \dQuote{bed15}}
  }
}
\references{ Official documentation:
  \url{https://genomewiki.ucsc.edu/index.php/Microarray_track}. }
\author{ Michael Lawrence }
\seealso{
  \code{\link{export.bed15}} for exporting bed15 tracks.
}
\keyword{classes}

\name{tracks-methods}
\docType{methods}
\alias{trackNames}
\alias{trackNames<-}
\alias{trackNames<--methods}
\alias{trackNames<-,UCSCTrackModes-method}
\alias{trackNames<-,UCSCView-method}
\alias{trackNames-methods}
\alias{trackNames,UCSCSession-method}
\alias{trackNames,UCSCTrackModes-method}
\alias{trackNames,UCSCView-method}
\title{ Accessing track names }
\description{
  Methods for getting and setting track names.
}
\section{Methods}{
\describe{
  The following methods are defined by \pkg{rtracklayer} for \bold{getting}
  track names via the generic \code{trackNames(object, ...)}.
  
  \item{object = "UCSCSession"}{ Get the tracks loaded in the session. }

  \item{object = "UCSCTrackModes"}{ Get the visible tracks according to
    the modes (all tracks not set to "hide").}

  \item{object = "UCSCView"}{ Get the visible tracks in the view. }

  The following methods are defined by \pkg{rtracklayer} for \bold{setting}
  track names via the generic \code{trackNames(object) <- value}.
  
  \item{object = "UCSCTrackModes"}{ Sets the tracks that should be
    visible in the modes. All specified tracks with mode "hide" in
    \code{object} are set to mode "full". Any tracks in \code{object} that
    are not specified in the value are set to "hide". No other modes are
    changed.}

  \item{object = "UCSCView"}{ Sets the visible tracks in the view. This
    opens a new web browser with only the specified tracks visible. }
}}
\keyword{methods}

\name{BamFile-methods}

\alias{export,ANY,BamFile,ANY-method}
\alias{export,GAlignments,BamFile,ANY-method}
\alias{export,GAlignmentPairs,BamFile,ANY-method}
\alias{import,BamFile,ANY,ANY-method}

\title{Export to BAM Files}
\description{
  Methods for import and export of
  \code{\link[GenomicAlignments]{GAlignments}} or
  \code{\link[GenomicAlignments]{GAlignmentPairs}} objects from and to BAM
  files, represented as \code{\link[Rsamtools]{BamFile}} objects.
}
\usage{
\S4method{import}{BamFile,ANY,ANY}(con, format, text, paired = FALSE,
                                   use.names = FALSE,
                                   param = ScanBamParam(...),
                                   genome = NA_character_, ...)
\S4method{export}{ANY,BamFile,ANY}(object, con, format, ...)
}
\arguments{
  \item{object}{The object to export, such as
    a \code{GAlignments} or \code{GAlignmentPairs}.
  }
  \item{con}{A path, URL, connection or \code{BamFile} object.}
  \item{format}{If not missing, should be \dQuote{bam}.}
  \item{text}{Not supported.}
  \item{paired}{If \code{TRUE}, return a GAlignmentPairs object,
    otherwise a GAlignments.
  }
  \item{use.names}{Whether to parse QNAME as the names on the result.}
  \item{param}{The \code{\link[Rsamtools]{ScanBamParam}} object
    governing the import.
  }
  \item{genome}{Single string or \code{Seqinfo} object identifying the genome}
  \item{\dots}{Arguments that are passed to \code{ScanBamParam}
    if \code{param} is missing.
  }
}
\details{
  BAM fields not formally present in the \code{GAlignments[Pairs]} object
  are extracted from the metadata columns, if present; otherwise, the
  missing value, \dQuote{"."}, is output. The file is sorted and
  indexed. This can be useful for subsetting BAM files, although
  \code{\link[Rsamtools]{filterBam}} may eventually become flexible
  enough to be the favored alternative.
}
\seealso{
  The \code{\link[GenomicAlignments]{readGAlignments}} and
  \code{\link[GenomicAlignments]{readGAlignmentPairs}} functions
  for reading BAM files.
}
\examples{
     library(Rsamtools)
     ex1_file <- system.file("extdata", "ex1.bam", package="Rsamtools")
     gal <- import(ex1_file, param=ScanBamParam(what="flag"))
     gal.minus <- gal[strand(gal) == "-"]
\dontrun{
     export(gal, BamFile("ex1-minus.bam"))
}
}
\author{Michael Lawrence}

\name{sequence<--methods}
\docType{methods}
\alias{sequence<-}
\alias{sequence<--methods}
\title{ Load a sequence }
\description{
  Methods for loading sequences.
}
\section{Methods}{
\describe{
  No methods are defined by \pkg{rtracklayer} for the
  \code{sequence(object, ...) <- value} generic.
}}
\keyword{methods}

\name{cpneTrack}
\alias{cpneTrack}
\docType{data}
\title{CPNE1 SNP track}
\description{
  A \code{GRanges} object (created by the \code{GGtools} package)
  with features from a subset of the SNPs on chromosome 20 from 60
  HapMap founders in the CEU cohort. Each SNP has an associated data
  value indicating its association with the expression of the CPNE1 gene
  according to a Cochran-Armitage 1df test. The top 5000 scoring SNPs
  were selected for the track.
}
\usage{data(cpneTrack)}
\format{
  Each feature (row) is a SNP. The association test scores are
  accessible via \code{\link[BiocGenerics]{score}}.
}
\source{
  Vince Carey and the \code{GGtools} package.
}
\examples{
data(cpneTrack)
plot(start(cpneTrack), score(cpneTrack))
}
\keyword{datasets}

\name{BigWigSelection-class}
\docType{class}
\alias{BigWigSelection-class}

% constructor
\alias{BigWigSelection}

\alias{coerce,IntegerRangesList,BigWigSelection-method}
\alias{coerce,GenomicRanges,BigWigSelection-method}

\title{Selection of ranges and columns}

\description{A \code{BigWigSelection} represents a query against a
  BigWig file, see \code{\link{import.bw}}. It is simply
  a \link[IRanges]{RangedSelection} that requires its \code{colnames}
  parameter to be "score", if non-empty, as that is the only column
  supported by BigWig.}

\section{Constructor}{
  \describe{
    \item{\code{BigWigSelection(ranges = GRanges(), colnames =
        "score")}}{Constructs a \code{BigWigSelection} with the given
        \code{ranges} and \code{colnames}. \code{ranges} can be either
        something coercible to a \code{\link[IRanges:IntegerRangesList-class]{IntegerRangesList}},
        a \code{character} identifying a genome (see
        \code{\link{GenomicSelection}}), or a
        \code{\linkS4class{BigWigFile}}, in which case the ranges are
        derived from the bounds of its sequences.
    }
  }
}

\section{Coercion}{
  \describe{
    \item{\code{as(from, "BigWigSelection")}}{Coerces \code{from} to a
      \code{BigWigSelection} object. Typically, \code{from} is a
      \code{\link[GenomicRanges]{GRanges}} or
      a \code{\link[IRanges]{IntegerRangesList}}, the ranges of
      which become the ranges in the
      new \code{BigWigSelection}.
    }
  }
}

\author{ Michael Lawrence }

\examples{
  rl <- IRangesList(chr1 = IRanges::IRanges(c(1, 5), c(3, 6)))

  BigWigSelection(rl)
  as(rl, "BigWigSelection") # same as above

  # do not select the 'score' column
  BigWigSelection(rl, character())
}

\keyword{methods}
\keyword{classes}

\name{BrowserView-class}
\docType{class}
\alias{BrowserView-class}
\alias{show,BrowserView-method}
\alias{visible}
\alias{visible,BrowserView-method}
\alias{visible<-}
\alias{visible<-,BrowserView-method}

\title{Class "BrowserView"}
\description{ An object representing a genome browser view of a particular
  segment of a genome. }
\section{Objects from the Class}{
  A virtual Class: No objects may be created from it directly. See
  \code{\link{browserView}} for obtaining
  an instance of an implementation for a particular genome browser.
}
\section{Slots}{
  \describe{
    \item{\code{session}:}{Object of class \code{"BrowserSession"} the
      browser session to which this view belongs. }
  }
}
\section{Methods}{
  This specifies the API implemented by each browser backend. Note that
  a backend is not guaranteed to support all operations. See
  the backend-specific documentation for more details. The only built-in
  backend is \code{\linkS4class{UCSCView}}.
  \describe{
    \item{\code{\link{browserSession}(object)}}{ Obtains the
      \code{\linkS4class{BrowserSession}} to which this view belongs. }
    \item{\code{\link{close}(object)}}{ Close this view. }
    \item{\code{\link{range}(object)}}{ Obtains the
      \code{\link[GenomicRanges:GRanges-class]{GRanges}} displayed by
      this view.} 
    \item{\code{\link{trackNames}(object)}}{ Gets the names of the visible
      tracks in the view. }
    \item{\code{\link{trackNames}(object) <- value}}{ Sets the visible
      tracks by their names. }
    \item{\code{\link{show}(object)}}{ Outputs a textual description of
      this view. }
    \item{\code{visible(object)}}{Get a named logical vector indicating
      whether each track is visible.
    }
    \item{\code{visible(object) <- value}}{Set a logical vector
      indicating the visibility of each track, with the same names and
      in the same order as that returned by \code{visible(object)}.
    }
  }
}
\author{ Michael Lawrence }
\seealso{
  \code{\link{browserView}} for obtaining instances of this class.
}
\keyword{classes}

\name{asGFF}
\alias{asGFF}
\alias{asGFF,GRangesList-method}

\title{Coerce to GFF structure}
\description{
  Coerce the structure of an object to one following GFF-like
  conventions, i.e., using the \code{Parent} GFF3 attribute to encode
  the hierarchical structure. This object is then suitable for export as GFF3.
}
\usage{
asGFF(x, ...)
\S4method{asGFF}{GRangesList}(x, parentType = "mRNA", childType = "exon")
}

\arguments{
  \item{x}{
    Generally, a tabular object to structure as GFF(3)
  }
  \item{parentType}{The value to store in the \code{type} column for the
    top-level (e.g., transcript) ranges.
  }
  \item{childType}{The value to store in the \code{type} column for the
    child (e.g., exon) ranges.
  }
  \item{\dots}{
    Arguments to pass to methods
  }
}

\value{
  For the \code{GRangesList} method:
  A \code{GRanges}, with the columns: \code{ID} (unique identifier),
  \code{Name} (from \code{names(x)}, and the names on each element of
  \code{x}, if any), \code{type} (as given by \code{parentType} and
  \code{childType}), and \code{Parent} (to relate each child range to
  its parent at the top-level).
}

\author{
  Michael Lawrence
}

\examples{
\dontrun{
library(TxDb.Hsapiens.UCSC.hg19.knownGene)
exons <- exonsBy(TxDb_Hsapiens_UCSC_hg19_knownGene)
mcols(asGFF(exons))
}
}

\name{liftOver}

\alias{liftOver}
\alias{liftOver,GenomicRanges,Chain-method}
\alias{liftOver,GRangesList,Chain-method}
\alias{liftOver,Pairs,Chain-method}
\alias{liftOver,ANY,ANY-method}

\title{
  Lift intervals between genome builds
}
\description{
  A reimplementation of the UCSC liftover tool for lifting features from
  one genome build to another. In our preliminary tests, it is
  significantly faster than the command line tool. Like the UCSC tool, a
  chain file is required input.
}
\usage{
liftOver(x, chain, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
    The intervals to lift-over, usually
    a \code{\link[GenomicRanges:GRanges]{GRanges}}.
  }
  \item{chain}{
    A \code{\linkS4class{Chain}} object, usually imported with
    \code{\link{import.chain}}, or something coercible to one.
  }
  \item{\dots}{
    Arguments for methods.
  }
}
\value{
  A \code{GRangesList} object. Each element contains the ranges mapped
  from the corresponding element in the input (may be one-to-many).
}
\references{
  \url{https://genome.ucsc.edu/cgi-bin/hgLiftOver}
}
\examples{
\dontrun{
chain <- import.chain("hg19ToHg18.over.chain")
library(TxDb.Hsapiens.UCSC.hg19.knownGene)
tx_hg19 <- transcripts(TxDb.Hsapiens.UCSC.hg19.knownGene)
tx_hg18 <- liftOver(tx_hg19, chain)
}
}
\author{
  Michael Lawrence
}

\name{Chain-class}
\docType{class}

%% Classes:
\alias{class:Chain}
\alias{Chain-class}
\alias{ChainBlock-class}
\alias{class:ChainBlock}
\alias{class:ChainFile}
\alias{ChainFile-class}

%% Constructor:
\alias{import.chain}
\alias{import.chain,ANY-method}
\alias{import,ChainFile,ANY,ANY-method}
\alias{ChainFile}

%% Generics and methods:
\alias{ranges,ChainBlock-method}
\alias{offset,ChainBlock-method}
\alias{score,ChainBlock-method}
\alias{space,ChainBlock-method}
\alias{reversed}
\alias{reversed,ChainBlock-method}

\title{Chain objects}

\description{
  A \code{Chain} object represents a UCSC chain alignment, typically
  imported from a \code{chain} file, and is essentially a list of
  \code{ChainBlock} objects. Each \code{ChainBlock} has a corresponding
  chromosome (its name in the list) and is a run-length
  encoded alignment, mapping a set of intervals on that chromosome to
  intervals on the same or other chromosomes. 
}

\note{
  A chain file essentially details many local alignments, so it is
  possible for the "from" ranges to map to overlapping regions in the
  other sequence. The "from" ranges are guaranteed to be disjoint (but
  do not necessarily cover the entire "from" sequence).
}

\section{Accessor Methods}{
  In the code snippets below, \code{x} and \code{object} are
  \code{ChainBlock} objects.
  
  \describe{
    \item{\code{ranges(x)}}{
      Get the \code{\link[IRanges]{IntegerRanges}} object holding
      the starts and ends of the "from" ranges. Each range is a
      contiguous block of positions aligned without gaps to the other
      sequence.
    }
    \item{\code{offset(x)}}{
      Integer offset from the "from" start to the
      "end" start (which could be in another chromosome).
    }
    \item{\code{score(x)}}{
      The score for each mapping.
    }
    \item{\code{space(x)}}{
      The space (chromosome) of the "to" range.
    }
    \item{\code{reversed(x)}}{
      Whether the mapping inverts the region, i.e.,
      the alignment is between different strands.
    }
  }
}

\section{Import}{
  A \code{Chain} object can be loaded from a UCSC chain format file
  simply by passing the path \code{import} function. If the file
  extension is not \dQuote{chain}, then either pass \dQuote{chain} to
  the \code{format} argument, or cast the path to a \code{ChainFile}
  object. The \code{import.chain} function is provided as a (slight)
  convenience. It is documented below, along with the extra
  \code{exclude} argument to the \code{import} method.
  
  \describe{
    \item{\code{import.chain(con, exclude = "_", ...)}}{
      Imports a chain file named \code{con} as a \code{Chain} object, a
      list of \code{ChainBlock}s. Alignments for chromosomes matching the
      \code{exclude} pattern are not imported.
    }
  }
}

\author{Michael Lawrence}

\seealso{
  \code{\link{liftOver}} for performing lift overs using a chain alignment
}

\keyword{methods}
\keyword{classes}

\name{GraphTrackLine-class}
\docType{class}
\alias{GraphTrackLine-class}
\alias{coerce,character,GraphTrackLine-method}
\alias{coerce,GraphTrackLine,character-method}
\alias{coerce,GraphTrackLine,BasicTrackLine-method}
\alias{coerce,BasicTrackLine,GraphTrackLine-method}
\alias{fileFormat,GraphTrackLine-method}

\title{Class "GraphTrackLine"}
\description{ A \acronym{UCSC} track line for graphical tracks. }
\section{Objects from the Class}{
  Objects can be created by calls of the form \code{new("GraphTrackLine",
    ...)} or parsed from a character vector track line with
  \code{as(text, "GraphTrackLine")} or converted from a
  \code{\linkS4class{BasicTrackLine}} using \code{as(basic, "GraphTrackLine")}.
}
\section{Slots}{
  \describe{
    \item{\code{altColor}:}{Object of class \code{"integer"} giving an
      alternate color, as from \code{\link{col2rgb}}.}
    \item{\code{autoScale}:}{Object of class \code{"logical"} indicating
      whether to automatically scale to min/max of the data.}
    \item{\code{alwaysZero}:}{Object of class \code{"logical"} indicating
      whether to fix the lower limit of the Y axis at zero.}
    \item{\code{gridDefault}:}{Object of class \code{"logical"}
      indicating whether a grid should be drawn. }
    \item{\code{maxHeightPixels}:}{Object of class \code{"numeric"} of
      length three (max, default, min), giving the allowable range for
      the vertical height of the graph.}
    \item{\code{graphType}:}{Object of class \code{"character"},
      specifying the graph type, either "bar" or "points". }
    \item{\code{viewLimits}:}{Object of class \code{"numeric"} and of
      length two specifying the data range (min, max) shown in the graph.}
    \item{\code{yLineMark}:}{Object of class \code{"numeric"} giving the
      position of a horizontal line.}
    \item{\code{yLineOnOff}:}{Object of class \code{"logical"}
      indicating whether the \code{yLineMark} should be visible. }
    \item{\code{windowingFunction}:}{Object of class \code{"character"},
      one of "maximum", "mean", "minimum", for removing points when the
      graph shrinks.}
    \item{\code{smoothingWindow}:}{Object of class \code{"numeric"}
      giving the window size of a smoother to pass over the graph. }
    \item{\code{type}:}{Scalar \code{"character"} indicating the type of
      the track, either \code{"wig"} or \code{"bedGraph"}.}
    \item{\code{name}:}{Object of class \code{"character"} specifying
      the name of the track. }
    \item{\code{description}:}{Object of class \code{"character"}
      describing the track. }
    \item{\code{visibility}:}{Object of class \code{"character"}
      indicating the default visible mode of the track, see
      \code{\linkS4class{UCSCTrackModes}}.}
    \item{\code{color}:}{Object of class \code{"integer"} representing the
      track color (as from \code{\link{col2rgb}}).
    }
    \item{\code{priority}:}{Object of class \code{"numeric"} specifying
      the rank of this track. }
  }
}
\section{Extends}{
  Class \code{"\linkS4class{TrackLine}"}, directly.
}
\section{Methods}{
  \describe{
    \item{as(object, "character")}{Export line to its string
      representation.}
    \item{as(object, \code{"\linkS4class{BasicTrackLine}"})}{Convert this
      line to a basic UCSC track line, using defaults for slots not held
      in common.}
    \item{\code{fileFormat(x)}}{Returns the value of \code{@type},
    either \dQuote{wig} or \dQuote{bedGraph}}
  }
}
\references{ Official documentation:
  \url{https://genome.ucsc.edu/goldenPath/help/wiggle.html}. }
\author{ Michael Lawrence }
\seealso{
  \code{\link{export.wig}}, \code{\link{export.bedGraph}} for exporting
  graphical tracks.
}
\keyword{classes}

\name{BEDFile-class}
\docType{class}

%% Classes:
\alias{class:BEDFile}
\alias{class:BED15File}
\alias{class:BEDGraphFile}
\alias{class:NarrowPeakFile}
\alias{class:BroadPeakFile}
\alias{BEDFile-class}
\alias{BED15File-class}
\alias{BEDGraphFile-class}
\alias{BEDPEFile-class}
\alias{NarrowPeakFile-class}
\alias{BroadPeakFile-class}

%% Constructor:
\alias{BEDFile}
\alias{BED15File}
\alias{BEDGraphFile}
\alias{BEDPEFile}
\alias{NarrowPeakFile}
\alias{BroadPeakFile}

%% Import:
\alias{import,BEDFile,ANY,ANY-method}
\alias{import,BED15File,ANY,ANY-method}
\alias{import,BEDPEFile,ANY,ANY-method}
\alias{import,NarrowPeakFile,ANY-method}
\alias{import,BroadPeakFile,ANY-method}
\alias{import,NarrowPeakFile,ANY,ANY-method}
\alias{import,BroadPeakFile,ANY,ANY-method}
\alias{import.bed}
\alias{import.bed,ANY-method}
\alias{import.bed15}
\alias{import.bed15,ANY-method}
\alias{import.bedGraph}
\alias{import.bedGraph,ANY-method}
\alias{import.NarrowPeak}
\alias{import.NarrowPeak,ANY-method}
\alias{import.BroadPeak}
\alias{import.BroadPeak,ANY-method}

%% Export:
\alias{export,ANY,BEDFile,ANY-method}
\alias{export,GenomicRanges,BEDFile,ANY-method}
\alias{export,GRangesList,BEDFile,ANY-method}
\alias{export,SimpleGRangesList,BEDFile,ANY-method}
\alias{export,CompressedGRangesList,BEDFile,ANY-method}
\alias{export,GenomicRangesList,BEDFile,ANY-method}
\alias{export,UCSCData,BEDFile,ANY-method}
\alias{export,GenomicRanges,BED15File,ANY-method}
\alias{export,ANY,BEDPEFile,ANY-method}
\alias{export,Pairs,BEDPEFile,ANY-method}
\alias{export,UCSCData,BED15File,ANY-method}
\alias{export.bed}
\alias{export.bed,ANY-method}
\alias{export.bed15}
\alias{export.bed15,ANY-method}
\alias{export.bedGraph}
\alias{export.bedGraph,ANY-method}

\title{BEDFile objects}

\description{
  These functions support the import and export of the UCSC BED
  format and its variants, including BEDGraph.
}

\usage{
\S4method{import}{BEDFile,ANY,ANY}(con, format, text, trackLine = TRUE,
                   genome = NA, colnames = NULL,
                   which = NULL, seqinfo = NULL, extraCols = character(),
                   sep = c("\t", ""), na.strings=character(0L))
import.bed(con, ...)
import.bed15(con, ...)
import.bedGraph(con,  ...)

\S4method{export}{ANY,BEDFile,ANY}(object, con, format, ...)
\S4method{export}{GenomicRanges,BEDFile,ANY}(object, con, format,
                  append = FALSE, index = FALSE,
                  ignore.strand = FALSE, trackLine = NULL)
\S4method{export}{UCSCData,BEDFile,ANY}(object, con, format,
                   trackLine = TRUE, ...)
export.bed(object, con, ...)
export.bed15(object, con, ...)
\S4method{export}{GenomicRanges,BED15File,ANY}(object, con, format,
                  expNames = NULL, trackLine = NULL, ...)
export.bedGraph(object, con, ...)
}

\arguments{
  \item{con}{A path, URL, connection or \code{BEDFile} object. For the
    functions ending in \code{.bed}, \code{.bedGraph} and \code{.bed15},
    the file format is indicated by the function name. For the base
    \code{export} and \code{import} functions, the format must be
    indicated another way. If \code{con} is a path, URL or connection,
    either the file extension or the \code{format} argument needs to be
    one of \dQuote{bed}, \dQuote{bed15}, \dQuote{bedGraph},
    \dQuote{bedpe}, \dQuote{narrowPeak}, or
    \dQuote{broadPeak}. Compressed
    files (\dQuote{gz}, \dQuote{bz2} and
    \dQuote{xz}) are handled transparently.
  }
  \item{object}{The object to export, should be a \code{GRanges} or
    something coercible to a \code{GRanges}. If targeting the BEDPE
    format, this should be something coercible to \code{Pairs}.
    If the object has a method
    for \code{asBED} (like \code{GRangesList}), it is called prior to
    coercion. This makes it possible to export a \code{GRangesList} or
    \code{TxDb} in a way that preserves the hierarchical
    structure. For exporting multiple tracks, in the UCSC track line
    metaformat, pass a \code{GenomicRangesList}, or something coercible to
    one.
  }
  \item{format}{If not missing, should be one of \dQuote{bed},
    \dQuote{bed15}, \dQuote{bedGraph}, \dQuote{bedpe},
  \dQuote{narrowPeak} or \dQuote{broadPeak}.}
  \item{text}{If \code{con} is
    missing, a character vector to use as the input}
  \item{trackLine}{For import, an imported track line will be stored
    in a \code{\linkS4class{TrackLine}} object, as part of the returned
    \code{\linkS4class{UCSCData}}.  For the UCSCData method on export,
    whether to output the UCSC track line stored on the object, for the
    other export methods, the actual TrackLine object to export.
  }
  \item{genome}{The identifier of a genome, or a \code{Seqinfo}, or \code{NA} if
    unknown. Typically, this is a UCSC identifier like \dQuote{hg19}. An
    attempt will be made to derive the \code{seqinfo} on the return
    value using either an installed BSgenome package or UCSC, if network
    access is available.
  }
  \item{colnames}{A character vector naming the columns to parse. These
    should name columns in the result, not those in the BED spec, so
    e.g. specify \dQuote{thick}, instead of \dQuote{thickStart}.
  }
  \item{which}{A \code{GRanges} or other range-based object supported
    by \code{\link[IRanges]{findOverlaps}}. Only the intervals in the file
    overlapping the given ranges are returned. This is much more efficient
    when the file is indexed with the tabix utility.
  }
  \item{index}{If \code{TRUE}, automatically compress and index the
    output file with bgzf and tabix. Note that tabix indexing will
    sort the data by chromosome and start. Tabix supports a
    single track in a file.
  }
  \item{ignore.strand}{Whether to output the strand when not required
    (by the existence of later fields).
  }
  \item{seqinfo}{If not \code{NULL}, the \code{Seqinfo} object to set on
    the result. Ignored if \code{genome} is a \code{Seqinfo} object.
    If the \code{genome} argument is not \code{NA}, it must
    agree with \code{genome(seqinfo)}.
  }
  \item{extraCols}{A character vector in the same form as
    \code{colClasses} from \code{\link{read.table}}.  It should indicate
    the name and class of each extra/special column to read from the BED
    file. As BED does not encode column names, these are assumed to be
    the last columns in the file. This enables parsing of the various
    BEDX+Y formats.
  }
  \item{sep}{A character vector with a single character indicating the
    field separator, like \code{read.table}. This defaults to
    \code{"\t"}, as BEDtools requires, but BED files are also allowed to
    be whitespace separated (\code{""}) according to the UCSC spec.
  }
  \item{na.strings}{Character vector with strings, appended to the
    standard \code{"."}, that represent an \code{NA} value.
  }
  \item{append}{If \code{TRUE}, and \code{con} points to a file path,
    the data is appended to the file. Obviously, if \code{con} is a
    connection, the data is always appended.
  }
  \item{expNames}{character vector naming columns in
    \code{mcols(object)} to export as data columns in the BED15
    file. These correspond to the sample names in the experiment. If
    \code{NULL} (the default), there is an attempt to extract these from
    \code{trackLine}. If the attempt fails, no scores are exported.
  }
  \item{...}{Arguments to pass down to methods to other methods. For
    import, the flow eventually reaches the \code{BEDFile} method on
    \code{import}. When \code{trackLine} is
    \code{TRUE} or the target format is BED15, the arguments are passed
    through \code{export.ucsc}, so track line parameters are supported.
  }
}

\value{
  For a \dQuote{bedpe} file, a \code{Pairs} object combining two
  \code{GRanges}. The \code{name} and \code{score} are carried over to
  the metadata columns.

  Otherwise, a \code{GRanges} with the metadata columns described in the
  details.
}

\details{
  The BED format is a tab-separated table of intervals, with annotations
  like name, score and even sub-intervals for representing alignments and
  gene models. Official (UCSC) child formats currently include BED15
  (adding a number matrix for e.g. expression data across multiple samples)
  and BEDGraph (a compressed means of storing a single score variable, e.g.
  coverage; overlapping features are not allowed). Many tools and
  organizations have extended the BED format with additional columns for
  particular use cases. The
  advantage of BED is its balance between simplicity and expressiveness. It
  is also relatively scalable, because only the first three columns (chrom,
  start and end) are required. Thus, BED is best suited for representing
  simple features. For specialized cases, one is usually better off with
  another format. For example, genome-scale vectors belong in
  \link[=BigWigFile]{BigWig}, alignments from high-throughput sequencing
  belong in \link[Rsamtools:BamFile]{BAM}, and gene models are more richly
  expressed in \link[=GFFFile]{GFF}.

  The following is the mapping of BED elements to a \code{GRanges} object.
  NA values are allowed only where indicated.
  These appear as a \dQuote{.} in the file. Only the first three columns
  (chrom, start and strand) are required. The other columns can only be
  included if all previous columns (to the left) are included. Upon export,
  default values are used to automatically pad the table, if necessary.
  \describe{
    \item{chrom, start, end}{the \code{ranges} component.}
    \item{name}{character vector (NA's allowed) in the \code{name}
      column; defaults to NA on export.
    }
    \item{score}{numeric vector in the \code{score}
      column, accessible via the \code{score} accessor. Defaults to 0
      on export. This is the only column present in BEDGraph (besides
      chrom, start and end), and it is required.
    }
    \item{strand}{strand factor (NA's allowed) in the \code{strand}
      column, accessible via the \code{strand} accessor; defaults to NA
      on export.
    }
    \item{thickStart, thickEnd}{\code{IntegerRanges} object in a
      column named \code{thick}; defaults to the ranges of the feature
      on export.
    }
    \item{itemRgb}{an integer matrix of color codes, as returned by
      \code{\link{col2rgb}}, or any valid input to
      \code{\link{col2rgb}}, in the \code{itemRgb} column; default is NA
      on export, which translates to black.
    }
    \item{blockSizes, blockStarts, blockCounts}{\code{IntegerRangesList}
      object in a column named \code{blocks}; defaults to empty upon BED15
      export.
    }
  }

  For BED15 files, there should be a column of scores in
  \code{mcols(object)} for each sample in the experiment. The columns
  are named according to the \code{expNames} (found in the file, or
  passed as an argument during export). \code{NA} scores are stored as
  \dQuote{-10000} in the file.
}

\section{BEDX+Y formats}{
  To import one of the multitude of BEDX+Y formats, such as those used
  to distribute ENCODE data through UCSC (narrowPeaks, etc), specify
  the \code{extraCols} argument to indicate the expected names and
  classes of the special columns. We assume that the
  last \code{length(extraCols)} columns are special, and that the
  preceding columns adhere to the BED format. \dQuote{narrowPeak}
  and \dQuote{broadPeak} types are handled explicitly by specifying
  these types as the \code{format} argument, rather than by
  using \code{extraCols}.
}

\section{BEDFile objects}{
  The \code{BEDFile} class extends \code{\link[BiocIO:BiocFile-class]{BiocFile}} and is a
  formal represention of a resource in the BED format.
  To cast a path, URL or connection to a \code{BEDFile}, pass it to
  the \code{BEDFile} constructor. Classes and constructors also exist
  for the subclasses \code{BED15File}, \code{BEDGraphFile} and
  \code{BEDPEFile}.
}

\author{Michael Lawrence}

\references{
  \url{https://genome.ucsc.edu/goldenPath/help/customTrack.html}
  \url{http://bedtools.readthedocs.org/en/latest/content/general-usage.html}
}

\examples{
  test_path <- system.file("tests", package = "rtracklayer")
  test_bed <- file.path(test_path, "test.bed")

  test <- import(test_bed)
  test

  test_bed_file <- BEDFile(test_bed)
  import(test_bed_file)

  test_bed_con <- file(test_bed)
  import(test_bed_con, format = "bed")

  import(test_bed, trackLine = FALSE)
  import(test_bed, genome = "hg19")
  import(test_bed, colnames = c("name", "strand", "thick"))

  which <- GRanges("chr7:1-127473000")
  import(test_bed, which = which)

  bed15_file <- file.path(test_path, "test.bed15")
  bed15 <- import(bed15_file)

\dontrun{
  test_bed_out <- file.path(tempdir(), "test.bed")
  export(test, test_bed_out)

  test_bed_out_file <- BEDFile(test_bed_out)
  export(test, test_bed_out_file)

  export(test, test_bed_out, name = "Alternative name")

  test_bed_gz <- paste(test_bed_out, ".gz", sep = "")
  export(test, test_bed_gz)

  export(test, test_bed_out, index = TRUE)
  export(test, test_bed_out, index = TRUE, trackLine = FALSE)

  bed_text <- export(test, format = "bed")
  test <- import(format = "bed", text = bed_text)

  test_bed15_out <- file.path(tempdir(), "test.bed15")
  export(bed15, test_bed15_out) # UCSCData knows the expNames
  export(as(bed15, "GRanges"), test_bed15_out, # have to specify expNames
         expNames=paste0("breast_", c("A", "B", "C")))
}
}

\keyword{methods}
\keyword{classes}

\name{TrackDb-class}
\docType{class}

%% Classes:
\alias{class:TrackDb}
\alias{TrackDb-class}

%% Accessors:
\alias{$,TrackDb-method}
\alias{$<-,TrackDb-method}
\alias{[[,TrackDb,ANY,ANY-method}
\alias{[[<-,TrackDb,ANY,ANY-method}

%% Data Access:
\alias{track<-,TrackDb,ANY-method}

\title{Track Databases}

\description{
  The \code{TrackDb} class is an abstraction around a database of
  tracks. Implementations include \code{\linkS4class{BrowserSession}}
  derivatives and \code{\linkS4class{QuickloadGenome}}. Here, a track is
  defined as an interval dataset.
}

\section{Accessor Methods}{
  Every implementation should support these methods:
  
  \describe{
    \item{\code{length(x)}}{
      number of tracks
    }
    \item{\code{names(x)}, \code{trackNames(x)}}{
      names of the tracks
    }
    \item{\code{mcols(x)}}{
      merged metadata on the tracks
    }
    \item{\code{track(x, name)}, \code{x$name}, \code{x[[name]]}}{
      get the track called \code{name}
    }
    \item{\code{track(x, name) <- value}, \code{x$name <- value},
      \code{x[[name]] <- value}}{
      store the track \code{value} under
      \code{name}. Different implementations will support different
      types for \code{value}. Generally, an interval data structure like
      \code{GenomicRanges}.
    }
  }
}

\author{Michael Lawrence}

\keyword{methods}
\keyword{classes}

\name{Quickload-class}
\docType{class}

%% Classes:
\alias{class:Quickload}
\alias{Quickload-class}

%% Constructor:
\alias{Quickload}

%% Accessors:
\alias{$,Quickload-method}
\alias{[[,Quickload,ANY,ANY-method}
\alias{genome,Quickload-method}
\alias{names,Quickload-method}
\alias{uri,Quickload-method}
\alias{length,Quickload-method}

%% Coercion:
\alias{coerce,character,Quickload-method}

%% Show:
\alias{show,Quickload-method}

\title{Quickload Access}

\description{
  The \code{Quickload} class represents a Quickload data source,
  essentially directory layout separating tracks and sequences by
  genome, along with a few metadata files. This interface abstracts
  those details and provides access to a Quickload at any URL supported
  by R (HTTP, FTP, and local files). This is an easy way to make data
  accessible to the Integrated Genome Browser (IGB).
}

\section{Constructor}{
  \describe{
    \item{\code{Quickload(uri = "quickload", create = FALSE)}}{
      Constructs a
      new \code{Quickload} object, representing a repository at
      \code{uri}. If \code{create} is \code{TRUE}, and \code{uri} is
      writeable (i.e., local), the repository is created if it does not
      already exist. If it does exist, then a message is emitted to
      indicate that the repository was not recreated.
    }
  }
}

\section{Accessor Methods}{
  In the code snippets below, \code{x} represents a \code{Quickload}
  object.
  
  \describe{
    \item{\code{x$genome}, \code{x[["genome"]]}}{Get
      the \code{\linkS4class{QuickloadGenome}} object for the genome named
      \code{genome}. This is where all the data is stored.
    }
    \item{\code{length(x)}}{number of genomes in the repository
    }
    \item{\code{uri(x)}}{
      Get the URI pointing to the Quickload repository.
    }
    \item{\code{genome(x)}, \code{names(x)}}{
      Get the identifiers of the genomes present in the repository.
    }
  }
}

\author{Michael Lawrence}

\examples{
ql <- Quickload(system.file("tests", "quickload", package = "rtracklayer"))
uri(ql)
genome(ql)
ql$T_species_Oct_2011
}

\keyword{methods}
\keyword{classes}

\name{genomeBrowsers}
\alias{genomeBrowsers}
\title{Get available genome browsers}
\description{
  Gets the identifiers of the loaded genome browser drivers.
}
\usage{
genomeBrowsers(where = topenv(parent.frame()))
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{where}{ The environment in which to search for drivers. }
}
\details{
  This searches the specified environment for classes that extend
  \code{\linkS4class{BrowserSession}}. The prefix of the class name,
  e.g. "ucsc" in "UCSCSession", is returned for each driver. 
}
\value{
  A character vector of driver identifiers.
}
\author{ Michael Lawrence }
\seealso{ \code{\link{browseGenome}} and \code{\link{browserSession}}
  that create \code{browserSession} implementations given an identifier
  returned from this function. }
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory.
\keyword{interface}

\name{ucscTrackModes-methods}
\docType{methods}
\alias{ucscTrackModes}
\alias{ucscTrackModes<-}
\alias{ucscTrackModes-methods}
\alias{ucscTrackModes,character-method}
\alias{ucscTrackModes,missing-method}
\alias{ucscTrackModes,UCSCView-method}
\alias{ucscTrackModes<--methods}
\alias{ucscTrackModes<-,UCSCView,UCSCTrackModes-method}
\alias{ucscTrackModes<-,UCSCView,character-method}
\alias{ucscTrackModes,ucscTracks-method}
\alias{ucscTrackModes,UCSCSession-method}

\title{ Accessing UCSC track modes }
\description{
  Generics for getting and setting UCSC track visibility modes ("hide",
  "dense", "full", "pack", "squish").
}
\section{Methods}{
\describe{
  The following methods are defined by \pkg{rtracklayer} for \bold{getting} the
  track modes through the generic \code{ucscTrackModes(object, ...)}.
  
  \item{object = "character"}{
    \code{function(object, hide = character(),
                   dense = character(), pack = character(),
                   squish = character(), full = character())}
    Creates an instance of \code{\linkS4class{UCSCTrackModes}} from
    \code{object}, a character vector of mode names, with the
    corresponding track ids given in the \code{names} attribute. Note
    that \code{object} can be a \code{UCSCTrackModes} instance, as
    \code{UCSCTrackModes} extends \code{character}.
    The other parameters are character vectors identifying the tracks for
    each mode and overriding the modes specified by
    \code{object}.
  }

  \item{object = "missing"}{
    The same interface as above, except \code{object} defaults to an empty
    character vector.
  }

  \item{object = "UCSCView"}{
    Gets modes for tracks in the view.
  }
  
  \item{object = "UCSCSession"}{
    Gets default modes for the tracks in the session. These are the
    modes that will be used as the default for a newly created view.
  }

  The following methods are defined by \pkg{rtracklayer} for \bold{setting} the
  track modes through the generic \code{ucscTrackModes(object) <- value}.

  \item{object = "UCSCView", value = "UCSCTrackModes"}{
    Sets the modes for the tracks in the view.
  }

  \item{object = "UCSCView", value = "character"}{
    Sets the modes from a character vector of mode names, with the
    corresponding track names given in the \code{names} attribute.
  }
}}

\seealso{\code{\link{trackNames}} and \code{\link{trackNames<-}} for just
  getting or setting which tracks are visible (not of mode "hide").}

\examples{
# Tracks "foo" and "bar" are fully shown, "baz" is hidden
modes <- ucscTrackModes(full = c("foo", "bar"), hide = "baz")
# Update the modes to hide track "bar"
modes2 <- ucscTrackModes(modes, hide = "bar")
}
\keyword{methods}

\name{UCSCSession-class}
\docType{class}
\alias{UCSCSession-class}
\alias{initialize,UCSCSession-method}
\alias{track,UCSCSession-method}
\alias{track<-,UCSCSession,SimpleGRangesList-method}
\alias{range<-}
\alias{range<-,UCSCSession-method}
\alias{track<-,UCSCSession,BiocFile-method}
\alias{genome,UCSCSession-method}
\alias{genome<-,UCSCSession-method}
\alias{range,UCSCSession-method}
\alias{seqinfo,UCSCSession-method}
\alias{range,ucscCart-method}
\alias{genome,ucscCart-method}
\alias{getTable,UCSCSession-method}

\title{Class "UCSCSession"}
\description{An implementation of \code{\linkS4class{BrowserSession}}
  for the \acronym{UCSC} genome browser.}
\section{Objects from the Class}{
  Objects can be created by calls of the form
  \code{\link{browserSession}("ucsc", url =
    "https://genome.ucsc.edu/cgi-bin", ...)}. The arguments in \code{...}
  correspond to libcurl options, see
  \code{\link[httr]{httr_options}}. Setting these options may be
  useful e.g. for getting past a proxy.
}
\section{Slots}{
  \describe{
    \item{\code{url}:}{Object of class \code{"character"} holding the
      base \acronym{URL} of the \acronym{UCSC} browser. }
    \item{\code{hguid}:}{Object of class \code{"numeric"} holding the
      user identification code.}
    \item{\code{views}:}{Object of class \code{"environment"} containing
    a list stored under the name "instances". The list holds the instances of
    \code{\linkS4class{BrowserView}} for this session. }
  }
}
\section{Extends}{
Class \code{"\linkS4class{BrowserSession}"}, directly.
}
\section{Methods}{
  \describe{
    \item{\code{\link{browserView}(object, range =
        range(object), track = trackNames(object), ...)}}{ Creates a
      \code{\linkS4class{BrowserView}} of \code{range} with visible
      tracks specified by \code{track}. \code{track} may be an instance
      of \code{\linkS4class{UCSCTrackModes}}. Arguments in \code{...} should
      match parameters
      to a \code{\link{ucscTrackModes}} method for creating a
      \code{UCSCTrackModes} instance that will be merged with and
        override modes indicated by the \code{track} parameter. }
    \item{\code{\link{browserViews}}(object)}{ Gets the
      \code{\linkS4class{BrowserView}} instances for this session. }
    \item{\code{\link{range}}(x)}{ Gets the
      \code{\link[GenomicRanges:GRanges-class]{GRanges}} last
      displayed in this session.}
    \item{\code{genome(x)}}{Gets the genome identifier of the session,
      i.e. \code{genome(range(x))}.}
    \item{\code{seqinfo}}{ Gets
      the \code{\link[Seqinfo]{Seqinfo}} object with the
      lengths of the chromosomes in the currenet genome. No
      circularity information is available.
    }
    \item{\code{range(x) <- value}}{ Sets \code{value}, usually a
      \code{GRanges} object or \code{IntegerRangesList}, as the range
      of session \code{x}. Note that
      this setting only lasts until a view is created or
      manipulated. This mechanism is useful, for example, when treating the UCSC
      browser as a database, rather than a genome viewer.}
    \item{\code{genome(x) <- value}}{ 
      Sets the genome identifier on the range of session \code{x}.}
    \item{\code{\link[Biostrings]{getSeq}(object, range, track = "Assembly")}}{
      Gets the sequence in \code{range} and \code{track}.
    }
    \item{\code{\link[=track<-]{track}(object, name = names(track),
        format = "auto", ...) <- value}}{
      Loads a track, stored under \code{name} and formatted as
      \code{format}. The "auto" format resolves to "bed" for qualitative
        data. For quantitative data, i.e., data with a numeric \code{score}
        column, "wig" or "bedGraph" is chosen, depending on how well the
        data compresses into wig. The arguments in \code{...} are passed on to
      \code{\link{export.ucsc}}, so they could be slots in a
      \code{\linkS4class{TrackLine}} subclass (and thus specify
        visual attributes like color) or parameters to pass
      on to the export function for \code{format}. The \code{value} may
        be either a range object (like a \code{GRanges}) or a file
        object (like a \code{BEDFile}).
    }
    \item{\code{\link{track}(object, name, range = range(object),
        table = NULL)}}{
      Retrieves a \code{\link[GenomicRanges]{GRanges}} with features in 
      \code{range} from track named \code{name}. Some built-in tracks
      have multiple series, each stored in a separate database
      table. A specific table may be retrieved by passing its name in
      the \code{table} parameter. See \code{\link{tableNames}} for a way
      to list the available tables.
    }
    \item{\code{getTable(object, name, range = base::range(object), table =
      NULL)}}{
      Retrieves the table indicated by the track \code{name} and
      \code{table} name, over \code{range}, as a \code{data.frame}. See
      \code{\link{getTable}}.
    }
    \item{\code{\link{trackNames}}(object)}{ Gets the
      names of the tracks stored in the session. }
    \item{\code{\link{ucscTrackModes}}(object)}{ Gets the
      default view modes for the tracks in the session. }
  }
}
\author{ Michael Lawrence }
\seealso{
  \code{\link{browserSession}} for creating instances of this class.
}
\keyword{classes}

\name{FastaFile-class}
\docType{class}

%% Classes:
\alias{class:FastaFile}
\alias{FastaFile-class}

%% Constructor:
\alias{FastaFile}

%% Import:
\alias{import,FastaFile,ANY,ANY-method}

%% Export:
\alias{export,ANY,FastaFile,ANY-method}
\alias{export,XStringSet,FastaFile,ANY-method}

\title{FastaFile objects}

\description{
  These functions support the import and export of the Fasta
  sequence format, using the Biostrings package.
}

\usage{
\S4method{import}{FastaFile,ANY,ANY}(con, format, text,
           type = c("DNA", "RNA", "AA", "B"), ...)

\S4method{export}{ANY,FastaFile,ANY}(object, con, format, ...)
\S4method{export}{XStringSet,FastaFile,ANY}(object, con, format, ...)
}

\arguments{
  \item{con}{A path or \code{FastaFile} object.  URLs and connections
    are not supported. If \code{con} is not a \code{FastaFile}, either
    the file extension or the \code{format} argument needs to be
    \dQuote{fasta}. Compressed files (\dQuote{gz}, \dQuote{bz2} and
    \dQuote{xz}) are handled transparently.
  }
  \item{object}{The object to export, should be an \code{XStringSet} or
    something coercible to a \code{DNAStringSet}, like a character vector.
  }
  \item{format}{If not missing, should be \dQuote{fasta}.
  }
  \item{text}{If \code{con} is missing, a character vector to use as the
    input
  }
  \item{type}{Type of biological sequence.}
  \item{...}{Arguments to pass down
    to \code{\link[Biostrings]{writeXStringSet}} (export)
    or the \code{\link[Biostrings]{readDNAStringSet}} family of
    functions (import).
  }
}

\seealso{
  These functions are implemented by the Biostrings
  \code{\link[Biostrings]{writeXStringSet}} (export) and the
  \code{\link[Biostrings]{readDNAStringSet}} family of functions
  (import).

  See \link[BSgenome]{export-methods} in the \pkg{BSgenome} package for
  exporting a \link[BSgenome]{BSgenome} object as a FASTA file.
}

\section{FastaFile objects}{
  The \code{FastaFile} class extends
  \code{\link[BiocIO:BiocFile-class]{BiocFile}} and is
  a formal represention of a resource in the Fasta format.
  To cast a path, URL or connection to a \code{FastaFile}, pass it to
  the \code{FastaFile} constructor.
}

\author{Michael Lawrence}

\keyword{methods}
\keyword{classes}

\name{TwoBitFile-class}
\docType{class}

%% Classes:
\alias{class:TwoBitFile}
\alias{TwoBitFile-class}
\alias{class:2BitFile}
\alias{2BitFile-class}

%% Constructor:
\alias{TwoBitFile}
\alias{2BitFile}

%% Accessors:
\alias{seqinfo,TwoBitFile-method}

%% Import:
\alias{import.2bit}
\alias{import.2bit,ANY-method}
\alias{import,TwoBitFile,ANY,ANY-method}
\alias{getSeq,TwoBitFile-method}

%% Export:
\alias{export.2bit}
\alias{export.2bit,ANY-method}
\alias{export,DNAStringSet,TwoBitFile,ANY-method}
\alias{export,DNAStringSet,character,ANY-method}
\alias{export,ANY,TwoBitFile,ANY-method}

\title{2bit Files}

\description{
  These functions support the import and export of the UCSC 2bit
  compressed sequence format. The main advantage is speed of subsequence
  retrieval, as it only loads the sequence in the requested
  intervals. Compared to the FA format supported by Rsamtools, 2bit
  offers the additional feature of masking and also has better support
  in Java (and thus most genome browsers). The supporting
  \code{TwoBitFile} class is a reference to a TwoBit file.
}

\usage{
\S4method{import}{TwoBitFile,ANY,ANY}(con, format, text,
           which = as(seqinfo(con), "GenomicRanges"), ...)
\S4method{getSeq}{TwoBitFile}(x, which = as(seqinfo(x), "GenomicRanges"))
import.2bit(con, ...)

\S4method{export}{ANY,TwoBitFile,ANY}(object, con, format, ...)
\S4method{export}{DNAStringSet,TwoBitFile,ANY}(object, con, format)
\S4method{export}{DNAStringSet,character,ANY}(object, con, format, ...)
export.2bit(object, con, ...)
}

\arguments{
  \item{con}{A path, URL or \code{TwoBitFile} object. Connections are
    not supported. For the functions ending in \code{.2bit}, the file
    format is indicated by the function name. For the \code{export}
    and \code{import} methods, the format must be indicated another
    way. If \code{con} is a path, or URL, either the file
    extension or the \code{format} argument needs to be \dQuote{twoBit}
    or \dQuote{2bit}.
  }
  \item{object,x}{The object to export, either a \code{DNAStringSet} or
    something coercible to a \code{DNAStringSet}, like a character vector.
  }
  \item{format}{If not missing, should be \dQuote{twoBit} or \dQuote{2bit}
    (case insensitive).
  }
  \item{text}{Not supported.}
  \item{which}{A range data structure coercible to \code{IntegerRangesList},
    like a \code{GRanges}, or a \code{TwoBitFile}. Only the intervals in
    the file overlapping the given ranges are returned. By default, the
    value is the \code{TwoBitFile} itself. Its \code{Seqinfo} object is
    extracted and coerced to a \code{IntegerRangesList} that represents the
    entirety of the file.
  }
  \item{...}{Arguments to pass down to methods to other methods. For
    import, the flow eventually reaches the \code{TwoBitFile} method on
    \code{import}. For export, the \code{TwoBitFile} methods on
    \code{export} are the sink.
  }
}

\value{
  For import, a \code{DNAStringSet}.
}

\note{
  The 2bit format only suports A, C, G, T and N (via an internal
  mask). To export sequences with additional IUPAC ambiguity codes,
  first pass the object through
  \code{\link[Biostrings]{replaceAmbiguities}} from the Biostrings
  package.
}

\section{\code{TwoBitFile} objects}{
  A \code{TwoBitFile} object, an extension of
  \code{\link[BiocIO:BiocFile-class]{BiocFile-class}}
  is a reference to a TwoBit file. To cast
  a path, URL or connection to a \code{TwoBitFile}, pass it to the
  \code{TwoBitFile} constructor.

  A TwoBit file embeds the sequence information, which can be retrieved
  with the following:
  \describe{
    \item{\code{seqinfo(x)}}{
      Gets the \code{\link[Seqinfo]{Seqinfo}} object indicating
      the lengths of the sequences for the intervals in the
      file. No circularity or genome information is available.
    }
  }
}

\seealso{
  \link[BSgenome]{export-methods} in the \pkg{BSgenome} package for
  exporting a \link[BSgenome]{BSgenome} object as a twoBit file.
}

\author{Michael Lawrence}

\examples{
  test_path <- system.file("tests", package = "rtracklayer")
  test_2bit <- file.path(test_path, "test.2bit")

  test <- import(test_2bit)
  test

  test_2bit_file <- TwoBitFile(test_2bit)
  import(test_2bit_file) # the whole file
  
  which_range <- IRanges(c(10, 40), c(30, 42))
  which <- GRanges(names(test), which_range)
  import(test_2bit, which = which)

  seqinfo(test_2bit_file)

\dontrun{
  test_2bit_out <- file.path(tempdir(), "test_out.2bit")
  export(test, test_2bit_out)

  ## just a character vector
  test_char <- as.character(test)
  export(test_char, test_2bit_out)
}
}

\keyword{methods}
\keyword{classes}

\name{UCSCTableQuery-class}
\docType{class}

\alias{UCSCTableQuery-class}

% Accessors:
\alias{genome,UCSCTableQuery-method}
\alias{genome<-,UCSCTableQuery-method}
\alias{hubUrl}
\alias{hubUrl,UCSCTableQuery-method}
\alias{hubUrl<-,UCSCTableQuery-method}
\alias{tableName}
\alias{tableName,UCSCTableQuery-method}
\alias{tableName<-}
\alias{tableName<-,UCSCTableQuery-method}
\alias{range,UCSCTableQuery-method}
\alias{range<-,UCSCTableQuery-method}
\alias{names,UCSCTableQuery-method}
\alias{names<-,UCSCTableQuery-method}
\alias{ucscSchema}
\alias{ucscSchema,UCSCTableQuery-method}
\alias{browserSession,UCSCTableQuery-method}
\alias{browserSession<-}
\alias{browserSession<-,UCSCTableQuery-method}
\alias{ucscTables}

% Query execution
\alias{tableNames}
\alias{tableNames,UCSCTableQuery-method}
\alias{getTable}
\alias{getTable,UCSCTableQuery-method}
\alias{track}
\alias{track,UCSCTableQuery-method}

% Constructor:
\alias{ucscTableQuery}
\alias{ucscTableQuery,UCSCSession-method}
\alias{ucscTableQuery,character-method}

% Show:
\alias{show,UCSCTableQuery-method}

\title{Querying UCSC Tables}
\description{The UCSC genome browser is backed by a large database,
  which is exposed by the Table Browser web interface. Tracks are
  stored as tables, so this is also the mechanism for retrieving tracks. The
  \code{UCSCTableQuery} class represents a query against the Table
  Browser. Storing the query fields in a formal class facilitates
  incremental construction and adjustment of a query.} 

\details{
  There are six supported fields for a table query:
  \describe{
    \item{provider}{
      The provider should be a session, a genome identifier, or a TrackHub URI.
      \code{session}: The \code{\linkS4class{UCSCSession}} instance from
      the tables are retrieved. Although all sessions are based on the
      same database, the set of user-uploaded tracks, which are represented
      as tables, is not the same, in general.
    }
    \item{tableName}{The name of the specific table to retrieve. May be
      \code{NULL}, in which case the behavior depends on how the query
      is executed, see below.
    }
    \item{range}{A genome identifier, a
      \code{\link[GenomicRanges]{GRanges}} or 
      a \code{\link[IRanges]{IntegerRangesList}} indicating
      the portion of the table to retrieve, in genome coordinates.
      Simply specifying the genome string is the easiest way to download
      data for the entire genome, and \code{\link{GRangesForUCSCGenome}}
      facilitates downloading data for e.g. an entire chromosome.
    }
    \item{hubUrl}{The URI of the specific TrackHub}
    \item{genome}{A genome identifier of the specific TrackHub, only need to provide it if the provider is up of TrackHub URI.}
    \item{names}{Names/accessions of the desired features}
  } 

  A common workflow for querying the UCSC database is to create an
  instance of \code{UCSCTableQuery} using the \code{ucscTableQuery}
  constructor, invoke \code{tableNames} to list the available tables for
  a track, and finally to retrieve the desired table either as a
  \code{data.frame} via \code{getTable} or as a track
  via \code{track}. See the examples.

  The reason for a formal query class is to facilitate multiple queries
  when the differences between the queries are small. For example, one
  might want to query multiple tables within the track and/or same
  genomic region, or query the same table for multiple regions. The
  \code{UCSCTableQuery} instance can be incrementally adjusted for each
  new query. Some caching is also performed, which enhances performance. 
}

\section{Constructor}{
  \describe{
    \item{\code{ucscTableQuery(x, range = seqinfo(x), table = NULL,
        names = NULL, hubUrl = NULL, genome = NULL)}}{
        Creates a \code{UCSCTableQuery} with the
        \code{UCSCSession}, genome identifier or TrackHub URI given as \code{x} and
        the table name given by the single string \code{table}. \code{range} should
        be a genome string identifier, a \code{GRanges} instance or
        \code{IntegerRangesList} instance, and it effectively defaults to
        \code{genome(x)}. If the genome is missing, it is taken from the
        provider. Feature names, such as gene identifiers, may be
        passed via \code{names} as a character vector.
    }
  }
}

\section{Executing Queries}{
  Below, \code{object} is a \code{UCSCTableQuery} instance.
  
  \describe{
    \item{\code{track(object)}}{
      Retrieves the indicated table as a track, i.e. a \code{GRanges}
      object. Note that not all tables are available as tracks.
    }
    \item{\code{getTable(object)}}{
      Retrieves the indicated table as a
      \code{data.frame}. Note that not all tables are output in
      parseable form, and that UCSC will truncate responses if they
      exceed certain limits (usually around 100,000 records). The safest
      (and most efficient) bet for large queries is to download the file
      via FTP and query it locally.
    }
    \item{\code{tableNames(object)}}{
      Gets the names of the tables available
      for the provider, table and range specified by the query.
    }
  }
}

\section{Accessor methods}{
  In the code snippets below, \code{x}/\code{object} is a
  \code{UCSCTableQuery} object.

  \describe{
    \item{\code{genome(x)}, \code{genome(x) <- value}}{Gets or sets
    the genome identifier (e.g. \dQuote{hg18}) of the object.
    }
    \item{\code{hubUrl(x)}, \code{hubUrl(x) <- value}}{Gets or sets
    the TrackHub URI.
    }
    \item{\code{tableName(x)}, \code{tableName(x) <- value}}{Get or
      set the single string indicating the name of the table to
      retrieve. May be \code{NULL}, in which case the table is
      automatically determined.
    }
    \item{\code{range(x)}, \code{range(x) <- value}}{Get or set the
      \code{GRanges} indicating the portion of the table to retrieve in
      genomic coordinates. Any missing information, such as the genome
      identifier, is filled in using \code{range(browserSession(x))}. It
      is also possible to set the genome identifier string or
      a \code{IntegerRangesList}.
    }
    \item{\code{names(x)}, \code{names(x) <- value}}{Get or set the
      names of the features to retrieve. If \code{NULL}, this filter is
      disabled.
    }
    \item{\code{ucscSchema(x)}}{ Get
      the \code{\linkS4class{UCSCSchema}} object describing the selected table.}
    \item{\code{ucscTables(genome, track)}}{Defunct. Typically, one
      would use \code{tableNames} to retrieve the names of available
      tables. To get the name of a UCSC table for a track, use the table
      browser interactively.}
    \item{\code{browserSession(x)}, \code{browserSession(x) <- value}}{Gets or sets
    the \code{\linkS4class{BrowserSession}} of the object.
    }
  }
}

\author{ Michael Lawrence }

\examples{
\dontrun{
# query using `session` provider
session <- browserSession()
genome(session) <- "mm9"
## choose the phastCons30way table for a portion of mm9 chr1
query <- ucscTableQuery(session, table = "phastCons30way",
                        range = GRangesForUCSCGenome("mm9", "chr12",
                                                     IRanges(57795963, 57815592)))
## list the table names
tableNames(query)
## retrieve the track data
track(query)  # a GRanges object
## get the multiz30way table
tableName(query) <- "multiz30way"
## get a data.frame
getTable(query)

# query using `genome identifier` provider
query <- ucscTableQuery("hg38", table = "snp141Common",
                        range = GRangesForUCSCGenome("hg38", "chr12",
                                                     IRanges(57795963, 57815592)))
ucscSchema(query)
getTable(query)

# query using `TrackHub URI` provider
query <- ucscTableQuery("https://hgdownload.soe.ucsc.edu/hubs/gtexAnalysis",
                        genome = "hg38", table = "gtexAwgAseLung")
track(query)
}
}

\keyword{methods}
\keyword{classes}

\name{ucscGenomes}
\alias{ucscGenomes}
\title{Get available genomes on UCSC}
\description{
  Get a \code{data.frame} describing the available UCSC genomes.
}
\usage{
ucscGenomes(organism=FALSE)
}
\arguments{
  \item{organism}{A \code{logical(1)} indicating whether scientific name
    should be appended.}
}
\value{
  A \code{data.frame} with the following columns:
  \item{db}{UCSC DB identifier (e.g. "hg18")}
  \item{species}{The name of the species (e.g. "Human")}
  \item{date}{The date the genome was built}
  \item{name}{The official name of the genome build}
  \item{organism}{The scientific name of the species (e.g. "Homo sapiens")}
}
\seealso{\code{\linkS4class{UCSCSession}} for details on specifying the
  genome.}
\details{
  For populating the organism column, the web url
  \url{https://genome.ucsc.edu/cgi-bin} is scraped for every assembly
  version to get the scientific name.
}
\examples{
ucscGenomes()
}
\author{ Michael Lawrence }

\keyword{interface}

\name{browserViews-methods}
\docType{methods}
\alias{browserViews}
\alias{browserViews-methods}
\alias{browserViews,UCSCSession-method}
\title{ Getting the browser views }
\description{
 Methods for getting browser views.
}
\section{Methods}{
\describe{
  The following methods are defined by \pkg{rtracklayer}. 
  \item{object = "UCSCSession"}{ Gets the instances of
    \code{\linkS4class{BrowserView}} in the session. }
}}
\seealso{\code{\link{browserView}} for creating a browser view.}
\examples{
\dontrun{
session <- browseGenome()
browserViews(session)
}
}
\keyword{methods}

\name{TrackLine-class}
\docType{class}
\alias{TrackLine-class}
\alias{coerce,character,TrackLine-method}
\alias{coerce,TrackLine,character-method}
\alias{show,TrackLine-method}

\title{Class "TrackLine"}
\description{An object representing a "track line" in the \acronym{UCSC}
  format. There are two concrete types of track lines:
  \code{\linkS4class{BasicTrackLine}} (used for most types of tracks) and
  \code{\linkS4class{GraphTrackLine}} (used for graphical tracks). This class
  only declares the common elements between the two.}
\section{Objects from the Class}{
  Objects can be created by calls of the form \code{new("TrackLine",
    ...)} or parsed from a character vector track line with
  \code{as(text, "TrackLine")}.
  But note that \acronym{UCSC} only understands one of the
  subclasses mentioned above.
}
\section{Slots}{
  \describe{
    \item{\code{name}:}{Object of class \code{"character"} specifying
      the name of the track. }
    \item{\code{description}:}{Object of class \code{"character"}
      describing the track. }
    \item{\code{visibility}:}{Object of class \code{"character"}
      indicating the default visible mode of the track, see
      \code{\linkS4class{UCSCTrackModes}}.}
    \item{\code{color}:}{Object of class \code{"integer"} representing the
      track color (as from \code{\link{col2rgb}}).
    }
    \item{\code{priority}:}{Object of class \code{"numeric"} specifying
      the rank of this track. }
  }
}
\section{Methods}{
  \describe{
    \item{as(object, "character")}{Export line to its string
      representation.}
  }
}
\references{
  \url{https://genome.ucsc.edu/goldenPath/help/customTrack.html#TRACK}
  for the official documentation. }
\author{ Michael Lawrence }

\seealso{
  \code{\linkS4class{BasicTrackLine}} (used for most types of tracks) and
  \code{\linkS4class{GraphTrackLine}} (used for Wiggle/bedGraph tracks).
}
\keyword{classes}
